@@ -10,7 +10,7 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	22 November 1989
- * Last Edited:	3 March 1997
+ * Last Edited:	24 December 1997
  *
  * Copyright 1997 by the University of Washington
  *
@@ -98,6 +98,12 @@
 #define SET_RFC822OUTPUT (long) 110
 #define GET_READPROGRESS (long) 111
 #define SET_READPROGRESS (long) 112
+#define GET_THREADERS (long) 113
+#define SET_THREADERS (long) 114
+#define GET_NAMESPACE (long) 115
+#define SET_NAMESPACE (long) 116
+#define GET_MAILPROXYCOPY (long) 117
+#define SET_MAILPROXYCOPY (long) 118
 	/* 2xx: environment */
 #define GET_USERNAME (long) 201
 #define SET_USERNAME (long) 202
@@ -120,6 +126,8 @@
 #define SET_TIMEOUT (long) 309
 #define GET_RSHTIMEOUT (long) 310
 #define SET_RSHTIMEOUT (long) 311
+#define GET_ALARMSAVE (long) 312
+#define SET_ALARMSAVE (long) 313
 
 	/* 4xx: network drivers */
 #define GET_MAXLOGINTRIALS (long) 400
@@ -140,6 +148,8 @@
 #define SET_NNTPPORT (long) 415
 #define GET_IMAPENVELOPE (long) 416
 #define SET_IMAPENVELOPE (long) 417
+#define GET_IMAPREFERRAL (long) 418
+#define SET_IMAPREFERRAL (long) 419
 	/* 5xx: local file drivers */
 #define GET_MBXPROTECTION (long) 500
 #define SET_MBXPROTECTION (long) 501
@@ -186,6 +196,8 @@
 #define DR_NOFAST (long) 32	/* "fast" data is slow (whole msg fetch) */
 #define DR_NAMESPACE (long) 64	/* driver has a special namespace */
 #define DR_LOWMEM (long) 128	/* low amounts of memory available */
+#define DR_LOCKING (long) 256	/* driver does locking */
+#define DR_CRLF (long) 512	/* driver internal form uses CRLF newlines */
 
 
 /* Cache management function codes */
@@ -194,7 +206,10 @@
 #define CH_SIZE (long) 11	/* (re-)size the cache */
 #define CH_MAKEELT (long) 30	/* return elt, make if needed */
 #define CH_ELT (long) 31	/* return elt if exists */
+#define CH_SORTCACHE (long) 35	/* return sortcache entry, make if needed */
 #define CH_FREE (long) 40	/* free space used by elt */
+				/* free space used by sortcache */
+#define CH_FREESORTCACHE (long) 43
 #define CH_EXPUNGE (long) 45	/* delete elt pointer from list */
 
 
@@ -242,6 +257,7 @@
 #define SE_FREE (long) 2	/* free search program after finished */
 #define SE_NOPREFETCH (long) 4	/* no search prefetching */
 #define SO_FREE (long) 8	/* free sort program after finished */
+#define SO_NOSERVER (long) 16	/* don't do server-based sort */
 
 
 /* Status options */
@@ -284,15 +300,26 @@
 #define SORTCC 5		/* cc */
 #define SORTSIZE 6		/* size */
 
+/* imapreferral_t codes */
+
+#define REFAUTHFAILED (long) 0	/* authentication referral -- not logged in */
+#define REFAUTH (long) 1	/* authentication referral -- logged in */
+#define REFSELECT (long) 2	/* select referral */
+#define REFCREATE (long) 3
+#define REFDELETE (long) 4
+#define REFRENAME (long) 5
+#define REFSUBSCRIBE (long) 6
+#define REFUNSUBSCRIBE (long) 7
+#define REFSTATUS (long) 8
+#define REFCOPY (long) 9
+#define REFAPPEND (long) 10
+
 /* In-memory sized-text */
 
 #define SIZEDTEXT struct mail_text
 
-
-/* Sized text */
-
 SIZEDTEXT {
-  char *data;			/* text */
+  unsigned char *data;		/* text */
   unsigned long size;		/* size of text in octets */
 };
 
@@ -307,15 +334,21 @@ STRINGLIST {
 };
 
 
-/* Sort program */
+/* Parse results from mail_valid_net_parse */
 
-#define SORTPGM struct sort_program
-
-SORTPGM {
-  unsigned int reverse : 1;	/* sort function is to be reversed */
-  short function;		/* sort function */
-  SORTPGM *next;		/* next function */
-};
+#define NETMAXHOST 65
+#define NETMAXUSER 65
+#define NETMAXMBX 256
+#define NETMAXSRV 21
+typedef struct net_mailbox {
+  char host[NETMAXHOST];	/* host name */
+  char user[NETMAXUSER];	/* user name */
+  char mailbox[NETMAXMBX];	/* mailbox name */
+  char service[NETMAXSRV];	/* service name */
+  unsigned long port;		/* TCP port number */
+  unsigned int anoflag : 1;	/* anonymous */
+  unsigned int dbgflag : 1;	/* debug flag */
+} NETMBX;
 
 /* Item in an address list */
 
@@ -354,8 +387,6 @@ typedef struct mail_envelope {
 /* Primary body types */
 /* If you change any of these you must also change body_types in rfc822.c */
 
-extern char *body_types[];	/* defined body type strings */
-
 #define TYPETEXT 0		/* unformatted text */
 #define TYPEMULTIPART 1		/* multiple part */
 #define TYPEMESSAGE 2		/* encapsulated message */
@@ -372,8 +403,6 @@ extern char *body_types[];	/* defined body type strings */
 /* If you change any of these you must also change body_encodings in rfc822.c
  */
 
-extern char *body_encodings[];	/* defined body encoding strings */
-
 #define ENC7BIT 0		/* 7 bit SMTP semantic data */
 #define ENC8BIT 1		/* 8 bit SMTP semantic data */
 #define ENCBINARY 2		/* 8 bit binary data */
@@ -465,6 +494,7 @@ typedef struct message_cache {
     unsigned long uid;		/* message unique ID */
     PARTTEXT special;		/* special text pointers */
     MESSAGE msg;		/* internal message pointers */
+    unsigned int sequence : 1;	/* saved sequence bit; */
   } private;
 			/* internal date */
   unsigned int day : 5;		/* day of month (1-31) */
@@ -543,8 +573,8 @@ STRINGDRIVER {
 
 
 SEARCHHEADER {			/* header search */
-  char *line;			/* header line */
-  char *text;			/* text in header */
+  SIZEDTEXT line;		/* header line */
+  SIZEDTEXT text;		/* text in header */
   SEARCHHEADER *next;		/* next in list */
 };
 
@@ -617,6 +647,45 @@ typedef struct mbx_status {
   unsigned long uidvalidity;	/* UID validity value */
 } MAILSTATUS;
 
+/* Sort program */
+
+typedef void (*postsort_t) (void *sc);
+
+#define SORTPGM struct sort_program
+
+SORTPGM {
+  unsigned int reverse : 1;	/* sort function is to be reversed */
+  unsigned int abort : 1;	/* abort sorting */
+  short function;		/* sort function */
+  unsigned long nmsgs;		/* number of messages being sorted */
+  struct {
+    unsigned long cached;	/* number of messages cached so far */
+    unsigned long sorted;	/* number of messages sorted so far */
+    unsigned long postsorted;	/* number of postsorted messages so far */
+  } progress;
+  postsort_t postsort;		/* post sorter */
+  SORTPGM *next;		/* next function */
+};
+
+
+/* Sort cache */
+
+#define SORTCACHE struct sort_cache
+
+SORTCACHE {
+  unsigned int sorted : 1;	/* message has been sorted */
+  unsigned int postsorted : 1;	/* message has been postsorted */
+  SORTPGM *pgm;			/* sort program */
+  unsigned long num;		/* message number (sequence or UID) */
+  unsigned long date;		/* delivery date */
+  unsigned long arrival;	/* arrival date */
+  unsigned long size;		/* message size */
+  char *from;			/* from string */
+  char *to;			/* to string */
+  char *cc;			/* cc string */
+  char *subject;		/* subject string */
+};
+
 /* Mail Access I/O stream */
 
 
@@ -655,16 +724,16 @@ typedef struct mail_stream {
   char *user_flags[NUSERFLAGS];	/* pointers to user flags in bit order */
   unsigned long cachesize;	/* size of message cache */
   MESSAGECACHE **cache;		/* message cache array */
+  SORTCACHE **sc;		/* sort cache array */
   unsigned long msgno;		/* message number of `current' message */
   ENVELOPE *env;		/* scratch buffer for envelope */
   BODY *body;			/* scratch buffer for body */
-  SIZEDTEXT header;		/* scratch buffer for header text */
   SIZEDTEXT text;		/* scratch buffer for text */
   union {			/* internal use only */
     struct {			/* search temporaries */
-      char *charset;		/* character set */
       STRINGLIST *string;	/* string(s) to search */
       long result;		/* search result */
+      char *text;		/* cache of fetched text */
     } search;
   } private;
 } MAILSTREAM;
@@ -691,12 +760,191 @@ typedef struct mail_overview {
     char *xref;			/* cross references */
   } optional;
 } OVERVIEW;
+
+/* Network access I/O stream */
 
 
+/* Structure for network driver dispatch */
+
+#define NETDRIVER struct net_driver
+
+
+/* Network transport I/O stream */
+
+typedef struct net_stream {
+  void *stream;			/* driver's I/O stream */
+  NETDRIVER *dtb;		/* network driver */
+} NETSTREAM;
+
+
+/* Network transport driver dispatch */
+
+NETDRIVER {
+  void *(*open) (char *host,char *service,unsigned long port);
+  void *(*aopen) (NETMBX *mb,char *service,char *usrbuf);
+  char *(*getline) (void *stream);
+  long (*getbuffer) (void *stream,unsigned long size,char *buffer);
+  long (*soutr) (void *stream,char *string);
+  long (*sout) (void *stream,char *string,unsigned long size);
+  void (*close) (void *stream);
+  char *(*host) (void *stream);
+  unsigned long (*port) (void *stream);
+  char *(*localhost) (void *stream);
+};
+
+
+/* Mailgets data identifier */
+
+typedef struct GETS_DATA {
+  MAILSTREAM *stream;
+  unsigned long msgno;
+  char *what;
+  STRINGLIST *stl;
+  unsigned long first;
+  unsigned long last;
+  long flags;
+} GETS_DATA;
+
+
+#define INIT_GETS(md,s,m,w,f,l) \
+  md.stream = s, md.msgno = m, md.what = w, md.first = f, md.last = l, \
+  md.stl = NIL, md.flags = NIL;
+
+/* Mail delivery I/O stream */
+
+typedef struct send_stream {
+  NETSTREAM *netstream;		/* network I/O stream */
+  char *reply;			/* last reply string */
+  unsigned int debug : 1;	/* stream debug flag */
+  union {			/* protocol specific */
+    struct {			/* SMTP specific */
+      unsigned int ok : 1;	/* supports ESMTP */
+      struct {			/* service extensions */
+	unsigned int send : 1;	/* supports SEND */
+	unsigned int soml : 1;	/* supports SOML */
+	unsigned int saml : 1;	/* supports SAML */
+	unsigned int expn : 1;	/* supports EXPN */
+	unsigned int help : 1;	/* supports HELP */
+	unsigned int turn : 1;	/* supports TURN */
+      } service;
+      struct {			/* 8-bit MIME transport */
+	unsigned int ok : 1;	/* supports 8-bit MIME */
+	unsigned int want : 1;	/* want 8-bit MIME */
+      } eightbit;
+      struct {			/* delivery status notification */
+	unsigned int ok : 1;	/* supports DSN */
+	unsigned int want : 1;	/* want DSN */
+	struct {		/* notification options */
+				/* notify on failure */
+	  unsigned int failure : 1;
+				/* notify on delay */
+	  unsigned int delay : 1;
+				/* notify on success */
+	  unsigned int success : 1;
+	} notify;
+	unsigned int full : 1;	/* return full headers */
+      } dsn;
+      struct {			/* size declaration */
+	unsigned int ok : 1;	/* supports SIZE */
+	unsigned long limit;	/* maximum size supported */
+      } size;
+    } esmtp;
+    struct {			/* NNTP specific */
+      unsigned int post : 1;	/* supports POST */
+    } nntp;
+  } protocol;
+} SENDSTREAM;
+
+/* Jacket into external interfaces */
+
+typedef long (*readfn_t) (void *stream,unsigned long size,char *buffer);
+typedef char *(*mailgets_t) (readfn_t f,void *stream,unsigned long size,
+			     GETS_DATA *md);
+typedef char *(*readprogress_t) (GETS_DATA *md,unsigned long octets);
+typedef void *(*mailcache_t) (MAILSTREAM *stream,unsigned long msgno,long op);
+typedef long (*mailproxycopy_t) (MAILSTREAM *stream,char *sequence,
+				 char *mailbox,long options);
+typedef long (*tcptimeout_t) (long time);
+typedef void *(*authchallenge_t) (void *stream,unsigned long *len);
+typedef long (*authrespond_t) (void *stream,char *s,unsigned long size);
+typedef long (*authclient_t) (authchallenge_t challenger,
+			      authrespond_t responder,NETMBX *mb,void *s,
+			      unsigned long *trial,char *user);
+typedef char *(*authresponse_t) (void *challenge,unsigned long clen,
+				 unsigned long *rlen);
+typedef char *(*authserver_t) (authresponse_t responder,int argc,char *argv[]);
+typedef void (*smtpverbose_t) (char *buffer);
+typedef void (*imapenvelope_t) (MAILSTREAM *stream,unsigned long msgno,
+				ENVELOPE *env);
+typedef char *(*imapreferral_t) (MAILSTREAM *stream,char *url,long code);
 typedef void (*overview_t) (MAILSTREAM *stream,unsigned long uid,OVERVIEW *ov);
+typedef unsigned long *(*sorter_t) (MAILSTREAM *stream,char *charset,
+				    SEARCHPGM *spg,SORTPGM *pgm,long flags);
+
+
+/* Globals */
+
+extern char *body_types[];	/* defined body type strings */
+extern char *body_encodings[];	/* defined body encoding strings */
+extern const char *days[];	/* day name strings */
+extern const char *months[];	/* month name strings */
+
+/* Threading */
+
+/* Thread node */
+
+#define THREADNODE struct thread_node
+
+THREADNODE {
+  unsigned long num;		/* message number */
+  SORTCACHE *sc;		/* (internal use) sortcache entry */
+  THREADNODE *branch;		/* branch at this point in tree */
+  THREADNODE *next;		/* next node */
+};
+
+
+/* Thread dispatch */
+
+#define THREADER struct threader_list
+
+THREADER {
+  char *name;			/* name of threader */
+  THREADNODE *(*dispatch) (MAILSTREAM *stream,char *charset,SEARCHPGM *spg,
+			   long flags,sorter_t sorter);
+  THREADER *next;
+};
+
+
+/* Namespaces */
+
+#define NAMESPACE struct mail_namespace
+
+NAMESPACE {
+  char *name;			/* name of this namespace */
+  int delimiter;		/* hierarchy delimiter */
+  PARAMETER *param;		/* namespace parameters */
+  NAMESPACE *next;		/* next namespace */
+};
+
+
+/* Authentication */
+
+#define AUTHENTICATOR struct mail_authenticator
+
+AUTHENTICATOR {
+  char *name;			/* name of this authenticator */
+  authclient_t client;		/* client function that supports it */
+  authserver_t server;		/* server function that supports it */
+  AUTHENTICATOR *next;		/* next authenticator */
+};
 
 /* Mail driver dispatch */
 
+				/* normal dispatching */
+#define SAFE_DISPATCH(dtb,ret,dsp,args) \
+  if (dtb) SAFE_FUNCTION(dtb,ret,(*dtb->dsp),args)
+#define SAFE_FUNCTION(dtb,ret,func,args) ret = func args;
+
 DRIVER {
   char *name;			/* driver name */
   unsigned long flags;		/* driver flags */
@@ -760,7 +1008,8 @@ DRIVER {
   unsigned long *(*sort) (MAILSTREAM *stream,char *charset,SEARCHPGM *spg,
 			  SORTPGM *pgm,long flags);
 				/* thread messages */
-  void *(*thread) (MAILSTREAM *stream,char *seq,long function,long flag);
+  THREADNODE *(*thread) (MAILSTREAM *stream,char *type,char *charset,
+			 SEARCHPGM *spg,long flag);
 				/* ping mailbox to see if still alive */
   long (*ping) (MAILSTREAM *stream);
 				/* check for new messages */
@@ -777,152 +1026,6 @@ DRIVER {
 };
 
 
-/* Parse results from mail_valid_net_parse */
-
-#define NETMAXHOST 65
-#define NETMAXUSER 65
-#define NETMAXMBX 256
-#define NETMAXSRV 21
-typedef struct net_mailbox {
-  char host[NETMAXHOST];	/* host name */
-  char user[NETMAXUSER];	/* user name */
-  char mailbox[NETMAXMBX];	/* mailbox name */
-  char service[NETMAXSRV];	/* service name */
-  unsigned long port;		/* TCP port number */
-  unsigned int anoflag : 1;	/* anonymous */
-  unsigned int dbgflag : 1;	/* debug flag */
-} NETMBX;
-
-/* Network access I/O stream */
-
-
-/* Structure for network driver dispatch */
-
-#define NETDRIVER struct net_driver
-
-
-/* Network transport I/O stream */
-
-typedef struct net_stream {
-  void *stream;			/* driver's I/O stream */
-  NETDRIVER *dtb;		/* network driver */
-} NETSTREAM;
-
-
-/* Network transport driver dispatch */
-
-NETDRIVER {
-  void *(*open) (char *host,char *service,unsigned long port);
-  void *(*aopen) (NETMBX *mb,char *service,char *usrbuf);
-  char *(*getline) (void *stream);
-  long (*getbuffer) (void *stream,unsigned long size,char *buffer);
-  long (*soutr) (void *stream,char *string);
-  long (*sout) (void *stream,char *string,unsigned long size);
-  void (*close) (void *stream);
-  char *(*host) (void *stream);
-  unsigned long (*port) (void *stream);
-  char *(*localhost) (void *stream);
-};
-
-/* Mail delivery I/O stream */
-
-typedef struct send_stream {
-  NETSTREAM *netstream;		/* network I/O stream */
-  char *reply;			/* last reply string */
-  unsigned int debug : 1;	/* stream debug flag */
-  union {			/* protocol specific */
-    struct {			/* SMTP specific */
-      unsigned int ok : 1;	/* supports ESMTP */
-      struct {			/* service extensions */
-	unsigned int send : 1;	/* supports SEND */
-	unsigned int soml : 1;	/* supports SOML */
-	unsigned int saml : 1;	/* supports SAML */
-	unsigned int expn : 1;	/* supports EXPN */
-	unsigned int help : 1;	/* supports HELP */
-	unsigned int turn : 1;	/* supports TURN */
-      } service;
-      struct {			/* 8-bit MIME transport */
-	unsigned int ok : 1;	/* supports 8-bit MIME */
-	unsigned int want : 1;	/* want 8-bit MIME */
-      } eightbit;
-      struct {			/* delivery status notification */
-	unsigned int ok : 1;	/* supports DSN */
-	unsigned int want : 1;	/* want DSN */
-	struct {		/* notification options */
-				/* notify on failure */
-	  unsigned int failure : 1;
-				/* notify on delay */
-	  unsigned int delay : 1;
-				/* notify on success */
-	  unsigned int success : 1;
-	} notify;
-	unsigned int full : 1;	/* return full headers */
-      } dsn;
-      struct {			/* size declaration */
-	unsigned int ok : 1;	/* supports SIZE */
-	unsigned long limit;	/* maximum size supported */
-      } size;
-    } esmtp;
-    struct {			/* NNTP specific */
-      unsigned int post : 1;	/* supports POST */
-    } nntp;
-  } protocol;
-} SENDSTREAM;
-
-/* Mailgets data identifier */
-
-typedef struct GETS_DATA {
-  MAILSTREAM *stream;
-  unsigned long msgno;
-  char *what;
-  STRINGLIST *stl;
-  unsigned long first;
-  unsigned long last;
-  long flags;
-} GETS_DATA;
-
-#define INIT_GETS(md,s,m,w,f,l) \
-  md.stream = s, md.msgno = m, md.what = w, md.first = f, md.last = l, \
-  md.stl = NIL, md.flags = NIL;
-
-
-/* Jacket into external interfaces */
-
-typedef long (*readfn_t) (void *stream,unsigned long size,char *buffer);
-typedef char *(*mailgets_t) (readfn_t f,void *stream,unsigned long size,
-			     GETS_DATA *md);
-typedef char *(*readprogress_t) (GETS_DATA *md,unsigned long octets);
-typedef void *(*mailcache_t) (MAILSTREAM *stream,unsigned long msgno,long op);
-typedef long (*tcptimeout_t) (long time);
-typedef void *(*authchallenge_t) (void *stream,unsigned long *len);
-typedef long (*authrespond_t) (void *stream,char *s,unsigned long size);
-typedef long (*authclient_t) (authchallenge_t challenger,
-			      authrespond_t responder,NETMBX *mb,void *s,
-			      unsigned long *trial,char *user);
-typedef char *(*authresponse_t) (void *challenge,unsigned long clen,
-				 unsigned long *rlen);
-typedef char *(*authserver_t) (authresponse_t responder,int argc,char *argv[]);
-typedef void (*smtpverbose_t) (char *buffer);
-typedef void (*imapenvelope_t) (MAILSTREAM *stream,unsigned long msgno,
-				ENVELOPE *env);
-
-
-#define AUTHENTICATOR struct mail_authenticator
-
-AUTHENTICATOR {
-  char *name;			/* name of this authenticator */
-  authclient_t client;		/* client function that supports it */
-  authserver_t server;		/* server function that supports it */
-  AUTHENTICATOR *next;		/* next authenticator */
-};
-
-
-/* Other symbols */
-
-extern const char *days[];	/* day name strings */
-extern const char *months[];	/* month name strings */
-
-
 #include "linkage.h"
 
 /* Compatibility support names for old interfaces */
@@ -1004,6 +1107,7 @@ long mail_create (MAILSTREAM *stream,char *mailbox);
 long mail_delete (MAILSTREAM *stream,char *mailbox);
 long mail_rename (MAILSTREAM *stream,char *old,char *newname);
 long mail_status (MAILSTREAM *stream,char *mbx,long flags);
+long mail_status_default (MAILSTREAM *stream,char *mbx,long flags);
 MAILSTREAM *mail_open (MAILSTREAM *oldstream,char *name,long options);
 MAILSTREAM *mail_close_full (MAILSTREAM *stream,long options);
 MAILHANDLE *mail_makehandle (MAILSTREAM *stream);
@@ -1053,6 +1157,7 @@ long mail_append_full (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
 		       STRING *message);
 void mail_gc (MAILSTREAM *stream,long gcflags);
 void mail_gc_msg (MESSAGE *msg,long gcflags);
+void mail_gc_body (BODY *body);
 
 BODY *mail_body (MAILSTREAM *stream,unsigned long msgno,char *section);
 char *mail_date (char *string,MESSAGECACHE *elt);
@@ -1068,32 +1173,40 @@ void mail_nodebug (MAILSTREAM *stream);
 long mail_match_lines (STRINGLIST *lines,STRINGLIST *msglines,long flags);
 unsigned long mail_filter (char *text,unsigned long len,STRINGLIST *lines,
 			   long flags);
-long mail_search_msg (MAILSTREAM *stream,unsigned long msgno,char *charset,
-		      SEARCHPGM *pgm);
+long mail_search_msg (MAILSTREAM *stream,unsigned long msgno,SEARCHPGM *pgm);
+long mail_search_header (SIZEDTEXT *hdr,STRINGLIST *st);
+long mail_search_text (MAILSTREAM *stream,unsigned long msgno,STRINGLIST *st,
+		       long flags);
+long mail_search_body (MAILSTREAM *stream,unsigned long msgno,BODY *body,
+		       char *prefix,unsigned long section,long flags);
+long mail_search_string (SIZEDTEXT *s,char *charset,STRINGLIST **st);
 long mail_search_keyword (MAILSTREAM *stream,MESSAGECACHE *elt,STRINGLIST *st);
-long mail_search_addr (ADDRESS *adr,char *charset,STRINGLIST *st);
-long mail_search_string (char *txt,char *charset,STRINGLIST *st);
+long mail_search_addr (ADDRESS *adr,STRINGLIST *st);
 char *mail_search_gets (readfn_t f,void *stream,unsigned long size,
 			MAILSTREAM *ms,unsigned long msgno,char *what,
 			long flags);
-long mail_search_text (char *txt,long len,char *charset,STRINGLIST *st);
 SEARCHPGM *mail_criteria (char *criteria);
 int mail_criteria_date (unsigned short *date);
 int mail_criteria_string (STRINGLIST **s);
 unsigned long *mail_sort (MAILSTREAM *stream,char *charset,SEARCHPGM *spg,
 			  SORTPGM *pgm,long flags);
+unsigned long *mail_sort_cache (MAILSTREAM *stream,SORTPGM *pgm,SORTCACHE **sc,
+				long flags);
 unsigned long *mail_sort_msgs (MAILSTREAM *stream,char *charset,SEARCHPGM *spg,
 			       SORTPGM *pgm,long flags);
+SORTCACHE **mail_sort_loadcache (MAILSTREAM *stream,SORTPGM *pgm);
 int mail_sort_compare (const void *a1,const void *a2);
-int mail_compare_msg (MAILSTREAM *stream,short function,unsigned long m1,
-		      unsigned long m2);
 int mail_compare_ulong (unsigned long l1,unsigned long l2);
 int mail_compare_cstring (char *s1,char *s2);
-int mail_compare_sstring (char *s1,char *s2);
-int mail_compare_address (ADDRESS *a1,ADDRESS *a2);
 unsigned long mail_longdate (MESSAGECACHE *elt);
-char *mail_skip_re (char *s);
-char *mail_skip_fwd (char *s);
+THREADNODE *mail_thread (MAILSTREAM *stream,char *type,char *charset,
+			 SEARCHPGM *spg,long flags);
+THREADNODE *mail_thread_msgs (MAILSTREAM *stream,char *type,char *charset,
+			      SEARCHPGM *spg,long flags,sorter_t sorter);
+THREADNODE *mail_thread_orderedsubject (MAILSTREAM *stream,char *charset,
+					SEARCHPGM *spg,long flags,
+					sorter_t sorter);
+int mail_thread_compare_date (const void *a1,const void *a2);
 long mail_sequence (MAILSTREAM *stream,char *sequence);
 long mail_uid_sequence (MAILSTREAM *stream,char *sequence);
 long mail_parse_flags (MAILSTREAM *stream,char *flag,unsigned long *uf);
@@ -1108,11 +1221,12 @@ PART *mail_newbody_part (void);
 MESSAGE *mail_newmsg (void);
 STRINGLIST *mail_newstringlist (void);
 SEARCHPGM *mail_newsearchpgm (void);
-SEARCHHEADER *mail_newsearchheader (char *line);
+SEARCHHEADER *mail_newsearchheader (char *line,char *text);
 SEARCHSET *mail_newsearchset (void);
 SEARCHOR *mail_newsearchor (void);
 SEARCHPGMLIST *mail_newsearchpgmlist (void);
 SORTPGM *mail_newsortpgm (void);
+THREADNODE *mail_newthreadnode (SORTCACHE *sc);
 void mail_free_body (BODY **body);
 void mail_free_body_data (BODY *body);
 void mail_free_body_parameter (PARAMETER **parameter);
@@ -1127,7 +1241,9 @@ void mail_free_searchheader (SEARCHHEADER **hdr);
 void mail_free_searchset (SEARCHSET **set);
 void mail_free_searchor (SEARCHOR **orl);
 void mail_free_searchpgmlist (SEARCHPGMLIST **pgl);
+void mail_free_namespace (NAMESPACE **n);
 void mail_free_sortpgm (SORTPGM **pgm);
+void mail_free_threadnode (THREADNODE **thr);
 void auth_link (AUTHENTICATOR *auth);
 char *mail_auth (char *mechanism,authresponse_t resp,int argc,char *argv[]);
 AUTHENTICATOR *mail_lookup_auth (unsigned long i);
