@@ -10,7 +10,7 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	6 June 1994
- * Last Edited:	19 February 1997
+ * Last Edited:	29 December 1997
  *
  * Copyright 1997 by the University of Washington
  *
@@ -38,10 +38,12 @@
 #include "osdep.h"
 #include <ctype.h>
 #include <stdio.h>
+#include <time.h>
 #include "pop3.h"
 #include "rfc822.h"
 #include "misc.h"
 #include "netmsg.h"
+#include "flstring.h"
 
 /* POP3 mail routines */
 
@@ -53,7 +55,7 @@ DRIVER pop3driver = {
 #ifdef INADEQUATE_MEMORY
   DR_LOWMEM |
 #endif
-  DR_MAIL|DR_NOFAST,		/* driver flags */
+  DR_MAIL|DR_NOFAST|DR_CRLF,	/* driver flags */
   (DRIVER *) NIL,		/* next driver */
   pop3_valid,			/* mailbox is valid for us */
   pop3_parameters,		/* manipulate parameters */
@@ -265,25 +267,27 @@ long pop3_status (MAILSTREAM *stream,char *mbx,long flags)
   unsigned long i;
   char tmp[MAILTMPLEN];
   MAILSTREAM *tstream;
-  NETMBX mb;
-  if (!LOCAL->netstream) return NIL;
-  mail_valid_net_parse (mbx,&mb);
-  if (!(tstream = (stream &&
-		   !strcmp (ucase (strcpy (tmp,net_host (LOCAL->netstream))),
-			    ucase (mb.host))) ?
-	stream : mail_open (NIL,mbx,OP_HALFOPEN|OP_SILENT))) return NIL;
-  status.flags = flags;		/* return status values */
-  status.messages = tstream->nmsgs;
-  status.recent = tstream->recent;
-  if (flags & SA_UNSEEN)	/* must search to get unseen messages */
-    for (i = 1,status.unseen = 0; i < tstream->nmsgs; i++)
-      if (!mail_elt (tstream,i)->seen) status.unseen++;
-  status.uidnext = tstream->uid_last + 1;
-  status.uidvalidity = tstream->uid_validity;
+  NETMBX mb,mb2;
+  long ret = NIL;
+  if (LOCAL->netstream && !mail_valid_net_parse (mbx,&mb) &&
+      mail_valid_net_parse (strcpy (tmp,stream->mailbox),&mb2) &&
+      (tstream = (stream && !strcmp (ucase (mb.host),ucase (mb2.host)) &&
+		  !strcmp (ucase (mb.user),ucase (mb2.user))) ?
+       stream : mail_open (NIL,mbx,OP_HALFOPEN|OP_SILENT))) {
+    status.flags = flags;	/* return status values */
+    status.messages = tstream->nmsgs;
+    status.recent = tstream->recent;
+    if (flags & SA_UNSEEN)	/* must search to get unseen messages */
+      for (i = 1,status.unseen = 0; i < tstream->nmsgs; i++)
+	if (!mail_elt (tstream,i)->seen) status.unseen++;
+    status.uidnext = tstream->uid_last + 1;
+    status.uidvalidity = tstream->uid_validity;
 				/* pass status to main program */
-  mm_status (tstream,mbx,&status);
-  if (stream != tstream) mail_close (tstream);
-  return T;			/* success */
+    mm_status (tstream,mbx,&status);
+    if (stream != tstream) mail_close (tstream);
+    ret = LONGT;
+  }
+  return ret;			/* success */
 }
 
 /* POP3 mail open
@@ -301,6 +305,8 @@ MAILSTREAM *pop3_open (MAILSTREAM *stream)
   if (!stream) return &pop3proto;
   mail_valid_net_parse (stream->mailbox,&mb);
   if (LOCAL) {			/* if recycle stream */
+    stream->uid_last = 0;	/* default UID validity */
+    stream->uid_validity = time (0);
     sprintf (tmp,"Closing connection to %s",LOCAL->host);
     if (!stream->silent) mm_log (tmp,(long) NIL);
     pop3_close (stream,NIL);	/* do close action */
@@ -433,7 +439,8 @@ void pop3_fetchfast (MAILSTREAM *stream,char *sequence,long flags)
 	  unsigned long hs;
 	  char *ht = (*stream->dtb->header) (stream,i,&hs,NIL);
 				/* need to make an envelope? */
-	  if (!*env) rfc822_parse_msg (env,NIL,ht,hs,NIL,BADHOST);
+	  if (!*env) rfc822_parse_msg (env,NIL,ht,hs,NIL,mylocalhost(),
+				       stream->dtb->flags);
 				/* need message size too, ugh */
 	  if (!elt->rfc822_size) {
 	    (*stream->dtb->text) (stream,i,&bs,FT_PEEK);
@@ -460,29 +467,20 @@ void pop3_fetchfast (MAILSTREAM *stream,char *sequence,long flags)
 char *pop3_header (MAILSTREAM *stream,unsigned long msgno,unsigned long *size,
 		   long flags)
 {
-  SIZEDTEXT *hdr;
   MESSAGECACHE *elt;
   if ((flags & FT_UID) && !(msgno = mail_msgno (stream,msgno))) return NIL;
-  elt = mail_elt (stream,msgno);
-  if (stream->scache) {		/* short caching? */
-    hdr = &stream->header;	/* point to header */
-    if (stream->msgno != msgno){/* different message, flush cache */
-      fs_give ((void **) &hdr->data);
-      hdr->size = 0;
-    }
-    stream->msgno = msgno;
-  }
-				/* long caching */
-  else hdr = &elt->private.msg.header.text;
-  if (!hdr->data) {		/* have header text? */
-    hdr->size = pop3_cache (stream,elt);
+				/* have header text? */
+  if (!(elt = mail_elt (stream,msgno))->private.msg.header.text.data) {
+    elt->private.msg.header.text.size = pop3_cache (stream,elt);
 				/* read the header */
-    fread (hdr->data = (char *) fs_get ((size_t) hdr->size + 1),
-	   (size_t) 1,(size_t) hdr->size,LOCAL->txt);
-    hdr->data[hdr->size] = '\0';
+    fread (elt->private.msg.header.text.data = (unsigned char *)
+	   fs_get ((size_t) elt->private.msg.header.text.size + 1),
+	   (size_t) 1,(size_t) elt->private.msg.header.text.size,LOCAL->txt);
+    elt->private.msg.header.text.data[elt->private.msg.header.text.size] ='\0';
   }
-  if (size) *size = hdr->size;	/* return size of text */
-  return hdr->data;
+				/* return size of text */
+  if (size) *size = elt->private.msg.header.text.size;
+  return (char *) elt->private.msg.header.text.data;
 }
 
 /* POP3 fetch body
@@ -501,8 +499,11 @@ long pop3_text (MAILSTREAM *stream,unsigned long msgno,STRING *bs,long flags)
   elt = mail_elt (stream,msgno);
   pop3_cache (stream,elt);	/* make sure cache loaded */
   if (!LOCAL->txt) return NIL;	/* error if don't have a file */
-  if (!(flags & FT_PEEK)) elt->seen = T;
-  INIT (bs,netmsg_string,(void *) LOCAL->txt,elt->rfc822_size);
+  if (!(flags & FT_PEEK)) {	/* mark seen if needed */
+    elt->seen = T;
+    mm_flags (stream,elt->msgno);
+  }
+  INIT (bs,file_string,(void *) LOCAL->txt,elt->rfc822_size);
   SETPOS (bs,LOCAL->hdrsize);	/* skip past header */
   return T;
 }
@@ -587,6 +588,9 @@ void pop3_expunge (MAILSTREAM *stream)
 
 long pop3_copy (MAILSTREAM *stream,char *sequence,char *mailbox,long options)
 {
+  mailproxycopy_t pc =
+    (mailproxycopy_t) mail_parameters (stream,GET_MAILPROXYCOPY,NIL);
+  if (pc) return (*pc) (stream,sequence,mailbox,options);
   mm_log ("Copy not valid for POP3",ERROR);
   return NIL;
 }
