@@ -10,7 +10,7 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	27 July 1988
- * Last Edited:	20 February 1997
+ * Last Edited:	29 December 1997
  *
  * Sponsorship:	The original version of this work was developed in the
  *		Symbolic Systems Resources Group of the Knowledge Systems
@@ -113,9 +113,9 @@ void rfc822_header (char *header,ENVELOPE *env,BODY *body)
 {
   if (env->remail) {		/* if remailing */
     long i = strlen (env->remail);
-				/* flush extra blank line */
-    if (i > 4 && env->remail[i-4] == '\015') env->remail[i-2] = '\0';
     strcpy (header,env->remail);/* start with remail header */
+				/* flush extra blank line */
+    if (i > 4 && header[i-4] == '\015') header[i-2] = '\0';
   }
   else *header = '\0';		/* else initialize header to null */
   rfc822_header_line (&header,"Newsgroups",env,env->newsgroups);
@@ -327,10 +327,11 @@ void rfc822_write_body_header (char **dst,BODY *body)
   if (stl = body->language) {
     strcpy (*dst += strlen (*dst),"Content-Language: ");
     do {
-      rfc822_cat (*dst,stl->text.data,tspecials);
+      rfc822_cat (*dst,(char *) stl->text.data,tspecials);
       if (stl = stl->next) strcat (*dst += strlen (*dst),", ");
     }
     while (stl);
+    strcpy (*dst += strlen (*dst),"\015\012");
   }
   if (body->disposition.type) {
     sprintf (*dst += strlen (*dst),"Content-Disposition: %s",
@@ -377,10 +378,11 @@ char *rfc822_default_subtype (unsigned short type)
  *	    header byte count
  *	    pointer to body stringstruct
  *	    pointer to local host name
+ *	    source driver flags
  */
 
 void rfc822_parse_msg (ENVELOPE **en,BODY **bdy,char *s,unsigned long i,
-		       STRING *bs,char *host)
+		       STRING *bs,char *host,unsigned long flags)
 {
   char c,*t,*d;
   char *tmp = (char *) fs_get ((size_t) i + 100);
@@ -426,7 +428,9 @@ void rfc822_parse_msg (ENVELOPE **en,BODY **bdy,char *s,unsigned long i,
 	else if ((tmp[1] == 'O') && (tmp[2] == 'N') && (tmp[3] == 'T') &&
 		 (tmp[4] == 'E') && (tmp[5] == 'N') && (tmp[6] == 'T') &&
 		 (tmp[7] == '-') && body &&
-		 (MIMEp || (search (s-1,i,"\012MIME-Version",(long) 13))))
+		 (MIMEp || (search ((unsigned char *) s-1,i,
+				    (unsigned char *)"\012MIME-Version",
+				    (long) 13))))
 	  rfc822_parse_content_header (body,tmp+8,d);
 	break;
       case 'D':			/* possible Date: */
@@ -513,16 +517,17 @@ void rfc822_parse_msg (ENVELOPE **en,BODY **bdy,char *s,unsigned long i,
   if (!env->sender) env->sender = rfc822_cpy_adr (env->from);
   if (!env->reply_to) env->reply_to = rfc822_cpy_adr (env->from);
 				/* now parse the body */
-  if (body) rfc822_parse_content (body,bs,host);
+  if (body) rfc822_parse_content (body,bs,host,flags);
 }
 
 /* Parse a message body content
  * Accepts: pointer to body structure
  *	    body string
  *	    pointer to local host name
+ *	    source driver flags
  */
 
-void rfc822_parse_content (BODY *body,STRING *bs,char *h)
+void rfc822_parse_content (BODY *body,STRING *bs,char *h,unsigned long flags)
 {
   char c,c1,*s,*s1;
   int f;
@@ -534,10 +539,8 @@ void rfc822_parse_content (BODY *body,STRING *bs,char *h)
 				/* note offset and sizes */
   body->contents.offset = GETPOS (bs);
 				/* note internal body size in all cases */
-  body->contents.text.size = i = SIZE (bs);
-  body->size.bytes = ((body->encoding == ENCBINARY) ||
-		      (body->type == TYPEMULTIPART)) ?
-			body->contents.text.size : strcrlflen (bs);
+  body->size.bytes = body->contents.text.size = i = SIZE (bs);
+  if (!(flags & DR_CRLF)) body->size.bytes = strcrlflen (bs);
   switch (body->type) {		/* see if anything else special to do */
   case TYPETEXT:		/* text content */
     if (!body->parameter) {	/* default parameters */
@@ -578,7 +581,7 @@ void rfc822_parse_content (BODY *body,STRING *bs,char *h)
       s[j] = '\0';		/* tie off string (not really necessary) */
 				/* now parse the body */
       rfc822_parse_msg (&body->nested.msg->env,&body->nested.msg->body,s,j,bs,
-			h);
+			h,flags);
       fs_give ((void **) &s);	/* free header string */
 				/* restore position */
       SETPOS (bs,body->contents.offset);
@@ -595,7 +598,7 @@ void rfc822_parse_content (BODY *body,STRING *bs,char *h)
 				/* find cookie */
     for (s1 = NIL,param = body->parameter; param && !s1; param = param->next)
       if (!strcmp (param->attribute,"BOUNDARY")) s1 = param->value;
-    if (!s1) s1 = "-";		/* yucky default */
+    if (!(s1 && *s1)) s1 = "-";	/* yucky default */
     j = strlen (s1);		/* length of cookie and header */
     c = '\012';			/* initially at beginning of line */
 
@@ -657,7 +660,7 @@ void rfc822_parse_content (BODY *body,STRING *bs,char *h)
 				/* make a scratch buffer */
     s1 = (char *) fs_get ((size_t) (k = MAILTMPLEN));
 				/* parse non-empty body parts */
-    for (part = body->nested.part; part; part = part->next)
+    for (part = body->nested.part; part; part = part->next) {
       if (i = part->body.mime.text.size) {
 				/* move to that part of the body */
 	SETPOS (bs,part->body.mime.offset);
@@ -704,7 +707,8 @@ void rfc822_parse_content (BODY *body,STRING *bs,char *h)
 				/* parse the header */
 	    rfc822_parse_content_header (&part->body,ucase (s1+8),s);
 	  }
-	}			/* skip header trailing (CR)LF */
+	}
+				/* skip header trailing (CR)LF */
 	if (i && (CHR (bs) =='\015')) {i--; c1 = SNX (bs);}
 	if (i && (CHR (bs) =='\012')) {i--; c1 = SNX (bs);}
 	j = bs->size;		/* save upper level size */
@@ -712,9 +716,12 @@ void rfc822_parse_content (BODY *body,STRING *bs,char *h)
 	bs->size = GETPOS (bs) + i;
 	part->body.mime.text.size -= i;
 				/* now parse it */
-	rfc822_parse_content (&part->body,bs,h);
+	rfc822_parse_content (&part->body,bs,h,flags);
 	bs->size = j;		/* restore current level size */
       }
+      else part->body.subtype =	/* default subtype if necessary */
+	cpystr (rfc822_default_subtype (part->body.type));
+    }
     fs_give ((void **) &s1);	/* finished with scratch buffer */
     break;
   default:			/* nothing special to do in any other case */
@@ -761,7 +768,7 @@ void rfc822_parse_content_header (BODY *body,char *name,char *s)
 	*name = '\0';		/* tie off subtype */
 	if (stl) stl = stl->next = mail_newstringlist ();
 	else stl = body->language = mail_newstringlist ();
-	stl->text.data = ucase (cpystr (s));
+	stl->text.data = (unsigned char *) ucase (cpystr (s));
 	stl->text.size = strlen (stl->text.data);
 	*name = c;		/* restore delimiter */
 	rfc822_skipws (&name);	/* skip whitespace */
@@ -1055,8 +1062,10 @@ ADDRESS *rfc822_parse_mailbox (char **string,char *defaulthost)
     if ((phrase = rfc822_parse_phrase (s)) &&
 	(adr = rfc822_parse_routeaddr (phrase,string,defaulthost))) {
       *phrase = '\0';		/* tie off phrase */
-				/* phrase is a personal name */
-      adr->personal = rfc822_cpy (s);
+      if (*s) {			/* phrase is a personal name */
+	if (adr->personal) fs_give ((void **) &adr->personal);
+	adr->personal = rfc822_cpy (s);
+      }
     }
     else adr = rfc822_parse_addrspec (s,string,defaulthost);
   }
@@ -1153,7 +1162,7 @@ ADDRESS *rfc822_parse_addrspec (char *string,char **ret,char *defaulthost)
       }
       else {
 	mm_log ("Invalid domain literal after @",PARSE);
-	adr->host = cpystr (BADHOST);
+	adr->host = cpystr (errhst);
       }
     }
     				/* search for end of host */
@@ -1166,7 +1175,7 @@ ADDRESS *rfc822_parse_addrspec (char *string,char **ret,char *defaulthost)
     }
     else {
       mm_log ("Missing or invalid host name after @",PARSE);
-      adr->host = cpystr (BADHOST);
+      adr->host = cpystr (errhst);
     }
   }
   else end = t;			/* make person name default start after mbx */
@@ -1365,39 +1374,6 @@ char *rfc822_skip_comment (char **s,long trim)
 /* Body contents utility and encoding/decoding routines */
 
 
-/* Return body contents in normal form
- * Accepts: pointer to destination
- *	    pointer to length of destination
- *	    returned destination length
- *	    source
- *	    length of source
- *	    source encoding
- * Returns: destination
- *
- * Originally, this routine was supposed to do decoding as well, but that was
- * moved to a higher level.  Now, it's merely a jacket into strcrlfcpy that
- * avoids the work for BINARY segments.
- */
-
-char *rfc822_contents (char **dst,unsigned long *dstl,unsigned long *len,
-		       char *src,unsigned long srcl,unsigned short encoding)
-{
-  *len = 0;			/* in case we return an error */
-  if (encoding == ENCBINARY) {	/* unmodified binary */
-    if ((*len = srcl) > *dstl) {/* resize if not enough space */
-      fs_give ((void **) dst);	/* fs_resize does an unnecessary copy */
-      *dst = (char *) fs_get ((size_t) (*dstl = srcl) + 1);
-    }
-				/* copy that many bytes */
-    memcpy (*dst,src,(size_t) srcl);
-    *(*dst + srcl) = '\0';	/* tie off destination */
-  }
-				/* all other cases return strcrlfcpy version */
-  else *len = strcrlfcpy (dst,dstl,src,srcl);
-  return *dst;			/* return the string */
-}
-
-
 /* Output RFC 822 message
  * Accepts: temporary buffer
  *	    envelope
@@ -1431,15 +1407,19 @@ void rfc822_encode_body_7bit (ENVELOPE *env,BODY *body)
 {
   void *f;
   PART *part;
+  PARAMETER **param;
   if (body) switch (body->type) {
   case TYPEMULTIPART:		/* multi-part */
-    if (!body->parameter) {	/* cookie not set up yet? */
+    for (param = &body->parameter;
+	 *param && strcmp ((*param)->attribute,"BOUNDARY");
+	 param = &(*param)->next);
+    if (!*param) {		/* cookie not set up yet? */
       char tmp[MAILTMPLEN];	/* make cookie not in BASE64 or QUOTEPRINT*/
       sprintf (tmp,"%ld-%ld-%ld=:%ld",gethostid (),random (),time (0),
 	       getpid ());
-      body->parameter = mail_newbody_parameter ();
-      body->parameter->attribute = cpystr ("BOUNDARY");
-      body->parameter->value = cpystr (tmp);
+      (*param) = mail_newbody_parameter ();
+      (*param)->attribute = cpystr ("BOUNDARY");
+      (*param)->value = cpystr (tmp);
     }
     part = body->nested.part;	/* encode body parts */
     do rfc822_encode_body_7bit (env,&part->body);
@@ -1464,8 +1444,8 @@ void rfc822_encode_body_7bit (ENVELOPE *env,BODY *body)
     case ENC8BIT:		/* encode 8BIT into QUOTED-PRINTABLE */
 				/* remember old 8-bit contents */
       f = (void *) body->contents.text.data;
-      body->contents.text.data = (char *)
-	rfc822_8bit ((unsigned char *) body->contents.text.data,
+      body->contents.text.data =
+	rfc822_8bit (body->contents.text.data,
 		     body->contents.text.size,&body->contents.text.size);
       body->encoding = ENCQUOTEDPRINTABLE;
       fs_give (&f);		/* flush old binary contents */
@@ -1473,7 +1453,7 @@ void rfc822_encode_body_7bit (ENVELOPE *env,BODY *body)
     case ENCBINARY:		/* encode binary into BASE64 */
 				/* remember old binary contents */
       f = (void *) body->contents.text.data;
-      body->contents.text.data = (char *)
+      body->contents.text.data =
 	rfc822_binary ((void *) body->contents.text.data,
 		       body->contents.text.size,&body->contents.text.size);
       body->encoding = ENCBASE64;
@@ -1494,15 +1474,19 @@ void rfc822_encode_body_8bit (ENVELOPE *env,BODY *body)
 {
   void *f;
   PART *part;
+  PARAMETER **param;
   if (body) switch (body->type) {
   case TYPEMULTIPART:		/* multi-part */
-    if (!body->parameter) {	/* cookie not set up yet? */
+    for (param = &body->parameter;
+	 *param && strcmp ((*param)->attribute,"BOUNDARY");
+	 param = &(*param)->next);
+    if (!*param) {		/* cookie not set up yet? */
       char tmp[MAILTMPLEN];	/* make cookie not in BASE64 or QUOTEPRINT*/
       sprintf (tmp,"%ld-%ld-%ld=:%ld",gethostid (),random (),time (0),
 	       getpid ());
-      body->parameter = mail_newbody_parameter ();
-      body->parameter->attribute = cpystr ("BOUNDARY");
-      body->parameter->value = cpystr (tmp);
+      (*param) = mail_newbody_parameter ();
+      (*param)->attribute = cpystr ("BOUNDARY");
+      (*param)->value = cpystr (tmp);
     }
     part = body->nested.part;	/* encode body parts */
     do rfc822_encode_body_8bit (env,&part->body);
@@ -1524,7 +1508,7 @@ void rfc822_encode_body_8bit (ENVELOPE *env,BODY *body)
     if (body->encoding == ENCBINARY) {
 				/* remember old binary contents */
       f = (void *) body->contents.text.data;
-      body->contents.text.data = (char *)
+      body->contents.text.data =
 	rfc822_binary ((void *) body->contents.text.data,
 		       body->contents.text.size,&body->contents.text.size);
       body->encoding = ENCBASE64;
@@ -1588,7 +1572,7 @@ void *rfc822_base64 (unsigned char *src,unsigned long srcl,unsigned long *len)
   char c;
   void *ret = fs_get ((size_t) (*len = 4 + ((srcl * 3) / 4)));
   char *d = (char *) ret;
-  short e = 0;
+  int e = 0;
   memset (ret,0,(size_t) *len);	/* initialize block */
   *len = 0;			/* in case we return an error */
   while (srcl--) {		/* until run out of characters */
@@ -1744,8 +1728,7 @@ unsigned char *rfc822_8bit (unsigned char *src,unsigned long srcl,
 {
   unsigned long lp = 0;
   unsigned char *ret = (unsigned char *)
-    fs_get ((((size_t) srcl) * (size_t) 3) + (((size_t) srcl) / MAXL) +
-	    (size_t) 2);
+    fs_get ((size_t) (3*srcl + (6*srcl)/MAXL + 3));
   unsigned char *d = ret;
   char *hex = "0123456789ABCDEF";
   unsigned char c;
