@@ -10,7 +10,7 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	1 November 1990
- * Last Edited:	4 March 1997
+ * Last Edited:	23 December 1997
  *
  * Copyright 1997 by the University of Washington
  *
@@ -47,6 +47,8 @@ extern int errno;		/* just in case */
 
 
 /* Autologout timer */
+#define KODTIMEOUT 60*5
+#define LOGINTIMEOUT 60*3
 #define TIMEOUT 60*30
 
 
@@ -63,6 +65,7 @@ extern int errno;		/* just in case */
 #define AUTHORIZATION 0
 #define TRANSACTION 1
 #define UPDATE 2
+#define LOGOUT 3
 
 /* Eudora food */
 
@@ -72,10 +75,12 @@ extern int errno;		/* just in case */
 
 /* Global storage */
 
-char *version = "4.37";		/* server version */
-int state = AUTHORIZATION;	/* server state */
+char *version = "4.44";		/* server version */
+short state = AUTHORIZATION;	/* server state */
+short critical = NIL;		/* non-zero if in critical code */
 MAILSTREAM *stream = NIL;	/* mailbox stream */
 int logtry = LOGTRY;		/* login tries */
+long idletime = 0;		/* time we went idle */
 unsigned long nmsgs = 0;	/* current number of messages */
 unsigned long last = 0;		/* highest message accessed */
 unsigned long il = 0;		/* initial last message */
@@ -94,6 +99,7 @@ void hupint ();
 void trmint ();
 int login (char *t,int argc,char *argv[]);
 long blat (char *text,long lines,unsigned long size);
+void rset ();
 
 /* Main program */
 
@@ -111,18 +117,27 @@ void main (int argc,char *argv[])
   fflush (stdout);		/* dump output buffer */
 				/* set up traps */
   server_traps (clkint,kodint,hupint,trmint);
-  while (state != UPDATE) {	/* command processing loop */
-    alarm (TIMEOUT);		/* get a command under timeout */
+				/* command processing loop */
+  while ((state != UPDATE) && (state != LOGOUT)) {
+    idletime = time (0);	/* get a command under timeout */
+    alarm ((state == TRANSACTION) ? TIMEOUT : LOGINTIMEOUT);
     while (!fgets (tmp,TMPLEN-1,stdin)) {
       if (errno==EINTR) errno=0;/* ignore if some interrupt */
       else {
+	char *e = errno ? strerror (errno) : "Command stream end of file";
+	alarm (0);		/* disable all interrupts */
 	syslog (LOG_INFO,
-		"Connection broken while reading line user=%.80s host=%.80s",
-		user ? user : "???",tcp_clienthost (tmp));
+		"%s while reading line user=%.80s host=%.80s",
+		e,user ? user : "???",tcp_clienthost ());
+	rset ();		/* try to gracefully close the stream */
+	if (state == TRANSACTION) mail_close (stream);
+	stream = NIL;
+	state = LOGOUT;
 	_exit (1);
       }
     }
     alarm (0);			/* make sure timeout disabled */
+    idletime = 0;		/* no longer idle */
 				/* find end of line */
     if (!strchr (tmp,'\012')) puts ("-ERR Command line too long\015");
     else if (!(s = strtok (tmp," \015\012"))) puts ("-ERR Null command\015");
@@ -138,6 +153,8 @@ void main (int argc,char *argv[])
 	  if (host) fs_give ((void **) &host);
 	  if (user) fs_give ((void **) &user);
 	  if (t && *t) {	/* if user name given */
+				/* skip leading whitespace (bogus clients!) */
+	    while (*t == ' ') ++t;
 				/* remote user name? */
 	    if (s = strchr (t,':')) {
 	      *s++ = '\0';	/* tie off host name */
@@ -242,21 +259,7 @@ void main (int argc,char *argv[])
 	else if (!strcmp (s,"NOOP")) puts ("+OK No-op to you too!\015");
 	else if (!strcmp (s,"LAST")) printf ("+OK %lu\015\012",last);
 	else if (!strcmp (s,"RSET")) {
-	  if (nmsgs) {		/* undelete and unmark all of our messages */
-	    for (i = 1; i <= nmsgs; i++) {
-				/* ugly and inefficient, but trustworthy */
-	      if (msg[i] < 0) {
-		sprintf (tmp,"%ld",msg[i] = -msg[i]);
-		mail_clearflag (stream,tmp,i <= il ? "\\Deleted" :
-			      "\\Deleted \\Seen");
-	      }
-	      else if (i > il) {
-		sprintf (tmp,"%ld",msg[i]);
-		mail_clearflag (stream,tmp,"\\Seen");
-	      }
-	    }
-	    last = il;
-	  }
+	  rset ();		/* reset the mailbox */
 	  puts ("+OK Reset state\015");
 	}
 	else if (!strcmp (s,"TOP")) {
@@ -289,7 +292,6 @@ void main (int argc,char *argv[])
 	  else puts ("-ERR Bad message number argument\015");
 	}
 	else if (!strcmp (s,"XTND")) puts ("-ERR Sorry I can't do that\015");
-
 	else puts ("-ERR Unknown command in TRANSACTION state\015");
 	break;
       default:
@@ -299,14 +301,13 @@ void main (int argc,char *argv[])
     }
     fflush (stdout);		/* make sure output finished */
   }
-				/* expunge mailbox if a stream open */
-  if (stream && nmsgs) mail_expunge (stream);
-  mail_close (stream);		/* clean up the stream */
+				/* only if update requested */
+  if (stream && (state == UPDATE)) stream = mail_close_full(stream,CL_EXPUNGE);
 				/* "now it's time to say sayonara..." */
   if (logtry) puts ("+OK Sayonara\015");
   fflush (stdout);		/* make sure output finished */
-  syslog (LOG_INFO,"Logout user %.80s host %.80s",user ? user : "???",
-	  tcp_clienthost (tmp));
+  syslog (LOG_INFO,"Logout user=%.80s host=%.80s",user ? user : "???",
+	  tcp_clienthost ());
   exit (0);			/* all done */
 }
 
@@ -315,12 +316,20 @@ void main (int argc,char *argv[])
 
 void clkint ()
 {
-  char tmp[MAILTMPLEN];
   puts ("-ERR Autologout; idle for too long\015");
   syslog (LOG_INFO,"Autologout user=%.80s host=%.80s",user ? user : "???",
-	  tcp_clienthost (tmp));
+	  tcp_clienthost ());
   fflush (stdout);		/* make sure output blatted */
-  _exit (0);			/* die die die */
+  if (critical) state = LOGOUT;	/* badly hosed if in critical code */
+  else {			/* try to gracefully close the stream */
+    if ((state == TRANSACTION) && !stream->lock) {
+      rset ();
+      mail_close (stream);
+    }
+    state = LOGOUT;
+    stream = NIL;
+    _exit (1);			/* die die die */
+  }
 }
 
 
@@ -329,12 +338,24 @@ void clkint ()
 
 void kodint ()
 {
-  char tmp[MAILTMPLEN];
-  puts ("-ERR Received Kiss of Death\015");
-  syslog (LOG_INFO,"Kiss of death user=%.80s host=%.80s",user ? user : "???",
-	  tcp_clienthost (tmp));
-  fflush (stdout);		/* make sure output blatted */
-  _exit (0);			/* die die die */
+				/* only if idle */
+  if (idletime && ((time (0) - idletime) > KODTIMEOUT)) {
+    alarm (0);			/* disable all interrupts */
+    server_traps (SIG_IGN,SIG_IGN,SIG_IGN,SIG_IGN);
+    puts ("-ERR Received Kiss of Death\015");
+    syslog (LOG_INFO,"Killed (lost mailbox lock) user=%.80s host=%.80s",
+	    user ? user : "???",tcp_clienthost ());
+    if (critical) state =LOGOUT;/* must defer if in critical code */
+    else {			/* try to gracefully close the stream */
+      if ((state == TRANSACTION) && !stream->lock) {
+	rset ();
+	mail_close (stream);
+      }
+      state = LOGOUT;
+      stream = NIL;
+      _exit (1);		/* die die die */
+    }
+  }
 }
 
 
@@ -343,10 +364,20 @@ void kodint ()
 
 void hupint ()
 {
-  char tmp[MAILTMPLEN];
+  alarm (0);			/* disable all interrupts */
+  server_traps (SIG_IGN,SIG_IGN,SIG_IGN,SIG_IGN);
   syslog (LOG_INFO,"Hangup user=%.80s host=%.80s",user ? user : "???",
-	  tcp_clienthost (tmp));
-  _exit (0);			/* die die die */
+	  tcp_clienthost ());
+  if (critical) state = LOGOUT;	/* must defer if in critical code */
+  else {			/* try to gracefully close the stream */
+    if ((state == TRANSACTION) && !stream->lock) {
+      rset ();
+      mail_close (stream);
+    }
+    state = LOGOUT;
+    stream = NIL;
+    _exit (1);			/* die die die */
+  }
 }
 
 
@@ -355,12 +386,21 @@ void hupint ()
 
 void trmint ()
 {
-  char tmp[MAILTMPLEN];
+  alarm (0);			/* disable all interrupts */
+  server_traps (SIG_IGN,SIG_IGN,SIG_IGN,SIG_IGN);
   puts ("-ERR Killed\015");
   syslog (LOG_INFO,"Killed user=%.80s host=%.80s",user ? user : "???",
-	  tcp_clienthost (tmp));
-  fflush (stdout);		/* make sure output blatted */
-  _exit (1);			/* die die die */
+	  tcp_clienthost ());
+  if (critical) state = LOGOUT;	/* must defer if in critical code */
+  else {			/* try to gracefully close the stream */
+    if ((state == TRANSACTION) && !stream->lock) {
+      rset ();
+      mail_close (stream);
+    }
+    state = LOGOUT;
+    stream = NIL;
+    _exit (1);			/* die die die */
+  }
 }
 
 /* Parse PASS command
@@ -380,28 +420,33 @@ int login (char *t,int argc,char *argv[])
     return AUTHORIZATION;
   }
   pass = cpystr (t);		/* copy password argument */
-  if (host) {			/* remote; build remote INBOX */
+				/* remote; build remote INBOX */
+  if (host && anonymous_login (argc,argv)) {
     syslog (LOG_INFO,"IMAP login to host=%.80s user=%.80s host=%.80s",host,
-	    user,tcp_clienthost (tmp));
+	    user,tcp_clienthost ());
     sprintf (tmp,"{%.128s}INBOX",host);
-    anonymous_login ();		/* try to become someone harmless */
   }
 				/* local; attempt login, select INBOX */
-  else if (server_login (user,pass,argc,argv)) {
-    syslog (LOG_INFO,"Login user=%.80s host=%.80s",user,tcp_clienthost (tmp));
+  else if (!host && server_login (user,pass,argc,argv)) {
+    syslog (LOG_INFO,"Login user=%.80s host=%.80s",user,tcp_clienthost ());
     strcpy (tmp,"INBOX");
   }
   else {
+    if (!host && (*t == ' ')) {	/* client bogus? */
+      while (*t == ' ') ++t;	/* yes, skip leading whitespace */
+				/* give it another try */
+      if (*t) return login (t,argc,argv);
+    }
     sleep (3);			/* slow the cracker down */
     if (--logtry) {		/* vague error message to confuse crackers */
       puts ("-ERR Bad login\015");
       syslog (LOG_INFO,"Login failure user=%.80s host=%.80s",user,
-	      tcp_clienthost(tmp));
+	      tcp_clienthost ());
       return AUTHORIZATION;
     }
     fputs ("-ERR Too many login failures\015\012",stdout);
     syslog (LOG_INFO,"Excessive login failures user=%.80s host=%.80s",user,
-	    tcp_clienthost (tmp));
+	    tcp_clienthost ());
     return UPDATE;
   }
 
@@ -464,6 +509,29 @@ long blat (char *text,long lines,unsigned long size)
   }
   return ret;
 }
+
+
+/* Reset mailbox
+ */
+
+void rset ()
+{
+  unsigned long i;
+  char tmp[20];
+  if (nmsgs) {			/* undelete and unmark all of our messages */
+    for (i = 1; i <= nmsgs; i++) { /*  */
+      if (msg[i] < 0) {		/* ugly and inefficient, but trustworthy */
+	sprintf (tmp,"%ld",msg[i] = -msg[i]);
+	mail_clearflag (stream,tmp,i <= il ? "\\Deleted" : "\\Deleted \\Seen");
+      }
+      else if (i > il) {
+	sprintf (tmp,"%ld",msg[i]);
+	mail_clearflag (stream,tmp,"\\Seen");
+      }
+    }
+    last = il;
+  }
+}
 
 /* Co-routines from MAIL library */
 
@@ -603,7 +671,7 @@ void mm_login (NETMBX *mb,char *username,char *password,long trial)
 
 void mm_critical (MAILSTREAM *stream)
 {
-  /* Not doing anything here for now */
+  critical = T;
 }
 
 
@@ -613,7 +681,7 @@ void mm_critical (MAILSTREAM *stream)
 
 void mm_nocritical (MAILSTREAM *stream)
 {
-  /* Not doing anything here for now */
+  critical = NIL;
 }
 
 
@@ -626,14 +694,21 @@ void mm_nocritical (MAILSTREAM *stream)
 
 long mm_diskerror (MAILSTREAM *stream,long errcode,long serious)
 {
-  char tmp[MAILTMPLEN];
-  syslog (LOG_ALERT,
-	  "Retrying after disk error user=%.80s host=%.80s mbx=%.80s: %.80s",
-	  user,tcp_clienthost (tmp),
+  if (serious) {		/* try your damnest if clobberage likely */
+    syslog (LOG_ALERT,
+	    "Retrying after disk error user=%.80s host=%.80s mbx=%.80s: %.80s",
+	    user,tcp_clienthost (),
+	    (stream && stream->mailbox) ? stream->mailbox : "???",
+	    strerror (errcode));
+    alarm (0);			/* make damn sure timeout disabled */
+    sleep (60);			/* give it some time to clear up */
+    return NIL;
+  }
+  syslog (LOG_ALERT,"Fatal disk error user=%.80s host=%.80s mbx=%.80s: %.80s",
+	  user,tcp_clienthost (),
 	  (stream && stream->mailbox) ? stream->mailbox : "???",
 	  strerror (errcode));
-  sleep (5);			/* can't do much better than this! */
-  return NIL;
+  return T;
 }
 
 
