@@ -10,7 +10,7 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	24 June 1992
- * Last Edited:	24 January 1997
+ * Last Edited:	15 December 1997
  *
  * Copyright 1997 by the University of Washington
  *
@@ -53,11 +53,11 @@
 #include <time.h>
 #include <sys\stat.h>
 #include <dos.h>
-#include <io.h>
 #include "bezrkdos.h"
 #include "rfc822.h"
 #include "dummy.h"
 #include "misc.h"
+#include "fdstring.h"
 
 /* Berkeley mail routines */
 
@@ -66,7 +66,8 @@
 
 DRIVER bezerkdriver = {
   "bezerk",			/* driver name */
-  DR_LOCAL|DR_MAIL|DR_LOWMEM,	/* driver flags */
+				/* driver flags */
+  DR_LOCAL|DR_MAIL|DR_LOWMEM|DR_CRLF,
   (DRIVER *) NIL,		/* next driver */
   bezerk_valid,			/* mailbox is valid for us */
   bezerk_parameters,		/* manipulate parameters */
@@ -346,74 +347,6 @@ void bezerk_close (MAILSTREAM *stream,long options)
   }
 }
 
-/* Berkeley string driver for file stringstructs */
-
-STRINGDRIVER bezerk_string = {
-  bezerk_string_init,		/* initialize string structure */
-  bezerk_string_next,		/* get next byte in string structure */
-  bezerk_string_setpos		/* set position in string structure */
-};
-
-
-/* Cache buffer for file stringstructs */
-
-#define DOSCHUNKLEN 4096
-char dos_chunk[DOSCHUNKLEN];
-
-
-/* Initialize BEZERK string structure for file stringstruct
- * Accepts: string structure
- *	    pointer to string
- *	    size of string
- */
-
-void bezerk_string_init (STRING *s,void *data,unsigned long size)
-{
-  BEZERKDATA *d = (BEZERKDATA *) data;
-  s->data = (void *) d->fd;	/* note fd */
-  s->data1 = d->pos;		/* note file offset */
-  s->size = size;		/* note size */
-  s->curpos = s->chunk = dos_chunk;
-  s->chunksize = (unsigned long) DOSCHUNKLEN;
-  s->offset = 0;		/* initial position */
-				/* and size of data */
-  s->cursize = min (s->chunksize,size);
-				/* move to that position in the file */
-  lseek (d->fd,d->pos,SEEK_SET);
-  read (d->fd,s->chunk,(size_t) s->cursize);
-}
-
-/* Get next character from file stringstruct
- * Accepts: string structure
- * Returns: character, string structure chunk refreshed
- */
-
-char bezerk_string_next (STRING *s)
-{
-  char c = *s->curpos++;	/* get next byte */
-				/* move to next chunk */
-  SETPOS (s,s->offset + s->chunksize);
-  return c;			/* return the byte */
-}
-
-
-/* Set string pointer position for file stringstruct
- * Accepts: string structure
- *	    new position
- */
-
-void bezerk_string_setpos (STRING *s,unsigned long i)
-{
-  s->offset = i;		/* set new offset */
-  s->curpos = s->chunk;		/* reset position */
-				/* set size of data */
-  if (s->cursize = s->size > s->offset ? min (s->chunksize,SIZE (s)) : 0) {
-				/* move to that position in the file */
-    lseek ((int) s->data,s->data1 + s->offset,SEEK_SET);
-    read ((int) s->data,s->curpos,(size_t) s->cursize);
-  }
-}
-
 /* Berkeley mail fetch message header
  * Accepts: MAIL stream
  *	    message # to fetch
@@ -451,18 +384,24 @@ char *bezerk_header (MAILSTREAM *stream,unsigned long msgno,
 long bezerk_text (MAILSTREAM *stream,unsigned long msgno,STRING *bs,long flags)
 {
   MESSAGECACHE *elt;
-  BEZERKDATA d;
+  FDDATA d;
   unsigned long hdrsize,hdrpos;
 				/* UID call "impossible" */
   if (flags & FT_UID) return NIL;
   elt = mail_elt (stream,msgno);/* if message not seen */
 				/* mark message as seen */
-  if (elt->seen && !(flags & FT_PEEK)) elt->seen = T;
+  if (elt->seen && !(flags & FT_PEEK)) {
+    elt->seen = T;
+    mm_flags (stream,msgno);
+  }
 				/* get location of text data */
   hdrpos = bezerk_hdrpos (stream,msgno,&hdrsize);
   d.fd = LOCAL->fd;		/* set initial stringstruct */
   d.pos = hdrpos + hdrsize;
-  INIT (bs,bezerk_string,(void *) &d,elt->rfc822_size - hdrsize);
+				/* flush old buffer */
+  if (LOCAL->buf) fs_give ((void **) &LOCAL->buf);
+  d.chunk = LOCAL->buf = (char *) fs_get ((size_t) d.chunksize = CHUNK);
+  INIT (bs,fd_string,(void *) &d,elt->rfc822_size - hdrsize);
   return T;			/* success */
 }
 
@@ -518,6 +457,8 @@ long bezerk_copy (MAILSTREAM *stream,char *sequence,char *mailbox,long options)
   MESSAGECACHE *elt;
   unsigned long i,j,k;
   int fd;
+  mailproxycopy_t pc =
+    (mailproxycopy_t) mail_parameters (stream,GET_MAILPROXYCOPY,NIL);
   if (!((options & CP_UID) ? mail_uid_sequence (stream,sequence) :
 	mail_sequence (stream,sequence))) return NIL;
 				/* make sure valid mailbox */
@@ -525,6 +466,7 @@ long bezerk_copy (MAILSTREAM *stream,char *sequence,char *mailbox,long options)
     if (errno == ENOENT)
       mm_notify (stream,"[TRYCREATE] Must create mailbox before append",
 		 (long) NIL);
+    else if (pc) return (*pc) (stream,sequence,mailbox,options);
     else if (mailboxfile (tmp,mailbox)) {
       sprintf (tmp,"Not a Bezerk-format mailbox: %s",mailbox);
       mm_log (tmp,ERROR);
@@ -592,18 +534,27 @@ long bezerk_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
   if (date) {			/* want to preserve date? */
 				/* yes, parse date into an elt */
     if (!mail_parse_date (&elt,date)) {
-      sprintf (tmp,"Bad date in append: %s",date);
+      sprintf (tmp,"Bad date in append: %.80ss",date);
       mm_log (tmp,ERROR);
       return NIL;
     }
   }
 				/* make sure valid mailbox */
   if (!bezerk_isvalid (mailbox,tmp) && errno) {
-    if (errno == ENOENT)
-      mm_notify (stream,"[TRYCREATE] Must create mailbox before append",
-		 (long) NIL);
+    if (errno == ENOENT) {
+      if (((mailbox[0] == 'I') || (mailbox[0] == 'i')) &&
+	  ((mailbox[1] == 'N') || (mailbox[1] == 'n')) &&
+	  ((mailbox[2] == 'B') || (mailbox[2] == 'b')) &&
+	  ((mailbox[3] == 'O') || (mailbox[3] == 'o')) &&
+	  ((mailbox[4] == 'X') || (mailbox[4] == 'x')) && !mailbox[5])
+	bezerk_create (NIL,"INBOX");
+      else {
+	mm_notify (stream,"[TRYCREATE] Must create mailbox before append",NIL);
+	return NIL;
+      }
+    }
     else if (mailboxfile (tmp,mailbox)) {
-      sprintf (tmp,"Not a Bezerk-format mailbox: %s",mailbox);
+      sprintf (tmp,"Not a Bezerk-format mailbox: %.80ss",mailbox);
       mm_log (tmp,ERROR);
     }
     else bezerk_badname (tmp,mailbox);
