@@ -10,9 +10,9 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	24 May 1993
- * Last Edited:	5 December 1996
+ * Last Edited:	31 December 1997
  *
- * Copyright 1996 by the University of Washington
+ * Copyright 1997 by the University of Washington
  *
  *  Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose and without fee is hereby granted, provided
@@ -42,7 +42,6 @@
 #include "osdep.h"
 #include <sys\stat.h>
 #include <dos.h>
-#include <io.h>
 #include "dummy.h"
 #include "misc.h"
 
@@ -138,7 +137,7 @@ void *dummy_parameters (long function,void *value)
   return value;
 }
 
-/* Dummy list mailboxes
+/* Dummy scan mailboxes
  * Accepts: mail stream
  *	    reference
  *	    pattern to search
@@ -151,8 +150,16 @@ void dummy_scan (MAILSTREAM *stream,char *ref,char *pat,char *contents)
 {
   char *s,test[LISTTMPLEN],file[LISTTMPLEN];
   long i = 0;
+  if (!pat || !*pat) {		/* empty pattern? */
+    if (dummy_canonicalize (test,ref,"*")) {
+				/* tie off name at root */
+      if (s = strchr (test,'\\')) *++s = '\0';
+      else test[0] = '\0';
+      dummy_listed (stream,'\\',test,LATT_NOINFERIORS,NIL);
+    }
+  }
 				/* get canonical form of name */
-  if (dummy_canonicalize (test,ref,pat)) {
+  else if (dummy_canonicalize (test,ref,pat)) {
 				/* found any wildcards? */
     if (s = strpbrk (test,"%*")) {
 				/* yes, copy name up to that point */
@@ -169,6 +176,8 @@ void dummy_scan (MAILSTREAM *stream,char *ref,char *pat,char *contents)
     else if (file[0] == '#') s = file;
 				/* do the work */
     dummy_list_work (stream,s,test,contents,0);
+    if (pmatch ("INBOX",test)	/* always an INBOX */
+      dummy_listed (stream,NIL,"INBOX",LATT_NOINFERIORS,contents);
   }
 }
 
@@ -194,10 +203,12 @@ void dummy_lsub (MAILSTREAM *stream,char *ref,char *pat)
   void *sdb = NIL;
   char *s,test[MAILTMPLEN],tmp[MAILTMPLEN];
 				/* get canonical form of name */
-  if (dummy_canonicalize (test,ref,pat) && ucase(test) && (s = sm_read (&sdb)))
-    do if (pmatch_full (ucase (strcpy (tmp,s)),test,'\\'))
-      mm_lsub (stream,'\\',s,NIL);
-  while (s = sm_read (&sdb));	/* until no more subscriptions */
+  if (dummy_canonicalize (test,ref,pat) && (s = sm_read (&sdb)))
+    do if (pmatch_full (ucase (strcpy (tmp,s)),test,'\\')) {
+      if (pmatch (tmp,"INBOX")) mm_lsub (stream,NIL,s,LATT_NOINFERIORS);
+      else mm_lsub (stream,'\\',s,NIL);
+				/* until no more subscriptions */
+    } while (s = sm_read (&sdb));
 }
 
 /* Dummy list mailboxes worker routine
@@ -213,7 +224,7 @@ void dummy_list_work (MAILSTREAM *stream,char *dir,char *pat,char *contents,
 {
   struct find_t f;
   struct stat sbuf;
-  char *s,tmp[LISTTMPLEN],tmpx[LISTTMPLEN];
+  char *s,tmp[LISTTMPLEN],tmpx[LISTTMPLEN],tmpy[LSTTMPLEN];
   char *base = (dir && (dir[0] == '\\')) ? NIL : myhomedir ();
 				/* build name */
   if (base) sprintf (tmpx,"%s\\",base);
@@ -227,7 +238,7 @@ void dummy_list_work (MAILSTREAM *stream,char *dir,char *pat,char *contents,
 				/* do nothing if can't open directory */
   if (!_dos_findfirst (tmp,_A_NORMAL|_A_SUBDIR,&f)) {
 				/* list it if not at top-level */
-    if (!level && dir && pmatch_full (dir,pat,'\\'))
+    if (!level && dir && pmatch_full (ucase (strcpy (tmpy,dir)),pat,'\\'))
       dummy_listed (stream,'\\',dir,LATT_NOSELECT,contents);
 				/* scan directory */
     if (tmpx[strlen (tmpx) - 1] == '\\') do if (*f.name != '.') {
@@ -244,7 +255,7 @@ void dummy_list_work (MAILSTREAM *stream,char *dir,char *pat,char *contents,
 				/* only interested in file type */
 	switch (sbuf.st_mode &= S_IFMT) {
 	case S_IFDIR:		/* directory? */
-	  if (pmatch_full (tmp,pat,'\\')) {
+	  if (pmatch_full (ucase (tmp),pat,'\\')) {
 	    dummy_listed (stream,'\\',tmp,LATT_NOSELECT,contents);
 	    strcat (tmp,"\\");	/* set up for dmatch call */
 	  }
@@ -256,7 +267,7 @@ void dummy_list_work (MAILSTREAM *stream,char *dir,char *pat,char *contents,
 	    dummy_list_work (stream,tmp,pat,contents,level+1);
 	  break;
 	case S_IFREG:		/* ordinary name */
-	  if (pmatch_full (tmp,pat,'\\'))
+	  if (pmatch_full (ucase (tmp),pat,'\\') && !pmatch ("INBOX",tmp))
 	    dummy_listed (stream,'\\',tmp,LATT_NOINFERIORS,contents);
 	  break;
 	}
@@ -293,7 +304,8 @@ long dummy_listed (MAILSTREAM *stream,char delimiter,char *name,
     memset (buf,'\0',ssiz);	/* no slop area the first time */
     while (sbuf.st_size) {	/* until end of file */
       read (fd,buf+ssiz,bsiz = min (sbuf.st_size,BUFSIZE));
-      if (search (buf,bsiz+ssiz,contents,csiz)) break;
+      if (search ((unsigned char *) buf,bsiz+ssiz,
+		  (unsigned char *) contents,csiz)) break;
       memcpy (buf,buf+BUFSIZE,ssiz);
       sbuf.st_size -= bsiz;	/* note that we read that much */
     }
@@ -426,11 +438,12 @@ MAILSTREAM *dummy_open (MAILSTREAM *stream)
 				/* say there are 0 messages */
 	mail_exists (stream,(long) 0);
 	mail_recent (stream,(long) 0);
+	stream->uid_validity = 1;
       }
       return stream;		/* return success */
     }
   }
-  if (!stream->silent) mm_log (tmp,ERROR);
+  mm_log (tmp,stream->silent ? WARN: ERROR);
   return NIL;			/* always fails */
 }
 
@@ -531,7 +544,7 @@ long dummy_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
       return NIL;
     }
   }
-  return (*default_proto ()->dtb->append) (stream,mailbox,flags,date,message);
+  return (*default_proto (T)->dtb->append) (stream,mailbox,flags,date,message);
 }
 
 /* Return bad file name error message
@@ -590,5 +603,6 @@ long dummy_canonicalize (char *tmp,char *ref,char *pat)
   if (dev[0] && (*(ref ? ref : pat) != '\\')) dev[2] = '\\';
 				/* build name */
   sprintf (tmp,"%s%s%s",dev,ref ? ref : "",pat);
+  ucase (tmp);			/* force upper case */
   return T;
 }
