@@ -10,9 +10,9 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	24 June 1992
- * Last Edited:	2 December 1996
+ * Last Edited:	15 December 1997
  *
- * Copyright 1996 by the University of Washington
+ * Copyright 1997 by the University of Washington
  *
  *  Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose and without fee is hereby granted, provided
@@ -43,11 +43,11 @@
 #include <time.h>
 #include <sys\stat.h>
 #include <dos.h>
-#include <io.h>
 #include "mtxdos.h"
 #include "rfc822.h"
 #include "dummy.h"
 #include "misc.h"
+#include "fdstring.h"
 
 /* MTX mail routines */
 
@@ -56,7 +56,8 @@
 
 DRIVER mtxdriver = {
   "mtx",			/* driver name */
-  DR_LOCAL|DR_MAIL|DR_LOWMEM,	/* driver flags */
+				/* driver flags */
+  DR_LOCAL|DR_MAIL|DR_LOWMEM|DR_CRLF,
   (DRIVER *) NIL,		/* next driver */
   mtx_valid,			/* mailbox is valid for us */
   mtx_parameters,		/* manipulate parameters */
@@ -241,6 +242,11 @@ MAILSTREAM *mtx_open (MAILSTREAM *stream)
     mtx_close (stream,NIL);	/* dump and save the changes */
     stream->dtb = &mtxdriver;	/* reattach this driver */
     mail_free_cache (stream);	/* clean up cache */
+    stream->uid_last = 0;	/* default UID validity */
+    stream->uid_validity = time (0);
+				/* flush user flags */
+    for (i = 0; i < NUSERFLAGS; i++)
+      if (stream->user_flags[i]) fs_give ((void **) &stream->user_flags[i]);
   }
   if (!mailboxfile (tmp,stream->mailbox))
     return (MAILSTREAM *) mtx_badname (tmp,stream->mailbox);
@@ -292,74 +298,6 @@ void mtx_close (MAILSTREAM *stream,long options)
   }
 }
 
-/* MTX string driver for file stringstructs */
-
-STRINGDRIVER mtx_string = {
-  mtx_string_init,		/* initialize string structure */
-  mtx_string_next,		/* get next byte in string structure */
-  mtx_string_setpos		/* set position in string structure */
-};
-
-
-/* Cache buffer for file stringstructs */
-
-#define DOSCHUNKLEN 4096
-char dos_chunk[DOSCHUNKLEN];
-
-
-/* Initialize MTX string structure for file stringstruct
- * Accepts: string structure
- *	    pointer to string
- *	    size of string
- */
-
-void mtx_string_init (STRING *s,void *data,unsigned long size)
-{
-  MTXDATA *d = (MTXDATA *) data;
-  s->data = (void *) d->fd;	/* note fd */
-  s->data1 = d->pos;		/* note file offset */
-  s->size = size;		/* note size */
-  s->curpos = s->chunk = dos_chunk;
-  s->chunksize = (unsigned long) DOSCHUNKLEN;
-  s->offset = 0;		/* initial position */
-				/* and size of data */
-  s->cursize = min (s->chunksize,size);
-				/* move to that position in the file */
-  lseek (d->fd,d->pos,SEEK_SET);
-  read (d->fd,s->chunk,(size_t) s->cursize);
-}
-
-/* Get next character from file stringstruct
- * Accepts: string structure
- * Returns: character, string structure chunk refreshed
- */
-
-char mtx_string_next (STRING *s)
-{
-  char c = *s->curpos++;	/* get next byte */
-				/* move to next chunk */
-  SETPOS (s,s->offset + s->chunksize);
-  return c;			/* return the byte */
-}
-
-
-/* Set string pointer position for file stringstruct
- * Accepts: string structure
- *	    new position
- */
-
-void mtx_string_setpos (STRING *s,unsigned long i)
-{
-  s->offset = i;		/* set new offset */
-  s->curpos = s->chunk;		/* reset position */
-				/* set size of data */
-  if (s->cursize = s->size > s->offset ? min (s->chunksize,SIZE (s)) : 0) {
-				/* move to that position in the file */
-    lseek ((int) s->data,s->data1 + s->offset,SEEK_SET);
-    read ((int) s->data,s->curpos,(size_t) s->cursize);
-  }
-}
-
 /* MTX mail fetch message header
  * Accepts: MAIL stream
  *	    message # to fetch
@@ -396,7 +334,7 @@ char *mtx_header (MAILSTREAM *stream,unsigned long msgno,unsigned long *length,
 long mtx_text (MAILSTREAM *stream,unsigned long msgno,STRING *bs,long flags)
 {
   MESSAGECACHE *elt;
-  MTXDATA d;
+  FDDATA d;
   unsigned long hdrsize,hdrpos;
 				/* UID call "impossible" */
   if (flags & FT_UID) return NIL;
@@ -405,12 +343,16 @@ long mtx_text (MAILSTREAM *stream,unsigned long msgno,STRING *bs,long flags)
     elt->seen = T;		/* mark message as seen */
 				/* recalculate status */
     mtx_update_status (stream,msgno);
+    mm_flags (stream,msgno);
   }
 				/* get location of text data */
   hdrpos = mtx_hdrpos (stream,msgno,&hdrsize);
   d.fd = LOCAL->fd;		/* set initial stringstruct */
   d.pos = hdrpos + hdrsize;
-  INIT (bs,mtx_string,(void *) &d,elt->rfc822_size - hdrsize);
+				/* flush old buffer */
+  if (LOCAL->buf) fs_give ((void **) &LOCAL->buf);
+  d.chunk = LOCAL->buf = (char *) fs_get ((size_t) d.chunksize = CHUNK);
+  INIT (bs,fd_string,(void *) &d,elt->rfc822_size - hdrsize);
   return T;			/* success */
 }
 
@@ -527,6 +469,8 @@ long mtx_copy (MAILSTREAM *stream,char *sequence,char *mailbox,long options)
   unsigned long i,j,k;
   long ret = LONGT;
   int fd;
+  mailproxycopy_t pc =
+    (mailproxycopy_t) mail_parameters (stream,GET_MAILPROXYCOPY,NIL);
   if (!((options & CP_UID) ? mail_uid_sequence (stream,sequence) :
 	mail_sequence (stream,sequence))) return NIL;
   if (!mtx_isvalid (mailbox,tmp)) switch (errno) {
@@ -537,9 +481,11 @@ long mtx_copy (MAILSTREAM *stream,char *sequence,char *mailbox,long options)
   case 0:			/* merely empty file? */
     break;
   case EINVAL:			/* name is bogus */
+    if (pc) return (*pc) (stream,sequence,mailbox,options);
     return mtx_badname (tmp,mailbox);
   default:			/* file exists, but not valid format */
-    sprintf (tmp,"Not a Mtx-format mailbox: %s",mailbox);
+    if (pc) return (*pc) (stream,sequence,mailbox,options);
+    sprintf (tmp,"Not a MTX-format mailbox: %s",mailbox);
     mm_log (tmp,ERROR);
     return NIL;
   }
@@ -605,22 +551,30 @@ long mtx_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
   if (date) {			/* want to preserve date? */
 				/* yes, parse date into an elt */
     if (!mail_parse_date (&elt,date)) {
-      sprintf (tmp,"Bad date in append: %s",date);
+      sprintf (tmp,"Bad date in append: %.80s",date);
       mm_log (tmp,ERROR);
       return NIL;
     }
   }
   if (!mtx_isvalid (mailbox,tmp)) switch (errno) {
   case ENOENT:			/* no such file? */
-    mm_notify (stream,"[TRYCREATE] Must create mailbox before append",
-	       (long) NIL);
-    return NIL;
+    if (((mailbox[0] == 'I') || (mailbox[0] == 'i')) &&
+	((mailbox[1] == 'N') || (mailbox[1] == 'n')) &&
+	((mailbox[2] == 'B') || (mailbox[2] == 'b')) &&
+	((mailbox[3] == 'O') || (mailbox[3] == 'o')) &&
+	((mailbox[4] == 'X') || (mailbox[4] == 'x')) && !mailbox[5])
+      mtx_create (NIL,"INBOX");
+    else {
+      mm_notify (stream,"[TRYCREATE] Must create mailbox before append",NIL);
+      return NIL;
+    }
+				/* falls through */
   case 0:			/* merely empty file? */
     break;
   case EINVAL:			/* name is bogus */
     return mtx_badname (tmp,mailbox);
   default:			/* file exists, but not valid format */
-    sprintf (tmp,"Not a Mtx-format mailbox: %s",mailbox);
+    sprintf (tmp,"Not a Mtx-format mailbox: %.80s",mailbox);
     mm_log (tmp,ERROR);
     return NIL;
   }
