@@ -10,7 +10,7 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	20 December 1989
- * Last Edited:	3 March 1997
+ * Last Edited:	17 March 1997
  *
  * Copyright 1997 by the University of Washington
  *
@@ -697,18 +697,17 @@ long bezerk_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
 		    STRING *message)
 {
   struct stat sbuf;
-  int i,fd,ti,zn;
-  char *s,*x,buf[BUFLEN],file[MAILTMPLEN],lock[MAILTMPLEN];
+  int fd,ti,zn;
+  long i,ok;
+  unsigned long j,n,uf,size;
+  char c,*x,buf[BUFLEN],tmp[MAILTMPLEN],file[MAILTMPLEN],lock[MAILTMPLEN];
   struct utimbuf times;
   MESSAGECACHE elt;
-  char c = '\012';
-  unsigned long j,n,ok = T;
   time_t t = time (0);
-  unsigned long size = SIZE (message);
-  unsigned long uf;
   long f = mail_parse_flags (stream ? stream : &bezerkproto,flags,&uf);
-				/* parse date if given */
-  if (date && !mail_parse_date (&elt,date)) {
+				/* parse date */
+  if (!date) rfc822_date (date = tmp);
+  if (!mail_parse_date (&elt,date)) {
     sprintf (buf,"Bad date in append: %s",date);
     mm_log (buf,ERROR);
     return NIL;
@@ -744,47 +743,34 @@ long bezerk_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
   fstat (fd,&sbuf);		/* get current file size */
   sprintf (buf,"From %s@%s ",myusername (),mylocalhost ());
 				/* write the date given */
-  if (date) mail_cdate (buf + strlen (buf),&elt);
-  else {
-    strcat (buf,ctime (&t));	/* otherwise write the time now */
+  mail_cdate (buf + strlen (buf),&elt);
 				/* replace trailing NL with CRLF */
-    strcpy (buf + strlen (buf) -1,"\015\012");
-  }
+  strcpy (buf + strlen (buf) -1,"\015\012");
   sprintf (buf + strlen (buf),"Status: %s\015\012X-Status: %s%s%s%s\015\012",
 	   f&fSEEN ? "R" : "",f&fDELETED ? "D" : "",
 	   f&fFLAGGED ? "F" : "",f&fANSWERED ? "A" : "",f&fDRAFT ? "T" : "");
-  i = strlen (buf);		/* initial buffer space used */
-  while (ok && size--) {	/* copy text, tossing out CR's */
-				/* if at start of line */
-    if ((c == '\012') && (size > 5)) {
-      n = GETPOS (message);	/* prepend a broket if needed */
-      if ((SNX (message) == 'F') && (SNX (message) == 'r') &&
-	  (SNX (message) == 'o') && (SNX (message) == 'm') &&
-	  (SNX (message) == ' ')) {
-				/* always write widget if unconditional */
-	if (bezerk_fromwidget) ok = bezerk_append_putc (fd,buf,&i,'>');
-	else {			/* hairier test, count length of this line */
-	  for (j = 6; (j < size) && (SNX (message) != '\012'); j++);
-	  if (j < size) {	/* copy line */
-	    SETPOS (message,n);	/* restore position */
-	    x = s = (char *) fs_get (j + 1);
-	    while (j--) *x++ = SNX (message);
-	    *x = '\0';		/* tie off line */
-	    VALID (s,x,ti,zn);	/* see if looks like need a widget */
-	    if (ti) ok = bezerk_append_putc (fd,buf,&i,'>');
-	    fs_give ((void **) &s);
-	  }
-	}
-      }
-      SETPOS (message,n);	/* restore position as needed */
+				/* copy text */
+  for (i = strlen (buf), ok = T, size = SIZE (message); ok && size; ) {
+    for (j = 0, c = '\0'; size && (c != '\n') && (j < MAILTMPLEN); size--)
+      tmp[j++] = c = SNX (message);
+				/* possible "From " line? */
+    if ((j > 4) && (tmp[0] == 'F') && (tmp[1] == 'r') && (tmp[2] == 'o') &&
+	(tmp[3] == 'm') && (tmp[4] == ' ')) {
+				/* see if need to write a widget */
+      if (!(ti = bezerk_fromwidget || (c != '\n'))) VALID (tmp,x,ti,zn);
+      if (ti) ok = bezerk_append_putc (fd,buf,&i,'>');
     }
-				/* copy another character */
-    ok = bezerk_append_putc (fd,buf,&i,c);
+				/* write the line */
+    if (ok) for (n = 0; n < j; n++) ok = bezerk_append_putc (fd,buf,&i,tmp[n]);
+				/* handle pathologically long line */
+    if (ok && (c != '\n') && size) do
+      ok = bezerk_append_putc (fd,buf,&i,c = SNX (message));
+    while (ok && --size && (c != '\n'));
   }
 				/* write trailing CRLF */
-  if (ok) ok = (bezerk_append_putc (fd,buf,&i,'\015') &&
-		bezerk_append_putc (fd,buf,&i,'\012'));
-  if (!(ok && (ok = (write (fd,buf,i) >= 0)) && (ok = !fsync (fd)))) {
+  if (!(ok && (bezerk_append_putc (fd,buf,&i,'\015') &&
+	       bezerk_append_putc (fd,buf,&i,'\012') &&
+	       (i ? (write (fd,buf,i) >= 0) : T) && !fsync (fd)))) {
     sprintf (buf,"Message append failed: %s",strerror (errno));
     mm_log (buf,ERROR);
     ftruncate (fd,sbuf.st_size);
@@ -805,7 +791,7 @@ long bezerk_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
  * Returns: T if append successful, else NIL
  */
 
-long bezerk_append_putc (int fd,char *s,int *i,char c)
+long bezerk_append_putc (int fd,char *s,long *i,char c)
 {
   s[(*i)++] = c;
   if (*i == BUFLEN) {		/* dump if buffer filled */
