@@ -10,9 +10,9 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	24 May 1993
- * Last Edited:	4 December 1996
+ * Last Edited:	31 December 1997
  *
- * Copyright 1996 by the University of Washington
+ * Copyright 1997 by the University of Washington
  *
  *  Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose and without fee is hereby granted, provided
@@ -43,7 +43,6 @@
 #include "osdep.h"
 #include <sys\stat.h>
 #include <dos.h>
-#include <io.h>
 #include "dummy.h"
 #include "misc.h"
 
@@ -136,8 +135,16 @@ void dummy_scan (MAILSTREAM *stream,char *ref,char *pat,char *contents)
 {
   char *s,test[MAILTMPLEN],file[MAILTMPLEN];
   long i = 0;
+  if (!pat || !*pat) {		/* empty pattern? */
+    if (dummy_canonicalize (test,ref,"*")) {
+				/* tie off name at root */
+      if (s = strchr (test,'\\')) *++s = '\0';
+      else test[0] = '\0';
+      dummy_listed (stream,'\\',test,LATT_NOINFERIORS,NIL);
+    }
+  }
 				/* get canonical form of name */
-  if (dummy_canonicalize (test,ref,pat)) {
+  else if (dummy_canonicalize (test,ref,pat)) {
 				/* found any wildcards? */
     if (s = strpbrk (test,"%*")) {
 				/* yes, copy name up to that point */
@@ -153,7 +160,7 @@ void dummy_scan (MAILSTREAM *stream,char *ref,char *pat,char *contents)
 				/* silly case */
     else if (file[0] == '#') s = file;
 				/* do the work */
-    dummy_list_work (stream,s,ucase (test),contents,0);
+    dummy_list_work (stream,s,test,contents,0);
     if (pmatch ("INBOX",test))	/* always an INBOX */
       dummy_listed (stream,NIL,"INBOX",LATT_NOINFERIORS,contents);
   }
@@ -181,10 +188,12 @@ void dummy_lsub (MAILSTREAM *stream,char *ref,char *pat)
   void *sdb = NIL;
   char *s,test[MAILTMPLEN],tmp[MAILTMPLEN];
 				/* get canonical form of name */
-  if (dummy_canonicalize (test,ref,pat) && ucase(test) && (s = sm_read (&sdb)))
-    do if (pmatch_full (ucase (strcpy (tmp,s)),test,'\\'))
-      mm_lsub (stream,'\\',s,NIL);
-  while (s = sm_read (&sdb));	/* until no more subscriptions */
+  if (dummy_canonicalize (test,ref,pat) && (s = sm_read (&sdb)))
+    do if (pmatch_full (ucase (strcpy (tmp,s)),test,'\\')) {
+      if (pmatch (tmp,"INBOX")) mm_lsub (stream,NIL,s,LATT_NOINFERIORS);
+      else mm_lsub (stream,'\\',s,NIL);
+				/* until no more subscriptions */
+    } while (s = sm_read (&sdb));
 }
 
 /* Dummy list mailboxes worker routine
@@ -259,7 +268,8 @@ void dummy_list_work (MAILSTREAM *stream,char *dir,char *pat,char *contents,
 	      dummy_list_work (stream,tmp,pat,contents,level+1);
 	    break;
 	  case S_IFREG:		/* ordinary name */
-	    if (pmatch_full (ucase (strcpy (tmpy,tmp)),pat,'\\'))
+	    if (pmatch_full (ucase (strcpy (tmpy,tmp)),pat,'\\') &&
+		!pmatch ("INBOX",tmpy))
 	      dummy_listed (stream,'\\',tmp,LATT_NOINFERIORS,contents);
 	    break;
 	  }
@@ -297,7 +307,8 @@ long dummy_listed (MAILSTREAM *stream,char delimiter,char *name,
     memset (buf,'\0',ssiz);	/* no slop area the first time */
     while (sbuf.st_size) {	/* until end of file */
       read (fd,buf+ssiz,bsiz = min (sbuf.st_size,BUFSIZE));
-      if (search (buf,bsiz+ssiz,contents,csiz)) break;
+      if (search ((unsigned char *) buf,bsiz+ssiz,
+		  (unsigned char *) contents,csiz)) break;
       memcpy (buf,buf+BUFSIZE,ssiz);
       sbuf.st_size -= bsiz;	/* note that we read that much */
     }
@@ -377,10 +388,6 @@ long dummy_delete (MAILSTREAM *stream,char *mailbox)
 {
   struct stat sbuf;
   char *s,tmp[MAILTMPLEN];
-  if (!strcmp (ucase (strcpy (tmp,mailbox)),"INBOX")) {
-    mm_log ("Can't delete INBOX",ERROR);
-    return NIL;
-  }
 				/* no trailing / (workaround BSD kernel bug) */
   if ((s = strrchr (dummy_file (tmp,mailbox),'\\')) && !s[1]) *s = '\0';
   if (stat (tmp,&sbuf) || ((sbuf.st_mode & S_IFMT) == S_IFDIR) ?
@@ -456,14 +463,16 @@ MAILSTREAM *dummy_open (MAILSTREAM *stream)
       sprintf (err,"%s (file %s) is not in valid mailbox format",
 	       stream->mailbox,tmp);
   }
-  if (!stream->silent) {	/* only if silence not requested */
-    if (err[0]) mm_log (err,ERROR);
-    else {
-      mail_exists (stream,0);	/* say there are 0 messages */
-      mail_recent (stream,0);	/* and certainly no recent ones! */
-    }
+  if (err[0]) {			/* if an error happened */
+    mm_log (err,stream->silent ? WARN : ERROR);
+    return NIL;
   }
-  return err[0] ? NIL : stream;	/* return success if no error */
+  else if (!stream->silent) {	/* only if silence not requested */
+    mail_exists (stream,0);	/* say there are 0 messages */
+    mail_recent (stream,0);	/* and certainly no recent ones! */
+    stream->uid_validity = 1;
+  }
+  return stream;		/* return success */
 }
 
 
@@ -563,7 +572,7 @@ long dummy_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
       return NIL;
     }
   }
-  return (*default_proto ()->dtb->append) (stream,mailbox,flags,date,message);
+  return (*default_proto (T)->dtb->append) (stream,mailbox,flags,date,message);
 }
 
 /* Dummy mail generate file string
@@ -622,5 +631,6 @@ long dummy_canonicalize (char *tmp,char *ref,char *pat)
   if (dev[0] && (*(ref ? ref : pat) != '\\')) dev[2] = '\\';
 				/* build name */
   sprintf (tmp,"%s%s%s",dev,ref ? ref : "",pat);
+  ucase (tmp);			/* force upper case */
   return T;
 }
