@@ -10,9 +10,9 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	3 October 1995
- * Last Edited:	3 December 1996
+ * Last Edited:	4 November 1997
  *
- * Copyright 1996 by the University of Washington
+ * Copyright 1997 by the University of Washington
  *
  *  Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose and without fee is hereby granted, provided
@@ -42,7 +42,6 @@ extern int errno;		/* just in case */
 #include "osdep.h"
 #include <fcntl.h>
 #include <time.h>
-#include <io.h>
 #include <sys/stat.h>
 #include <sys/utime.h>
 #include "mbxnt.h"
@@ -56,7 +55,7 @@ extern int errno;		/* just in case */
 
 DRIVER mbxdriver = {
   "mbx",			/* driver name */
-  DR_NAMESPACE|DR_LOCAL|DR_MAIL,/* driver flags */
+  DR_LOCAL|DR_MAIL|DR_CRLF,	/* driver flags */
   (DRIVER *) NIL,		/* next driver */
   mbx_valid,			/* mailbox is valid for us */
   mbx_parameters,		/* manipulate parameters */
@@ -201,7 +200,7 @@ void mbx_lsub (MAILSTREAM *stream,char *ref,char *pat)
 
 long mbx_create (MAILSTREAM *stream,char *mailbox)
 {
-  char *s,*t,mbx[MAILTMPLEN],tmp[HDRSIZE];
+  char *s,mbx[MAILTMPLEN],tmp[HDRSIZE];
   long ret = NIL;
   int i,fd;
   if (!(s = mbx_file (mbx,mailbox))) {
@@ -220,7 +219,8 @@ long mbx_create (MAILSTREAM *stream,char *mailbox)
     return NIL;
   }
   memset (tmp,'\0',HDRSIZE);	/* initialize header */
-  for (s = tmp,t = "*mbx*\015\0120000000100000000\015\012"; *t;) *s++ = *t++;
+  sprintf (s = tmp,"*mbx*\015\012%08lx00000000\015\012",time (0));
+  s += strlen (s);		/* move to end of pointer */
   for (i = 0; i < NUSERFLAGS; ++i) *s++ ='\015',*s++ = '\012';
   if (write (fd,tmp,HDRSIZE) != HDRSIZE) {
     sprintf (tmp,"Can't initialize mailbox node %s: %s",mbx,strerror (errno));
@@ -281,6 +281,8 @@ long mbx_rename (MAILSTREAM *stream,char *old,char *newname)
       mm_log (tmp,ERROR);
       ret = NIL;		/* set failure */
     }
+    flock (fd,LOCK_UN);		/* release lock on the file */
+    close (fd);			/* pacify NTFS */
     if (rename (file,s)) {	/* rename the file */
       sprintf (tmp,"Can't rename mailbox %s to %s: %s",old,newname,
 	       strerror (errno));
@@ -288,14 +290,16 @@ long mbx_rename (MAILSTREAM *stream,char *old,char *newname)
       ret = NIL;		/* set failure */
     }
   }
-  else if (unlink (file)) {
-    sprintf (tmp,"Can't delete mailbox %s: %s",old,strerror (errno));
-    mm_log (tmp,ERROR);
-    ret = NIL;			/* set failure */
+  else {
+    flock (fd,LOCK_UN);		/* release lock on the file */
+    close (fd);			/* pacify NTFS */
+    if (unlink (file)) {
+      sprintf (tmp,"Can't delete mailbox %s: %s",old,strerror (errno));
+      mm_log (tmp,ERROR);
+      ret = NIL;		/* set failure */
+    }
   }
-  flock (fd,LOCK_UN);		/* release lock on the file */
   unlockfd (ld,lock);		/* release exclusive parse/append permission */
-  close (fd);			/* close the file */
 				/* recreate file if renamed INBOX */
   if (ret && !strcmp (ucase (strcpy (tmp,old)),"INBOX"))
     mbx_create (NIL,"#mbx\\INBOX");
@@ -309,13 +313,18 @@ long mbx_rename (MAILSTREAM *stream,char *old,char *newname)
 
 MAILSTREAM *mbx_open (MAILSTREAM *stream)
 {
-  int fd,ld;
+  int fd,ld,i;
   char tmp[MAILTMPLEN];
   if (!stream) return &mbxproto;/* return prototype for OP_PROTOTYPE call */
   if (LOCAL) {			/* close old file if stream being recycled */
     mbx_close (stream,NIL);	/* dump and save the changes */
     stream->dtb = &mbxdriver;	/* reattach this driver */
     mail_free_cache (stream);	/* clean up cache */
+    stream->uid_last = 0;	/* default UID validity */
+    stream->uid_validity = 0;
+				/* flush user flags */
+    for (i = 0; i < NUSERFLAGS; i++)
+      if (stream->user_flags[i]) fs_give ((void **) &stream->user_flags[i]);
   }
   if (stream->rdonly ||
       (fd = open (mbx_file (tmp,stream->mailbox),O_BINARY|O_RDWR,NIL)) < 0) {
@@ -445,24 +454,22 @@ long mbx_text (MAILSTREAM *stream,unsigned long msgno,STRING *bs,long flags)
     elt->seen = T;		/* mark message as seen */
 				/* recalculate status */
     mbx_update_status (stream,msgno,T);
+    mm_flags (stream,msgno);
   }
 				/* find header position */
   i = mbx_hdrpos (stream,msgno,&j);
 				/* go to text position */
   lseek (LOCAL->fd,i + j,L_SET);
 				/* is buffer big enough? */
-  if (elt->private.msg.text.text.size > LOCAL->buflen) {
+  if ((i = elt->rfc822_size - j) > LOCAL->buflen) {
     fs_give ((void **) &LOCAL->buf);
-    LOCAL->buflen = elt->private.msg.text.text.size;
-    LOCAL->buf = (char *) fs_get (LOCAL->buflen + 1);
+    LOCAL->buf = (char *) fs_get ((LOCAL->buflen = i) + 1);
   }
-				/* slurp the data */
-  if (read (LOCAL->fd,LOCAL->buf,elt->private.msg.text.text.size) !=
-      (long) elt->private.msg.text.text.size) return NIL;
-				/* tie off string */
-  LOCAL->buf[elt->private.msg.text.text.size] = '\0';
+			
+  read (LOCAL->fd,LOCAL->buf,i);/* slurp the data */
+  LOCAL->buf[i] = '\0';		/* tie off string */
 				/* set up stringstruct */
-  INIT (bs,mail_string,LOCAL->buf,elt->private.msg.text.text.size);
+  INIT (bs,mail_string,LOCAL->buf,i);
   return T;			/* success */
 }
 
@@ -606,10 +613,14 @@ void mbx_expunge (MAILSTREAM *stream)
       do {			/* read from source position */
 	m = min (k,LOCAL->buflen);
 	lseek (LOCAL->fd,j,L_SET);
-	read (LOCAL->fd,LOCAL->buf,(unsigned int) m);
+	read (LOCAL->fd,LOCAL->buf,m);
 	pos = j - delta;	/* write to destination position */
-	lseek (LOCAL->fd,pos,L_SET);
-	write (LOCAL->fd,LOCAL->buf,(unsigned int) m);
+	while (T) {
+	  lseek (LOCAL->fd,pos,L_SET);
+	  if (write (LOCAL->fd,LOCAL->buf,m) > 0) break;
+	  mm_notify (stream,strerror (errno),WARN);
+	  mm_diskerror (stream,errno,T);
+	}
 	pos += m;		/* new position */
 	j += m;			/* next chunk, perhaps */
       } while (k -= m);		/* until done */
@@ -621,13 +632,13 @@ void mbx_expunge (MAILSTREAM *stream)
   }
   if (n) {			/* truncate file after last message */
     if (pos != (LOCAL->filesize -= delta)) {
-      sprintf (LOCAL->buf,"Calculated size mismatch %ld != %ld, delta = %ld",
+      sprintf (LOCAL->buf,"Calculated size mismatch %lu != %lu, delta = %lu",
 	       pos,LOCAL->filesize,delta);
       mm_log (LOCAL->buf,WARN);
       LOCAL->filesize = pos;	/* fix it then */
     }
     ftruncate (LOCAL->fd,LOCAL->filesize);
-    sprintf (LOCAL->buf,"Expunged %ld messages",n);
+    sprintf (LOCAL->buf,"Expunged %lu messages",n);
 				/* output the news */
     mm_log (LOCAL->buf,(long) NIL);
   }
@@ -653,25 +664,95 @@ void mbx_expunge (MAILSTREAM *stream)
 
 long mbx_copy (MAILSTREAM *stream,char *sequence,char *mailbox,long options)
 {
-  unsigned long i;
   struct stat sbuf;
+  struct utimbuf times;
   MESSAGECACHE *elt;
-  if (!(((options & CP_UID) ? mail_uid_sequence (stream,sequence) :
-	 mail_sequence (stream,sequence)) &&
-	mbx_copy_messages (stream,mailbox))) return NIL;
-				/* delete all requested messages */
-  if (options & CP_MOVE) for (i = 1; i <= stream->nmsgs; i++)
-    if ((elt = mbx_elt (stream,i))->sequence) {
-      elt->deleted = T;		/* mark message deleted */
-				/* recalculate status */
-      mbx_update_status (stream,i,NIL);
-    }
-  if (!stream->rdonly) {	/* make sure the update takes */
-    fsync (LOCAL->fd);
-    fstat (LOCAL->fd,&sbuf);	/* get current write time */
-    LOCAL->filetime = sbuf.st_mtime;
+  unsigned long i,j,k;
+  long ret = LONGT;
+  int fd,ld;
+  char file[MAILTMPLEN],lock[MAILTMPLEN];
+  mailproxycopy_t pc =
+    (mailproxycopy_t) mail_parameters (stream,GET_MAILPROXYCOPY,NIL);
+				/* make sure valid mailbox */
+  if (!mbx_isvalid (mailbox,LOCAL->buf)) switch (errno) {
+  case ENOENT:			/* no such file? */
+    mm_notify (stream,"[TRYCREATE] Must create mailbox before copy",NIL);
+    return NIL;
+  case 0:			/* merely empty file? */
+    break;
+  case EINVAL:
+    if (pc) return (*pc) (stream,sequence,mailbox,options);
+    sprintf (LOCAL->buf,"Invalid MBX-format mailbox name: %s",mailbox);
+    mm_log (LOCAL->buf,ERROR);
+    return NIL;
+  default:
+    if (pc) return (*pc) (stream,sequence,mailbox,options);
+    sprintf (LOCAL->buf,"Not a MBX-format mailbox: %s",mailbox);
+    mm_log (LOCAL->buf,ERROR);
+    return NIL;
   }
-  return LONGT;
+  if (!((options & CP_UID) ? mail_uid_sequence (stream,sequence) :
+	mail_sequence (stream,sequence))) return NIL;
+				/* got file? */  
+  if ((fd = open (mbx_file (file,mailbox),O_BINARY|O_RDWR|O_CREAT,
+		  S_IREAD|S_IWRITE)) < 0) {
+    sprintf (LOCAL->buf,"Unable to open copy mailbox: %s",strerror (errno));
+    mm_log (LOCAL->buf,ERROR);
+    return NIL;
+  }
+  mm_critical (stream);		/* go critical */
+				/* get parse/append permission */
+  if ((ld = lockname (lock,mailbox,LOCK_EX)) < 0) {
+    mm_log ("Unable to lock copy mailbox",ERROR);
+    return NIL;
+  }
+  fstat (fd,&sbuf);		/* get current file size */
+  lseek (fd,sbuf.st_size,L_SET);/* move to end of file */
+
+				/* for each requested message */
+  for (i = 1; ret && (i <= stream->nmsgs); i++) 
+    if ((elt = mail_elt (stream,i))->sequence) {
+      lseek (LOCAL->fd,elt->private.special.offset +
+	     elt->private.special.text.size,L_SET);
+      mail_date(LOCAL->buf,elt);/* build target header */
+      sprintf (LOCAL->buf+strlen(LOCAL->buf),",%lu;%08lx%04x-00000000\015\012",
+	       elt->rfc822_size,elt->user_flags,(fSEEN * elt->seen) +
+	       (fDELETED * elt->deleted) + (fFLAGGED * elt->flagged) +
+	       (fANSWERED * elt->answered) + (fDRAFT * elt->draft));
+				/* write target header */
+      if (ret = (write (fd,LOCAL->buf,strlen (LOCAL->buf)) > 0))
+	for (k = elt->rfc822_size; ret && (j = min (k,LOCAL->buflen)); k -= j){
+	  read (LOCAL->fd,LOCAL->buf,j);
+	  ret = write (fd,LOCAL->buf,j) >= 0;
+	}
+    }
+				/* make sure all the updates take */
+  if (!(ret && (ret = !fsync (fd)))) {
+    sprintf (LOCAL->buf,"Unable to write message: %s",strerror (errno));
+    mm_log (LOCAL->buf,ERROR);
+    ftruncate (fd,sbuf.st_size);
+  }
+  times.actime = sbuf.st_atime;	/* preserve atime and mtime */
+  times.modtime = sbuf.st_mtime;
+  utime (file,&times);		/* set the times */
+  unlockfd (ld,lock);		/* release exclusive parse/append permission */
+  close (fd);			/* close the file */
+  mm_nocritical (stream);	/* release critical */
+				/* delete all requested messages */
+  if (ret && (options & CP_MOVE)) {
+    for (i = 1; i <= stream->nmsgs; i++)
+      if ((elt = mbx_elt (stream,i))->sequence) {
+	elt->deleted = T;	/* mark message deleted */
+				/* recalculate status */
+	mbx_update_status (stream,i,NIL);
+      }
+    if (!stream->rdonly) {	/* make sure the update takes */
+      fsync (LOCAL->fd);
+      fstat (LOCAL->fd,&sbuf);	/* get current write time */
+      LOCAL->filetime = sbuf.st_mtime;
+    }
+  }
+  return ret;
 }
 
 /* MBX mail append message from stringstruct
@@ -698,7 +779,7 @@ long mbx_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
   if (date) {			/* want to preserve date? */
 				/* yes, parse date into an elt */
     if (!mail_parse_date (&elt,date)) {
-      sprintf (tmp,"Bad date in append: %s",date);
+      sprintf (tmp,"Bad date in append: %.80s",date);
       mm_log (tmp,ERROR);
       return NIL;
     }
@@ -707,22 +788,31 @@ long mbx_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
 				/* make sure valid mailbox */
   if (!mbx_isvalid (mailbox,tmp)) switch (errno) {
   case ENOENT:			/* no such file? */
-    mm_notify (stream,"[TRYCREATE] Must create mailbox before append",NIL);
-    return NIL;
+    if (((mailbox[0] == 'I') || (mailbox[0] == 'i')) &&
+	((mailbox[1] == 'N') || (mailbox[1] == 'n')) &&
+	((mailbox[2] == 'B') || (mailbox[2] == 'b')) &&
+	((mailbox[3] == 'O') || (mailbox[3] == 'o')) &&
+	((mailbox[4] == 'X') || (mailbox[4] == 'x')) && !mailbox[5])
+      mbx_create (NIL,"INBOX");
+    else {
+      mm_notify (stream,"[TRYCREATE] Must create mailbox before append",NIL);
+      return NIL;
+    }
+				/* falls through */
   case 0:			/* merely empty file? */
     break;
   case EINVAL:
-    sprintf (tmp,"Invalid MBX-format mailbox name: %s",mailbox);
+    sprintf (tmp,"Invalid MBX-format mailbox name: %.80s",mailbox);
     mm_log (tmp,ERROR);
     return NIL;
   default:
-    sprintf (tmp,"Not a MBX-format mailbox: %s",mailbox);
+    sprintf (tmp,"Not a MBX-format mailbox: %.80s",mailbox);
     mm_log (tmp,ERROR);
     return NIL;
   }
   if ((fd = open (mbx_file (file,mailbox),O_BINARY|O_RDWR|O_CREAT,
 		  S_IREAD|S_IWRITE)) < 0) {
-    sprintf (tmp,"Can't open append mailbox: %s",strerror (errno));
+    sprintf (tmp,"Can't open append mailbox: %.80s",strerror (errno));
     mm_log (tmp,ERROR);
     return NIL;
   }
@@ -740,7 +830,7 @@ long mbx_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
   if (date) mail_date(tmp,&elt);/* write preseved date */
   else internal_date (tmp);	/* get current date in IMAP format */
 				/* add remainder of header */
-  sprintf (tmp+26,",%ld;%08lx%04x-00000000\015\012",size,uf,f);
+  sprintf (tmp+26,",%lu;%08lx%04x-00000000\015\012",size,uf,f);
 				/* write header */
   if ((write (fd,tmp,strlen (tmp)) < 0) || ((write (fd,s,size)) < 0) ||
       fsync (fd)) {
@@ -795,7 +885,7 @@ long mbx_parse (MAILSTREAM *stream)
   short silent = stream->silent;
   fstat (LOCAL->fd,&sbuf);	/* get status */
   if (sbuf.st_size < curpos) {	/* sanity check */
-    sprintf (tmp,"Mailbox shrank from %ld to %ld!",curpos,sbuf.st_size);
+    sprintf (tmp,"Mailbox shrank from %lu to %lu!",curpos,sbuf.st_size);
     mm_log (tmp,ERROR);
     mbx_close (stream,NIL);
     return NIL;
@@ -824,7 +914,7 @@ long mbx_parse (MAILSTREAM *stream)
 				/* get to that position in the file */
     lseek (LOCAL->fd,curpos,L_SET);
     if ((i = read (LOCAL->fd,LOCAL->buf,64)) <= 0) {
-      sprintf (tmp,"Unable to read internal header at %ld, size = %ld: %s",
+      sprintf (tmp,"Unable to read internal header at %lu, size = %lu: %s",
 	       curpos,sbuf.st_size,i ? strerror (errno) : "no data read");
       mm_log (tmp,ERROR);
       mbx_close (stream,NIL);
@@ -832,7 +922,7 @@ long mbx_parse (MAILSTREAM *stream)
     }
     LOCAL->buf[i] = '\0';	/* tie off buffer just in case */
     if (!((s = strchr (LOCAL->buf,'\015')) && (s[1] == '\012'))) {
-      sprintf (tmp,"Unable to find CRLF at %ld in %ld bytes, text: %s",
+      sprintf (tmp,"Unable to find CRLF at %lu in %lu bytes, text: %s",
 	       curpos,i,LOCAL->buf);
       mm_log (tmp,ERROR);
       mbx_close (stream,NIL);
@@ -841,7 +931,7 @@ long mbx_parse (MAILSTREAM *stream)
     *s = '\0';			/* tie off header line */
     i = (s + 2) - LOCAL->buf;	/* note start of text offset */
     if (!((s = strchr (LOCAL->buf,',')) && (t = strchr (s+1,';')))) {
-      sprintf (tmp,"Unable to parse internal header at %ld: %s",curpos,
+      sprintf (tmp,"Unable to parse internal header at %lu: %s",curpos,
 	       LOCAL->buf);
       mm_log (tmp,ERROR);
       mbx_close (stream,NIL);
@@ -852,13 +942,13 @@ long mbx_parse (MAILSTREAM *stream)
 				/* swell the cache */
     mail_exists (stream,++nmsgs);
 				/* intantiate an elt for this message */
-    (elt = mail_elt (stream,nmsgs))->valid = T;
+    elt = mail_elt (stream,nmsgs);
 				/* note file offset of header */
     elt->private.special.offset = curpos;
 				/* in case error */
     elt->private.special.text.size = 0;
 				/* header size not known yet */
-    elt->private.msg.header.text.size = elt->private.msg.text.text.size = 0;
+    elt->private.msg.header.text.size = 0;
 				/* parse the header components */
     if (mail_parse_date (elt,LOCAL->buf) &&
 	(elt->rfc822_size = strtoul (x = s,&s,10)) && (!(s && *s)))
@@ -870,11 +960,8 @@ long mbx_parse (MAILSTREAM *stream)
 	    isxdigit (t[15]) && isxdigit (t[16]) && isxdigit (t[17]) &&
 	    isxdigit (t[18]) && isxdigit (t[19]) && isxdigit (t[20]))
 	  elt->private.special.text.size = i;
-    if (elt->private.special.text.size) elt->private.msg.header.offset =
-      elt->private.msg.text.offset = 
-	elt->private.special.offset + elt->private.special.text.size;
     else {			/* oops */
-      sprintf (tmp,"Unable to parse internal header elements at %ld: %s,%s;%s",
+      sprintf (tmp,"Unable to parse internal header elements at %lu: %s,%s;%s",
 	       curpos,LOCAL->buf,x,t);
       mm_log (tmp,ERROR);
       mbx_close (stream,NIL);
@@ -902,7 +989,7 @@ long mbx_parse (MAILSTREAM *stream)
     if ((elt->private.uid = strtoul (t+13,NIL,16)) &&
 	((elt->private.uid <= lastuid) ||
 	 (elt->private.uid > stream->uid_last))) {
-      sprintf (tmp,"Invalid UID %08x in message %ld, rebuilding UIDs",
+      sprintf (tmp,"Invalid UID %08x in message %lu, rebuilding UIDs",
 	       elt->private.uid,elt->msgno);
       mm_log (tmp,WARN);
       elt->private.uid = 0;	/* lose this UID */
@@ -919,7 +1006,8 @@ long mbx_parse (MAILSTREAM *stream)
     }
     lastuid = elt->private.uid;	/* update last parsed UID */
   }
-  mbx_update_header (stream);	/* update header */
+				/* update header */
+  if (!stream->rdonly) mbx_update_header (stream);
   fsync (LOCAL->fd);		/* make sure all the UID updates take */
 				/* update parsed file size and time */
   LOCAL->filesize = sbuf.st_size;
@@ -931,82 +1019,6 @@ long mbx_parse (MAILSTREAM *stream)
   return LONGT;			/* return the winnage */
 }
 
-/* MBX copy messages
- * Accepts: MAIL stream
- *	    mailbox copy vector
- *	    mailbox name
- * Returns: T if success, NIL if failed
- */
-
-long mbx_copy_messages (MAILSTREAM *stream,char *mailbox)
-{
-  struct stat sbuf;
-  struct utimbuf times;
-  MESSAGECACHE *elt;
-  unsigned long i,j,k;
-  long ret = LONGT;
-  int fd,ld;
-  char file[MAILTMPLEN],lock[MAILTMPLEN];
-				/* make sure valid mailbox */
-  if (!mbx_isvalid (mailbox,LOCAL->buf)) switch (errno) {
-  case ENOENT:			/* no such file? */
-    mm_notify (stream,"[TRYCREATE] Must create mailbox before copy",NIL);
-    return NIL;
-  case 0:			/* merely empty file? */
-    break;
-  case EINVAL:
-    sprintf (LOCAL->buf,"Invalid MBX-format mailbox name: %s",mailbox);
-    mm_log (LOCAL->buf,ERROR);
-    return NIL;
-  default:
-    sprintf (LOCAL->buf,"Not a MBX-format mailbox: %s",mailbox);
-    mm_log (LOCAL->buf,ERROR);
-    return NIL;
-  }
-				/* got file? */  
-  if ((fd = open (mbx_file (file,mailbox),O_BINARY|O_RDWR|O_CREAT,
-		  S_IREAD|S_IWRITE)) < 0) {
-    sprintf (LOCAL->buf,"Unable to open copy mailbox: %s",strerror (errno));
-    mm_log (LOCAL->buf,ERROR);
-    return NIL;
-  }
-  mm_critical (stream);		/* go critical */
-				/* get parse/append permission */
-  if ((ld = lockname (lock,mailbox,LOCK_EX)) < 0) {
-    mm_log ("Unable to lock copy mailbox",ERROR);
-    return NIL;
-  }
-  fstat (fd,&sbuf);		/* get current file size */
-  lseek (fd,sbuf.st_size,L_SET);/* move to end of file */
-
-				/* for each requested message */
-  for (i = 1; ret && (i <= stream->nmsgs); i++) 
-    if ((elt = mail_elt (stream,i))->sequence) {
-      lseek (LOCAL->fd,elt->private.special.offset,L_SET);
-				/* number of bytes to copy */
-      k = elt->private.special.text.size + elt->rfc822_size;
-      do {			/* read from source position */
-	j = min (k,LOCAL->buflen);
-	read (LOCAL->fd,LOCAL->buf,(unsigned int) j);
-	if ((write (fd,LOCAL->buf,(unsigned int) j) < 0) || fsync (fd)) {
-	  sprintf (LOCAL->buf,"Unable to write message: %s",strerror (errno));
-	  mm_log (LOCAL->buf,ERROR);
-	  ftruncate (fd,sbuf.st_size);
-	  j = k;
-	  ret = NIL;		/* note error */
-	  break;
-	}
-      } while (k -= j);		/* until done */
-    }
-  times.actime = sbuf.st_atime;	/* preserve atime and mtime */
-  times.modtime = sbuf.st_mtime;
-  utime (file,&times);		/* set the times */
-  unlockfd (ld,lock);		/* release exclusive parse/append permission */
-  close (fd);			/* close the file */
-  mm_nocritical (stream);	/* release critical */
-  return ret;
-}
-
 /* MBX get cache element with status updating from file
  * Accepts: MAIL stream
  *	    message number
@@ -1015,13 +1027,36 @@ long mbx_copy_messages (MAILSTREAM *stream,char *mailbox)
 
 MESSAGECACHE *mbx_elt (MAILSTREAM *stream,unsigned long msgno)
 {
-  unsigned long sysflags;
-  char c;
   MESSAGECACHE *elt = mail_elt (stream,msgno);
-  unsigned long oldsysflags = (elt->seen ? fSEEN : NIL) |
-    (elt->deleted ? fDELETED : NIL) | (elt->flagged ? fFLAGGED : NIL) |
-      (elt->answered ? fANSWERED : NIL) | (elt->draft ? fDRAFT : NIL);
-  unsigned long olduserflags = elt->user_flags;
+  struct {			/* old flags */
+    unsigned int seen : 1;
+    unsigned int deleted : 1;
+    unsigned int flagged : 1;
+    unsigned int answered : 1;
+    unsigned int draft : 1;
+    unsigned long user_flags;
+  } old;
+  old.seen = elt->seen; old.deleted = elt->deleted; old.flagged = elt->flagged;
+  old.answered = elt->answered; old.draft = elt->draft;
+  old.user_flags = elt->user_flags;
+  mbx_read_flags (stream,elt);
+  if ((old.seen != elt->seen) || (old.deleted != elt->deleted) ||
+      (old.flagged != elt->flagged) || (old.answered != elt->answered) ||
+      (old.draft != elt->draft) || (old.user_flags != elt->user_flags))
+    mm_flags (stream,msgno);	/* let top level know */
+  return elt;
+}
+
+/* MBX read flags from file
+ * Accepts: MAIL stream
+ * Returns: cache element
+ */
+
+void mbx_read_flags (MAILSTREAM *stream,MESSAGECACHE *elt)
+{
+  unsigned long i;
+				/* noop if readonly and have valid flags */
+  if (stream->rdonly && elt->valid) return;
 				/* set the seek pointer */
   lseek (LOCAL->fd,(off_t) elt->private.special.offset +
 	 elt->private.special.text.size - 23,L_SET);
@@ -1032,26 +1067,24 @@ MESSAGECACHE *mbx_elt (MAILSTREAM *stream,unsigned long msgno)
   }
   LOCAL->buf[12] = '\0';	/* tie off buffer */
 				/* calculate system flags */
-  sysflags = strtoul (LOCAL->buf+8,NIL,16);
-  elt->seen = sysflags & fSEEN ? T : NIL;
-  elt->deleted = sysflags & fDELETED ? T : NIL;
-  elt->flagged = sysflags & fFLAGGED ? T : NIL;
-  elt->answered = sysflags & fANSWERED ? T : NIL;
-  elt->draft = sysflags & fDRAFT ? T : NIL;
-  c = LOCAL->buf[8];		/* remember first system flags byte */
+  i = strtoul (LOCAL->buf+8,NIL,16);
+  elt->seen = i & fSEEN ? T : NIL;
+  elt->deleted = i & fDELETED ? T : NIL;
+  elt->flagged = i & fFLAGGED ? T : NIL;
+  elt->answered = i & fANSWERED ? T : NIL;
+  elt->draft = i & fDRAFT ? T : NIL;
   LOCAL->buf[8] = '\0';		/* tie off flags */
 				/* get user flags value */
   elt->user_flags = strtoul (LOCAL->buf,NIL,16);
-  LOCAL->buf[8] = c;		/* restore first system flags byte */
-  if ((oldsysflags != sysflags) || (olduserflags != elt->user_flags))
-    mm_flags (stream,msgno);	/* let top level know */
-  return elt;
+  elt->valid = T;		/* have valid flags now */
 }
 
 /* MBX update header
  * Accepts: MAIL stream
  */
 
+#define NTKLUDGEOFFSET 7
+
 void mbx_update_header (MAILSTREAM *stream)
 {
   int i;
@@ -1062,9 +1095,15 @@ void mbx_update_header (MAILSTREAM *stream)
   for (i = 0; i < NUSERFLAGS; ++i)
     sprintf (s += strlen (s),"%s\015\012",
 	     stream->user_flags[i] ? stream->user_flags[i] : "");
-  lseek (LOCAL->fd,0,L_SET);	/* rewind file */
+  while (T) {
+				/* rewind file */
+    lseek (LOCAL->fd,NTKLUDGEOFFSET,L_SET);
 				/* write new header */
-  write (LOCAL->fd,LOCAL->buf,HDRSIZE);
+    if (write (LOCAL->fd,LOCAL->buf + NTKLUDGEOFFSET,
+	       HDRSIZE - NTKLUDGEOFFSET) > 0) break;
+    mm_notify (stream,strerror (errno),WARN);
+    mm_diskerror (stream,errno,T);
+  }
 }
 
 
@@ -1078,7 +1117,9 @@ void mbx_update_status (MAILSTREAM *stream,unsigned long msgno,long syncflag)
 {
   MESSAGECACHE *elt = mail_elt (stream,msgno);
   struct stat sbuf;
-  if (!stream->rdonly) {	/* not if readonly you don't */
+				/* readonly */
+  if (stream->rdonly) mbx_read_flags (stream,elt);
+  else {			/* readwrite */
     if (LOCAL->newkeyword) {	/* defined a new keyword? */
       mbx_update_header (stream);
       LOCAL->newkeyword = NIL;	/* don't do this again */
@@ -1088,11 +1129,14 @@ void mbx_update_status (MAILSTREAM *stream,unsigned long msgno,long syncflag)
 	     (fSEEN * elt->seen) + (fDELETED * elt->deleted) +
 	     (fFLAGGED * elt->flagged) + (fANSWERED * elt->answered) +
 	     (fDRAFT * elt->draft),elt->private.uid);
-				/* get to that place in the file */
-    lseek (LOCAL->fd,(off_t) elt->private.special.offset +
-	   elt->private.special.text.size - 23,L_SET);
+    while (T) {			/* get to that place in the file */
+      lseek (LOCAL->fd,(off_t) elt->private.special.offset +
+	     elt->private.special.text.size - 23,L_SET);
 				/* write new flags and UID */
-    write (LOCAL->fd,LOCAL->buf,21);
+      if (write (LOCAL->fd,LOCAL->buf,21) > 0) break;
+      mm_notify (stream,strerror (errno),WARN);
+      mm_diskerror (stream,errno,T);
+    }
     if (syncflag) {		/* sync if requested */
       fsync (LOCAL->fd);
       fstat (LOCAL->fd,&sbuf);	/* get new write time */
@@ -1142,7 +1186,6 @@ unsigned long mbx_hdrpos (MAILSTREAM *stream,unsigned long msgno,
 	if (*s++ == '\012') {	/* have the sequence? */
 				/* yes, note for later */
 	  elt->private.msg.header.text.size = *size = siz;
-	  elt->private.msg.text.text.size = elt->rfc822_size - siz;
 	  return ret;
 	}
 	q = 0;			/* lost... */
