@@ -10,9 +10,9 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	22 May 1990
- * Last Edited:	4 December 1996
+ * Last Edited:	15 December 1997
  *
- * Copyright 1996 by the University of Washington
+ * Copyright 1997 by the University of Washington
  *
  *  Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose and without fee is hereby granted, provided
@@ -42,7 +42,6 @@ extern int errno;		/* just in case */
 #include "osdep.h"
 #include <fcntl.h>
 #include <time.h>
-#include <io.h>
 #include <sys/stat.h>
 #include <sys/utime.h>
 #include "tenexnt.h"
@@ -253,6 +252,8 @@ long tenex_rename (MAILSTREAM *stream,char *old,char *newname)
       mm_log (tmp,ERROR);
       ret = NIL;		/* set failure */
     }
+    flock (fd,LOCK_UN);		/* release lock on the file */
+    close (fd);			/* pacify NTFS */
     if (rename (file,s)) {	/* rename the file */
       sprintf (tmp,"Can't rename mailbox %s to %s: %s",old,newname,
 	       strerror (errno));
@@ -260,14 +261,16 @@ long tenex_rename (MAILSTREAM *stream,char *old,char *newname)
       ret = NIL;		/* set failure */
     }
   }
-  else if (unlink (file)) {
-    sprintf (tmp,"Can't delete mailbox %s: %s",old,strerror (errno));
-    mm_log (tmp,ERROR);
-    ret = NIL;			/* set failure */
+  else {
+    flock (fd,LOCK_UN);		/* release lock on the file */
+    close (fd);			/* pacify NTFS */
+    if (unlink (file)) {
+      sprintf (tmp,"Can't delete mailbox %s: %s",old,strerror (errno));
+      mm_log (tmp,ERROR);
+      ret = NIL;		/* set failure */
+    }
   }
-  flock (fd,LOCK_UN);		/* release lock on the file */
   unlockfd (ld,lock);		/* release exclusive parse/append permission */
-  close (fd);			/* close the file */
   return ret;			/* return success */
 }
 
@@ -278,7 +281,7 @@ long tenex_rename (MAILSTREAM *stream,char *old,char *newname)
 
 MAILSTREAM *tenex_open (MAILSTREAM *stream)
 {
-  int fd,ld;
+  int fd,ld,i;
   char tmp[MAILTMPLEN];
 				/* return prototype for OP_PROTOTYPE call */
   if (!stream) return &tenexproto;
@@ -286,6 +289,11 @@ MAILSTREAM *tenex_open (MAILSTREAM *stream)
     tenex_close (stream,NIL);	/* dump and save the changes */
     stream->dtb = &tenexdriver;	/* reattach this driver */
     mail_free_cache (stream);	/* clean up cache */
+    stream->uid_last = 0;	/* default UID validity */
+    stream->uid_validity = time (0);
+				/* flush user flags */
+    for (i = 0; i < NUSERFLAGS; i++)
+      if (stream->user_flags[i]) fs_give ((void **) &stream->user_flags[i]);
   }
   if (stream->rdonly ||
       (fd = open (mailboxfile (tmp,stream->mailbox),O_BINARY|O_RDWR,NIL)) < 0){
@@ -412,10 +420,10 @@ char *tenex_header (MAILSTREAM *stream,unsigned long msgno,
       LOCAL->buf = (char *) fs_get (LOCAL->buflen = i + 1);
     }
 				/* slurp the data */
-    if (read (LOCAL->fd,LOCAL->buf,*length = i) != (long) i) return NIL;
+    read (LOCAL->fd,LOCAL->buf,*length = i);
   }
   else {
-    s = (char *) fs_get (i + 1); /* get readin buffer */
+    s = (char *) fs_get (i + 1);/* get readin buffer */
     s[i] = '\0';		/* tie off string */
     read (LOCAL->fd,s,i);	/* slurp the data */
 				/* make CRLF copy of string */
@@ -447,6 +455,7 @@ long tenex_text (MAILSTREAM *stream,unsigned long msgno,STRING *bs,long flags)
     elt->seen = T;		/* mark message as seen */
 				/* recalculate status */
     tenex_update_status (stream,msgno,T);
+    mm_flags (stream,msgno);
   }
 				/* find header position */
   i = tenex_hdrpos (stream,msgno,&j);
@@ -463,10 +472,9 @@ long tenex_text (MAILSTREAM *stream,unsigned long msgno,STRING *bs,long flags)
   else {			/* get readin buffer */
     s = (char *) fs_get ((i = tenex_size (stream,msgno) - j) + 1);
     s[i] = '\0';		/* tie off string */
-    if (read (LOCAL->fd,s,i) != (long) i) return NIL;
+    read (LOCAL->fd,s,i);	/* slurp the data */
 				/* make CRLF copy of string */
-    elt->private.msg.text.text.size =
-      unix_crlfcpy (&LOCAL->buf,&LOCAL->buflen,s,i);
+    i = unix_crlfcpy (&LOCAL->buf,&LOCAL->buflen,s,i);
     fs_give ((void **) &s);	/* free readin buffer */
   }
 				/* set up stringstruct */
@@ -583,6 +591,11 @@ void tenex_expunge (MAILSTREAM *stream)
   }
 				/* get exclusive access */
   if ((ld = lockname (lock,stream->mailbox,LOCK_EX)) < 0) {
+    mm_log ("Unable to lock expunge mailbox",ERROR);
+    return;
+  }
+				/* get exclusive access */
+  if (flock (LOCAL->fd,LOCK_EX|LOCK_NB)) {
     flock (LOCAL->fd,LOCK_SH);	/* recover previous lock */
     mm_log("Can't expunge because mailbox is in use by another process",ERROR);
     unlockfd (ld,lock);		/* release exclusive parse/append permission */
@@ -607,10 +620,14 @@ void tenex_expunge (MAILSTREAM *stream)
       do {			/* read from source position */
 	m = min (k,LOCAL->buflen);
 	lseek (LOCAL->fd,j,L_SET);
-	read (LOCAL->fd,LOCAL->buf,(unsigned int) m);
+	read (LOCAL->fd,LOCAL->buf,m);
 	pos = j - delta;	/* write to destination position */
-	lseek (LOCAL->fd,pos,L_SET);
-	write (LOCAL->fd,LOCAL->buf,(unsigned int) m);
+	while (T) {
+	  lseek (LOCAL->fd,pos,L_SET);
+	  if (write (LOCAL->fd,LOCAL->buf,m) > 0) break;
+	  mm_notify (stream,strerror (errno),WARN);
+	  mm_diskerror (stream,errno,T);
+	}
 	pos += m;		/* new position */
 	j += m;			/* next chunk, perhaps */
       } while (k -= m);		/* until done */
@@ -654,25 +671,90 @@ void tenex_expunge (MAILSTREAM *stream)
 
 long tenex_copy (MAILSTREAM *stream,char *sequence,char *mailbox,long options)
 {
-  unsigned long i;
   struct stat sbuf;
+  struct utimbuf times;
   MESSAGECACHE *elt;
-  if (!(((options & CP_UID) ? mail_uid_sequence (stream,sequence) :
-	 mail_sequence (stream,sequence)) &&
-	tenex_copy_messages (stream,mailbox))) return NIL;
-				/* delete all requested messages */
-  if (options & CP_MOVE) for (i = 1; i <= stream->nmsgs; i++)
-    if ((elt = tenex_elt (stream,i))->sequence) {
-      elt->deleted = T;		/* mark message deleted */
-				/* recalculate status */
-      tenex_update_status (stream,i,NIL);
-    }
-  if (!stream->rdonly) {	/* make sure the update takes */
-    fsync (LOCAL->fd);
-    fstat (LOCAL->fd,&sbuf);	/* get current write time */
-    LOCAL->filetime = sbuf.st_mtime;
+  unsigned long i,j,k;
+  long ret = LONGT;
+  int fd,ld;
+  char file[MAILTMPLEN],lock[MAILTMPLEN];
+  mailproxycopy_t pc =
+    (mailproxycopy_t) mail_parameters (stream,GET_MAILPROXYCOPY,NIL);
+				/* make sure valid mailbox */
+  if (!tenex_isvalid (mailbox,LOCAL->buf)) switch (errno) {
+  case ENOENT:			/* no such file? */
+    mm_notify (stream,"[TRYCREATE] Must create mailbox before copy",NIL);
+    return NIL;
+  case 0:			/* merely empty file? */
+    break;
+  case EINVAL:
+    if (pc) return (*pc) (stream,sequence,mailbox,options);
+    sprintf (LOCAL->buf,"Invalid Tenex-format mailbox name: %s",mailbox);
+    mm_log (LOCAL->buf,ERROR);
+    return NIL;
+  default:
+    if (pc) return (*pc) (stream,sequence,mailbox,options);
+    sprintf (LOCAL->buf,"Not a Tenex-format mailbox: %s",mailbox);
+    mm_log (LOCAL->buf,ERROR);
+    return NIL;
   }
-  return LONGT;
+  if (!((options & CP_UID) ? mail_uid_sequence (stream,sequence) :
+	mail_sequence (stream,sequence))) return NIL;
+				/* got file? */  
+  if ((fd = open (mailboxfile (file,mailbox),O_BINARY|O_RDWR|O_CREAT,
+		  S_IREAD|S_IWRITE)) < 0) {
+    sprintf (LOCAL->buf,"Unable to open copy mailbox: %s",strerror (errno));
+    mm_log (LOCAL->buf,ERROR);
+    return NIL;
+  }
+  mm_critical (stream);		/* go critical */
+				/* get exclusive parse/append permission */
+  if ((ld = lockname (lock,mailbox,LOCK_EX)) < 0) {
+    mm_log ("Unable to lock copy mailbox",ERROR);
+    return NIL;
+  }
+  fstat (fd,&sbuf);		/* get current file size */
+  lseek (fd,sbuf.st_size,L_SET);/* move to end of file */
+
+				/* for each requested message */
+  for (i = 1; ret && (i <= stream->nmsgs); i++) 
+    if ((elt = mail_elt (stream,i))->sequence) {
+      lseek (LOCAL->fd,elt->private.special.offset,L_SET);
+				/* number of bytes to copy */
+      k = elt->private.special.text.size + tenex_size (stream,i);
+      do {			/* read from source position */
+	j = min (k,LOCAL->buflen);
+	read (LOCAL->fd,LOCAL->buf,j);
+	if (write (fd,LOCAL->buf,j) < 0) ret = NIL;
+      } while (ret && (k -= j));/* until done */
+    }
+				/* delete all requested messages */
+  if (ret && (options & CP_MOVE)) {
+    sprintf (LOCAL->buf,"Unable to write message: %s",strerror (errno));
+    mm_log (LOCAL->buf,ERROR);
+    ftruncate (fd,sbuf.st_size);
+  }
+  times.actime = sbuf.st_atime;	/* preserve atime and mtime */
+  times.modtime = sbuf.st_mtime;
+  utime (file,&times);		/* set the times */
+  unlockfd (ld,lock);		/* release exclusive parse/append permission */
+  close (fd);			/* close the file */
+  mm_nocritical (stream);	/* release critical */
+				/* delete all requested messages */
+  if (ret && (options & CP_MOVE)) {
+    for (i = 1; i <= stream->nmsgs; i++)
+      if ((elt = tenex_elt (stream,i))->sequence) {
+	elt->deleted = T;	/* mark message deleted */
+				/* recalculate status */
+	tenex_update_status (stream,i,NIL);
+      }
+    if (!stream->rdonly) {	/* make sure the update takes */
+      fsync (LOCAL->fd);
+      fstat (LOCAL->fd,&sbuf);	/* get current write time */
+      LOCAL->filetime = sbuf.st_mtime;
+    }
+  }
+  return ret;
 }
 
 /* Tenex mail append message from stringstruct
@@ -700,7 +782,7 @@ long tenex_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
   if (date) {			/* want to preserve date? */
 				/* yes, parse date into an elt */
     if (!mail_parse_date (&elt,date)) {
-      sprintf (tmp,"Bad date in append: %s",date);
+      sprintf (tmp,"Bad date in append: %.80s",date);
       mm_log (tmp,ERROR);
       return NIL;
     }
@@ -709,16 +791,25 @@ long tenex_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
 				/* make sure valid mailbox */
   if (!tenex_isvalid (mailbox,tmp)) switch (errno) {
   case ENOENT:			/* no such file? */
-    mm_notify (stream,"[TRYCREATE] Must create mailbox before append",NIL);
-    return NIL;
+    if (((mailbox[0] == 'I') || (mailbox[0] == 'i')) &&
+	((mailbox[1] == 'N') || (mailbox[1] == 'n')) &&
+	((mailbox[2] == 'B') || (mailbox[2] == 'b')) &&
+	((mailbox[3] == 'O') || (mailbox[3] == 'o')) &&
+	((mailbox[4] == 'X') || (mailbox[4] == 'x')) && !mailbox[5])
+      tenex_create (NIL,"INBOX");
+    else {
+      mm_notify (stream,"[TRYCREATE] Must create mailbox before append",NIL);
+      return NIL;
+    }
+				/* falls through */
   case 0:			/* merely empty file? */
     break;
   case EINVAL:
-    sprintf (tmp,"Invalid Tenex-format mailbox name: %s",mailbox);
+    sprintf (tmp,"Invalid Tenex-format mailbox name: %.80s",mailbox);
     mm_log (tmp,ERROR);
     return NIL;
   default:
-    sprintf (tmp,"Not a Tenex-format mailbox: %s",mailbox);
+    sprintf (tmp,"Not a Tenex-format mailbox: %.80s",mailbox);
     mm_log (tmp,ERROR);
     return NIL;
   }
@@ -837,7 +928,7 @@ long tenex_parse (MAILSTREAM *stream)
 				/* swell the cache */
     mail_exists (stream,++nmsgs);
 				/* intantiate an elt for this message */
-    (elt = mail_elt (stream,nmsgs))->valid = T;
+    elt = mail_elt (stream,nmsgs);
     elt->private.uid = ++stream->uid_last;
 				/* note file offset of header */
     elt->private.special.offset = curpos;
@@ -898,82 +989,6 @@ long tenex_parse (MAILSTREAM *stream)
   return LONGT;			/* return the winnage */
 }
 
-/* Tenex copy messages
- * Accepts: MAIL stream
- *	    mailbox copy vector
- *	    mailbox name
- * Returns: T if success, NIL if failed
- */
-
-long tenex_copy_messages (MAILSTREAM *stream,char *mailbox)
-{
-  struct stat sbuf;
-  struct utimbuf times;
-  MESSAGECACHE *elt;
-  unsigned long i,j,k;
-  long ret = LONGT;
-  int fd,ld;
-  char file[MAILTMPLEN],lock[MAILTMPLEN];
-				/* make sure valid mailbox */
-  if (!tenex_isvalid (mailbox,LOCAL->buf)) switch (errno) {
-  case ENOENT:			/* no such file? */
-    mm_notify (stream,"[TRYCREATE] Must create mailbox before copy",NIL);
-    return NIL;
-  case 0:			/* merely empty file? */
-    break;
-  case EINVAL:
-    sprintf (LOCAL->buf,"Invalid Tenex-format mailbox name: %s",mailbox);
-    mm_log (LOCAL->buf,ERROR);
-    return NIL;
-  default:
-    sprintf (LOCAL->buf,"Not a Tenex-format mailbox: %s",mailbox);
-    mm_log (LOCAL->buf,ERROR);
-    return NIL;
-  }
-				/* got file? */  
-  if ((fd = open (mailboxfile (file,mailbox),O_BINARY|O_RDWR|O_CREAT,
-		  S_IREAD|S_IWRITE)) < 0) {
-    sprintf (LOCAL->buf,"Unable to open copy mailbox: %s",strerror (errno));
-    mm_log (LOCAL->buf,ERROR);
-    return NIL;
-  }
-  mm_critical (stream);		/* go critical */
-				/* get exclusive parse/append permission */
-  if ((ld = lockname (lock,mailbox,LOCK_EX)) < 0) {
-    mm_log ("Unable to lock copy mailbox",ERROR);
-    return NIL;
-  }
-  fstat (fd,&sbuf);		/* get current file size */
-  lseek (fd,sbuf.st_size,L_SET);/* move to end of file */
-
-				/* for each requested message */
-  for (i = 1; ret && (i <= stream->nmsgs); i++) 
-    if ((elt = mail_elt (stream,i))->sequence) {
-      lseek (LOCAL->fd,elt->private.special.offset,L_SET);
-				/* number of bytes to copy */
-      k = elt->private.special.text.size + tenex_size (stream,i);
-      do {			/* read from source position */
-	j = min (k,LOCAL->buflen);
-	read (LOCAL->fd,LOCAL->buf,(unsigned int) j);
-	if ((write (fd,LOCAL->buf,(unsigned int) j) < 0) || fsync (fd)) {
-	  sprintf (LOCAL->buf,"Unable to write message: %s",strerror (errno));
-	  mm_log (LOCAL->buf,ERROR);
-	  ftruncate (fd,sbuf.st_size);
-	  j = k;
-	  ret = NIL;		/* note error */
-	  break;
-	}
-      } while (k -= j);		/* until done */
-    }
-  times.actime = sbuf.st_atime;	/* preserve atime and mtime */
-  times.modtime = sbuf.st_mtime;
-  utime (file,&times);		/* set the times */
-  unlockfd (ld,lock);		/* release exclusive parse/append permission */
-  close (fd);			/* close the file */
-  mm_nocritical (stream);	/* release critical */
-  return ret;
-}
-
 /* Tenex get cache element with status updating from file
  * Accepts: MAIL stream
  *	    message number
@@ -982,13 +997,37 @@ long tenex_copy_messages (MAILSTREAM *stream,char *mailbox)
 
 MESSAGECACHE *tenex_elt (MAILSTREAM *stream,unsigned long msgno)
 {
-  unsigned long i,j,sysflags;
-  char c;
   MESSAGECACHE *elt = mail_elt (stream,msgno);
-  unsigned long oldsysflags = (elt->seen ? fSEEN : NIL) |
-    (elt->deleted ? fDELETED : NIL) | (elt->flagged ? fFLAGGED : NIL) |
-      (elt->answered ? fANSWERED : NIL) | (elt->draft ? fDRAFT : NIL);
-  unsigned long olduserflags = elt->user_flags;
+  struct {			/* old flags */
+    unsigned int seen : 1;
+    unsigned int deleted : 1;
+    unsigned int flagged : 1;
+    unsigned int answered : 1;
+    unsigned int draft : 1;
+    unsigned long user_flags;
+  } old;
+  old.seen = elt->seen; old.deleted = elt->deleted; old.flagged = elt->flagged;
+  old.answered = elt->answered; old.draft = elt->draft;
+  old.user_flags = elt->user_flags;
+  tenex_read_flags (stream,elt);
+  if ((old.seen != elt->seen) || (old.deleted != elt->deleted) ||
+      (old.flagged != elt->flagged) || (old.answered != elt->answered) ||
+      (old.draft != elt->draft) || (old.user_flags != elt->user_flags))
+    mm_flags (stream,msgno);	/* let top level know */
+  return elt;
+}
+
+
+/* Tenex read flags from file
+ * Accepts: MAIL stream
+ * Returns: cache element
+ */
+
+void tenex_read_flags (MAILSTREAM *stream,MESSAGECACHE *elt)
+{
+  unsigned long i,j;
+				/* noop if readonly and have valid flags */
+  if (stream->rdonly && elt->valid) return;
 				/* set the seek pointer */
   lseek (LOCAL->fd,(off_t) elt->private.special.offset +
 	 elt->private.special.text.size - 13,L_SET);
@@ -998,23 +1037,16 @@ MESSAGECACHE *tenex_elt (MAILSTREAM *stream,unsigned long msgno)
     fatal (LOCAL->buf);
   }
 				/* calculate system flags */
-  sysflags = (((LOCAL->buf[10]-'0') * 8) + LOCAL->buf[11]-'0') &
-    (fSEEN|fDELETED|fFLAGGED|fANSWERED|fDRAFT);
-  elt->seen = sysflags & fSEEN ? T : NIL;
-  elt->deleted = sysflags & fDELETED ? T : NIL;
-  elt->flagged = sysflags & fFLAGGED ? T : NIL;
-  elt->answered = sysflags & fANSWERED ? T : NIL;
-  elt->draft = sysflags & fDRAFT ? T : NIL;
-  c = LOCAL->buf[10];		/* remember first system flags byte */
+  i = (((LOCAL->buf[10]-'0') * 8) + LOCAL->buf[11]-'0');
+  elt->seen = i & fSEEN ? T : NIL; elt->deleted = i & fDELETED ? T : NIL;
+  elt->flagged = i & fFLAGGED ? T : NIL;
+  elt->answered = i & fANSWERED ? T : NIL; elt->draft = i & fDRAFT ? T : NIL;
   LOCAL->buf[10] = '\0';	/* tie off flags */
   j = strtoul(LOCAL->buf,NIL,8);/* get user flags value */
-  LOCAL->buf[10] = c;		/* restore first system flags byte */
 				/* set up all valid user flags (reversed!) */
   while (j) if (((i = 29 - find_rightmost_bit (&j)) < NUSERFLAGS) &&
 		stream->user_flags[i]) elt->user_flags |= 1 << i;
-  if ((oldsysflags != sysflags) || (olduserflags != elt->user_flags))
-    mm_flags (stream,msgno);	/* let top level know */
-  return elt;
+  elt->valid = T;		/* have valid flags now */
 }
 
 /* Tenex update status string
@@ -1028,7 +1060,9 @@ void tenex_update_status (MAILSTREAM *stream,unsigned long msgno,long syncflag)
   MESSAGECACHE *elt = mail_elt (stream,msgno);
   struct stat sbuf;
   unsigned long j,k = 0;
-  if (!stream->rdonly) {	/* not if readonly you don't */
+				/* readonly */
+  if (stream->rdonly) tenex_read_flags (stream,elt);
+  else {			/* readwrite */
     j = elt->user_flags;	/* get user flags */
 				/* reverse bits (dontcha wish we had CIRC?) */
     while (j) k |= 1 << (29 - find_rightmost_bit (&j));
@@ -1037,15 +1071,19 @@ void tenex_update_status (MAILSTREAM *stream,unsigned long msgno,long syncflag)
 	     fOLD + (fSEEN * elt->seen) + (fDELETED * elt->deleted) +
 	     (fFLAGGED * elt->flagged) + (fANSWERED * elt->answered) +
 	     (fDRAFT * elt->draft));
-				/* get to that place in the file */
-    lseek (LOCAL->fd,(off_t) elt->private.special.offset +
-	   elt->private.special.text.size - 13,L_SET);
+    while (T) {			/* get to that place in the file */
+      lseek (LOCAL->fd,(off_t) elt->private.special.offset +
+	     elt->private.special.text.size - 13,L_SET);
 				/* write new flags */
-    write (LOCAL->fd,LOCAL->buf,12);
-				/* sync if requested */
-    if (syncflag) fsync (LOCAL->fd);
-    fstat (LOCAL->fd,&sbuf);	/* get new write time */
-    LOCAL->filetime = sbuf.st_mtime;
+      if (write (LOCAL->fd,LOCAL->buf,12) > 0) break;
+      mm_notify (stream,strerror (errno),WARN);
+      mm_diskerror (stream,errno,T);
+    }
+    if (syncflag) {		/* sync if requested */
+      fsync (LOCAL->fd);
+      fstat (LOCAL->fd,&sbuf);	/* get new write time */
+      LOCAL->filetime = sbuf.st_mtime;
+    }
   }
 }
 
