@@ -10,7 +10,7 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	20 December 1989
- * Last Edited:	3 March 1997
+ * Last Edited:	8 April 1997
  *
  * Copyright 1997 by the University of Washington
  *
@@ -475,7 +475,6 @@ char *bezerk_header (MAILSTREAM *stream,unsigned long msgno,
 long bezerk_text (MAILSTREAM *stream,unsigned long msgno,STRING *bs,
 		  long flags)
 {
-  unsigned long i;
   FILECACHE *m;
   MESSAGECACHE *elt;
 				/* UID call "impossible" */
@@ -725,19 +724,17 @@ long bezerk_copy (MAILSTREAM *stream,char *sequence,char *mailbox,long options)
 long bezerk_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
 		    STRING *message)
 {
-  struct stat sbuf;
-  int i,fd,ti,zn;
-  char *s,*x,buf[BUFLEN],file[MAILTMPLEN],lock[MAILTMPLEN];
-  time_t tp[2];
   MESSAGECACHE elt;
-  char c = '\n';
-  unsigned long j,n,ok = T;
-  time_t t = time (0);
-  unsigned long size = SIZE (message);
-  unsigned long uf;
+  struct stat sbuf;
+  int fd,ti,zn;
+  long i,ok;
+  unsigned long j,n,uf,size;
+  char c,*x,buf[BUFLEN],tmp[MAILTMPLEN],file[MAILTMPLEN],lock[MAILTMPLEN];
+  time_t tp[2],t = time (0);
   long f = mail_parse_flags (stream ? stream : &bezerkproto,flags,&uf);
-				/* parse date if given */
-  if (date && !mail_parse_date (&elt,date)) {
+				/* parse date */
+  if (!date) rfc822_date (date = tmp);
+  if (!mail_parse_date (&elt,date)) {
     sprintf (buf,"Bad date in append: %s",date);
     mm_log (buf,ERROR);
     return NIL;
@@ -749,8 +746,8 @@ long bezerk_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
       mm_notify (stream,"[TRYCREATE] Must create mailbox before append",NIL);
       return NIL;
     }
-    else break;
-  case 0:			/* merely empty file? */
+				/* falls through */
+  case 0:			/* INBOX ENOENT or empty file? */
     break;
   case EINVAL:
     sprintf (buf,"Invalid Berkeley-format mailbox name: %s",mailbox);
@@ -767,47 +764,36 @@ long bezerk_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
     mm_log (buf,ERROR);
     return NIL;
   }
-
   mm_critical (stream);		/* go critical */
   fstat (fd,&sbuf);		/* get current file size */
+
   sprintf (buf,"From %s@%s ",myusername (),mylocalhost ());
 				/* write the date given */
-  if (date) mail_cdate (buf + strlen (buf),&elt);
-  else strcat (buf,ctime (&t));	/* otherwise write the time now */
+  mail_cdate (buf + strlen (buf),&elt);
   sprintf (buf + strlen (buf),"Status: %s\nX-Status: %s%s%s%s\n",
-	   f&fSEEN ? "R" : "",f&fDELETED ? "D" : "",
-	   f&fFLAGGED ? "F" : "",f&fANSWERED ? "A" : "",f&fDRAFT ? "T" : "");
-  i = strlen (buf);		/* initial buffer space used */
-  while (ok && size--) {	/* copy text, tossing out CR's */
-				/* if at start of line */
-    if ((c == '\n') && (size > 5)) {
-      n = GETPOS (message);	/* prepend a broket if needed */
-      if ((SNX (message) == 'F') && (SNX (message) == 'r') &&
-	  (SNX (message) == 'o') && (SNX (message) == 'm') &&
-	  (SNX (message) == ' ')) {
-				/* always write widget if unconditional */
-	if (bezerk_fromwidget) ok = bezerk_append_putc (fd,buf,&i,'>');
-	else {			/* hairier test, count length of this line */
-	  for (j = 6; (j < size) && (SNX (message) != '\n'); j++);
-	  if (j < size) {	/* copy line */
-	    SETPOS (message,n);	/* restore position */
-	    x = s = (char *) fs_get (j + 1);
-	    while (j--) if ((c = SNX (message)) != '\015') *x++ = c;
-	    *x = '\0';		/* tie off line */
-	    VALID (s,x,ti,zn);	/* see if looks like need a widget */
-	    if (ti) ok = bezerk_append_putc (fd,buf,&i,'>');
-	    fs_give ((void **) &s);
-	  }
-	}
-      }
-      SETPOS (message,n);	/* restore position as needed */
+	   f&fSEEN ? "R" : "",f&fDELETED ? "D" : "",f&fFLAGGED ? "F" : "",
+	   f&fANSWERED ? "A" : "",f&fDRAFT ? "T" : "");
+				/* copy text, tossing out CRs */
+  for (i = strlen (buf), ok = T, size = SIZE (message); ok && size; ) {
+    for (j = 0, c = '\0'; size && (c != '\n') && (j < MAILTMPLEN); size--)
+      if ((c = SNX (message)) != '\015') tmp[j++] = c;
+				/* possible "From " line? */
+    if ((j > 4) && (tmp[0] == 'F') && (tmp[1] == 'r') && (tmp[2] == 'o') &&
+	(tmp[3] == 'm') && (tmp[4] == ' ')) {
+				/* see if need to write a widget */
+      if (!(ti = bezerk_fromwidget || (c != '\n'))) VALID (tmp,x,ti,zn);
+      if (ti) ok = bezerk_append_putc (fd,buf,&i,'>');
     }
-				/* copy another character */
-    if ((c = SNX (message)) != '\015') ok = bezerk_append_putc (fd,buf,&i,c);
+				/* write the line */
+    if (ok) for (n = 0; n < j; n++) ok = bezerk_append_putc (fd,buf,&i,tmp[n]);
+				/* handle pathologically long line */
+    if (ok && (c != '\n') && size) do {
+      if ((c = SNX (message)) != '\015') ok = bezerk_append_putc (fd,buf,&i,c);
+    } while (ok && --size && (c != '\n'));
   }
 				/* write trailing newline */
-  if (ok) ok = bezerk_append_putc (fd,buf,&i,'\n');
-  if (!(ok && (ok = (write (fd,buf,i) >= 0)) && (ok = !fsync (fd)))) {
+  if (!(ok && (ok = bezerk_append_putc (fd,buf,&i,'\n') &&
+	       (i ? (write (fd,buf,i) >= 0) : T) && !fsync (fd)))) {
     sprintf (buf,"Message append failed: %s",strerror (errno));
     mm_log (buf,ERROR);
     ftruncate (fd,sbuf.st_size);
@@ -819,7 +805,8 @@ long bezerk_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
   mm_nocritical (stream);	/* release critical */
   return ok;			/* return success */
 }
-
+
+
 /* Berkeley mail append character
  * Accepts: file descriptor
  *	    output buffer
@@ -828,7 +815,7 @@ long bezerk_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
  * Returns: T if append successful, else NIL
  */
 
-long bezerk_append_putc (int fd,char *s,int *i,char c)
+long bezerk_append_putc (int fd,char *s,long *i,char c)
 {
   s[(*i)++] = c;
   if (*i == BUFLEN) {		/* dump if buffer filled */
