@@ -10,7 +10,7 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	3 October 1995
- * Last Edited:	3 February 1997
+ * Last Edited:	29 December 1997
  *
  * Copyright 1997 by the University of Washington
  *
@@ -54,7 +54,7 @@ extern int errno;		/* just in case */
 
 DRIVER mbxdriver = {
   "mbx",			/* driver name */
-  DR_LOCAL|DR_MAIL,		/* driver flags */
+  DR_LOCAL|DR_MAIL|DR_CRLF,	/* driver flags */
   (DRIVER *) NIL,		/* next driver */
   mbx_valid,			/* mailbox is valid for us */
   mbx_parameters,		/* manipulate parameters */
@@ -66,7 +66,7 @@ DRIVER mbxdriver = {
   mbx_create,			/* create mailbox */
   mbx_delete,			/* delete mailbox */
   mbx_rename,			/* rename mailbox */
-  NIL,				/* status of mailbox */
+  mbx_status,			/* status of mailbox */
   mbx_open,			/* open mailbox */
   mbx_close,			/* close mailbox */
   mbx_flags,			/* fetch message "fast" attributes */
@@ -219,7 +219,7 @@ long mbx_create (MAILSTREAM *stream,char *mailbox)
     return NIL;
   }
   memset (tmp,'\0',HDRSIZE);	/* initialize header */
-  strcpy (s = tmp,"*mbx*\015\0120000000100000000\015\012");
+  sprintf (s = tmp,"*mbx*\015\012%08lx00000000\015\012",time (0));
   for (i = 0; i < NUSERFLAGS; ++i)
     sprintf (s += strlen (s),"%s\015\012",(t = default_user_flag(i)) ? t : "");
   if (write (fd,tmp,HDRSIZE) != HDRSIZE) {
@@ -302,6 +302,47 @@ long mbx_rename (MAILSTREAM *stream,char *old,char *newname)
   return ret;			/* return success */
 }
 
+/* MBX Mail status
+ * Accepts: mail stream
+ *	    mailbox name
+ *	    status flags
+ * Returns: T on success, NIL on failure
+ */
+
+long mbx_status (MAILSTREAM *stream,char *mbx,long flags)
+{
+  MAILSTATUS status;
+  unsigned long i;
+  MAILSTREAM *tstream = NIL;
+  MAILSTREAM *systream = NIL;
+				/* make temporary stream (unless this mbx) */
+  if (!stream && !(stream = tstream =
+		   mail_open (NIL,mbx,OP_READONLY|OP_SILENT))) return NIL;
+  status.flags = flags;		/* return status values */
+  status.messages = stream->nmsgs;
+  status.recent = stream->recent;
+  if (flags & SA_UNSEEN)	/* must search to get unseen messages */
+    for (i = 1,status.unseen = 0; i <= stream->nmsgs; i++)
+      if (!mail_elt (stream,i)->seen) status.unseen++;
+  status.uidnext = stream->uid_last + 1;
+  status.uidvalidity = stream->uid_validity;
+  if (!status.recent &&		/* calculate post-snarf results */
+      (systream = mail_open (NIL,sysinbox (),OP_READONLY|OP_SILENT))) {
+    status.messages += systream->nmsgs;
+    status.recent += systream->recent;
+    if (flags & SA_UNSEEN)	/* must search to get unseen messages */
+      for (i = 1; i <= systream->nmsgs; i++)
+	if (!mail_elt (systream,i)->seen) status.unseen++;
+				/* kludge but probably good enough */
+    status.uidnext += systream->nmsgs;
+  }
+				/* pass status to main program */
+  mm_status (stream,mbx,&status);
+  if (tstream) mail_close (tstream);
+  if (systream) mail_close (systream);
+  return T;			/* success */
+}
+
 /* MBX mail open
  * Accepts: stream to open
  * Returns: stream on success, NIL on failure
@@ -309,7 +350,7 @@ long mbx_rename (MAILSTREAM *stream,char *old,char *newname)
 
 MAILSTREAM *mbx_open (MAILSTREAM *stream)
 {
-  int fd,ld;
+  int fd,ld,i;
   char tmp[MAILTMPLEN];
 				/* return prototype for OP_PROTOTYPE call */
   if (!stream) return user_flags (&mbxproto);
@@ -317,6 +358,11 @@ MAILSTREAM *mbx_open (MAILSTREAM *stream)
     mbx_close (stream,NIL);	/* dump and save the changes */
     stream->dtb = &mbxdriver;	/* reattach this driver */
     mail_free_cache (stream);	/* clean up cache */
+    stream->uid_last = 0;	/* default UID validity */
+    stream->uid_validity = 0;
+				/* flush user flags */
+    for (i = 0; i < NUSERFLAGS; i++)
+      if (stream->user_flags[i]) fs_give ((void **) &stream->user_flags[i]);
   }
   if (stream->rdonly ||
       (fd = open (mbx_file (tmp,stream->mailbox),O_RDWR,NIL)) < 0) {
@@ -394,7 +440,7 @@ void mbx_close (MAILSTREAM *stream,long options)
 
 void mbx_flags (MAILSTREAM *stream,char *sequence,long flags)
 {
-  long i;
+  unsigned long i;
   if (mbx_ping (stream) && 	/* ping mailbox, get new status for messages */
       ((flags & FT_UID) ? mail_uid_sequence (stream,sequence) :
        mail_sequence (stream,sequence)))
@@ -448,6 +494,7 @@ long mbx_text (MAILSTREAM *stream,unsigned long msgno,STRING *bs,long flags)
     elt->seen = T;		/* mark message as seen */
 				/* recalculate status */
     mbx_update_status (stream,msgno,T);
+    mm_flags (stream,msgno);
   }
 				/* find header position */
   i = mbx_hdrpos (stream,msgno,&j);
@@ -458,8 +505,7 @@ long mbx_text (MAILSTREAM *stream,unsigned long msgno,STRING *bs,long flags)
     fs_give ((void **) &LOCAL->buf);
     LOCAL->buf = (char *) fs_get ((LOCAL->buflen = i) + 1);
   }
-				/* slurp the data */
-  if (read (LOCAL->fd,LOCAL->buf,i) != i) return NIL;
+  read (LOCAL->fd,LOCAL->buf,i);/* slurp the data */
   LOCAL->buf[i] = '\0';		/* tie off string */
 				/* set up stringstruct */
   INIT (bs,mail_string,LOCAL->buf,i);
@@ -511,7 +557,7 @@ void mbx_flagmsg (MAILSTREAM *stream,MESSAGECACHE *elt)
 
 long mbx_ping (MAILSTREAM *stream)
 {
-  long i = 1;
+  unsigned long i = 1;
   long r = NIL;
   int ld;
   char lock[MAILTMPLEN];
@@ -567,7 +613,7 @@ void mbx_check (MAILSTREAM *stream)
 void mbx_snarf (MAILSTREAM *stream)
 {
   unsigned long i = 0;
-  unsigned long r,hdrlen,txtlen;
+  unsigned long j,r,hdrlen,txtlen;
   struct stat sbuf;
   char *hdr,*txt,lock[MAILTMPLEN],tmp[MAILTMPLEN];
   MESSAGECACHE *elt;
@@ -577,8 +623,8 @@ void mbx_snarf (MAILSTREAM *stream)
   if ((!strcmp (sysinbox (),stream->mailbox)) ||
       ((ld = lockfd (LOCAL->fd,lock,LOCK_EX)) < 0)) return;
   mm_critical (stream);		/* go critical */
-  stat (sysinbox (),&sbuf);	/* see if anything there */
-  if (sbuf.st_size) {		/* non-empty? */
+				/* see if anything there */
+  if (!stat (sysinbox (),&sbuf) && sbuf.st_size) {
     fstat (LOCAL->fd,&sbuf);	/* yes, get current file size */
 				/* sizes match and can get sysibx mailbox? */
     if ((sbuf.st_size == LOCAL->filesize) && 
@@ -591,19 +637,22 @@ void mbx_snarf (MAILSTREAM *stream)
 				/* snarf message from system INBOX */
 	hdr = cpystr (mail_fetchheader_full (sysibx,i,NIL,&hdrlen,NIL));
 	txt = mail_fetchtext_full (sysibx,i,&txtlen,FT_PEEK);
+	if (j = hdrlen + txtlen) {
 				/* calculate header line */
-	mail_date (LOCAL->buf,elt = mail_elt (sysibx,i));
-	sprintf (LOCAL->buf + strlen (LOCAL->buf),
-		 ",%lu;00000000%04x-00000000\015\012",
-		 hdrlen + txtlen,(fSEEN * elt->seen) +
-		 (fDELETED * elt->deleted) + (fFLAGGED * elt->flagged) +
-		 (fANSWERED * elt->answered) + (fDRAFT * elt->draft));
+	  mail_date (LOCAL->buf,elt = mail_elt (sysibx,i));
+	  sprintf (LOCAL->buf + strlen (LOCAL->buf),
+		   ",%lu;00000000%04x-00000000\015\012",j,(fSEEN * elt->seen) +
+		   (fDELETED * elt->deleted) + (fFLAGGED * elt->flagged) +
+		   (fANSWERED * elt->answered) + (fDRAFT * elt->draft));
 				/* copy message */
-	if ((write (LOCAL->fd,LOCAL->buf,strlen (LOCAL->buf)) < 0) ||
-	    (write (LOCAL->fd,hdr,hdrlen) < 0) ||
-	    (write (LOCAL->fd,txt,txtlen) < 0) || fsync (LOCAL->fd)) r = 0;
+	  if ((write (LOCAL->fd,LOCAL->buf,strlen (LOCAL->buf)) < 0) ||
+	      (write (LOCAL->fd,hdr,hdrlen) < 0) ||
+	      (write (LOCAL->fd,txt,txtlen) < 0)) r = 0;
+	}
 	fs_give ((void **) &hdr);
       }
+				/* make sure all the updates take */
+      if (fsync (LOCAL->fd)) r = 0;
       if (r) {			/* delete all the messages we copied */
 	if (r == 1) strcpy (tmp,"1");
 	else sprintf (tmp,"%lu:%lu",1,r);
@@ -689,7 +738,7 @@ void mbx_expunge (MAILSTREAM *stream)
       do {			/* read from source position */
 	m = min (k,LOCAL->buflen);
 	lseek (LOCAL->fd,j,L_SET);
-	read (LOCAL->fd,LOCAL->buf,(unsigned int) m);
+	read (LOCAL->fd,LOCAL->buf,m);
 	pos = j - delta;	/* write to destination position */
 	lseek (LOCAL->fd,pos,L_SET);
 	write (LOCAL->fd,LOCAL->buf,m);
@@ -736,25 +785,92 @@ void mbx_expunge (MAILSTREAM *stream)
 
 long mbx_copy (MAILSTREAM *stream,char *sequence,char *mailbox,long options)
 {
-  long i;
   struct stat sbuf;
+  time_t tp[2];
   MESSAGECACHE *elt;
-  if (!(((options & CP_UID) ? mail_uid_sequence (stream,sequence) :
-	 mail_sequence (stream,sequence)) &&
-	mbx_copy_messages (stream,mailbox))) return NIL;
-				/* delete all requested messages */
-  if (options & CP_MOVE) for (i = 1; i <= stream->nmsgs; i++)
-    if ((elt = mbx_elt (stream,i))->sequence) {
-      elt->deleted = T;		/* mark message deleted */
-				/* recalculate status */
-      mbx_update_status (stream,i,NIL);
-    }
-  if (!stream->rdonly) {	/* make sure the update takes */
-    fsync (LOCAL->fd);
-    fstat (LOCAL->fd,&sbuf);	/* get current write time */
-    LOCAL->filetime = sbuf.st_mtime;
+  unsigned long i,j,k;
+  long ret = LONGT;
+  int fd,ld;
+  char file[MAILTMPLEN],lock[MAILTMPLEN];
+  mailproxycopy_t pc =
+    (mailproxycopy_t) mail_parameters (stream,GET_MAILPROXYCOPY,NIL);
+				/* make sure valid mailbox */
+  if (!mbx_isvalid (mailbox,LOCAL->buf)) switch (errno) {
+  case ENOENT:			/* no such file? */
+    mm_notify (stream,"[TRYCREATE] Must create mailbox before copy",NIL);
+    return NIL;
+  case EINVAL:
+    if (pc) return (*pc) (stream,sequence,mailbox,options);
+    sprintf (LOCAL->buf,"Invalid MBX-format mailbox name: %s",mailbox);
+    mm_log (LOCAL->buf,ERROR);
+    return NIL;
+  default:
+    if (pc) return (*pc) (stream,sequence,mailbox,options);
+    sprintf (LOCAL->buf,"Not a MBX-format mailbox: %s",mailbox);
+    mm_log (LOCAL->buf,ERROR);
+    return NIL;
   }
-  return LONGT;
+  if (!((options & CP_UID) ? mail_uid_sequence (stream,sequence) :
+	mail_sequence (stream,sequence))) return NIL;
+				/* got file? */  
+  if ((fd=open(mbx_file(file,mailbox),O_RDWR|O_CREAT,S_IREAD|S_IWRITE))<0) {
+    sprintf (LOCAL->buf,"Unable to open copy mailbox: %s",strerror (errno));
+    mm_log (LOCAL->buf,ERROR);
+    return NIL;
+  }
+  mm_critical (stream);		/* go critical */
+				/* get parse/append permission */
+  if ((ld = lockfd (fd,lock,LOCK_EX)) < 0) {
+    mm_log ("Unable to lock copy mailbox",ERROR);
+    return NIL;
+  }
+  fstat (fd,&sbuf);		/* get current file size */
+  lseek (fd,sbuf.st_size,L_SET);/* move to end of file */
+
+				/* for each requested message */
+  for (i = 1; ret && (i <= stream->nmsgs); i++) 
+    if ((elt = mail_elt (stream,i))->sequence) {
+      lseek (LOCAL->fd,elt->private.special.offset +
+	     elt->private.special.text.size,L_SET);
+      mail_date(LOCAL->buf,elt);/* build target header */
+      sprintf (LOCAL->buf+strlen(LOCAL->buf),",%lu;%08lx%04x-00000000\015\012",
+	       elt->rfc822_size,elt->user_flags,(fSEEN * elt->seen) +
+	       (fDELETED * elt->deleted) + (fFLAGGED * elt->flagged) +
+	       (fANSWERED * elt->answered) + (fDRAFT * elt->draft));
+				/* write target header */
+      if (ret = (write (fd,LOCAL->buf,strlen (LOCAL->buf)) > 0))
+	for (k = elt->rfc822_size; ret && (j = min (k,LOCAL->buflen)); k -= j){
+	  read (LOCAL->fd,LOCAL->buf,j);
+	  ret = write (fd,LOCAL->buf,j) >= 0;
+	}
+    }
+				/* make sure all the updates take */
+  if (!(ret && (ret = !fsync (fd)))) {
+    sprintf (LOCAL->buf,"Unable to write message: %s",strerror (errno));
+    mm_log (LOCAL->buf,ERROR);
+    ftruncate (fd,sbuf.st_size);
+  }
+  tp[0] = sbuf.st_atime;	/* preserve atime and mtime */
+  tp[1] = sbuf.st_mtime;
+  utime (file,tp);		/* set the times */
+  close (fd);			/* close the file */
+  unlockfd (ld,lock);		/* release exclusive parse/append permission */
+  mm_nocritical (stream);	/* release critical */
+				/* delete all requested messages */
+  if (ret && (options & CP_MOVE)) {
+    for (i = 1; i <= stream->nmsgs; i++)
+      if ((elt = mbx_elt (stream,i))->sequence) {
+	elt->deleted = T;	/* mark message deleted */
+				/* recalculate status */
+	mbx_update_status (stream,i,NIL);
+      }
+    if (!stream->rdonly) {	/* make sure the update takes */
+      fsync (LOCAL->fd);
+      fstat (LOCAL->fd,&sbuf);	/* get current write time */
+      LOCAL->filetime = sbuf.st_mtime;
+    }
+  }
+  return ret;
 }
 
 /* MBX mail append message from stringstruct
@@ -791,8 +907,17 @@ long mbx_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
 				/* make sure valid mailbox */
   if (!mbx_isvalid (mailbox,tmp)) switch (errno) {
   case ENOENT:			/* no such file? */
-    mm_notify (stream,"[TRYCREATE] Must create mailbox before append",NIL);
-    return NIL;
+    if (((mailbox[0] == 'I') || (mailbox[0] == 'i')) &&
+	((mailbox[1] == 'N') || (mailbox[1] == 'n')) &&
+	((mailbox[2] == 'B') || (mailbox[2] == 'b')) &&
+	((mailbox[3] == 'O') || (mailbox[3] == 'o')) &&
+	((mailbox[4] == 'X') || (mailbox[4] == 'x')) && !mailbox[5])
+      mbx_create (NIL,"INBOX");
+    else {
+      mm_notify (stream,"[TRYCREATE] Must create mailbox before append",NIL);
+      return NIL;
+    }
+				/* falls through */
   case 0:			/* merely empty file? */
     break;
   case EINVAL:
@@ -935,13 +1060,13 @@ long mbx_parse (MAILSTREAM *stream)
 				/* swell the cache */
     mail_exists (stream,++nmsgs);
 				/* intantiate an elt for this message */
-    (elt = mail_elt (stream,nmsgs))->valid = T;
+    elt = mail_elt (stream,nmsgs);
 				/* note file offset of header */
     elt->private.special.offset = curpos;
 				/* in case error */
     elt->private.special.text.size = 0;
 				/* header size not known yet */
-    elt->private.msg.header.text.size = elt->private.msg.text.text.size = 0;
+    elt->private.msg.header.text.size = 0;
 				/* parse the header components */
     if (mail_parse_date (elt,LOCAL->buf) &&
 	(elt->rfc822_size = strtoul (x = s,&s,10)) && (!(s && *s)))
@@ -999,7 +1124,8 @@ long mbx_parse (MAILSTREAM *stream)
     }
     lastuid = elt->private.uid;	/* update last parsed UID */
   }
-  mbx_update_header (stream);	/* update header */
+				/* update header */
+  if (!stream->rdonly) mbx_update_header (stream);
   fsync (LOCAL->fd);		/* make sure all the UID updates take */
 				/* update parsed file size and time */
   LOCAL->filesize = sbuf.st_size;
@@ -1011,83 +1137,6 @@ long mbx_parse (MAILSTREAM *stream)
   return LONGT;			/* return the winnage */
 }
 
-/* MBX copy messages
- * Accepts: MAIL stream
- *	    mailbox copy vector
- *	    mailbox name
- * Returns: T if success, NIL if failed
- */
-
-long mbx_copy_messages (MAILSTREAM *stream,char *mailbox)
-{
-  struct stat sbuf;
-  time_t tp[2];
-  MESSAGECACHE *elt;
-  unsigned long i,j,k;
-  long ret = LONGT;
-  int fd,ld;
-  char file[MAILTMPLEN],lock[MAILTMPLEN];
-				/* make sure valid mailbox */
-  if (!mbx_isvalid (mailbox,LOCAL->buf)) switch (errno) {
-  case ENOENT:			/* no such file? */
-    mm_notify (stream,"[TRYCREATE] Must create mailbox before copy",NIL);
-    return NIL;
-  case 0:			/* merely empty file? */
-    break;
-  case EINVAL:
-    sprintf (LOCAL->buf,"Invalid MBX-format mailbox name: %s",mailbox);
-    mm_log (LOCAL->buf,ERROR);
-    return NIL;
-  default:
-    sprintf (LOCAL->buf,"Not a MBX-format mailbox: %s",mailbox);
-    mm_log (LOCAL->buf,ERROR);
-    return NIL;
-  }
-				/* got file? */  
-  if ((fd=open(mbx_file(file,mailbox),O_RDWR|O_CREAT,S_IREAD|S_IWRITE))<0) {
-    sprintf (LOCAL->buf,"Unable to open copy mailbox: %s",strerror (errno));
-    mm_log (LOCAL->buf,ERROR);
-    return NIL;
-  }
-  mm_critical (stream);		/* go critical */
-				/* get parse/append permission */
-  if ((ld = lockfd (fd,lock,LOCK_EX)) < 0) {
-    mm_log ("Unable to lock copy mailbox",ERROR);
-    return NIL;
-  }
-  fstat (fd,&sbuf);		/* get current file size */
-  lseek (fd,sbuf.st_size,L_SET);/* move to end of file */
-
-				/* for each requested message */
-  for (i = 1; ret && (i <= stream->nmsgs); i++) 
-    if ((elt = mail_elt (stream,i))->sequence) {
-      lseek (LOCAL->fd,elt->private.special.offset +
-	     elt->private.special.text.size,L_SET);
-      mail_date(LOCAL->buf,elt);/* build target header */
-      sprintf (LOCAL->buf+strlen(LOCAL->buf),",%lu;%08lx%04x-00000000\015\012",
-	       elt->rfc822_size,elt->user_flags,(fSEEN * elt->seen) +
-	       (fDELETED * elt->deleted) + (fFLAGGED * elt->flagged) +
-	       (fANSWERED * elt->answered) + (fDRAFT * elt->draft));
-				/* write target header */
-      if (ret = (write (fd,LOCAL->buf,strlen (LOCAL->buf)) > 0) && !fsync (fd))
-	for (k = elt->rfc822_size; ret && (j = min (k,LOCAL->buflen)); k -= j)
-	  ret = (read (LOCAL->fd,LOCAL->buf,j) == j) &&
-	    (write (fd,LOCAL->buf,j) == j) && !fsync (fd);
-    }
-  if (!ret) {			/* if error */
-    sprintf (LOCAL->buf,"Unable to write message: %s",strerror (errno));
-    mm_log (LOCAL->buf,ERROR);
-    ftruncate (fd,sbuf.st_size);
-  }
-  tp[0] = sbuf.st_atime;	/* preserve atime and mtime */
-  tp[1] = sbuf.st_mtime;
-  utime (file,tp);		/* set the times */
-  close (fd);			/* close the file */
-  unlockfd (ld,lock);		/* release exclusive parse/append permission */
-  mm_nocritical (stream);	/* release critical */
-  return ret;
-}
-
 /* MBX get cache element with status updating from file
  * Accepts: MAIL stream
  *	    message number
@@ -1096,13 +1145,36 @@ long mbx_copy_messages (MAILSTREAM *stream,char *mailbox)
 
 MESSAGECACHE *mbx_elt (MAILSTREAM *stream,unsigned long msgno)
 {
-  unsigned long sysflags;
-  char c;
   MESSAGECACHE *elt = mail_elt (stream,msgno);
-  unsigned long oldsysflags = (elt->seen ? fSEEN : NIL) |
-    (elt->deleted ? fDELETED : NIL) | (elt->flagged ? fFLAGGED : NIL) |
-      (elt->answered ? fANSWERED : NIL) | (elt->draft ? fDRAFT : NIL);
-  unsigned long olduserflags = elt->user_flags;
+  struct {			/* old flags */
+    unsigned int seen : 1;
+    unsigned int deleted : 1;
+    unsigned int flagged : 1;
+    unsigned int answered : 1;
+    unsigned int draft : 1;
+    unsigned long user_flags;
+  } old;
+  old.seen = elt->seen; old.deleted = elt->deleted; old.flagged = elt->flagged;
+  old.answered = elt->answered; old.draft = elt->draft;
+  old.user_flags = elt->user_flags;
+  mbx_read_flags (stream,elt);
+  if ((old.seen != elt->seen) || (old.deleted != elt->deleted) ||
+      (old.flagged != elt->flagged) || (old.answered != elt->answered) ||
+      (old.draft != elt->draft) || (old.user_flags != elt->user_flags))
+    mm_flags (stream,msgno);	/* let top level know */
+  return elt;
+}
+
+/* MBX read flags from file
+ * Accepts: MAIL stream
+ * Returns: cache element
+ */
+
+void mbx_read_flags (MAILSTREAM *stream,MESSAGECACHE *elt)
+{
+  unsigned long i;
+				/* noop if readonly and have valid flags */
+  if (stream->rdonly && elt->valid) return;
 				/* set the seek pointer */
   lseek (LOCAL->fd,(off_t) elt->private.special.offset +
 	 elt->private.special.text.size - 23,L_SET);
@@ -1113,20 +1185,16 @@ MESSAGECACHE *mbx_elt (MAILSTREAM *stream,unsigned long msgno)
   }
   LOCAL->buf[12] = '\0';	/* tie off buffer */
 				/* calculate system flags */
-  sysflags = strtoul (LOCAL->buf+8,NIL,16);
-  elt->seen = sysflags & fSEEN ? T : NIL;
-  elt->deleted = sysflags & fDELETED ? T : NIL;
-  elt->flagged = sysflags & fFLAGGED ? T : NIL;
-  elt->answered = sysflags & fANSWERED ? T : NIL;
-  elt->draft = sysflags & fDRAFT ? T : NIL;
-  c = LOCAL->buf[8];		/* remember first system flags byte */
+  i = strtoul (LOCAL->buf+8,NIL,16);
+  elt->seen = i & fSEEN ? T : NIL;
+  elt->deleted = i & fDELETED ? T : NIL;
+  elt->flagged = i & fFLAGGED ? T : NIL;
+  elt->answered = i & fANSWERED ? T : NIL;
+  elt->draft = i & fDRAFT ? T : NIL;
   LOCAL->buf[8] = '\0';		/* tie off flags */
 				/* get user flags value */
   elt->user_flags = strtoul (LOCAL->buf,NIL,16);
-  LOCAL->buf[8] = c;		/* restore first system flags byte */
-  if ((oldsysflags != sysflags) || (olduserflags != elt->user_flags))
-    mm_flags (stream,msgno);	/* let top level know */
-  return elt;
+  elt->valid = T;		/* have valid flags now */
 }
 
 /* MBX update header
@@ -1159,7 +1227,9 @@ void mbx_update_status (MAILSTREAM *stream,unsigned long msgno,long syncflag)
 {
   MESSAGECACHE *elt = mail_elt (stream,msgno);
   struct stat sbuf;
-  if (!stream->rdonly) {	/* not if readonly you don't */
+				/* readonly */
+  if (stream->rdonly) mbx_read_flags (stream,elt);
+  else {			/* readwrite */
     if (LOCAL->newkeyword) {	/* defined a new keyword? */
       mbx_update_header (stream);
       LOCAL->newkeyword = NIL;	/* don't do this again */
@@ -1192,7 +1262,7 @@ void mbx_update_status (MAILSTREAM *stream,unsigned long msgno,long syncflag)
 unsigned long mbx_hdrpos (MAILSTREAM *stream,unsigned long msgno,
 			  unsigned long *size)
 {
-  long siz;
+  unsigned long siz;
   long i = 0;
   int q = 0;
   char *s,tmp[MAILTMPLEN];
