@@ -10,9 +10,9 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	22 May 1990
- * Last Edited:	15 January 1997
+ * Last Edited:	29 December 1997
  *
- * Copyright 1996 by the University of Washington
+ * Copyright 1997 by the University of Washington
  *
  *  Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose and without fee is hereby granted, provided
@@ -54,7 +54,7 @@ extern int errno;		/* just in case */
 
 DRIVER mtxdriver = {
   "mtx",			/* driver name */
-  DR_LOCAL|DR_MAIL,		/* driver flags */
+  DR_LOCAL|DR_MAIL|DR_CRLF,	/* driver flags */
   (DRIVER *) NIL,		/* next driver */
   mtx_valid,			/* mailbox is valid for us */
   mtx_parameters,		/* manipulate parameters */
@@ -66,7 +66,7 @@ DRIVER mtxdriver = {
   mtx_create,			/* create mailbox */
   mtx_delete,			/* delete mailbox */
   mtx_rename,			/* rename mailbox */
-  NIL,				/* status of mailbox */
+  mtx_status,			/* status of mailbox */
   mtx_open,			/* open mailbox */
   mtx_close,			/* close mailbox */
   mtx_flags,			/* fetch message "fast" attributes */
@@ -275,6 +275,47 @@ long mtx_rename (MAILSTREAM *stream,char *old,char *newname)
   return ret;			/* return success */
 }
 
+/* Mtx Mail status
+ * Accepts: mail stream
+ *	    mailbox name
+ *	    status flags
+ * Returns: T on success, NIL on failure
+ */
+
+long mtx_status (MAILSTREAM *stream,char *mbx,long flags)
+{
+  MAILSTATUS status;
+  unsigned long i;
+  MAILSTREAM *tstream = NIL;
+  MAILSTREAM *systream = NIL;
+				/* make temporary stream (unless this mbx) */
+  if (!stream && !(stream = tstream =
+		   mail_open (NIL,mbx,OP_READONLY|OP_SILENT))) return NIL;
+  status.flags = flags;		/* return status values */
+  status.messages = stream->nmsgs;
+  status.recent = stream->recent;
+  if (flags & SA_UNSEEN)	/* must search to get unseen messages */
+    for (i = 1,status.unseen = 0; i <= stream->nmsgs; i++)
+      if (!mail_elt (stream,i)->seen) status.unseen++;
+  status.uidnext = stream->uid_last + 1;
+  status.uidvalidity = stream->uid_validity;
+  if (!status.recent &&		/* calculate post-snarf results */
+      (systream = mail_open (NIL,sysinbox (),OP_READONLY|OP_SILENT))) {
+    status.messages += systream->nmsgs;
+    status.recent += systream->recent;
+    if (flags & SA_UNSEEN)	/* must search to get unseen messages */
+      for (i = 1; i <= systream->nmsgs; i++)
+	if (!mail_elt (systream,i)->seen) status.unseen++;
+				/* kludge but probably good enough */
+    status.uidnext += systream->nmsgs;
+  }
+				/* pass status to main program */
+  mm_status (stream,mbx,&status);
+  if (tstream) mail_close (tstream);
+  if (systream) mail_close (systream);
+  return T;			/* success */
+}
+
 /* MTX mail open
  * Accepts: stream to open
  * Returns: stream on success, NIL on failure
@@ -282,7 +323,7 @@ long mtx_rename (MAILSTREAM *stream,char *old,char *newname)
 
 MAILSTREAM *mtx_open (MAILSTREAM *stream)
 {
-  int fd,ld;
+  int fd,ld,i;
   char tmp[MAILTMPLEN];
 				/* return prototype for OP_PROTOTYPE call */
   if (!stream) return user_flags (&mtxproto);
@@ -290,6 +331,11 @@ MAILSTREAM *mtx_open (MAILSTREAM *stream)
     mtx_close (stream,NIL);	/* dump and save the changes */
     stream->dtb = &mtxdriver;	/* reattach this driver */
     mail_free_cache (stream);	/* clean up cache */
+    stream->uid_last = 0;	/* default UID validity */
+    stream->uid_validity = time (0);
+				/* flush user flags */
+    for (i = 0; i < NUSERFLAGS; i++)
+      if (stream->user_flags[i]) fs_give ((void **) &stream->user_flags[i]);
   }
   user_flags (stream);		/* set up user flags */
   if (stream->rdonly ||
@@ -322,7 +368,6 @@ MAILSTREAM *mtx_open (MAILSTREAM *stream)
   LOCAL->filetime = 0;		/* time not set up yet */
   LOCAL->mustcheck = LOCAL->shouldcheck = NIL;
   stream->sequence++;		/* bump sequence number */
-  stream->uid_validity = time (0);
 				/* parse mailbox */
   stream->nmsgs = stream->recent = 0;
   if (mtx_ping (stream) && !stream->nmsgs)
@@ -421,6 +466,7 @@ long mtx_text (MAILSTREAM *stream,unsigned long msgno,STRING *bs,long flags)
     elt->seen = T;		/* mark message as seen */
 				/* recalculate status */
     mtx_update_status (stream,msgno,T);
+    mm_flags (stream,msgno);
   }
 				/* find header position */
   i = mtx_hdrpos (stream,msgno,&j);
@@ -431,8 +477,7 @@ long mtx_text (MAILSTREAM *stream,unsigned long msgno,STRING *bs,long flags)
     fs_give ((void **) &LOCAL->buf);
     LOCAL->buf = (char *) fs_get ((LOCAL->buflen = i) + 1);
   }
-				/* slurp the data */
-  if (read (LOCAL->fd,LOCAL->buf,i) != i) return NIL;
+  read (LOCAL->fd,LOCAL->buf,i);/* slurp the data */
   LOCAL->buf[i] = '\0';		/* tie off string */
 				/* set up stringstruct */
   INIT (bs,mail_string,LOCAL->buf,i);
@@ -482,7 +527,7 @@ void mtx_flagmsg (MAILSTREAM *stream,MESSAGECACHE *elt)
 
 long mtx_ping (MAILSTREAM *stream)
 {
-  long i = 1;
+  unsigned long i = 1;
   long r = NIL;
   int ld;
   char lock[MAILTMPLEN];
@@ -538,7 +583,7 @@ void mtx_check (MAILSTREAM *stream)
 void mtx_snarf (MAILSTREAM *stream)
 {
   unsigned long i = 0;
-  unsigned long r,hdrlen,txtlen;
+  unsigned long j,r,hdrlen,txtlen;
   struct stat sbuf;
   char *hdr,*txt,lock[MAILTMPLEN],tmp[MAILTMPLEN];
   MESSAGECACHE *elt;
@@ -548,8 +593,8 @@ void mtx_snarf (MAILSTREAM *stream)
   if ((!strcmp (sysinbox (),stream->mailbox)) ||
       ((ld = lockfd (LOCAL->fd,lock,LOCK_EX)) < 0)) return;
   mm_critical (stream);		/* go critical */
-  stat (sysinbox (),&sbuf);	/* see if anything there */
-  if (sbuf.st_size) {		/* non-empty? */
+				/* see if anything there */
+  if (!stat (sysinbox (),&sbuf) && sbuf.st_size) {
     fstat (LOCAL->fd,&sbuf);	/* yes, get current file size */
 				/* sizes match and can get sysibx mailbox? */
     if ((sbuf.st_size == LOCAL->filesize) && 
@@ -562,19 +607,22 @@ void mtx_snarf (MAILSTREAM *stream)
 				/* snarf message from system INBOX */
 	hdr = cpystr (mail_fetchheader_full (sysibx,i,NIL,&hdrlen,NIL));
 	txt = mail_fetchtext_full (sysibx,i,&txtlen,FT_PEEK);
+	if (j = hdrlen + txtlen) {
 				/* calculate header line */
-	mail_date (LOCAL->buf,elt = mail_elt (sysibx,i));
-	sprintf (LOCAL->buf + strlen (LOCAL->buf),
-		 ",%ld;0000000000%02o\015\012",
-		 hdrlen + txtlen,(fSEEN * elt->seen) +
-		 (fDELETED * elt->deleted) + (fFLAGGED * elt->flagged) +
-		 (fANSWERED * elt->answered) + (fDRAFT * elt->draft));
+	  mail_date (LOCAL->buf,elt = mail_elt (sysibx,i));
+	  sprintf (LOCAL->buf + strlen (LOCAL->buf),
+		   ",%ld;0000000000%02o\015\012",j,(fSEEN * elt->seen) +
+		   (fDELETED * elt->deleted) + (fFLAGGED * elt->flagged) +
+		   (fANSWERED * elt->answered) + (fDRAFT * elt->draft));
 				/* copy message */
-	if ((write (LOCAL->fd,LOCAL->buf,strlen (LOCAL->buf)) < 0) ||
-	    (write (LOCAL->fd,hdr,hdrlen) < 0) ||
-	    (write (LOCAL->fd,txt,txtlen) < 0) || fsync (LOCAL->fd)) r = 0;
+	  if ((write (LOCAL->fd,LOCAL->buf,strlen (LOCAL->buf)) < 0) ||
+	      (write (LOCAL->fd,hdr,hdrlen) < 0) ||
+	      (write (LOCAL->fd,txt,txtlen) < 0)) r = 0;
+	}
 	fs_give ((void **) &hdr);
       }
+				/* make sure all the updates take */
+      if (fsync (LOCAL->fd)) r = 0;
       if (r) {			/* delete all the messages we copied */
 	if (r == 1) strcpy (tmp,"1");
 	else sprintf (tmp,"%lu:%lu",1,r);
@@ -660,7 +708,7 @@ void mtx_expunge (MAILSTREAM *stream)
       do {			/* read from source position */
 	m = min (k,LOCAL->buflen);
 	lseek (LOCAL->fd,j,L_SET);
-	read (LOCAL->fd,LOCAL->buf,(unsigned int) m);
+	read (LOCAL->fd,LOCAL->buf,m);
 	pos = j - delta;	/* write to destination position */
 	lseek (LOCAL->fd,pos,L_SET);
 	write (LOCAL->fd,LOCAL->buf,m);
@@ -707,25 +755,89 @@ void mtx_expunge (MAILSTREAM *stream)
 
 long mtx_copy (MAILSTREAM *stream,char *sequence,char *mailbox,long options)
 {
-  unsigned long i;
   struct stat sbuf;
+  time_t tp[2];
   MESSAGECACHE *elt;
-  if (!(((options & CP_UID) ? mail_uid_sequence (stream,sequence) :
-	 mail_sequence (stream,sequence)) &&
-	mtx_copy_messages (stream,mailbox))) return NIL;
-				/* delete all requested messages */
-  if (options & CP_MOVE) for (i = 1; i <= stream->nmsgs; i++)
-    if ((elt = mtx_elt (stream,i))->sequence) {
-      elt->deleted = T;		/* mark message deleted */
-				/* recalculate status */
-      mtx_update_status (stream,i,NIL);
-    }
-  if (!stream->rdonly) {	/* make sure the update takes */
-    fsync (LOCAL->fd);
-    fstat (LOCAL->fd,&sbuf);	/* get current write time */
-    LOCAL->filetime = sbuf.st_mtime;
+  unsigned long i,j,k;
+  long ret = LONGT;
+  int fd,ld;
+  char file[MAILTMPLEN],lock[MAILTMPLEN];
+  mailproxycopy_t pc =
+    (mailproxycopy_t) mail_parameters (stream,GET_MAILPROXYCOPY,NIL);
+				/* make sure valid mailbox */
+  if (!mtx_isvalid (mailbox,LOCAL->buf)) switch (errno) {
+  case ENOENT:			/* no such file? */
+    mm_notify (stream,"[TRYCREATE] Must create mailbox before copy",NIL);
+    return NIL;
+  case 0:			/* merely empty file? */
+    break;
+  case EINVAL:
+    if (pc) return (*pc) (stream,sequence,mailbox,options);
+    sprintf (LOCAL->buf,"Invalid MTX-format mailbox name: %s",mailbox);
+    mm_log (LOCAL->buf,ERROR);
+    return NIL;
+  default:
+    if (pc) return (*pc) (stream,sequence,mailbox,options);
+    sprintf (LOCAL->buf,"Not a MTX-format mailbox: %s",mailbox);
+    mm_log (LOCAL->buf,ERROR);
+    return NIL;
   }
-  return LONGT;
+  if (!((options & CP_UID) ? mail_uid_sequence (stream,sequence) :
+	mail_sequence (stream,sequence))) return NIL;
+				/* got file? */  
+  if ((fd=  open (mtx_file (file,mailbox),O_RDWR|O_CREAT,S_IREAD|S_IWRITE))<0){
+    sprintf (LOCAL->buf,"Unable to open copy mailbox: %s",strerror (errno));
+    mm_log (LOCAL->buf,ERROR);
+    return NIL;
+  }
+  mm_critical (stream);		/* go critical */
+				/* get exclusive parse/append permission */
+  if ((ld = lockfd (fd,lock,LOCK_EX)) < 0) {
+    mm_log ("Unable to lock copy mailbox",ERROR);
+    return NIL;
+  }
+  fstat (fd,&sbuf);		/* get current file size */
+  lseek (fd,sbuf.st_size,L_SET);/* move to end of file */
+
+				/* for each requested message */
+  for (i = 1; ret && (i <= stream->nmsgs); i++) 
+    if ((elt = mail_elt (stream,i))->sequence) {
+      lseek (LOCAL->fd,elt->private.special.offset,L_SET);
+				/* number of bytes to copy */
+      k = elt->private.special.text.size + elt->rfc822_size;
+      do {			/* read from source position */
+	j = min (k,LOCAL->buflen);
+	read (LOCAL->fd,LOCAL->buf,j);
+	if (write (fd,LOCAL->buf,j) < 0) ret = NIL;
+      } while (ret && (k -= j));/* until done */
+    }
+				/* make sure all the updates take */
+  if (!(ret && (ret = !fsync (fd)))) {
+    sprintf (LOCAL->buf,"Unable to write message: %s",strerror (errno));
+    mm_log (LOCAL->buf,ERROR);
+    ftruncate (fd,sbuf.st_size);
+  }
+  tp[0] = sbuf.st_atime;	/* preserve atime and mtime */
+  tp[1] = sbuf.st_mtime;
+  utime (file,tp);		/* set the times */
+  close (fd);			/* close the file */
+  unlockfd (ld,lock);		/* release exclusive parse/append permission */
+  mm_nocritical (stream);	/* release critical */
+				/* delete all requested messages */
+  if (ret && (options & CP_MOVE)) {
+    for (i = 1; i <= stream->nmsgs; i++)
+      if ((elt = mtx_elt (stream,i))->sequence) {
+	elt->deleted = T;	/* mark message deleted */
+				/* recalculate status */
+	mtx_update_status (stream,i,NIL);
+      }
+    if (!stream->rdonly) {	/* make sure the update takes */
+      fsync (LOCAL->fd);
+      fstat (LOCAL->fd,&sbuf);	/* get current write time */
+      LOCAL->filetime = sbuf.st_mtime;
+    }
+  }
+  return ret;
 }
 
 /* MTX mail append message from stringstruct
@@ -747,14 +859,14 @@ long mtx_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
   long size = SIZE (message);
   long ret = LONGT;
   unsigned long uf = 0,j;
-  short f = mail_parse_flags (stream ? stream : user_flags (&mtxproto),
-			      flags,&j);
+  long f = mail_parse_flags (stream ? stream : user_flags (&mtxproto),
+			     flags,&j);
 				/* reverse bits (dontcha wish we had CIRC?) */
   while (j) uf |= 1 << (29 - find_rightmost_bit (&j));
   if (date) {			/* want to preserve date? */
 				/* yes, parse date into an elt */
     if (!mail_parse_date (&elt,date)) {
-      sprintf (tmp,"Bad date in append: %s",date);
+      sprintf (tmp,"Bad date in append: %.80s",date);
       mm_log (tmp,ERROR);
       return NIL;
     }
@@ -763,16 +875,25 @@ long mtx_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
 				/* make sure valid mailbox */
   if (!mtx_isvalid (mailbox,tmp)) switch (errno) {
   case ENOENT:			/* no such file? */
-    mm_notify (stream,"[TRYCREATE] Must create mailbox before append",NIL);
-    return NIL;
+    if (((mailbox[0] == 'I') || (mailbox[0] == 'i')) &&
+	((mailbox[1] == 'N') || (mailbox[1] == 'n')) &&
+	((mailbox[2] == 'B') || (mailbox[2] == 'b')) &&
+	((mailbox[3] == 'O') || (mailbox[3] == 'o')) &&
+	((mailbox[4] == 'X') || (mailbox[4] == 'x')) && !mailbox[5])
+      mtx_create (NIL,"INBOX.MTX");
+    else {
+      mm_notify (stream,"[TRYCREATE] Must create mailbox before append",NIL);
+      return NIL;
+    }
+				/* falls through */
   case 0:			/* merely empty file? */
     break;
   case EINVAL:
-    sprintf (tmp,"Invalid MTX-format mailbox name: %s",mailbox);
+    sprintf (tmp,"Invalid MTX-format mailbox name: %.80s",mailbox);
     mm_log (tmp,ERROR);
     return NIL;
   default:
-    sprintf (tmp,"Not a MTX-format mailbox: %s",mailbox);
+    sprintf (tmp,"Not a MTX-format mailbox: %.80s",mailbox);
     mm_log (tmp,ERROR);
     return NIL;
   }
@@ -796,7 +917,7 @@ long mtx_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
   if (date) mail_date(tmp,&elt);/* write preseved date */
   else internal_date (tmp);	/* get current date in IMAP format */
 				/* add remainder of header */
-  sprintf (tmp+26,",%ld;%010lo%02o\015\012",size,uf,f);
+  sprintf (tmp+26,",%ld;%010lo%02lo\015\012",size,uf,f);
 				/* write header */
   if ((write (fd,tmp,strlen (tmp)) < 0) || ((write (fd,s,size)) < 0) ||
       fsync (fd)) {
@@ -890,14 +1011,14 @@ long mtx_parse (MAILSTREAM *stream)
 				/* swell the cache */
     mail_exists (stream,++nmsgs);
 				/* intantiate an elt for this message */
-    (elt = mail_elt (stream,nmsgs))->valid = T;
+    elt = mail_elt (stream,nmsgs);
     elt->private.uid = ++stream->uid_last;
 				/* note file offset of header */
     elt->private.special.offset = curpos;
 				/* in case error */
     elt->private.special.text.size = 0;
 				/* header size not known yet */
-    elt->private.msg.header.text.size = elt->private.msg.text.text.size = 0;
+    elt->private.msg.header.text.size = 0;
 				/* parse the header components */
     if (mail_parse_date (elt,LOCAL->buf) &&
 	(elt->rfc822_size = strtoul (x = s,&s,10)) && (!(s && *s)) &&
@@ -909,7 +1030,7 @@ long mtx_parse (MAILSTREAM *stream)
     if (elt->private.special.text.size) elt->private.msg.header.offset =
       elt->private.msg.text.offset = 
 	elt->private.special.offset + elt->private.special.text.size;
-    else {
+    else {			/* oops */
       sprintf (tmp,"Unable to parse internal header elements at %ld: %s,%s;%s",
 	       curpos,LOCAL->buf,x,t);
       mm_log (tmp,ERROR);
@@ -953,81 +1074,6 @@ long mtx_parse (MAILSTREAM *stream)
   return LONGT;			/* return the winnage */
 }
 
-/* MTX copy messages
- * Accepts: MAIL stream
- *	    mailbox copy vector
- *	    mailbox name
- * Returns: T if success, NIL if failed
- */
-
-long mtx_copy_messages (MAILSTREAM *stream,char *mailbox)
-{
-  struct stat sbuf;
-  time_t tp[2];
-  MESSAGECACHE *elt;
-  unsigned long i,j,k;
-  long ret = LONGT;
-  int fd,ld;
-  char file[MAILTMPLEN],lock[MAILTMPLEN];
-				/* make sure valid mailbox */
-  if (!mtx_isvalid (mailbox,LOCAL->buf)) switch (errno) {
-  case ENOENT:			/* no such file? */
-    mm_notify (stream,"[TRYCREATE] Must create mailbox before copy",NIL);
-    return NIL;
-  case 0:			/* merely empty file? */
-    break;
-  case EINVAL:
-    sprintf (LOCAL->buf,"Invalid MTX-format mailbox name: %s",mailbox);
-    mm_log (LOCAL->buf,ERROR);
-    return NIL;
-  default:
-    sprintf (LOCAL->buf,"Not a MTX-format mailbox: %s",mailbox);
-    mm_log (LOCAL->buf,ERROR);
-    return NIL;
-  }
-				/* got file? */  
-  if ((fd=open(mtx_file(file,mailbox),O_RDWR|O_CREAT,S_IREAD|S_IWRITE))<0) {
-    sprintf (LOCAL->buf,"Unable to open copy mailbox: %s",strerror (errno));
-    mm_log (LOCAL->buf,ERROR);
-    return NIL;
-  }
-  mm_critical (stream);		/* go critical */
-				/* get exclusive parse/append permission */
-  if ((ld = lockfd (fd,lock,LOCK_EX)) < 0) {
-    mm_log ("Unable to lock copy mailbox",ERROR);
-    return NIL;
-  }
-  fstat (fd,&sbuf);		/* get current file size */
-  lseek (fd,sbuf.st_size,L_SET);/* move to end of file */
-
-				/* for each requested message */
-  for (i = 1; ret && (i <= stream->nmsgs); i++) 
-    if ((elt = mail_elt (stream,i))->sequence) {
-      lseek (LOCAL->fd,elt->private.special.offset,L_SET);
-				/* number of bytes to copy */
-      k = elt->private.special.text.size + elt->rfc822_size;
-      do {			/* read from source position */
-	j = min (k,LOCAL->buflen);
-	read (LOCAL->fd,LOCAL->buf,(unsigned int) j);
-	if ((write (fd,LOCAL->buf,j) < 0) || fsync (fd)) {
-	  sprintf (LOCAL->buf,"Unable to write message: %s",strerror (errno));
-	  mm_log (LOCAL->buf,ERROR);
-	  ftruncate (fd,sbuf.st_size);
-	  j = k;
-	  ret = NIL;		/* note error */
-	  break;
-	}
-      } while (k -= j);		/* until done */
-    }
-  tp[0] = sbuf.st_atime;	/* preserve atime and mtime */
-  tp[1] = sbuf.st_mtime;
-  utime (file,tp);		/* set the times */
-  close (fd);			/* close the file */
-  unlockfd (ld,lock);		/* release exclusive parse/append permission */
-  mm_nocritical (stream);	/* release critical */
-  return ret;
-}
-
 /* MTX get cache element with status updating from file
  * Accepts: MAIL stream
  *	    message number
@@ -1036,13 +1082,36 @@ long mtx_copy_messages (MAILSTREAM *stream,char *mailbox)
 
 MESSAGECACHE *mtx_elt (MAILSTREAM *stream,unsigned long msgno)
 {
-  unsigned long i,j,sysflags;
-  char c;
   MESSAGECACHE *elt = mail_elt (stream,msgno);
-  unsigned long oldsysflags = (elt->seen ? fSEEN : NIL) |
-    (elt->deleted ? fDELETED : NIL) | (elt->flagged ? fFLAGGED : NIL) |
-      (elt->answered ? fANSWERED : NIL) | (elt->draft ? fDRAFT : NIL);
-  unsigned long olduserflags = elt->user_flags;
+  struct {			/* old flags */
+    unsigned int seen : 1;
+    unsigned int deleted : 1;
+    unsigned int flagged : 1;
+    unsigned int answered : 1;
+    unsigned int draft : 1;
+    unsigned long user_flags;
+  } old;
+  old.seen = elt->seen; old.deleted = elt->deleted; old.flagged = elt->flagged;
+  old.answered = elt->answered; old.draft = elt->draft;
+  old.user_flags = elt->user_flags;
+  mtx_read_flags (stream,elt);
+  if ((old.seen != elt->seen) || (old.deleted != elt->deleted) ||
+      (old.flagged != elt->flagged) || (old.answered != elt->answered) ||
+      (old.draft != elt->draft) || (old.user_flags != elt->user_flags))
+    mm_flags (stream,msgno);	/* let top level know */
+  return elt;
+}
+
+/* MTX read flags from file
+ * Accepts: MAIL stream
+ * Returns: cache element
+ */
+
+void mtx_read_flags (MAILSTREAM *stream,MESSAGECACHE *elt)
+{
+  unsigned long i,j;
+				/* noop if readonly and have valid flags */
+  if (stream->rdonly && elt->valid) return;
 				/* set the seek pointer */
   lseek (LOCAL->fd,(off_t) elt->private.special.offset +
 	 elt->private.special.text.size - 14,L_SET);
@@ -1052,23 +1121,16 @@ MESSAGECACHE *mtx_elt (MAILSTREAM *stream,unsigned long msgno)
     fatal (LOCAL->buf);
   }
 				/* calculate system flags */
-  sysflags = (((LOCAL->buf[10]-'0') * 8) + LOCAL->buf[11]-'0') &
-    (fSEEN|fDELETED|fFLAGGED|fANSWERED|fDRAFT);
-  elt->seen = sysflags & fSEEN ? T : NIL;
-  elt->deleted = sysflags & fDELETED ? T : NIL;
-  elt->flagged = sysflags & fFLAGGED ? T : NIL;
-  elt->answered = sysflags & fANSWERED ? T : NIL;
-  elt->draft = sysflags & fDRAFT ? T : NIL;
-  c = LOCAL->buf[10];		/* remember first system flags byte */
+  i = (((LOCAL->buf[10]-'0') * 8) + LOCAL->buf[11]-'0');
+  elt->seen = i & fSEEN ? T : NIL; elt->deleted = i & fDELETED ? T : NIL;
+  elt->flagged = i & fFLAGGED ? T : NIL;
+  elt->answered = i & fANSWERED ? T : NIL; elt->draft = i & fDRAFT ? T : NIL;
   LOCAL->buf[10] = '\0';	/* tie off flags */
   j = strtoul(LOCAL->buf,NIL,8);/* get user flags value */
-  LOCAL->buf[10] = c;		/* restore first system flags byte */
 				/* set up all valid user flags (reversed!) */
   while (j) if (((i = 29 - find_rightmost_bit (&j)) < NUSERFLAGS) &&
 		stream->user_flags[i]) elt->user_flags |= 1 << i;
-  if ((oldsysflags != sysflags) || (olduserflags != elt->user_flags))
-    mm_flags (stream,msgno);	/* let top level know */
-  return elt;
+  elt->valid = T;		/* have valid flags now */
 }
 
 /* MTX update status string
@@ -1082,7 +1144,9 @@ void mtx_update_status (MAILSTREAM *stream,unsigned long msgno,long syncflag)
   MESSAGECACHE *elt = mail_elt (stream,msgno);
   struct stat sbuf;
   unsigned long j,k = 0;
-  if (!stream->rdonly) {	/* not if readonly you don't */
+				/* readonly */
+  if (stream->rdonly) mtx_read_flags (stream,elt);
+  else {			/* readwrite */
     j = elt->user_flags;	/* get user flags */
 				/* reverse bits (dontcha wish we had CIRC?) */
     while (j) k |= 1 << (29 - find_rightmost_bit (&j));
@@ -1114,7 +1178,7 @@ void mtx_update_status (MAILSTREAM *stream,unsigned long msgno,long syncflag)
 unsigned long mtx_hdrpos (MAILSTREAM *stream,unsigned long msgno,
 			  unsigned long *size)
 {
-  long siz;
+  unsigned long siz;
   long i = 0;
   int q = 0;
   char *s,tmp[MAILTMPLEN];
