@@ -3,7 +3,8 @@
  *
  *	This file contains basic support for handling addresses.
  *
- * TkRat software and its included text is Copyright 1996 by Martin Forssen.
+ * TkRat software and its included text is Copyright 1996,1997,1998
+ * by Martin Forssn
  *
  * The full text of the legal notice is contained in the file called
  * COPYRIGHT, included with this distribution.
@@ -11,12 +12,17 @@
 
 #include "rat.h"
 
+#define PADDED(n)	(n+(4-((int)(n)%4))%4)
+
 /*
  * This struct defines an alias.
  */
 typedef struct {
+    char *book;		/* Address book this alias comes from */
     char *fullname;	/* Long name of alias (phrase part) */
     char *content;	/* Content that alias expands to */
+    char *comment;	/* Comment to alias */
+    unsigned long mark;	/* Id of last use */
 } AliasInfo;
 
 /*
@@ -35,6 +41,70 @@ Tcl_HashTable userTable;
  */
 static int numAddresses = 0;
 
+/*
+ * A mark used to prevent loops when resolving aliases
+ */
+static unsigned long aliasMark = 0;
+
+/*
+ * Internal functions
+ */
+static void AddressStrcpy(char *dst, const char *src);
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * AddressStrcpy --
+ *
+ *      Copy an address list and eliminate all space around te address
+ *	entities.
+ *
+ * Results:
+ *	None
+ *
+ * Side effects:
+ *	None
+ *
+ *
+ *----------------------------------------------------------------------
+ */
+
+static void
+AddressStrcpy(char *dst, const char *src)
+{
+    char *mark = dst;
+    int quoted = 0;
+    int skip = 1;
+
+    for (; *src; src++) {
+	if ('\\' == *src) {
+	    *dst++ = *src++;
+	} else if ('"' == *src) {
+	    skip = 0;
+	    if (quoted) {
+		quoted = 0;
+	    } else {
+		quoted = 1;
+	    }
+	} else if (!quoted) {
+	    if (',' == *src) {
+		dst = mark;
+		skip = 1;
+	    } else if (isspace(*src)) {
+		if (skip) {
+		    continue;
+		}
+	    } else {
+		mark = dst+1;
+		skip = 0;
+	    }
+	}
+	*dst++ = *src;
+    }
+    *mark = '\0';
+}
+
 
 /*
  *----------------------------------------------------------------------
@@ -246,10 +316,28 @@ RatDeleteAddress(ClientData clientData)
 int
 RatAddressIsMe(Tcl_Interp *interp, ADDRESS *adrPtr, int useUP)
 {
+    char *from;
+
+    if (adrPtr == NULL) {
+	return 0;
+    }
     if (adrPtr->mailbox && !strcasecmp(adrPtr->mailbox, currentMailboxName)
 	    && (adrPtr->host && !strcasecmp(adrPtr->host, currentHost))) {
 	return 1;
-    } else if (useUP) {
+    }
+    if (NULL != (from = Tcl_GetVar2(interp,"option","from",TCL_GLOBAL_ONLY))
+	    && *from != '\0') {
+	ADDRESS *a = NULL;
+	rfc822_parse_adrlist(&a, from, currentHost);
+	if (a && adrPtr->mailbox && adrPtr->host
+		&& !strcasecmp(a->mailbox, adrPtr->mailbox)
+		&& !strcasecmp(a->host, adrPtr->host)) {
+	    mail_free_address(&a);
+	    return 1;
+	}
+	mail_free_address(&a);
+    }
+    if (useUP) {
 	Tcl_CmdInfo cmdInfo;
 	if (Tcl_GetCommandInfo(interp, "RatUP_IsMe", &cmdInfo)) {
 	    Tcl_DString cmd;
@@ -260,10 +348,12 @@ RatAddressIsMe(Tcl_Interp *interp, ADDRESS *adrPtr, int useUP)
 	    Tcl_DStringAppendElement(&cmd,adrPtr->host?adrPtr->host:"");
 	    Tcl_DStringAppendElement(&cmd,adrPtr->personal?adrPtr->personal:"");
 	    Tcl_DStringAppendElement(&cmd,adrPtr->adl?adrPtr->adl:"");
-	    Tcl_Eval(interp, Tcl_DStringValue(&cmd));
-	    Tcl_GetBoolean(interp, interp->result, &isMe);
+	    if (TCL_OK == Tcl_Eval(interp, Tcl_DStringValue(&cmd))
+		    && TCL_OK == Tcl_GetBoolean(interp, interp->result, &isMe)){
+		Tcl_DStringFree(&cmd);
+		return isMe;
+	    }
 	    Tcl_DStringFree(&cmd);
-	    return isMe;
 	}
     }
     return 0;
@@ -335,9 +425,9 @@ RatAddressTranslate(Tcl_Interp *interp, ADDRESS *adrPtr)
     Tcl_DStringAppendElement(&cmd,adrPtr->host?adrPtr->host:"");
     Tcl_DStringAppendElement(&cmd,adrPtr->personal?adrPtr->personal:"");
     Tcl_DStringAppendElement(&cmd,adrPtr->adl?adrPtr->adl:"");
-    Tcl_Eval(interp, Tcl_DStringValue(&cmd));
-    if (   TCL_OK != Tcl_SplitList(interp, interp->result, &argc, &argv)
-	|| 4 != argc) {
+    if (TCL_OK != Tcl_Eval(interp, Tcl_DStringValue(&cmd))
+    	    || TCL_OK != Tcl_SplitList(interp, interp->result, &argc, &argv)
+	    || 4 != argc) {
 	char buf[2048];
 	sprintf(buf, "%s: %s", Tcl_GetVar2(interp, "t", "translate_error",
 		TCL_GLOBAL_ONLY), Tcl_DStringValue(&cmd));
@@ -419,27 +509,45 @@ RatAlias(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
     if (!strcmp(argv[1], "add")) {
 	AliasInfo *aliasPtr;
 	Tcl_HashEntry *entryPtr;
-	char *key;
-	int new;
+	char *key, *book, *comment, *cPtr;
+	int new, s;
 
-	if (argc != 5) {
+	if (argc < 5 || argc > 7) {
 	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
-		    " add name fullname content\"", (char *) NULL);
+		    " add book name fullname content comment\"", (char *) NULL);
 	    return TCL_ERROR;
 	}
-	if (!strlen(argv[2])) {
+	if (5 == argc) {
+	    s = 2;
+	    book = "Personal";
+	    comment = "";
+	} else {
+	    s = 3;
+	    book = argv[2];
+	    comment = argv[6];
+	}
+	if (!strlen(argv[s])) {
 	    Tcl_SetResult(interp, "The name can not be an empty string",
 		    TCL_STATIC);
 	    return TCL_OK;
 	}
-	key = cpystr(argv[2]);
+	key = argv[s];
 	lcase(key);
-	aliasPtr = (AliasInfo*)ckalloc(sizeof(AliasInfo)+2+strlen(argv[3])+
-		strlen(argv[4]));
-	aliasPtr->fullname = (char*)aliasPtr+sizeof(AliasInfo);
-	aliasPtr->content = aliasPtr->fullname + strlen(argv[3]) + 1;
-	strcpy(aliasPtr->fullname, argv[3]);
-	strcpy(aliasPtr->content, argv[4]);
+	aliasPtr = (AliasInfo*)ckalloc(sizeof(AliasInfo)+16+strlen(book)+
+		strlen(argv[s+1])+strlen(argv[s+2])+strlen(comment));
+	cPtr = (char*)PADDED((char*)aliasPtr+sizeof(AliasInfo));
+	aliasPtr->book = cPtr;
+	cPtr = PADDED(cPtr + strlen(book) + 1);
+	aliasPtr->fullname = cPtr;
+	cPtr = PADDED(cPtr + strlen(argv[s+1]) + 1);
+	aliasPtr->content = cPtr;
+	cPtr = PADDED(cPtr + strlen(argv[s+2]) + 1);
+	aliasPtr->comment = cPtr;
+	aliasPtr->mark = 0;
+	strcpy(aliasPtr->book, book);
+	strcpy(aliasPtr->fullname, argv[s+1]);
+	AddressStrcpy(aliasPtr->content, argv[s+2]);
+	strcpy(aliasPtr->comment, comment);
 	entryPtr = Tcl_CreateHashEntry(&aliasTable, key, &new);
 	Tcl_SetHashValue(entryPtr, (ClientData)aliasPtr);
 	return TCL_OK;
@@ -459,6 +567,26 @@ RatAlias(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 	}
 	return TCL_OK;
 
+    } else if (!strcmp(argv[1], "get")) {
+	Tcl_HashEntry *entryPtr;
+	AliasInfo *aliasPtr;
+
+	if (argc != 3) {
+	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
+		    " get alias\"", (char *) NULL);
+	    return TCL_ERROR;
+	}
+	if (!(entryPtr = Tcl_FindHashEntry(&aliasTable, argv[2]))) {
+	    Tcl_SetResult(interp, "Illegal alias", TCL_STATIC);
+	    return TCL_ERROR;
+	}
+	aliasPtr = (AliasInfo*) Tcl_GetHashValue(entryPtr);
+	Tcl_AppendElement(interp, aliasPtr->book);
+	Tcl_AppendElement(interp, aliasPtr->fullname);
+	Tcl_AppendElement(interp, aliasPtr->content);
+	Tcl_AppendElement(interp, aliasPtr->comment);
+	return TCL_OK;
+
     } else if (!strcmp(argv[1], "list")) {
 	Tcl_HashEntry *entryPtr;
 	Tcl_HashSearch search;
@@ -475,8 +603,10 @@ RatAlias(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 	for (entryPtr = Tcl_FirstHashEntry(&aliasTable, &search);
 		entryPtr; entryPtr = Tcl_NextHashEntry(&search)) {
 	    aliasPtr = (AliasInfo*) Tcl_GetHashValue(entryPtr);
+	    Tcl_DStringAppendElement(&list, aliasPtr->book);
 	    Tcl_DStringAppendElement(&list, aliasPtr->fullname);
 	    Tcl_DStringAppendElement(&list, aliasPtr->content);
+	    Tcl_DStringAppendElement(&list, aliasPtr->comment);
 	    Tcl_SetVar2(interp, argv[2], Tcl_GetHashKey(&aliasTable, entryPtr),
 		    Tcl_DStringValue(&list), 0);
 	    Tcl_DStringFree(&list);
@@ -499,13 +629,13 @@ RatAlias(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 	char buf[1024];
 	FILE *fp;
 
-	if (argc != 3) {
+	if (argc != 4) {
 	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
-		    " save filename\"", (char *) NULL);
+		    " save book filename\"", (char *) NULL);
 	    return TCL_ERROR;
 	}
 
-	if (NULL == (fp = fopen(Tcl_TildeSubst(interp,argv[2],&buffer), "w"))) {
+	if (NULL == (fp = fopen(Tcl_TildeSubst(interp,argv[3],&buffer), "w"))) {
 	    Tcl_DStringFree(&buffer);
 	    sprintf(buf, "Failed to open file \"%s\": %s", argv[2],
 		    Tcl_PosixError(interp));
@@ -518,10 +648,15 @@ RatAlias(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 	for (entryPtr = Tcl_FirstHashEntry(&aliasTable, &search);
 		entryPtr; entryPtr = Tcl_NextHashEntry(&search)) {
 	    aliasPtr = (AliasInfo*) Tcl_GetHashValue(entryPtr);
+	    if (strcmp(argv[2], aliasPtr->book)) {
+		continue;
+	    }
+	    Tcl_DStringAppendElement(&list, aliasPtr->book);
 	    Tcl_DStringAppendElement(&list,
 		    Tcl_GetHashKey(&aliasTable, entryPtr));
 	    Tcl_DStringAppendElement(&list, aliasPtr->fullname);
 	    Tcl_DStringAppendElement(&list, aliasPtr->content);
+	    Tcl_DStringAppendElement(&list, aliasPtr->comment);
 	    fputs("RatAlias add ", fp);
 	    fputs(Tcl_DStringValue(&list), fp);
 	    fputs("\n", fp);
@@ -544,7 +679,7 @@ RatAlias(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 	Tcl_DString newList;
 	AliasInfo *aliasPtr;
 	char buf[1024], *sPtr, *cPtr;
-	int level, iterations, rescan, useTable, new;
+	int level, rescan, useTable, new, inGroup;
 
 	if (argc != 3) {
 	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
@@ -560,10 +695,22 @@ RatAlias(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 	    return TCL_OK;
 	}
 
+	/*
+	 * Create unique mark
+	 */
+	if (0 == ++aliasMark) {
+	    Tcl_HashSearch search;
+	    aliasMark++;
+
+	    for (entryPtr = Tcl_FirstHashEntry(&aliasTable, &search);
+		    entryPtr; entryPtr = Tcl_NextHashEntry(&search)) {
+		aliasPtr = (AliasInfo*) Tcl_GetHashValue(entryPtr);
+		aliasPtr->mark = 0;
+	    }
+	}
+
 	Tcl_GetBoolean(interp, Tcl_GetVar2(interp, "option", "lookup_name",
 			               TCL_GLOBAL_ONLY), &useTable);
-	Tcl_GetInt(interp, Tcl_GetVar2(interp, "option", "alias_iterations",
-			               TCL_GLOBAL_ONLY), &iterations);
 
 	if (!strcmp(argv[1], "expand1")) {
 	    level = 0;
@@ -578,10 +725,26 @@ RatAlias(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 	    }
 	}
 	do {
-	    rescan = 0;
+	    inGroup = rescan = 0;
 	    rfc822_parse_adrlist(&baseAdrPtr, Tcl_DStringValue(&newList), "");
 	    Tcl_DStringSetLength(&newList, 0);
 	    for (adrPtr = baseAdrPtr; adrPtr; adrPtr = adrPtr->next) {
+		if (!adrPtr->mailbox) {
+		    if (inGroup) {
+			Tcl_DStringAppend(&newList, ";", 1);
+			inGroup = 0;
+		    }
+		    if (adrPtr->next && adrPtr->next->mailbox) {
+			Tcl_DStringAppend(&newList, ", ", 2);
+		    }
+		    continue;
+		}
+		if (!inGroup && !adrPtr->host) {
+		    Tcl_DStringAppend(&newList, adrPtr->mailbox, -1);
+		    Tcl_DStringAppend(&newList, ":", 1);
+		    inGroup = 1;
+		    continue;
+		}
 		/*
 		 * Check if this matches an alias. If the host is an empty
 		 * string we just try the mailbox name as key, otherwise we
@@ -613,11 +776,15 @@ RatAlias(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 			if ((cPtr = strchr(buf, ','))) {
 			    *cPtr = '\0';
 			}
-			Tcl_SetHashValue(userPtr, (ClientData)strdup(buf));
+			Tcl_SetHashValue(userPtr, (ClientData)cpystr(buf));
 		    }
 		}
 		if (entryPtr) {
 		    aliasPtr = (AliasInfo*) Tcl_GetHashValue(entryPtr);
+		    if (aliasPtr->mark == aliasMark) {
+			Tcl_SetResult(interp, "Alias loop detected",TCL_STATIC);
+			return TCL_ERROR;
+		    }
 		    if (level) {
 			rescan = 1;
 			Tcl_DStringAppend(&newList, aliasPtr->content, -1);
@@ -629,6 +796,7 @@ RatAlias(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 			    Tcl_DStringAppend(&newList, buf, strlen(buf));
 			}
 		    }
+		    aliasPtr->mark = aliasMark;
 		} else if ((!adrPtr->host || !*adrPtr->host) &&
 			   !adrPtr->personal && !level && userPtr) {
 		    sPtr = RatAddressMail(adrPtr);
@@ -644,23 +812,19 @@ RatAlias(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 			Tcl_DStringAppend(&newList, buf, -1);
 		    }
 		}
-		if (adrPtr->next) {
+		if (adrPtr->next && adrPtr->next->mailbox) {
 		    Tcl_DStringAppend(&newList, ", ", 2);
 		}
 	    }
 	    mail_free_address(&baseAdrPtr);
-	    if (0 == --iterations && rescan) {
-		RatLog(interp, RAT_WARN, 
-			"Too many alias iterations (alias loop?)", 0);
-		rescan = 0;
-	    }
 	} while(rescan);
 	Tcl_DStringResult(interp, &newList);
 	return TCL_OK;
 	
     } else {
 	Tcl_AppendResult(interp, "bad option \"", argv[1],
-		"\": must be one of add, list, read, save, expand1 or expand2",
+		"\": must be one of add, delete, get, list, read, save,",
+		" expand1 or expand2",
 		(char *) NULL);
 	return TCL_ERROR;
     }
