@@ -5,18 +5,26 @@
  *	use in wish and similar Tk-based applications.
  *
  *
- * TkRat software and its included text is Copyright 1996 by Martin Forssen.
+ * TkRat software and its included text is Copyright 1996,1997,1998
+ * by Martin Forssn
  *
  * The full text of the legal notice is contained in the file called
  * COPYRIGHT, included with this distribution.
  */
 
 #include <pwd.h>
-#include <signal.h>
 #include <tk.h>
-#include "rat.h"
+#include "ratFolder.h"
+#include "ratPGP.h"
 #include <locale.h>
 
+/*
+ * Figure out if we are using a version later than 8.0b1
+ */
+#if TCL_MAJOR_VERSION >= 8 && (TCL_MINOR_VERSION >0 || TCL_RELEASE_LEVEL >=1)
+#    define TCL_CREATEOK
+#endif
+
 /*
  * The following variable is a special hack that is needed in order for
  * Sun shared libraries to be used for Tcl.
@@ -69,36 +77,63 @@ static int hasDisplay = 1;
  */
 static int noX = 0;
 
+/*
+ * Buffer for delayed output
+ */
+static char ratDelayBuffer[3];
+
 /*
  * Communication with the child
  */
 static FILE *toSender = NULL;
 static int fromSender;
-static Tcl_File fromSenderFile;
 static int sendSequence = 0;
+#ifndef TCL_CREATEOK
+static Tcl_File fromSenderFile;
+#endif
 
 static Tcl_FileProc RatHandleSender;
 static int RatCreateSender(Tcl_Interp *interp);
 
+/*
+ * List of sent messages
+ */
+typedef struct SentMsg {
+    int id;
+    char *handler;
+    struct SentMsg *nextPtr;
+} SentMsg;
+static SentMsg *sentMsg = NULL;
+
 /*
  * The location of all library files
  */
 static char *autoPath = NULL;
 
 /*
- * Timer procedure.
+ * Local functions
  */
 static Tk_TimerProc RatChildHandler;
-
-/*
- * Lock procedure
- */
 static Tcl_VarTraceProc RatReject;
-
-/*
- * Applications initialization procedure
- */
 static Tcl_AppInitProc RatAppInit;
+static Tcl_VarTraceProc RatHostWatcher;
+static Tcl_VarTraceProc RatInitCurrent;
+static Tcl_CmdProc RatBgExec;
+static Tcl_CmdProc RatSend;
+static Tcl_CmdProc RatGetEncoding;
+static Tcl_CmdProc RatCleanup;
+static Tcl_CmdProc RatTildeSubst;
+static Tcl_CmdProc RatTime;
+static Tcl_CmdProc RatLock;
+static Tcl_CmdProc RatIsLocked;
+static Tcl_CmdProc RatType;
+static Tcl_CmdProc RatDSE;
+static Tcl_CmdProc RatExpire;
+static Tcl_CmdProc RatLL;
+static Tcl_CmdProc RatDbaseCheck;
+static Tcl_CmdProc RatMangleNumberCmd;
+static Tcl_CmdProc RatEncodingCompatCmd;
+static Tcl_CmdProc RatFormatDateCmd;
 
 
 /*
@@ -164,22 +199,46 @@ main(int argc, char **argv)
 	autoPath = dir;
     } else {
 	if (!autoPath) {
-	    autoPath = (char*)malloc(strlen(LIBDIR)+1);
-	    strcpy(autoPath, LIBDIR);
+	    if (NULL == (dir = getenv("LIBDIR"))) {
+		fprintf(stderr, "Bad installation. I miss the LIBDIR env "
+				"variable\n");
+		exit(-1);
+	    }
+	    autoPath = (char*)malloc(strlen(dir)+1);
+	    strcpy(autoPath, dir);
 	}
 	fargv[1] = (char*)malloc(strlen(autoPath)+7);
 	sprintf(fargv[1], "%s/tkrat", autoPath);
     }
+    if (NULL == getenv("COMPRESS")) {
+	fprintf(stderr, "Bad installation. I miss the COMPRESS env variable\n");
+	exit(-1);
+    }
+    if (NULL == getenv("CSUFFIX")) {
+	fprintf(stderr, "Bad installation. I miss the CSUFFIX env variable\n");
+	exit(-1);
+    }
+    if (NULL == getenv("CONFIG_DIR")) {
+	fprintf(stderr,
+		"Bad installation. I miss the CONFIG_DIR env variable\n");
+	exit(-1);
+    }
+    if (NULL == getenv("PGP")) {
+	fprintf(stderr, "Bad installation. I miss the PGP env variable\n");
+	exit(-1);
+    }
     if (!interactive) {
 	if (stat(fargv[1], &sbuf)) {
-	    fprintf(stderr, "You have installed the files incorrectly. I expect to find them in\n");
+	    fprintf(stderr, "You have installed the files incorrectly. "
+			    "I expect to find them in\n");
 	    if (dir) {
-		fprintf(stderr, "\"%s\" as specified in the RATLIBDIR environment variable.\n",
-			dir);
+		fprintf(stderr, "\"%s\" as specified in the RATLIBDIR "
+				"environment variable.\n", dir);
 	    } else {
 		fprintf(stderr, "\"%s\" as specified at build time.\n",
 			autoPath);
-		fprintf(stderr, "You can use the RATLIBDIR environment variable to indicate where\n");
+		fprintf(stderr, "You can use the RATLIBDIR environment "
+				"variable to indicate where\n");
 		fprintf(stderr, "they really are installed.\n");
 	    }
 	    exit(1);
@@ -219,6 +278,9 @@ RatAppInit(Tcl_Interp *interp)
     Tk_Window main;
     Tcl_DString ds;
     char *s;
+#if TCL_MAJOR_VERSION < 8
+    char buf[1024];
+#endif
 
     if (Tcl_Init(interp) == TCL_ERROR) {
 	return TCL_ERROR;
@@ -239,11 +301,6 @@ RatAppInit(Tcl_Interp *interp)
      * Initialize some variables
      */
     Tcl_SetVar(interp, "ratSenderSending", "0", TCL_GLOBAL_ONLY);
-    Tcl_SetVar(interp, "compressProg", COMPRESS, TCL_GLOBAL_ONLY);
-    Tcl_SetVar(interp, "compressSuffix", CSUFFIX, TCL_GLOBAL_ONLY);
-    Tcl_SetVar(interp, "cpProg", CP_PROG, TCL_GLOBAL_ONLY);
-    Tcl_SetVar(interp, "ratLibDir", autoPath, TCL_GLOBAL_ONLY);
-    Tcl_SetVar(interp, "ratPrefix", PREFIX, TCL_GLOBAL_ONLY);
     Tcl_DStringInit(&ds);
     Tcl_DStringAppendElement(&ds, autoPath);
     s = Tcl_GetVar(interp, "auto_path", TCL_GLOBAL_ONLY);
@@ -252,6 +309,17 @@ RatAppInit(Tcl_Interp *interp)
     Tcl_SetVar(interp, "auto_path", Tcl_DStringValue(&ds), TCL_GLOBAL_ONLY);
     Tcl_DStringFree(&ds);
 
+    /*
+     * Make sure we know who we are and that we keep track of any changes
+     */
+    RatInitCurrent(NULL, interp, NULL, NULL, 0);
+    Tcl_TraceVar2(interp, "option", "charset", TCL_GLOBAL_ONLY|TCL_TRACE_WRITES,
+	    RatInitCurrent, NULL);
+    Tcl_TraceVar2(interp, "option", "domain", TCL_GLOBAL_ONLY|TCL_TRACE_WRITES,
+	    RatHostWatcher, NULL);
+    Tcl_TraceVar2(interp, "option", "masquerade_as",
+	    TCL_GLOBAL_ONLY|TCL_TRACE_WRITES, RatHostWatcher, NULL);
+
     /*
      * Make sure that env(USER) and env(HOME) are set. If not then we
      * initialize them.
@@ -308,6 +376,20 @@ RatAppInit(Tcl_Interp *interp)
     Tcl_CreateCommand(interp, "RatDbaseCheck", RatDbaseCheck, NULL, NULL);
     Tcl_CreateCommand(interp, "RatSplitAdr", RatSplitAddresses, NULL, NULL);
     Tcl_CreateCommand(interp, "RatMailcapReload", RatMailcapReload, NULL, NULL);
+    Tcl_CreateCommand(interp, "RatPGP", RatPGPCmd, NULL, NULL);
+    Tcl_CreateCommand(interp, "RatMangleNumber", RatMangleNumberCmd, NULL,NULL);
+    Tcl_CreateCommand(interp, "RatEncodingCompat", RatEncodingCompatCmd,
+	    NULL, NULL);
+    Tcl_CreateCommand(interp, "RatFormatDate", RatFormatDateCmd, NULL, NULL);
+
+    /*
+     * Create tcl8 compatibility functions
+     */
+#if TCL_MAJOR_VERSION < 8
+    sprintf(buf, "rename file RatOldFile");
+    Tcl_Eval(interp, buf);
+    Tcl_CreateCommand(interp, "file", RatFile, NULL, NULL);
+#endif	/* TCL_MAJOR_VERSION < 8 */
 
     /*
      * Specify a user-specific startup file to invoke if the application
@@ -349,8 +431,9 @@ RatAppInit(Tcl_Interp *interp)
  *----------------------------------------------------------------------
  */
 
-void
-RatInitCurrent(Tcl_Interp *interp)
+static char*
+RatInitCurrent(ClientData clientData, Tcl_Interp *interp, char *name1,
+	       char *name2, int flags)
 {
     char *charset = Tcl_GetVar2(interp, "option", "charset", TCL_GLOBAL_ONLY);
     struct passwd *passwdPtr;
@@ -372,15 +455,18 @@ RatInitCurrent(Tcl_Interp *interp)
 		strcat(buf, domain);
 	    }
 	}
-	currentHost = strdup(buf);
+	currentHost = (char*)buf;
     }
+    currentHost = cpystr(currentHost);
+    Tcl_SetVar(interp, "ratCurrentHost", currentHost, TCL_GLOBAL_ONLY);
     passwdPtr = getpwuid(getuid());
-    currentMailboxName = strdup(passwdPtr->pw_name);
+    currentMailboxName = cpystr(passwdPtr->pw_name);
     strcpy((char*)buf, passwdPtr->pw_gecos);
     if (strchr((char*)buf, ',')) {
 	*strchr((char*)buf, ',') = '\0';
     }
     currentPersonalName = (char*)RatEncodeHeaderLine(buf, charset, 0, 1);
+    return NULL;
 }
 
 /*
@@ -446,27 +532,66 @@ RatLog(Tcl_Interp *interp, RatLogLevel level, char *message, int keep)
  *      ratMangleNumber.
  *
  * Results:
- *      The return value is normally TCL_OK and the result can be found
- *      in interp->result. If something goes wrong TCL_ERROR is returned
- *      and an error message will be left in interp->result.
+ *      Returns a pointer to a static buffer containg the string
+ *	representation of the number.
  *
  * Side effects:
- *      Calls RatMangleNumber.
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+char*
+RatMangleNumber(int number)
+{
+    static char buf[32];     /* Scratch area */
+
+    if (number < 1000) {
+	sprintf(buf, "%d", number);
+    } else if (number < 10000) {
+	sprintf(buf, "%.1fk", number/1000.0);
+    } else if (number < 1000000) {
+	sprintf(buf, "%dk", (number+500)/1000);
+    } else if (number < 10000000) {
+	sprintf(buf, "%.1fM", number/1000000.0);
+    } else {
+	sprintf(buf, "%dM", (number+500000)/1000000);
+    }
+    return buf;
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * RatMangleNumberCmd --
+ *
+ *	See ../doc/interface for a descriptions of arguments and result.
+ *
+ * Results:
+ *      A list of strings to display to the user.
+ *
+ * Side effects:
+ *      None.
  *
  *
  *----------------------------------------------------------------------
  */
 
-int
-RatMangleNumber(Tcl_Interp *interp, int number)
+static int
+RatMangleNumberCmd(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 {
-    char buf[1024];     /* Scratch area */
+    int number;
 
-    sprintf(buf, "RatMangleNumber %d", number);
-
-    return Tcl_GlobalEval(interp, buf);
+    if (2 != argc || TCL_OK != Tcl_GetInt(interp, argv[1], &number)) {
+	Tcl_AppendResult(interp, "Usage: ", argv[0], " number", (char*) NULL);
+	return TCL_ERROR;
+    }
+    Tcl_SetResult(interp, RatMangleNumber(number), TCL_VOLATILE);
+    return TCL_OK;
 }
 
+
 /*
  *----------------------------------------------------------------------
  *
@@ -486,7 +611,7 @@ RatMangleNumber(Tcl_Interp *interp, int number)
  *----------------------------------------------------------------------
  */
 
-int
+static int
 RatBgExec(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 {
     static RatBgInfo *ratBgList = NULL;
@@ -640,6 +765,8 @@ RatGenId(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
  *	    FAILED id prefix deferred text
  *	    SAVE file save_to to from cc subject
  *	    SENT id deferred
+ *	    PGP pgp_specific_data
+ *	The server will respond to all PGP commands
  *
  * Results:
  *      The return value is normally TCL_OK and the result can be found
@@ -653,11 +780,12 @@ RatGenId(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
  *----------------------------------------------------------------------
  */
 
-int
+static int
 RatSend(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 {
-    char *tmp, *cacheDir, *handler = NULL, buf[16];
+    char *tmp, *cacheDir, *handler = NULL, buf[1024];
     Tcl_DString tildeBuffer;
+    SentMsg **smPtrPtr;
 
     if (argc != 2 && argc != 3) {
 	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
@@ -665,8 +793,8 @@ RatSend(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 	return TCL_ERROR;
     }
 
-    cacheDir = Tcl_GetVar2(interp, "option", "send_cache", TCL_GLOBAL_ONLY);
     Tcl_DStringInit(&tildeBuffer);
+    cacheDir = Tcl_GetVar2(interp, "option", "send_cache", TCL_GLOBAL_ONLY);
     cacheDir = Tcl_TranslateFileName(interp, cacheDir, &tildeBuffer);
 
     if (!strcmp("kill", argv[1])) {
@@ -677,7 +805,7 @@ RatSend(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
     } else if (!strcmp("init", argv[1]) || !strcmp("sendDeferred", argv[1])) {
 	Tcl_DString fileList;
 	char **listArgv;
-	int listArgc, i;
+	int listArgc, i, sent;
 
 	Tcl_DStringInit(&fileList);
 	if (TCL_OK != RatHoldList(interp, cacheDir, &fileList)
@@ -692,16 +820,40 @@ RatSend(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 	    sprintf(buf, "%d", listArgc);
 	    Tcl_SetVar(interp, "ratDeferred", buf, TCL_GLOBAL_ONLY);
 	} else {
+	    Tcl_DString cmd;
+
 	    if (TCL_OK != RatCreateSender(interp)) {
 		free(listArgv);
 		goto error;
 	    }
-	    for (i=0; i<listArgc; i++) {
-		Tcl_SetVar(interp, "ratSenderSending", "1", TCL_GLOBAL_ONLY);
-		fprintf(toSender, "SEND %d {%s/%s} 1\n", sendSequence++,
-			cacheDir, listArgv[i]);
+	    Tcl_SetVar(interp, "ratSenderSending", "1", TCL_GLOBAL_ONLY);
+	    Tcl_DStringInit(&cmd);
+	    Tcl_DStringAppendElement(&cmd, "SEND");
+	    for (i=0, sent=0; i<listArgc; i++) {
+		for (smPtrPtr=&sentMsg;
+			*smPtrPtr && strcmp((*smPtrPtr)->handler, listArgv[i]);
+			smPtrPtr = &(*smPtrPtr)->nextPtr);
+		if (NULL != *smPtrPtr) continue;
+		*smPtrPtr =
+			(SentMsg*)malloc(sizeof(SentMsg)+strlen(listArgv[i])+1);
+		(*smPtrPtr)->id = sendSequence;
+		(*smPtrPtr)->handler = (char*)*smPtrPtr+sizeof(SentMsg);
+		strcpy((*smPtrPtr)->handler, listArgv[i]);
+		(*smPtrPtr)->nextPtr = NULL;
+		Tcl_DStringStartSublist(&cmd);
+		sprintf(buf, "%d", sendSequence++);
+		Tcl_DStringAppendElement(&cmd, buf);
+		sprintf(buf, "%s/%s", cacheDir, listArgv[i]);
+		Tcl_DStringAppendElement(&cmd, buf);
+		Tcl_DStringAppendElement(&cmd, "1");
+		Tcl_DStringEndSublist(&cmd);
+		sent++;
 	    }
+	    fprintf(toSender, "%s\n", Tcl_DStringValue(&cmd));
 	    fflush(toSender);
+	    Tcl_DStringFree(&cmd);
+	    sprintf(buf, "%d", sent);
+	    Tcl_SetResult(interp, buf, TCL_VOLATILE);
 	}
 	free(listArgv);
 
@@ -726,7 +878,7 @@ RatSend(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 	if (TCL_OK != RatHoldInsert(interp, cacheDir, argv[2], "")) {
 	    goto error;
 	}
-	handler = strdup(interp->result);
+	handler = cpystr(interp->result);
 
 	if (!strcmp(argv[1], "direct")) {
 	    if (TCL_OK != RatCreateSender(interp)) {
@@ -734,7 +886,13 @@ RatSend(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 		goto error;
 	    }
 	    Tcl_SetVar(interp, "ratSenderSending", "1", TCL_GLOBAL_ONLY);
-	    fprintf(toSender, "SEND %d %s 0\n", sendSequence++, handler);
+	    for (smPtrPtr=&sentMsg; *smPtrPtr;smPtrPtr = &(*smPtrPtr)->nextPtr);
+	    *smPtrPtr = (SentMsg*)malloc(sizeof(SentMsg)+strlen(handler)+1);
+	    (*smPtrPtr)->id = sendSequence;
+	    (*smPtrPtr)->handler = (char*)*smPtrPtr+sizeof(SentMsg);
+	    strcpy((*smPtrPtr)->handler, handler);
+	    (*smPtrPtr)->nextPtr = NULL;
+	    fprintf(toSender, "SEND {%d %s 0}\n", sendSequence++, handler);
 	    fflush(toSender);
 	} else {
 	    tmp = Tcl_GetVar(interp, "ratDeferred", TCL_GLOBAL_ONLY);
@@ -783,9 +941,6 @@ RatCreateSender(Tcl_Interp *interp)
      */
     pipe(toPipe);
     pipe(fromPipe);
-    if (!currentHost) {
-	RatInitCurrent(interp);
-    }
     if (0 == (senderPid = fork())) {
 	close(toPipe[1]);
 	close(fromPipe[0]);
@@ -805,9 +960,14 @@ RatCreateSender(Tcl_Interp *interp)
     close(fromPipe[1]);
     toSender = fdopen(toPipe[1], "w");
     fromSender = fromPipe[0];
+#ifndef TCL_CREATEOK
     fromSenderFile = Tcl_GetFile((ClientData)fromSender, TCL_UNIX_FD);
     Tcl_CreateFileHandler(fromSenderFile, TCL_READABLE, RatHandleSender,
 	    (ClientData)interp);
+#else
+    Tcl_CreateFileHandler(fromSender, TCL_READABLE, RatHandleSender,
+	    (ClientData)interp);
+#endif
     return TCL_OK;
 }
 
@@ -833,14 +993,20 @@ RatHandleSender(ClientData clientData, int mask)
 {
     Tcl_Interp *interp = (Tcl_Interp*)clientData;
     char buf[1024], **argv, **destArgv, *msg, *tmp;
-    int i, argc, destArgc, fd, deferred = 0, id = 0;
+    int i, argc, destArgc, fd, deferred = 0, id = 0, flags;
+    Tcl_DString cmd;
     struct stat sbuf;
+    SentMsg *smPtr, **smPtrPtr;
 
     i = 0;
     do {
 	if (1 != read(fromSender, &buf[i], 1)) {
 	    Tcl_SetVar(interp, "ratSenderSending", "0", TCL_GLOBAL_ONLY);
+#ifndef TCL_CREATEOK
 	    Tcl_DeleteFileHandler(fromSenderFile);
+#else
+	    Tcl_DeleteFileHandler(fromSender);
+#endif
 	    fclose(toSender);
 	    toSender = NULL;
 	    return;
@@ -875,22 +1041,55 @@ RatHandleSender(ClientData clientData, int mask)
 	if (!strcmp(destArgv[1], "file")) {
 	    Tcl_Channel channel;
 	    int perm;
+	    struct tm *tmPtr;
+	    time_t now;
 
-	    if (4 == destArgc) {
-		Tcl_GetInt(interp, destArgv[3], &perm);
+	    if (5 == destArgc) {
+		Tcl_GetInt(interp, destArgv[4], &perm);
 	    } else {
 		Tcl_GetInt(interp, Tcl_GetVar2(interp, "option", "permissions",
 			TCL_GLOBAL_ONLY), &perm);
 	    }
-	    channel = Tcl_OpenFileChannel(interp,destArgv[2],"a", perm);
-	    Tcl_Write(channel, msg, sbuf.st_size);
-	    Tcl_Close(interp, channel);
+	    channel = Tcl_OpenFileChannel(interp,destArgv[3],"a", perm);
+	    if (NULL != channel) {
+		now = time(NULL);
+		tmPtr = gmtime(&now);
+		sprintf(buf, "From %s@%s %s %s %2d %02d:%02d GMT 19%02d\n",
+			currentMailboxName, currentHost,
+			dayName[tmPtr->tm_wday], monthName[tmPtr->tm_mon],
+			tmPtr->tm_mday, tmPtr->tm_hour, tmPtr->tm_min,
+			tmPtr->tm_year);
+		strcat(buf, "Status: RO\n");
+		Tcl_Write(channel, buf, strlen(buf));
+		Tcl_Write(channel, msg, sbuf.st_size);
+		Tcl_Close(interp, channel);
+	    } else {
+		sprintf(buf, "%s: %s", "Failed to save copy of outgoing"
+			"message", Tcl_PosixError(interp));
+		RatLog(interp, RAT_ERROR, buf, 0);
+	    }
 	} else if (!strcmp(destArgv[1], "dbase")) {
-	    RatDbInsert(interp, argv[3], argv[4], argv[5], argv[6], time(NULL),
-		    "RO", destArgv[2], atoi(destArgv[4]), destArgv[3], msg,
-		    sbuf.st_size);
+	    struct tm *tmPtr;
+	    time_t now;
+
+	    now = time(NULL);
+	    tmPtr = gmtime(&now);
+	    sprintf(buf, "From %s@%s %s %s %2d %02d:%02d GMT 19%02d\n",
+		    currentMailboxName, currentHost,
+		    dayName[tmPtr->tm_wday], monthName[tmPtr->tm_mon],
+		    tmPtr->tm_mday, tmPtr->tm_hour, tmPtr->tm_min,
+		    tmPtr->tm_year);
+
+	    if (TCL_OK != RatDbInsert(interp, argv[3], argv[4], argv[5],
+		    argv[6], time(NULL), "RO", destArgv[3], atoi(destArgv[5]),
+		    destArgv[4], buf, msg, sbuf.st_size)) {
+		sprintf(buf, "%s: %s", "Failed to save copy of outgoing"
+			"message", interp->result);
+		RatLog(interp, RAT_ERROR, buf, 0);
+	    }
 	} else if (!strcmp(destArgv[1], "imap")) {
-	    AppendToIMAP(interp, destArgv[2], destArgv[3], msg, sbuf.st_size);
+	    AppendToIMAP(interp, destArgv[3], destArgv[4], argv[7],
+			 argv[8], msg, sbuf.st_size);
 	} else if (!strcmp(destArgv[1], "mh")) {
 	    RatLog(interp, RAT_ERROR, "Unfortunately I can not save outgoing "
 		    "mail in mh-folders yet.", 0);
@@ -901,6 +1100,33 @@ RatHandleSender(ClientData clientData, int mask)
 	unlink(argv[1]);
 	free(msg);
 	free(destArgv);
+    } else if (!strcmp(argv[0], "PGP")) {
+	if (!strcmp("getpass", argv[1])) {
+	    tmp = RatPGPPhrase(interp);
+	    if (tmp) {
+		Tcl_ScanElement(tmp, &flags);
+		Tcl_ConvertElement(tmp, buf, flags);
+		fprintf(toSender, "PGP PHRASE %s\n", buf);
+		memset(buf, '\0', strlen(buf));
+		memset(tmp, '\0', strlen(tmp));
+		free(tmp);
+	    } else {
+		fprintf(toSender, "PGP NOPHRASE\n");
+	    }
+	    fflush(toSender);
+	} else if (!strcmp("error", argv[1])) {
+	    ClearPGPPass(NULL);
+	    Tcl_DStringInit(&cmd);
+	    Tcl_DStringAppend(&cmd, "RatPGPError", -1);
+	    Tcl_DStringAppendElement(&cmd, argv[2]);
+	    if (TCL_OK != Tcl_Eval(interp, Tcl_DStringValue(&cmd))) {
+		fprintf(toSender, "PGP ABORT\n");
+	    } else {
+		fprintf(toSender, "PGP %s\n", interp->result);
+	    }
+	    fflush(toSender);
+	    Tcl_DStringFree(&cmd);
+	}
     } else if (!strcmp(argv[0], "SENT")) {
 	id = atoi(argv[1]);
 	deferred = atoi(argv[2]);
@@ -914,6 +1140,13 @@ RatHandleSender(ClientData clientData, int mask)
 	if (id == sendSequence-1) {
 	    Tcl_SetVar(interp, "ratSenderSending", "0", TCL_GLOBAL_ONLY);
 	}
+	for (smPtrPtr=&sentMsg; *smPtrPtr && (*smPtrPtr)->id != id;
+		smPtrPtr = &(*smPtrPtr)->nextPtr);
+	if (*smPtrPtr) {
+	    smPtr = *smPtrPtr;
+	    *smPtrPtr = smPtr->nextPtr;
+	    free(smPtr);
+	}
     }
     free(argv);
 }
@@ -937,7 +1170,7 @@ RatHandleSender(ClientData clientData, int mask)
  *----------------------------------------------------------------------
  */
 
-int
+static int
 RatGetEncoding(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 {
     FILE *fp;
@@ -994,7 +1227,7 @@ RatGetEncoding(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
  *----------------------------------------------------------------------
  */
 
-int
+static int
 RatCleanup(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 {
     RatDbClose();
@@ -1018,7 +1251,7 @@ RatCleanup(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
  *----------------------------------------------------------------------
  */
 
-int
+static int
 RatTildeSubst(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 {
     Tcl_DString buffer;
@@ -1053,7 +1286,7 @@ RatTildeSubst(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
  *----------------------------------------------------------------------
  */
 
-int
+static int
 RatTime(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 {
     time_t goal;
@@ -1140,7 +1373,7 @@ RatSearch(char *searchFor, char *searchIn)
  *----------------------------------------------------------------------
  */
 
-int
+static int
 RatLock(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 {
     char *value, *correct;
@@ -1218,7 +1451,7 @@ RatReject(ClientData clientData, Tcl_Interp *interp, char *name1,
  *----------------------------------------------------------------------
  */
 
-int
+static int
 RatIsLocked(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 {
     if (argc != 2) {
@@ -1256,7 +1489,7 @@ RatIsLocked(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
  *----------------------------------------------------------------------
  */
 
-int
+static int
 RatType(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 {
     int listArgc, elemArgc;
@@ -1323,7 +1556,8 @@ RatType(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
     if (!(channel = Tcl_OpenCommandChannel(interp, 2, cmdArgv, TCL_STDOUT))) {
 	return TCL_ERROR;
     }
-    length = Tcl_Read(channel, buf, sizeof(buf));
+    length = Tcl_Read(channel, buf, sizeof(buf)-1);
+    buf[length] = '\0';
     Tcl_Close(interp, channel);
     fileType = strchr(buf, ':')+1;
     Tcl_SplitList(interp, Tcl_GetVar2(interp, "option", "typetable",
@@ -1422,6 +1656,69 @@ RatStringPuts(void *stream_x, char *string)
 
     return(1L);                                 /* T for c-client */
 }
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * RatInitDelaySoutr --
+ *
+ *	Initializes the buffer for RatDelaySoutr by clearing it.
+ *
+ * Results:
+ *      None.
+ *
+ * Side effects:
+ *      Modifies the ratDelayBuffer array.
+ *
+ *
+ *----------------------------------------------------------------------
+ */
+void
+RatInitDelaySoutr(void)
+{
+    ratDelayBuffer[0] = '\0';
+    ratDelayBuffer[1] = '\0';
+    ratDelayBuffer[2] = '\0';
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * RatDelaySoutr --
+ *
+ *	A output function to use with rfc822_output that writes to
+ *	a file destriptor. This function is special in this that it
+ *	always delay writing the last two characters. This allows one to
+ *	filter final newlines.
+ *
+ * Results:
+ *      Always returns 1L.
+ *
+ * Side effects:
+ *      Modifies the ratDelayBuffer array.
+ *
+ *
+ *----------------------------------------------------------------------
+ */
+
+long
+RatDelaySoutr(void *stream_x, char *string)
+{
+    int len1, len2;
+    len1 = strlen(ratDelayBuffer);
+    len2 = strlen(string);
+
+    if (len1+len2 <= 2) {
+	strcat(ratDelayBuffer, string);
+	return 1;
+    }
+    write((int)stream_x, ratDelayBuffer, len1);
+    write((int)stream_x, string, len2-2);
+    ratDelayBuffer[0] = string[len2-2];
+    ratDelayBuffer[1] = string[len2-1];
+    return 1;
+}
+
 
 /*
  *----------------------------------------------------------------------
@@ -1448,7 +1745,6 @@ RatParseMsg(Tcl_Interp *interp, char *message)
     int length;		/* Length of header */
     int bodyOffset = 0;	/* Offset of body from start of header */
     MESSAGE *msgPtr;	/* Pointer to message to return */
-    char *buf;		/* scratch buffer */
     STRING bodyString;	/* Body data */
 
     for (length = 0; message[length]; length++) {
@@ -1468,17 +1764,15 @@ RatParseMsg(Tcl_Interp *interp, char *message)
     msgPtr->text.text.data = message;
     msgPtr->text.text.size = strlen(message);
     msgPtr->text.offset = bodyOffset;
-    buf = (char*)ckalloc(strlen(message)+1);
     INIT(&bodyString, mail_string, (void*) (message+bodyOffset),
 	    strlen(message)-bodyOffset);
     rfc822_parse_msg(&msgPtr->env, &msgPtr->body, message, length, &bodyString,
-	    currentHost);
+	    currentHost, NIL);
     if (msgPtr->body->type != TYPEMULTIPART 
 	    && msgPtr->body->type != TYPEMESSAGE) {
 	msgPtr->body->contents.text.data = cpystr(message+bodyOffset);
 	msgPtr->body->contents.text.size = strlen(message+bodyOffset);
     }
-    ckfree(buf);
     return msgPtr;
 }
 
@@ -1500,7 +1794,7 @@ RatParseMsg(Tcl_Interp *interp, char *message)
  *----------------------------------------------------------------------
  */
 
-int
+static int
 RatDSE(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 {
     sprintf(interp->result, "%d", RatDbDaysSinceExpire(interp));
@@ -1524,7 +1818,7 @@ RatDSE(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
  *----------------------------------------------------------------------
  */
 
-int
+static int
 RatExpire(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 {
     if (argc != 3) {
@@ -1564,6 +1858,42 @@ RatIsEmpty (char *string)
     }
     return 1;
 }
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * RatEncodingCompatCmd --
+ *
+ *	See ../doc/interface for a descriptions of arguments and result.
+ *
+ * Results:
+ *	True if the encodings are compatibe otherwise false.
+ *
+ * Side effects:
+ *      None.
+ *
+ *
+ *----------------------------------------------------------------------
+ */
+
+static int
+RatEncodingCompatCmd(ClientData dummy, Tcl_Interp *interp, int argc,
+	char *argv[])
+{
+    if (3 != argc) {
+	Tcl_AppendResult(interp, "Usage: ", argv[0], " wanted avail",
+		(char*) NULL);
+	return TCL_ERROR;
+    }
+    if (RatEncodingCompat(interp, argv[1], argv[2])) {
+	Tcl_SetResult(interp, "1", TCL_STATIC);
+    } else {
+	Tcl_SetResult(interp, "0", TCL_STATIC);
+    }
+    return TCL_OK;
+}
+
+
 
 /*
  *----------------------------------------------------------------------
@@ -1585,19 +1915,13 @@ RatIsEmpty (char *string)
 int
 RatEncodingCompat(Tcl_Interp *interp, char *wanted, char *avail)
 {
-    char *cmd, *argv[3];
-    int result = 0;
-
-    argv[0] = "RatEncodingCompat";
-    argv[1] = wanted;
-    argv[2] = avail;
-    cmd = Tcl_Merge(3, argv);
-
-    if (TCL_OK == Tcl_Eval(interp, cmd)) {
-	Tcl_GetBoolean(interp, interp->result, &result);
+    if (!strcasecmp(wanted, avail)) {
+	return 1;
     }
-    free(cmd);
-    return result;
+    if (!strcasecmp("us-ascii", wanted) && !strncasecmp("iso-8859-", avail, 9)){
+	return 1;
+    }
+    return 0;
 }
 
 
@@ -1658,7 +1982,7 @@ RatLindex(Tcl_Interp *interp, char *list, int index)
  *----------------------------------------------------------------------
  */
 
-int
+static int
 RatLL(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 {
     char *cPtr;
@@ -1679,6 +2003,7 @@ RatLL(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
     sprintf(interp->result, "%d", l);
     return TCL_OK;
 }
+
 
 /*
  *----------------------------------------------------------------------
@@ -1697,7 +2022,7 @@ RatLL(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
  *----------------------------------------------------------------------
  */
 
-int
+static int
 RatDbaseCheck(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
 {
     int fix;
@@ -1708,3 +2033,232 @@ RatDbaseCheck(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
     }
     return RatDbCheck(interp, fix);
 }
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * RatHostWatcher --
+ *
+ *	A trace function that gets called when the user modifies any of
+ *	the hostname options.
+ *
+ * Results:
+ *      NULL.
+ *
+ * Side effects:
+ *      The RatInitCurrent() function will be called.
+ *
+ *
+ *----------------------------------------------------------------------
+ */
+
+static char*
+RatHostWatcher(ClientData clientData, Tcl_Interp *interp, char *name1,
+	       char *name2, int flags)
+{
+    RatInitCurrent(NULL, interp, NULL, NULL, 0);
+    Tcl_Eval(interp, "CheckHostName");
+    return NULL;
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * RatFormatDateCmd --
+ *
+ *	See ../doc/interface for a descriptions of arguments and result.
+ *
+ * Results:
+ *      A list of strings to display to the user.
+ *
+ * Side effects:
+ *      None.
+ *
+ *
+ *----------------------------------------------------------------------
+ */
+
+static int
+RatFormatDateCmd(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
+{
+    if (7 != argc) {
+	Tcl_AppendResult(interp, "Usage: ", argv[0], \
+		" year month day hour min sec", (char*) NULL);
+	return TCL_ERROR;
+    }
+    Tcl_SetResult(interp, RatFormatDate(interp, atoi(argv[2])-1, atoi(argv[3])),
+	    TCL_VOLATILE);
+    return TCL_OK;
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * RatFormatDate --
+ *
+ *	Print the data in a short format.
+ *
+ * Results:
+ *      A pointer to a static area.
+ *
+ * Side effects:
+ *      None.
+ *
+ *
+ *----------------------------------------------------------------------
+ */
+
+char*
+RatFormatDate(Tcl_Interp *interp, int month, int day)
+{
+    static char buf[8];
+    static char *months[12];
+    static int initialized = 0;
+
+    if (!initialized) {
+	int i, argc;
+	char **argv;
+
+	Tcl_SplitList(interp,
+		Tcl_GetVar2(interp, "t", "months", TCL_GLOBAL_ONLY),
+		&argc, &argv);
+	for (i=0; i<12; i++) {
+	    months[i] = argv[i];
+	}
+	initialized = 1;
+    }
+
+    sprintf(buf, "%2d %s", day, months[month]);
+    return buf;
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * RatGetTimeZone --
+ *
+ *	Determines the current timezone.  The method varies wildly
+ *	between different platform implementations, so its hidden in
+ *	this function.
+ *
+ *	This function is shamelessy stolen from tcl8.0p2
+ *
+ * Results:
+ *	The return value is the local time zone, measured in
+ *	minutes away from GMT (-ve for east, +ve for west).
+ *
+ * Side effects:
+ *	None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+int
+RatGetTimeZone(unsigned long currentTime)
+{
+    /*
+     * Determine how a timezone is obtained from "struct tm".  If there is no
+     * time zone in this struct (very lame) then use the timezone variable.
+     * This is done in a way to make the timezone variable the method of last
+     * resort, as some systems have it in addition to a field in "struct tm".
+     * The gettimeofday system call can also be used to determine the time
+     * zone.
+     */
+    
+#if defined(HAVE_TM_TZADJ)
+#   define TCL_GOT_TIMEZONE
+    time_t      curTime = (time_t) currentTime;
+    struct tm  *timeDataPtr = localtime(&curTime);
+    int         timeZone;
+
+    timeZone = timeDataPtr->tm_tzadj  / 60;
+    if (timeDataPtr->tm_isdst) {
+        timeZone += 60;
+    }
+    
+    return timeZone;
+#endif
+
+#if defined(HAVE_TM_GMTOFF) && !defined (TCL_GOT_TIMEZONE)
+#   define TCL_GOT_TIMEZONE
+    time_t     curTime = (time_t) currentTime;
+    struct tm *timeDataPtr = localtime(&curTime);
+    int        timeZone;
+
+    timeZone = -(timeDataPtr->tm_gmtoff / 60);
+    if (timeDataPtr->tm_isdst) {
+        timeZone += 60;
+    }
+    
+    return timeZone;
+#endif
+
+#if defined(USE_DELTA_FOR_TZ)
+#define TCL_GOT_TIMEZONE 1
+    /*
+     * This hack replaces using global var timezone or gettimeofday
+     * in situations where they are buggy such as on AIX when libbsd.a
+     * is linked in.
+     */
+
+    int timeZone;
+    time_t tt;
+    struct tm *stm;
+    tt = 849268800L;      /*    1996-11-29 12:00:00  GMT */
+    stm = localtime(&tt); /* eg 1996-11-29  6:00:00  CST6CDT */
+    /* The calculation below assumes a max of +12 or -12 hours from GMT */
+    timeZone = (12 - stm->tm_hour)*60 + (0 - stm->tm_min);
+    return timeZone;  /* eg +360 for CST6CDT */
+#endif
+
+    /*
+     * Must prefer timezone variable over gettimeofday, as gettimeofday does
+     * not return timezone information on many systems that have moved this
+     * information outside of the kernel.
+     */
+    
+#if defined(HAVE_TIMEZONE_VAR) && !defined (TCL_GOT_TIMEZONE)
+#   define TCL_GOT_TIMEZONE
+    static int setTZ = 0;
+    int        timeZone;
+
+    if (!setTZ) {
+        tzset();
+        setTZ = 1;
+    }
+
+    /*
+     * Note: this is not a typo in "timezone" below!  See tzset
+     * documentation for details.
+     */
+
+    timeZone = timezone / 60;
+
+    return timeZone;
+#endif
+
+#if !defined(NO_GETTOD) && !defined (TCL_GOT_TIMEZONE)
+#   define TCL_GOT_TIMEZONE
+    struct timeval  tv;
+    struct timezone tz;
+    int timeZone;
+
+    gettimeofday(&tv, &tz);
+    timeZone = tz.tz_minuteswest;
+    if (tz.tz_dsttime) {
+        timeZone += 60;
+    }
+    
+    return timeZone;
+#endif
+
+#ifndef TCL_GOT_TIMEZONE
+    /*
+     * Cause compile error, we don't know how to get timezone.
+     */
+    error: autoconf did not figure out how to determine the timezone. 
+#endif
+
+}
