@@ -3,7 +3,8 @@
  *
  *	This file handles the delivery status notifications.
  *
- * TkRat software and its included text is Copyright 1996 by Martin Forssen.
+ * TkRat software and its included text is Copyright 1996,1997,1998
+ * by Martin Forssn
  *
  * The full text of the legal notice is contained in the file called
  * COPYRIGHT, included with this distribution.
@@ -197,6 +198,8 @@ RatDSNFinish(Tcl_Interp *interp, DSNhandle handle)
     Tcl_Write(channel, Tcl_DStringValue(dsPtr), Tcl_DStringLength(dsPtr));
     Tcl_Write(channel, "\n", 1);
     Tcl_Close(interp, channel);
+    Tcl_DStringFree(dsPtr);
+    free(dsPtr);
 }
 
 /*
@@ -275,6 +278,7 @@ RatDSNHandle (Tcl_Interp *interp, char *msg)
 		    (void)unlink(buf);
 		}
 	    }
+	    free(argv);
 	    changed++;
 	    continue;
 	}
@@ -294,7 +298,9 @@ RatDSNHandle (Tcl_Interp *interp, char *msg)
 	Tcl_DStringStartSublist(&ds);
 	for (i=0; i<rargc; i++) {
 	    for (j=rfound=0; !rfound && j<statusPtr->numRecipients; j++) {
-		if (!strcasecmp(statusPtr->recTypePtrPtr[j], "rfc822")
+		if (statusPtr->recTypePtrPtr[j]
+			&& statusPtr->actionPtrPtr[j]
+			&& !strcasecmp(statusPtr->recTypePtrPtr[j], "rfc822")
 			&& !strcmp(statusPtr->recipientPtrPtr[j],
 			        RatLindex(interp, rargv[i], 1))
 			&& strcmp(statusPtr->actionPtrPtr[j],
@@ -319,11 +325,11 @@ RatDSNHandle (Tcl_Interp *interp, char *msg)
 			Tcl_DString msgDS;
 			Tcl_Channel msgCh;
 
-			msgFile = strdup(buf);
+			msgFile = cpystr(buf);
 			Tcl_DStringInit(&msgDS);
 			Tcl_GetCommandInfo(interp, msg, &cmdInfo);
 			RatMessageGet(interp,(MessageInfo*)cmdInfo.clientData,
-				&msgDS);
+				&msgDS, NULL, NULL);
 			msgCh = Tcl_OpenFileChannel(interp, msgFile, "w", perm);
 			Tcl_Write(msgCh, Tcl_DStringValue(&msgDS),
 				Tcl_DStringLength(&msgDS));
@@ -557,7 +563,7 @@ RatDSNGet (ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
     len = Tcl_Read(channel, data, len);
     data[len] = '\0';
     Tcl_Close(interp, channel);
-    msg = RatFrMessageCreate(interp, data, len);
+    msg = RatFrMessageCreate(interp, data, len, NULL);
     ckfree(data);
 
     if (!strcmp(argv[1], "msg")) {
@@ -578,7 +584,7 @@ RatDSNGet (ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
 	    }
 	}
 	RatFreeDeliveryStatus(statusPtr);
-	Tcl_DeleteCommand(interp, msg);
+	RatMessageDelete(interp, msg);
     }
     return TCL_OK;
 }
@@ -727,7 +733,7 @@ RatParseDS(Tcl_Interp *interp, char *body, int length)
 	Tcl_DStringAppendElement(&sPtr->msgFields, value);
 	Tcl_DStringEndSublist(&sPtr->msgFields);
 	if (!strcasecmp("original-envelope-id", name)) {
-	    sPtr->envid = strdup(value);
+	    sPtr->envid = cpystr(value);
 	}
     }
 
@@ -766,13 +772,13 @@ RatParseDS(Tcl_Interp *interp, char *body, int length)
 	    Tcl_DStringAppendElement(&sPtr->rListPtrPtr[i], value);
 	    Tcl_DStringEndSublist(&sPtr->rListPtrPtr[i]);
 	    if (!strcasecmp("original-recipient", name)) {
-		sPtr->recTypePtrPtr[i] = strdup(value);
+		sPtr->recTypePtrPtr[i] = cpystr(value);
 		cPtr = strchr(sPtr->recTypePtrPtr[i], ';');
 		*cPtr++ = '\0';
 		sPtr->recipientPtrPtr[i] = cPtr;
 	    }
 	    if (!strcasecmp("action", name)) {
-		sPtr->actionPtrPtr[i] = strdup(value);
+		sPtr->actionPtrPtr[i] = cpystr(value);
 	    }
 	}
     }
