@@ -3,7 +3,8 @@
  *
  *      This file contains code which implements standard c-client folders.
  *
- * TkRat software and its included text is Copyright 1996 by Martin Forssen.
+ * TkRat software and its included text is Copyright 1996,1997,1998
+ * by Martin Forssn
  *
  * The full text of the legal notice is contained in the file called
  * COPYRIGHT, included with this distribution.
@@ -36,7 +37,6 @@ static RatUpdateProc Db_UpdateProc;
 static RatInsertProc Db_InsertProc;
 static RatSetFlagProc Db_SetFlagProc;
 static RatGetFlagProc Db_GetFlagProc;
-static RatExpungeProc Db_ExpungeProc;
 static RatCreateProc Db_CreateProc;
 static int GetAddressInfo(Tcl_Interp *interp, Tcl_DString *dsPtr, char *adr,
 	DbAdrInfo info);
@@ -134,7 +134,6 @@ RatDbFolderCreate(Tcl_Interp *interp, int argc, char *argv[])
     infoPtr->insertProc = Db_InsertProc;
     infoPtr->setFlagProc = Db_SetFlagProc;
     infoPtr->getFlagProc = Db_GetFlagProc;
-    infoPtr->expungeProc = Db_ExpungeProc;
     infoPtr->infoProc = Db_InfoProc;
     infoPtr->createProc = Db_CreateProc;
     infoPtr->private = (ClientData) dbPtr;
@@ -145,10 +144,6 @@ RatDbFolderCreate(Tcl_Interp *interp, int argc, char *argv[])
     dbPtr->exType = cpystr(argv[4]);
     Tcl_DStringFree(&expression);
 
-    if (!currentHost) {
-	RatInitCurrent(interp);
-    }
-
     return infoPtr;
 }
 
@@ -172,6 +167,13 @@ RatDbFolderCreate(Tcl_Interp *interp, int argc, char *argv[])
 static int
 Db_CloseProc(RatFolderInfoPtr infoPtr, Tcl_Interp *interp)
 {
+    DbFolderInfo *dbPtr = (DbFolderInfo *) infoPtr->private;
+    free(dbPtr->listPtr);
+    free(dbPtr->searchExpr);
+    free(dbPtr->keywords);
+    free(dbPtr->exDate);
+    free(dbPtr->exType);
+    free(dbPtr);
     return TCL_OK;
 }
 
@@ -193,11 +195,32 @@ Db_CloseProc(RatFolderInfoPtr infoPtr, Tcl_Interp *interp)
  *----------------------------------------------------------------------
  */
 static int
-Db_UpdateProc(RatFolderInfoPtr infoPtr, Tcl_Interp *interp)
+Db_UpdateProc(RatFolderInfoPtr infoPtr, Tcl_Interp *interp, int expunge)
 {
     DbFolderInfo *dbPtr = (DbFolderInfo *) infoPtr->private;
-    int *listPtr, number, numberOfNew;
+    int *listPtr, number;
 
+    if (expunge) {
+	RatDbEntry *entryPtr;
+	int i, dst;
+
+	if (TCL_OK != RatDbExpunge(interp)) {
+	    return -1;
+	}
+	infoPtr->size = 0;
+	for (i=dst=0; i<infoPtr->number; i++) {
+	    if ((entryPtr = RatDbGetEntry(dbPtr->listPtr[i]))) {
+		dbPtr->listPtr[dst] = dbPtr->listPtr[i];
+		infoPtr->msgCmdPtr[dst++] = infoPtr->msgCmdPtr[i];
+		infoPtr->size += atoi(entryPtr->content[RSIZE]);
+	    } else {
+		if (infoPtr->msgCmdPtr[i]) {
+		    RatMessageDelete(interp, infoPtr->msgCmdPtr[i]);
+		}
+	    }
+	}
+	infoPtr->number = dst;
+    }
     if (TCL_OK != RatDbSearch(interp, dbPtr->searchExpr, &number, &listPtr)) {
 	Tcl_VarEval(interp, "RatLog 4 \"$t(dbase_error): ", interp->result,
 		"\" time", (char*)NULL);
@@ -207,13 +230,10 @@ Db_UpdateProc(RatFolderInfoPtr infoPtr, Tcl_Interp *interp)
 	return -1;
     }
     if (number >= infoPtr->number) {
-	numberOfNew = number - infoPtr->number;
 	ckfree(dbPtr->listPtr);
 	dbPtr->listPtr = listPtr;
-    } else {
-	numberOfNew = 0;
     }
-    return numberOfNew;
+    return infoPtr->number;
 }
 
 
@@ -338,57 +358,13 @@ Db_GetFlagProc(RatFolderInfoPtr infoPtr, Tcl_Interp *interp, int index,
     return 0;
 }
 
-
-/*
- *----------------------------------------------------------------------
- *
- * Db_ExpungeProc --
- *
- *      See the documentation for expungeProc in ratFolder.h
- *
- * Results:
- *	A standard Tcl result.
- *
- * Side effects:
- *	See the documentation for expungeProc in ratFolder.h
- *
- *
- *----------------------------------------------------------------------
- */
-static int
-Db_ExpungeProc(RatFolderInfoPtr infoPtr, Tcl_Interp *interp)
-{
-    DbFolderInfo *dbPtr = (DbFolderInfo *) infoPtr->private;
-    RatDbEntry *entryPtr;
-
-    int i, dst;
-
-    if (TCL_OK != RatDbExpunge(interp)) {
-	return TCL_ERROR;
-    }
-    infoPtr->size = 0;
-    for (i=dst=0; i<infoPtr->number; i++) {
-        if ((entryPtr = RatDbGetEntry(dbPtr->listPtr[i]))) {
-	    dbPtr->listPtr[dst] = dbPtr->listPtr[i];
-	    infoPtr->msgCmdPtr[dst++] = infoPtr->msgCmdPtr[i];
-	    infoPtr->size += atoi(entryPtr->content[RSIZE]);
-	} else {
-	    if (infoPtr->msgCmdPtr[i]) {
-		Tcl_DeleteCommand(interp, infoPtr->msgCmdPtr[i]);
-	    }
-	}
-    }
-    infoPtr->number = dst;
-    return TCL_OK;
-}
-
 
 /*
  *----------------------------------------------------------------------
  *
  * GetAddressInfo --
  *
- *      Gets infor from an address. The info argument decides what we
+ *      Gets info from an address. The info argument decides what we
  *	wants to extract. The requested data will be appeded to dsPtr.
  *
  * Results:
@@ -406,7 +382,7 @@ GetAddressInfo(Tcl_Interp *interp, Tcl_DString *dsPtr, char *adr,
 	DbAdrInfo info)
 {
     ADDRESS *addressPtr = NULL;
-    char *s = strdup(adr);
+    char *s = cpystr(adr);
     int ret;
 
     rfc822_parse_adrlist(&addressPtr, s, currentHost);
@@ -450,10 +426,11 @@ Db_InfoProc(Tcl_Interp *interp, ClientData clientData, RatFolderInfoType type,
     static Tcl_DString ds;
     static int initialized = 0;
     static char buf[1024];
-    int i, seen, deleted, marked, answered, me, dbIndex;
+    int i, seen, deleted, marked, answered, me, dbIndex, zone;
     RatFolderInfo *infoPtr = (RatFolderInfo*)clientData;
     DbFolderInfo *dbPtr = (DbFolderInfo*)infoPtr->private;
     ADDRESS *addressPtr, *address2Ptr;
+    MESSAGECACHE elt;
     RatDbEntry *entryPtr;
     struct tm *tmPtr;
     Tcl_CmdInfo info;
@@ -484,6 +461,10 @@ Db_InfoProc(Tcl_Interp *interp, ClientData clientData, RatFolderInfoType type,
 		GetAddressInfo(interp, &ds, entryPtr->content[TO], Db_Name);
 	    }
 	    return Tcl_DStringValue(&ds);
+	case RAT_FOLDER_MAIL_REAL:
+	    Tcl_DStringSetLength(&ds, 0);
+	    GetAddressInfo(interp, &ds, entryPtr->content[FROM], Db_Mail);
+	    return Tcl_DStringValue(&ds);
 	case RAT_FOLDER_MAIL:
 	    Tcl_DStringSetLength(&ds, 0);
 	    if (GetAddressInfo(interp, &ds, entryPtr->content[FROM], Db_Mail)) {
@@ -523,30 +504,39 @@ Db_InfoProc(Tcl_Interp *interp, ClientData clientData, RatFolderInfoType type,
 	case RAT_FOLDER_SIZE:
 	    return entryPtr->content[RSIZE];
 	case RAT_FOLDER_SIZE_F:
-	    Tcl_DStringSetLength(&ds, 256);
-	    sprintf(Tcl_DStringValue(&ds), "RatMangleNumber %s",
-		    entryPtr->content[RSIZE]);
-	    Tcl_GlobalEval(interp, Tcl_DStringValue(&ds));
-	    Tcl_DStringGetResult(interp, &ds);
-	    return Tcl_DStringValue(&ds);
+	    return RatMangleNumber(atoi(entryPtr->content[RSIZE]));
 	case RAT_FOLDER_DATE_F:
+	    time = atoi(entryPtr->content[DATE]);
+	    tmPtr = localtime(&time);
+	    return RatFormatDate(interp, tmPtr->tm_mon, tmPtr->tm_mday);
+	case RAT_FOLDER_DATE_N:
+	    return entryPtr->content[DATE];
+	case RAT_FOLDER_DATE_IMAP4:
 	    Tcl_DStringSetLength(&ds, 256);
 	    time = atoi(entryPtr->content[DATE]);
 	    tmPtr = localtime(&time);
-	    sprintf(Tcl_DStringValue(&ds), "RatFormatDate %d %d %d %d %d %d",
-		    tmPtr->tm_year+1970, tmPtr->tm_mon+1,
-		    tmPtr->tm_mday, tmPtr->tm_hour,
-		    tmPtr->tm_min, tmPtr->tm_sec);
-	    Tcl_GlobalEval(interp, Tcl_DStringValue(&ds));
-	    Tcl_DStringGetResult(interp, &ds);
-	    return Tcl_DStringValue(&ds);
-	case RAT_FOLDER_DATE_N:
-	    return entryPtr->content[DATE];
+	    elt.day = tmPtr->tm_mday;
+	    elt.month = tmPtr->tm_mon+1;
+	    elt.year = tmPtr->tm_year+1900-BASEYEAR;
+	    elt.hours = tmPtr->tm_hour;
+	    elt.minutes = tmPtr->tm_min;
+	    elt.seconds = tmPtr->tm_sec;
+	    zone = RatGetTimeZone(time);
+	    if (zone >= 0) {
+		elt.zoccident = 1;
+		elt.zhours = zone/(60*60);
+		elt.zminutes = (zone%(60*60))/60;
+	    } else {
+		elt.zoccident = 0;
+		elt.zhours = (-1*zone)/(60*60);
+		elt.zminutes = ((-1*zone)%(60*60))/60;
+	    }
+	    return mail_date(Tcl_DStringValue(&ds), &elt);
 	case RAT_FOLDER_STATUS:
 	    seen = deleted = marked = answered = me = 0;
 	    for (i=0; entryPtr->content[STATUS][i]; i++) {
 		switch (entryPtr->content[STATUS][i]) {
-		    case 'R': seen = 1;	break;
+		    case 'R': seen = 1;		break;
 		    case 'D': deleted = 1;	break;
 		    case 'F': marked = 1;	break;
 		    case 'A': answered = 1;	break;
@@ -610,6 +600,25 @@ Db_InfoProc(Tcl_Interp *interp, ClientData clientData, RatFolderInfoType type,
 	    }
 	    Tcl_GetCommandInfo(interp, infoPtr->msgCmdPtr[index], &info);
 	    return (*infoPtr->infoProc)(interp, info.clientData, type, index);
+	case RAT_FOLDER_FLAGS:
+	    Tcl_DStringSetLength(&ds, 0);
+	    for (i=0; entryPtr->content[STATUS][i]; i++) {
+		switch (entryPtr->content[STATUS][i]) {
+		    case 'R':
+			Tcl_DStringAppend(&ds, " " RAT_SEEN_STR, -1); break;
+		    case 'D':
+			Tcl_DStringAppend(&ds, " " RAT_DELETED_STR, -1); break;
+		    case 'F':
+			Tcl_DStringAppend(&ds, " " RAT_FLAGGED_STR, -1); break;
+		    case 'A':
+			Tcl_DStringAppend(&ds, " " RAT_ANSWERED_STR, -1); break;
+		}
+	    }
+	    if (Tcl_DStringLength(&ds)) {
+		return Tcl_DStringValue(&ds)+1;
+	    } else {
+		return "";
+	    }
 	case RAT_FOLDER_END:
 	    break;
     }
