@@ -1,7 +1,8 @@
 /*
  * ratDbase.c --
  *
- * TkRat software and its included text is Copyright 1996 by Martin Forssen.
+ * TkRat software and its included text is Copyright 1996,1997,1998
+ * by Martin Forssn
  *
  * The full text of the legal notice is contained in the file called
  * COPYRIGHT, included with this distribution.
@@ -58,7 +59,7 @@
  *	contained in a .seq file found in said directory.
  */
 
-#include "rat.h"
+#include "ratFolder.h"
 
 #define DBASE_VERSION 3		/* Version of the database format */
 #define EXTRA_ENTRIES 100	/* How many extra entries we should allocate
@@ -101,7 +102,7 @@ static void	Update(ClientData clientData);
 static int	IsRlocked(char *ignore);
 static void	RatDbBuildList(Tcl_Interp *interp, Tcl_DString *dsPtr,
 	char *prefix, char *dir, Tcl_HashTable *tablePtr, int fix);
-static int	NoLFPrint(FILE *fp, unsigned char *s);
+static int	NoLFPrint(FILE *fp, char *s);
 
 
 
@@ -154,7 +155,7 @@ Read(Tcl_Interp *interp)
 	    Tcl_SetResult(interp, "option(dbase_dir) has no value", TCL_STATIC);
 	    return TCL_ERROR;
 	}
-	dbDir = strdup(Tcl_TildeSubst(interp, value, &ds));
+	dbDir = cpystr(Tcl_TildeSubst(interp, value, &ds));
 	Tcl_DStringFree(&ds);
     }
     if (0 == ident) {
@@ -311,7 +312,7 @@ Read(Tcl_Interp *interp)
 		char buf[64];
 		sprintf(buf, "%d", atoi(entryPtr[i].content[EX_TYPE])*24*60*60+
 			    atoi(entryPtr[i].content[DATE]));
-		entryPtr[i].content[EX_TIME] = strdup(buf);
+		entryPtr[i].content[EX_TIME] = cpystr(buf);
 		entryPtr[i].content[EX_TYPE] = "backup";
 	    }
 	}
@@ -448,118 +449,136 @@ Unlock(Tcl_Interp *interp)
 static int
 Sync(Tcl_Interp *interp)
 {
-    char buf[1024];	/* Scratch area */
-    struct stat sbuf;	/* Buffer for stat() calls */
-    FILE *fpChanges;	/* index.changes file pointer */
-    FILE *fpIndex;	/* Index file pointer */
-    char command;	/* Command in changes file */
-    int cmdArg;		/* Argument to command */
-    int i;		/* Loop counter */
-    int doWrite = 0;	/* 1 if we should write the changes to the disk */
-    int needRewrite = 0;/* 1 if we need to rewrite the indexfile */
-    int numEntries = 0;	/* How many entries there actually are */
+    char buf[1024];	   /* Scratch area */
+    struct stat sbuf;	   /* Buffer for stat() calls */
+    FILE *fpChanges;	   /* index.changes file pointer */
+    FILE *fpIndex;	   /* Index file pointer */
+    char command;	   /* Command in changes file */
+    int cmdArg;		   /* Argument to command */
+    int i;		   /* Loop counter */
+    int doWrite = 0;	   /* 1 if we should write the changes to the disk */
+    int needRewrite = 0;   /* 1 if we need to rewrite the indexfile */
+    int numEntries = 0;	   /* How many entries there actually are */
+    char *indexBuf = NULL; /* New part of index file */
+    int indexOffset = 0;   /* Offset of new part of index file */
+    char *cPtr;
 
     sprintf(buf, "%s/index.changes", dbDir);
-    if (0 == stat(buf, &sbuf)) {
-	if (changeSize < sbuf.st_size) {
-	    /*
-	     * Read and perform changes mentioned in index.changes file
-	     */
-	    if (0 == (fpChanges = fopen(buf, "r"))) {
-		Tcl_AppendResult(interp, "error opening file (for reading) \"",
-			buf, "\": ", Tcl_PosixError(interp), (char *) NULL);
-		return TCL_ERROR;
-	    }
-	    if (0 != fseek(fpChanges, changeSize, SEEK_SET)) {
-		Tcl_AppendResult(interp, "error seeking in file \"", buf,
-			"\": ", Tcl_PosixError(interp), (char *) NULL);
-		fclose(fpChanges);
-		return TCL_ERROR;
-	    }
-	    while(1) {
-		if (2 != fscanf(fpChanges, "%c %d ", &command, &cmdArg) ||
-			('d'!=command && 'a'!=command && 's'!=command)
-			|| ('s' == command &&
-			buf != fgets(buf, sizeof(buf), fpChanges))) {
-		    if (feof(fpChanges)) {
-			break;
-		    }
-		    Tcl_SetResult(interp, "syntax error in changes file",
-			    TCL_STATIC);
-		    fclose(fpChanges);
-		    return TCL_ERROR;
-		}
-		numChanges++;
-		if ('d' == command) {
-		    if (cmdArg < 0 || cmdArg >= numRead) {
-			continue;
-		    }
-		    needRewrite = 1;
-		    entryPtr[cmdArg].content[FROM] = NULL;
-		} else if ('s' == command) {
-		    if (cmdArg < 0 || cmdArg >= numRead) {
-			continue;
-		    }
-		    needRewrite = 1;
-		    buf[strlen(buf)-1] = '\0';
-		    if ( (int) strlen(buf) <=
-			    (int) strlen(entryPtr[cmdArg].content[STATUS])){
-			strcpy(entryPtr[cmdArg].content[STATUS], buf);
-		    } else {
-			/*
-			 * This code may leak the memory occupied by the
-			 * previous status string. I believe this loss can
-			 * be lived with (it should be quite rare).
-			 */
-			entryPtr[cmdArg].content[STATUS] =
-				(char *) ckalloc(strlen(buf)+1);
-			strcpy(entryPtr[cmdArg].content[STATUS], buf);
-		    }
-		} else {
-		    if (numRead == numAlloc) {
-			numAlloc += EXTRA_ENTRIES;
-			entryPtr = (RatDbEntry*)REALLOC(entryPtr,
-				numAlloc*sizeof(RatDbEntry));
-		    }
-		    sprintf(buf, "%s/index", dbDir);
-		    if (0 == (fpIndex = fopen(buf, "r"))) {
-			Tcl_AppendResult(interp,
-				"error opening file (for reading) \"", buf,
-				"\": ", Tcl_PosixError(interp), (char *) NULL);
-			fclose(fpChanges);
-			return TCL_ERROR;
-		    }
-		    if (0 != fseek(fpIndex, cmdArg, SEEK_SET)) {
-			Tcl_AppendResult(interp, "error seeking in file \"",buf,
-				"\": ", Tcl_PosixError(interp), (char *) NULL);
-			fclose(fpChanges);
-			return TCL_ERROR;
-		    }
-		    for (i=0; i<RATDBETYPE_END; i++) {
-			if (buf != fgets(buf, sizeof(buf), fpIndex)) {
-			    Tcl_AppendResult(interp, "error reading \"",buf,
-				    "\": ", Tcl_PosixError(interp),(char*)NULL);
-			    fclose(fpChanges);
-			    fclose(fpIndex);
-			    return TCL_ERROR;
-			}
-			buf[strlen(buf)-1] = '\0';
-			entryPtr[numRead].content[i] = strdup(buf);
-		    }
-		    numRead++;
-		    fclose(fpIndex);
-		}
-	    }
-	    fclose(fpChanges);
-	    changeSize = sbuf.st_size;
-	}
-    } else {
+    if (0 > stat(buf, &sbuf)) {
 	if (ENOENT != errno) {
 	    Tcl_AppendResult(interp, "error stating file \"", buf,
 		    "\": ", Tcl_PosixError(interp), (char *) NULL);
 	    return TCL_ERROR;
 	}
+	return TCL_OK;
     }
+    if (changeSize >= sbuf.st_size) {
+	return TCL_OK;
+    }
+
+    /*
+     * Read and perform changes mentioned in index.changes file
+     */
+    if (0 == (fpChanges = fopen(buf, "r"))) {
+	Tcl_AppendResult(interp, "error opening file (for reading) \"",
+		buf, "\": ", Tcl_PosixError(interp), (char *) NULL);
+	return TCL_ERROR;
+    }
+    if (0 != fseek(fpChanges, changeSize, SEEK_SET)) {
+	Tcl_AppendResult(interp, "error seeking in file \"", buf,
+		"\": ", Tcl_PosixError(interp), (char *) NULL);
+	fclose(fpChanges);
+	return TCL_ERROR;
+    }
+    changeSize = sbuf.st_size;
+    while(1) {
+	if (2 != fscanf(fpChanges, "%c %d ", &command, &cmdArg) ||
+		('d'!=command && 'a'!=command && 's'!=command)
+		|| ('s' == command &&
+		buf != fgets(buf, sizeof(buf), fpChanges))) {
+	    if (feof(fpChanges)) {
+		break;
+	    }
+	    Tcl_SetResult(interp, "syntax error in changes file",
+		    TCL_STATIC);
+	    fclose(fpChanges);
+	    return TCL_ERROR;
+	}
+	numChanges++;
+	if ('d' == command) {
+	    if (cmdArg < 0 || cmdArg >= numRead) {
+		continue;
+	    }
+	    needRewrite = 1;
+	    entryPtr[cmdArg].content[FROM] = NULL;
+	} else if ('s' == command) {
+	    if (cmdArg < 0 || cmdArg >= numRead) {
+		continue;
+	    }
+	    needRewrite = 1;
+	    buf[strlen(buf)-1] = '\0';
+	    if ( (int) strlen(buf) <=
+		    (int) strlen(entryPtr[cmdArg].content[STATUS])){
+		strcpy(entryPtr[cmdArg].content[STATUS], buf);
+	    } else {
+		/*
+		 * This code may leak the memory occupied by the
+		 * previous status string. I believe this loss can
+		 * be lived with (it should be quite rare).
+		 */
+		entryPtr[cmdArg].content[STATUS] =
+			(char *) ckalloc(strlen(buf)+1);
+		strcpy(entryPtr[cmdArg].content[STATUS], buf);
+	    }
+	} else {
+	    if (numRead == numAlloc) {
+		numAlloc += EXTRA_ENTRIES;
+		entryPtr = (RatDbEntry*)REALLOC(entryPtr,
+			numAlloc*sizeof(RatDbEntry));
+	    }
+	    if (!indexBuf) {
+		sprintf(buf, "%s/index", dbDir);
+		if (NULL == (fpIndex = fopen(buf, "r"))) {
+		    Tcl_AppendResult(interp,
+			    "error opening file (for reading) \"", buf,
+			    "\": ", Tcl_PosixError(interp), (char *) NULL);
+		    fclose(fpChanges);
+		    return TCL_ERROR;
+		}
+		if (0 != fseek(fpIndex, cmdArg, SEEK_SET)) {
+		    Tcl_AppendResult(interp, "error seeking in file \"",buf,
+			    "\": ", Tcl_PosixError(interp), (char *) NULL);
+		    fclose(fpIndex);
+		    fclose(fpChanges);
+		    return TCL_ERROR;
+		}
+		(void)fstat(fileno(fpIndex), &sbuf);
+		/*
+		 * Purify will probably report this as a leak, but that
+		 * is not true.
+		 */
+		indexBuf = (char*)malloc(sbuf.st_size - cmdArg + 1);
+		fread(indexBuf, sbuf.st_size - cmdArg, 1, fpIndex);
+		fclose(fpIndex);
+		indexBuf[sbuf.st_size - cmdArg] = '\0';
+		indexOffset = cmdArg;
+	    }
+	    cPtr = indexBuf + (cmdArg - indexOffset);
+	    for (i=0; i<RATDBETYPE_END; i++) {
+		entryPtr[numRead].content[i] = cPtr;
+		for (; *cPtr != '\n' && *cPtr; cPtr++);
+		if (!*cPtr) {
+		    Tcl_AppendResult(interp, "error reading \"",buf,
+			    "\": ", Tcl_PosixError(interp),(char*)NULL);
+		    fclose(fpChanges);
+		    return TCL_ERROR;
+		}
+		*cPtr++ = '\0';
+	    }
+	    numRead++;
+	}
+    }
+    fclose(fpChanges);
 
     /* 
      * If the number of changes is at least 20 and we are the only agent
@@ -733,9 +752,9 @@ IsRlocked(char *ignore)
 
     dirPtr = opendir(dbDir);
     while (0 != (direntPtr = readdir(dirPtr))) {
-	if (!strncmp("rlock.", (char*)&direntPtr->d_name, 6)
-		&& (ignore && strcmp((char*)&direntPtr->d_name, ignore))) {
-	    sprintf(buf, "%s/%s", dbDir, (char*)&direntPtr->d_name);
+	if (!strncmp("rlock.", (char*)direntPtr->d_name, 6)
+		&& (ignore && strcmp((char*)direntPtr->d_name, ignore))) {
+	    sprintf(buf, "%s/%s", dbDir, (char*)direntPtr->d_name);
 	    (void)stat(buf, &sbuf);
 	    if (deadline < sbuf.st_mtime) {
 		result = 1;
@@ -778,7 +797,7 @@ IsRlocked(char *ignore)
 int
 RatDbInsert(Tcl_Interp *interp, char *to, char *from, char *cc, char *subject,
 	    int date, char *flags, char *keywords, int exDate, char *exType,
-	    char *mail, int length)
+	    char *fromline, char *mail, int length)
 {
     char fname[1024];		/* filename of new entry */
     char buf[1024];		/* Scratch area */
@@ -919,6 +938,7 @@ RatDbInsert(Tcl_Interp *interp, char *to, char *from, char *cc, char *subject,
 		"\": ", Tcl_PosixError(interp), (char *) NULL);
 	goto losing;
     }
+    Tcl_Write(dataChannel, fromline, strlen(fromline));
     Tcl_Write(dataChannel, mail, length);
     if (TCL_OK != Tcl_Close(interp, dataChannel)) {
 	Tcl_AppendResult(interp, "error closing file \"", buf,
@@ -1048,7 +1068,7 @@ RatDbSetStatus(Tcl_Interp *interp, int index, char *status)
  *      The number of items found is returned in numFoundPtr if it is
  *	non zero a pointer to a list of found ones is put into *foundPtrPtr.
  *	It is the callers resopnsibility to free this list with a
- *	call to malloc(). The rotine normally returns TCL_OK except when
+ *	call to free(). The rotine normally returns TCL_OK except when
  *	there is an error. In that case TCL_ERROR is returned and the cause
  *	may be found in interp->error.
  *
@@ -1268,7 +1288,9 @@ RatDbGetEntry(int index)
  *	returns a MESSAGE structure.
  *
  * Results:
- *	A pointer to a MESSAGE* structure. 
+ *	A pointer to a MESSAGE* structure. It alse fills in the pointer
+ *	to a buffer among the arguments with the address that needs to be
+ *	freed when the message is deleted.
  *
  * Side effects:
  *      None.
@@ -1277,7 +1299,7 @@ RatDbGetEntry(int index)
  */
 
 MESSAGE*
-RatDbGetMessage(Tcl_Interp *interp, int index)
+RatDbGetMessage(Tcl_Interp *interp, int index, char **buffer)
 {
     char fname[1024];	/* Filename of message */
     int messfd;		/* Message file descriptor */
@@ -1315,7 +1337,7 @@ RatDbGetMessage(Tcl_Interp *interp, int index)
 	close(messfd);
 	return NULL;
     }
-    message = (char*)ckalloc(sbuf.st_size+1);
+    *buffer = message = (char*)ckalloc(sbuf.st_size+1);
     read(messfd, message, sbuf.st_size);
     message[sbuf.st_size] = '\0';
     (void)close(messfd);
@@ -1650,7 +1672,7 @@ RatDbDaysSinceExpire(Tcl_Interp *interp)
 	    Tcl_SetResult(interp, "option(dbase_dir) has no value", TCL_STATIC);
 	    return TCL_ERROR;
 	}
-	dbDir = strdup(Tcl_TildeSubst(interp, value, &ds));
+	dbDir = cpystr(Tcl_TildeSubst(interp, value, &ds));
 	Tcl_DStringFree(&ds);
     }
 
@@ -1661,7 +1683,11 @@ RatDbDaysSinceExpire(Tcl_Interp *interp)
 	    return 0;
 	}
     }
-    return (time(NULL)-sbuf.st_mtime)/(24*60*60);
+    if (sbuf.st_mtime > time(NULL)) {
+	return 0;
+    } else {
+	return (time(NULL)-sbuf.st_mtime)/(24*60*60);
+    }
 }
 
 
@@ -1731,8 +1757,8 @@ RatDbExpire(Tcl_Interp *interp, char *infolder, char *backupDirectory)
 	    return TCL_ERROR;
 	}
     }
-    compressProg = Tcl_GetVar(interp, "compressProg", TCL_GLOBAL_ONLY);
-    compressSuffix = Tcl_GetVar(interp, "compressSuffix", TCL_GLOBAL_ONLY);
+    compressProg = getenv("COMPRESS");
+    compressSuffix = getenv("CSUFFIX");
     backupDir = Tcl_GetVar2(interp, "option", "dbase_backup", TCL_GLOBAL_ONLY);
     if (!compressProg || !compressSuffix || !backupDir) {
 	Tcl_AppendResult(interp, "Internal error: compressProg, ",
@@ -1743,7 +1769,7 @@ RatDbExpire(Tcl_Interp *interp, char *infolder, char *backupDirectory)
 
     strcpy(buf, "RatLog 2 $t(db_expire) explicit");
     Tcl_Eval(interp, buf);
-    statusId = strdup(interp->result);
+    statusId = cpystr(interp->result);
     Lock(interp);
     for (i=0; !error && i < numRead; i++) {
 	if (!entryPtr[i].content[FROM]) {	/* Entry deleted */
@@ -1773,7 +1799,7 @@ RatDbExpire(Tcl_Interp *interp, char *infolder, char *backupDirectory)
 	    RatGenId(NULL, interp, 0, NULL);
 	    sprintf(buf2, "%s/message.%s", backupDirectory, interp->result);
 
-	} else if (!strcmp("inbox", entryPtr[i].content[EX_TYPE])) {
+	} else if (!strcmp("incoming", entryPtr[i].content[EX_TYPE])) {
 	    move = 1;
 	    numInbox++;
 	    sprintf(buf, "%s/dbase/%s", dbDir, entryPtr[i].content[FILENAME]);
@@ -1875,7 +1901,7 @@ RatDbExpire(Tcl_Interp *interp, char *infolder, char *backupDirectory)
 	free(statusId);
 	strcpy(buf, "RatLog 2 $t(packing_backup) explicit");
 	Tcl_Eval(interp, buf);
-	statusId = strdup(interp->result);
+	statusId = cpystr(interp->result);
 	tmPtr = localtime(&now);
 	sprintf(buf2, ">%s/backup_%02d%02d%02d%s", backupDirectory,
 		tmPtr->tm_year+1900, tmPtr->tm_mon+1, tmPtr->tm_mday+1,
@@ -1949,7 +1975,7 @@ RatDbExpire(Tcl_Interp *interp, char *infolder, char *backupDirectory)
 	    if (*(++cPtr) == '\r') {
 		cPtr++;
 	    }
-	    msg = RatFrMessageCreate(interp, cPtr, sbuf.st_size);
+	    msg = RatFrMessageCreate(interp, cPtr, sbuf.st_size, NULL);
 	    if (TCL_OK == Tcl_VarEval(interp, infolder, " insert ", msg,NULL)){
 		unlink(buf2);
 	    }
@@ -2192,7 +2218,7 @@ RatDbCheck(Tcl_Interp *interp, int fix)
 	if (NULL == value) {
 	    return TCL_OK;
 	}
-	dbDir = strdup(Tcl_TildeSubst(interp, value, &ds));
+	dbDir = cpystr(Tcl_TildeSubst(interp, value, &ds));
 	Tcl_DStringFree(&ds);
     }
     if (0 == ident) {
@@ -2201,6 +2227,16 @@ RatDbCheck(Tcl_Interp *interp, int fix)
 	sprintf(ident, "%s:%d", buf, (int)getpid());
     }
 
+    /*
+     * Check that the database directory exists. If not we return zeros
+     */
+    if (0 > stat(dbDir, &sbuf) ||
+	    !S_ISDIR(sbuf.st_mode)) {
+	Tcl_SetResult(interp, "0 0 0 0 0 {}", TCL_STATIC);
+	return TCL_OK;
+    }
+
+
     /*
      * Lock the database. We should also check that nobody else has
      * a read lock as well.
@@ -2441,17 +2477,20 @@ RatDbCheck(Tcl_Interp *interp, int fix)
 			flags = cpystr(elemArgv[1]);
 		    }
 		} else if (!strcasecmp(elemArgv[0], "date")) {
-		    mail_parse_date(&elt, elemArgv[1]);
-		    tm.tm_sec = elt.seconds;
-		    tm.tm_min = elt.minutes;
-		    tm.tm_hour = elt.hours;
-		    tm.tm_mday = elt.day;
-		    tm.tm_mon = elt.month - 1;
-		    tm.tm_year = elt.year+70;
-		    tm.tm_wday = 0;
-		    tm.tm_yday = 0;
-		    tm.tm_isdst = -1;
-		    date = (int)mktime(&tm);
+		    if (T == mail_parse_date(&elt, elemArgv[1])) {
+			tm.tm_sec = elt.seconds;
+			tm.tm_min = elt.minutes;
+			tm.tm_hour = elt.hours;
+			tm.tm_mday = elt.day;
+			tm.tm_mon = elt.month - 1;
+			tm.tm_year = elt.year+70;
+			tm.tm_wday = 0;
+			tm.tm_yday = 0;
+			tm.tm_isdst = -1;
+			date = (int)mktime(&tm);
+		    } else {
+			date = 0;
+		    }
 		}
 		ckfree(elemArgv);
 	    }
@@ -2464,13 +2503,13 @@ RatDbCheck(Tcl_Interp *interp, int fix)
 	    itemPtr->entry.content[CC] = cc ? cc : "";
 	    itemPtr->entry.content[SUBJECT] = subject ? subject : "";
 	    sprintf(buf, "%d", date);
-	    itemPtr->entry.content[DATE] = extraPtrPtr[extraNum++] =strdup(buf);
+	    itemPtr->entry.content[DATE] = extraPtrPtr[extraNum++] =cpystr(buf);
 	    itemPtr->entry.content[KEYWORDS] = "LostMessage";
 	    sprintf(buf, "%d", itemPtr->fileSize);
-	    itemPtr->entry.content[RSIZE] = extraPtrPtr[extraNum++]=strdup(buf);
+	    itemPtr->entry.content[RSIZE] = extraPtrPtr[extraNum++]=cpystr(buf);
 	    itemPtr->entry.content[STATUS] = flags ? flags : "";
 	    sprintf(buf, "%ld", time(NULL)+60*60*24*100);
-	    itemPtr->entry.content[EX_TIME]=extraPtrPtr[extraNum++]=strdup(buf);
+	    itemPtr->entry.content[EX_TIME]=extraPtrPtr[extraNum++]=cpystr(buf);
 	    itemPtr->entry.content[EX_TYPE] = "backup";
 	    itemPtr->entry.content[FILENAME] = Tcl_GetHashKey(&items, entryPtr);
 	}
@@ -2580,11 +2619,11 @@ RatDbCheck(Tcl_Interp *interp, int fix)
  */
 
 static int
-NoLFPrint(FILE *fp, unsigned char *s)
+NoLFPrint(FILE *fp, char *s)
 {
     unsigned char *cPtr;
 
-    for (cPtr = s; cPtr && *cPtr; cPtr++) {
+    for (cPtr = (unsigned char*)s; cPtr && *cPtr; cPtr++) {
 	if ('\n' == *cPtr) {
 	    if (isspace(cPtr[1])) {
 		cPtr++;
