@@ -6,7 +6,8 @@
  *      a folder handler, which when invoked calls the RatFolderCmd()
  *      procedure with a pointer to a RatFolderInfo structure as clientData.
  *
- * TkRat software and its included text is Copyright 1996 by Martin Forssen.
+ * TkRat software and its included text is Copyright 1996,1997,1998
+ * by Martin Forssn
  *
  * The full text of the legal notices is contained in the file called
  * COPYRIGHT, included with this distribution.
@@ -14,6 +15,22 @@
 
 #include "ratFolder.h"
 
+/*
+ * This structure is used to hold the data while sorting
+ */
+typedef struct SortData {
+    char *subject;
+    char *sender;
+    time_t date;
+    struct SortData *nextPtr;
+} SortData;
+
+/*
+ * The different sort methods
+ */
+typedef enum {SORT_NONE, SORT_SUBJECT, SORT_SUBJDATE,
+	      SORT_SENDER, SORT_SENDERDATE, SORT_DATE} SortOrder;
+
 /*
  * The number of folders opened. This is used when making
  * the folder entities.
@@ -21,20 +38,20 @@
 static int numFolders = 0;
 
 /*
- * Global variables used to controll the RatFolderSortCompare function
+ * Global variables used to controll the sorting functions
  */
-static char **rat_SortList;
-static int rat_SortOrder;
+static SortData *baseSortDataPtr;
 
-static int RatFolderOpen _ANSI_ARGS_((ClientData clientData,
-	Tcl_Interp *interp, int argc, char *argv[]));
-static int RatFolderCmd _ANSI_ARGS_((ClientData clientData,
-	Tcl_Interp *interp, int argc, char *argv[]));
-static void RatFolderSort _ANSI_ARGS_((Tcl_Interp *interp,
-	RatFolderInfo *infoPtr));
-static int RatFolderSortCompare _ANSI_ARGS_((const void *arg1,
-	const void *arg2));
-static int RatFolderSubjSame _ANSI_ARGS_((char *subj1, char *subj2));
+
+static int RatFolderOpen(ClientData clientData,
+	Tcl_Interp *interp, int argc, char *argv[]);
+static int RatFolderCmd(ClientData clientData,
+	Tcl_Interp *interp, int argc, char *argv[]);
+static void RatFolderSort(Tcl_Interp *interp, RatFolderInfo *infoPtr);
+static int RatFolderSortCompareDate(const void *arg1, const void *arg2);
+static int RatFolderSortCompareSubject(const void *arg1, const void *arg2);
+static int RatFolderSortCompareSender(const void *arg1, const void *arg2);
+static char *RatFolderSubjCopy (char *subject);
 
 
 /*
@@ -99,12 +116,8 @@ static int
 RatFolderOpen(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
 {
     RatFolderInfo *infoPtr;
-    char buf[256];
     int i;
 
-    if (!currentHost) {
-	RatInitCurrent(interp);
-    }
     if (argc < 2) {
 	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
 		" prot [args]\"", (char *) NULL);
@@ -123,18 +136,22 @@ RatFolderOpen(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
     if (NULL == infoPtr) {
 	return TCL_ERROR;
     }
+    infoPtr->cmdName = ckalloc(16);
     infoPtr->allocated = infoPtr->number;
     infoPtr->msgCmdPtr = (char **) ckalloc(infoPtr->allocated*sizeof(char*));
+    infoPtr->privatePtr = (ClientData**)ckalloc(
+	    infoPtr->allocated*sizeof(ClientData));
     for (i=0; i<infoPtr->allocated; i++) {
 	infoPtr->msgCmdPtr[i] = (char *) NULL;
+	infoPtr->privatePtr[i] = (ClientData*) NULL;
     }
     infoPtr->presentationOrder = (int*) ckalloc(infoPtr->allocated*sizeof(int));
     infoPtr->hidden = (int*) ckalloc(infoPtr->allocated*sizeof(int));
     RatFolderSort(interp, infoPtr);
-    sprintf(buf, "RatFolder%d", numFolders++);
-    Tcl_CreateCommand(interp, buf, RatFolderCmd,
+    sprintf(infoPtr->cmdName, "RatFolder%d", numFolders++);
+    Tcl_CreateCommand(interp, infoPtr->cmdName, RatFolderCmd,
     	    (ClientData) infoPtr, (Tcl_CmdDeleteProc *) NULL);
-    Tcl_SetResult(interp, buf, TCL_VOLATILE);
+    Tcl_SetResult(interp, infoPtr->cmdName, TCL_VOLATILE);
     return TCL_OK;
 }
 
@@ -175,41 +192,46 @@ RatFolderCmd(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
     length = strlen(argv[1]);
     if ((c == 'u') && (strncmp(argv[1], "update", length) == 0)
 	    && (length > 1)) {
-	int i, number, oldNumber;
+	int i, number, expunge, oldNumber, oldVisible, delta;
 
-	if (argc != 2) {
+	if (argc != 3 || TCL_OK != Tcl_GetBoolean(interp, argv[2], &expunge)) {
 	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
-		    " update\"", (char *) NULL);
+		    " update expunge\"", (char *) NULL);
 	    return TCL_ERROR;
 	}
-	oldNumber = infoPtr->visible;
-	number = (*infoPtr->updateProc)(infoPtr, interp);
+	oldVisible = infoPtr->visible;
+	oldNumber = infoPtr->number;
+	number = (*infoPtr->updateProc)(infoPtr, interp, expunge);
 	if (number < 0) {
 	    RatLog(interp, RAT_WARN, interp->result, 0);
 	    number = 0;
 	} else {
-	    infoPtr->number += number;
+	    delta = number - oldNumber;
+	    infoPtr->number = number;
+	    if (delta <0) {
+		delta = 0;
+	    }
 	    if (infoPtr->number > infoPtr->allocated) {
 		infoPtr->allocated = infoPtr->number;
 		infoPtr->msgCmdPtr = (char **) REALLOC(infoPtr->msgCmdPtr,
 			infoPtr->allocated*sizeof(char*));
+		infoPtr->privatePtr = (ClientData**)REALLOC(infoPtr->privatePtr,
+			infoPtr->allocated*sizeof(ClientData*));
 		infoPtr->presentationOrder = (int *) REALLOC(
 			infoPtr->presentationOrder,
 			infoPtr->allocated*sizeof(int));
 		infoPtr->hidden = (int *) REALLOC(infoPtr->hidden,
 			infoPtr->allocated*sizeof(int));
 	    }
-	    for (i=infoPtr->number-number; i<infoPtr->number; i++) {
+	    for (i=infoPtr->number-delta; i<infoPtr->number; i++) {
 		infoPtr->msgCmdPtr[i] = (char *) NULL;
+		infoPtr->privatePtr[i] = (ClientData*) NULL;
 	    }
 	    RatFolderSort(interp, infoPtr);
-	    number = infoPtr->visible - oldNumber;
-	    if (number < 0) {
-		number = 0;
-	    }
 	}
+	delta = infoPtr->visible - oldVisible;
 	Tcl_ResetResult(interp);
-	sprintf(interp->result, "%d", number);
+	sprintf(interp->result, "%d", (delta>0 ? delta : 0));
 	return TCL_OK;
 
     } if ((c == 'c') && (strncmp(argv[1], "close", length) == 0)
@@ -223,15 +245,17 @@ RatFolderCmd(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
 	ckfree(infoPtr->name);
 	for(i=0; i < infoPtr->number; i++) {
 	    if (NULL != infoPtr->msgCmdPtr[i]) {
-		(void)Tcl_DeleteCommand(interp, infoPtr->msgCmdPtr[i]);
+		(void)RatMessageDelete(interp, infoPtr->msgCmdPtr[i]);
 	    }
 	}
 	ret = (*infoPtr->closeProc)(infoPtr, interp);
+	(void)Tcl_DeleteCommand(interp, argv[0]);
+	ckfree(infoPtr->cmdName);
 	ckfree(infoPtr->msgCmdPtr);
+	ckfree(infoPtr->privatePtr);
 	ckfree(infoPtr->presentationOrder);
 	ckfree(infoPtr->hidden);
 	ckfree(infoPtr);
-	(void)Tcl_DeleteCommand(interp, argv[0]);
 	return ret;
 
     } if ((c == 's') && (strncmp(argv[1], "setName", length) == 0)
@@ -419,7 +443,7 @@ RatFolderCmd(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
 	}
 
 	Tcl_ResetResult(interp);
-	for (i=0; i<infoPtr->number; i++) {
+	for (i=0; i<infoPtr->visible; i++) {
 	    if ((*infoPtr->getFlagProc)(infoPtr, interp,
 		    infoPtr->presentationOrder[i], flag)) {
 		sprintf(buf, "%d", i);
@@ -428,17 +452,6 @@ RatFolderCmd(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
 	}
 	return TCL_OK;
 
-    } if ((c == 'e') && (strncmp(argv[1], "expunge", length) == 0)
-	    && (length > 1)) {
-	if (argc != 2) {
-	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
-		    " expunge\"", (char *) NULL);
-	    return TCL_ERROR;
-	}
-	(*infoPtr->expungeProc)(infoPtr, interp);
-	RatFolderSort(interp, infoPtr);
-	return TCL_OK;
-
     } if ((c == 'i') && (strncmp(argv[1], "insert", length) == 0)
 	    && (length > 2)) {
 	Tcl_CmdInfo cmdInfo;
@@ -454,6 +467,8 @@ RatFolderCmd(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
 	    infoPtr->allocated += argc;
 	    infoPtr->msgCmdPtr = (char**)REALLOC(infoPtr->msgCmdPtr,
 	    	    infoPtr->allocated*sizeof(char*));
+	    infoPtr->privatePtr = (ClientData**)REALLOC(infoPtr->privatePtr,
+	    	    infoPtr->allocated*sizeof(ClientData*));
 	    infoPtr->presentationOrder = (int*)REALLOC(
 		    infoPtr->presentationOrder, infoPtr->allocated*sizeof(int));
 	}
@@ -488,8 +503,9 @@ RatFolderCmd(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
 	for (msgNo=0; msgNo < infoPtr->number; msgNo++) {
 	    if (infoPtr->msgCmdPtr[msgNo]
 	    	    && !strcmp(infoPtr->msgCmdPtr[msgNo], argv[2])) {
-	        for (i=0; msgNo != infoPtr->presentationOrder[i]; i++);
+		for (i=0; msgNo != infoPtr->presentationOrder[i]; i++);
 	        sprintf(interp->result, "%d", i);
+		break;
 	    }
 	}
 	return TCL_OK;
@@ -536,13 +552,15 @@ RatFolderCmd(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
  *      Sorts the folder according to the users wishes. The user may
  *	communicates their will via the folder_sort variable. Currently
  *	The following methods are implemented:
+ *	  subjectonly		- Alphabetically on subject
+ *	  sender		- Alphabetically on sender name
  *	  folder		- Sorts in native folder order
  *	  reverseFolder		- The reverse of the above
  *	  date			- By date sent
  *	  reverseDate		- By reverse date sent
- *	  subjectsByDate	- Group messages with the same subject
+ *	  subject		- Group messages with the same subject
  *				  and sort the groups by the earliest date
- *				  in each group. NOT IMPLEMENTED YET
+ *				  in each group.
  *
  * Results:
  *	None.
@@ -559,13 +577,46 @@ static void
 RatFolderSort(Tcl_Interp *interp, RatFolderInfo *infoPtr)
 {
     char *sortMethod=Tcl_GetVar2(interp,"option","folder_sort",TCL_GLOBAL_ONLY);
-    int i, j, snarf, numParm, notused, listArgc, *p=infoPtr->presentationOrder,
-	seen;
-    char *s, **pArgv, **listArgv, *argv[4];
+    int i, j, k, snarf, numParm, seen, *tmpPtr,
+	*p=infoPtr->presentationOrder,
+	reverse=0, needDate=0, needSubject=0, needSender=0,
+	*uniqList, uniqListUsed, *subList, *lengthList;
+    SortOrder order = SORT_NONE;
+    SortData *dataPtr, *dPtr;
+    char *s, **pArgv;
 
     Tcl_GetBoolean(interp, Tcl_GetVar2(interp, "option", "dsn_snarf_reports",
 	    TCL_GLOBAL_ONLY), &snarf);
 
+    if (!strcmp(sortMethod, "subject")) {
+	order = SORT_SUBJDATE;
+	needDate = 1;
+	needSubject = 1;
+    } else if (!strcmp(sortMethod, "sender")) {
+	order = SORT_SENDERDATE;
+	needDate = 1;
+	needSender = 1;
+    } else if (!strcmp(sortMethod, "subjectonly")) {
+	order = SORT_SUBJECT;
+	needSubject = 1;
+    } else if (!strcmp(sortMethod, "senderonly")) {
+	order = SORT_SENDER;
+	needSender = 1;
+    } else if (!strcmp(sortMethod, "date")) {
+	needDate = 1;
+	order = SORT_DATE;
+    } else if (!strcmp(sortMethod, "folder")) {
+	order = SORT_NONE;
+    } else if (!strcmp(sortMethod, "reverseFolder")) {
+	order = SORT_NONE;
+	reverse = 1;
+    } else if (!strcmp(sortMethod, "reverseDate")) {
+	needDate = 1;
+	order = SORT_DATE;
+	reverse = 1;
+    }
+
+    dataPtr = (SortData*)malloc(infoPtr->number*sizeof(*dataPtr));
     infoPtr->size = 0;
     for (i=0; i<infoPtr->number; i++) {
 	infoPtr->hidden[i] = 0;
@@ -597,229 +648,241 @@ RatFolderSort(Tcl_Interp *interp, RatFolderInfo *infoPtr)
 			(*infoPtr->setFlagProc)(infoPtr,interp,i,RAT_DELETED,1);
 		    }
 		    (*infoPtr->setFlagProc)(infoPtr,interp,i,RAT_SEEN,seen);
+		    break;
 		}
 	    }
 	    ckfree(pArgv);
 	}
+	if (needSubject) {
+	    dataPtr[i].subject = RatFolderSubjCopy(
+		    (*infoPtr->infoProc)(interp, (ClientData)infoPtr,
+		    RAT_FOLDER_SUBJECT, i));
+	    lcase(dataPtr[i].subject);
+	}
+	if (needSender) {
+	    dataPtr[i].sender = cpystr((*infoPtr->infoProc)(interp,
+		    (ClientData)infoPtr, RAT_FOLDER_NAME, i));
+	    lcase(dataPtr[i].sender);
+	}
+	if (needDate) {
+	    dataPtr[i].date = atoi((*infoPtr->infoProc)(interp,
+			(ClientData)infoPtr, RAT_FOLDER_DATE_N, i));
+	}
     }
 
-    if (!strcmp(sortMethod, "subject")) {
-	int j, start, end, burrow, *datePtr, *snugglePtr;
-	char **subjArgv, *s;
-
-	subjArgv = (char**)malloc(infoPtr->number*sizeof(char*));
-	datePtr = (int*)malloc(infoPtr->number*sizeof(int));
-	snugglePtr = (int*)malloc(infoPtr->number*sizeof(int));
-	for (i=0; i<infoPtr->number; i++) {
-	    s = (*infoPtr->infoProc)(interp, (ClientData)infoPtr,
-		    RAT_FOLDER_SUBJECT, i);
-	    if (s && *s) {
-		subjArgv[i] =  strdup(s);
-	    } else {
-		subjArgv[i] =  "";
-	    }
-	    s = (*infoPtr->infoProc)(interp, (ClientData)infoPtr,
-		    RAT_FOLDER_DATE_N, i);
-	    if (s) {
-		datePtr[i] = atoi(s);
-	    } else {
-		datePtr[i] = 0;
-	    }
-	}
-	for (i=0; i<infoPtr->number; i++) {
-	    snugglePtr[i] = 0;
-	    burrow = 0;
-	    start = 0;
-	    end = i+1;
-	    for (j=0; j<i; j++) {
-		if (RatFolderSubjSame(subjArgv[i], subjArgv[p[j]])) {
-		    snugglePtr[i] = 1;
-		    snugglePtr[p[j]] = 1;
-		    burrow = 1;
-		    start = j;
-		    for (j++; j < i &&
-			 RatFolderSubjSame(subjArgv[i],subjArgv[p[j]]); j++);
-		    end = j;
-		    break;
-		}
-	    } 
-	    for (j=start; j<end && j<i; j++) {
-		if (datePtr[i]<datePtr[p[j]] && (burrow || !snugglePtr[p[j]])) {
-		    break;
-		}
-	    }
-	    if (j < i) {
-		memmove(&p[j+1], &p[j], (i-j)*sizeof(int));
-	    }
-	    p[j] = i;
-	}
-	for (i=0; i<infoPtr->number; i++) {
-	    if (*subjArgv[i]) {
-		ckfree(subjArgv[i]);
-	    }
-	}
-	ckfree(subjArgv);
-	ckfree(datePtr);
-
-    } else if (!strcmp(sortMethod, "reverseFolder")) {
-	for (i=0; i<infoPtr->number; i++) {
-	    p[i] = infoPtr->number-i-1;
-	}
-
-    } else if (!strcmp(sortMethod, "date")) {
+    baseSortDataPtr = dataPtr;
+    switch (order) {
+    case SORT_NONE:
 	for (i=0; i<infoPtr->number; i++) {
 	    p[i] = i;
 	}
-	infoPtr->visible = infoPtr->number;
-	argv[0] = "foo";
-	argv[1] = "list";
-	argv[2] = "%D";
-	RatFolderCmd((ClientData)infoPtr, interp, 3, argv);
-	Tcl_SplitList(interp, interp->result, &listArgc, &listArgv);
-	rat_SortList = listArgv;
-	rat_SortOrder = 0;
-	qsort((void*)p, infoPtr->number,
-		sizeof(int), RatFolderSortCompare);
-	ckfree(listArgv);
-	Tcl_ResetResult(interp);
-
-    } else if (!strcmp(sortMethod, "reverseDate")) {
-	for (i=0; i<infoPtr->number; i++) {
-	    infoPtr->presentationOrder[i] = i;
+	break;
+    case SORT_SUBJDATE:
+    case SORT_SENDERDATE:
+	/*
+	 * This algorithm is complicated:
+	 * - First we build a list of unique subjects in uniqList. Each entry
+	 *   in this list contains the index of the first message with this
+	 *   subject. The messages are linked with the nextPtr field in
+	 *   the SortData structs.
+	 * - Then we sort each found subject. This is done by placing the
+	 *   indexes of the messages in subList. And sort that. When it
+	 *   is sorted we rebuild the subject chains via the nextPtr;
+	 * - After that we sort the first message in each subject. This is done
+	 *   by reusing the uniqList. We replace each entry in it with a pointer
+	 *   to the first entry in the set. Actually we do this in the preceding
+	 *   step. Then we sort this list.
+	 * - Finally we build to result array.
+	 */
+	uniqList = (int*)malloc(2*infoPtr->number*sizeof(*uniqList));
+	subList = &uniqList[infoPtr->number];
+	lengthList = &uniqList[2*infoPtr->number];
+	for (i=uniqListUsed=0; i<infoPtr->number; i++) {
+	    for (j=0; j<uniqListUsed; j++) {
+		if (order == SORT_SUBJDATE ? !strcmp(dataPtr[i].subject,
+			dataPtr[uniqList[j]].subject) : !strcmp(
+			dataPtr[i].sender, dataPtr[uniqList[j]].sender)) {
+		    dataPtr[i].nextPtr = dataPtr[uniqList[j]].nextPtr;
+		    dataPtr[uniqList[j]].nextPtr = &dataPtr[i];
+		    break;
+		}
+	    }
+	    if (j == uniqListUsed) {
+		dataPtr[i].nextPtr = NULL;
+		uniqList[uniqListUsed++] = i;
+	    }
 	}
-	infoPtr->visible = infoPtr->number;
-	argv[0] = "foo";
-	argv[1] = "list";
-	argv[2] = "%D";
-	RatFolderCmd((ClientData)infoPtr, interp, 3, argv);
-	Tcl_SplitList(interp, interp->result, &listArgc, &listArgv);
-	rat_SortList = listArgv;
-	rat_SortOrder = 1;
-	qsort((void*)p, infoPtr->number, sizeof(int), RatFolderSortCompare);
-	ckfree(listArgv);
-	Tcl_ResetResult(interp);
-
-    } else {	/* default to folder native order */
+	for (i=0; i<uniqListUsed; i++) {
+	    if (NULL != dataPtr[uniqList[i]].nextPtr) {
+		for (j = 0, dPtr = &dataPtr[uniqList[i]]; dPtr;
+			dPtr = dPtr->nextPtr) {
+		    subList[j++] = dPtr-dataPtr;
+		}
+		qsort((void*)subList, j, sizeof(int), RatFolderSortCompareDate);
+		for (k=0; k<j-1; k++) {
+		    dataPtr[subList[k]].nextPtr = &dataPtr[subList[k+1]];
+		}
+		dataPtr[subList[k]].nextPtr = NULL;
+		uniqList[i] = subList[0];
+	    }
+	}
+	qsort((void*)uniqList, uniqListUsed, sizeof(int),
+		RatFolderSortCompareDate);
+	for (i=k=0; i<uniqListUsed; i++) {
+	    for (dPtr = &dataPtr[uniqList[i]]; dPtr; dPtr = dPtr->nextPtr) {
+		p[k++] = dPtr-baseSortDataPtr;
+	    }
+	}
+	free(uniqList);
+	break;
+    case SORT_SENDER:
 	for (i=0; i<infoPtr->number; i++) {
 	    p[i] = i;
 	}
+	qsort((void*)p, infoPtr->number,sizeof(int),RatFolderSortCompareSender);
+	break;
+    case SORT_SUBJECT:
+	for (i=0; i<infoPtr->number; i++) {
+	    p[i] = i;
+	}
+	qsort((void*)p,infoPtr->number,sizeof(int),RatFolderSortCompareSubject);
+	break;
+    case SORT_DATE:
+	for (i=0; i<infoPtr->number; i++) {
+	    p[i] = i;
+	}
+	qsort((void*)p, infoPtr->number, sizeof(int), RatFolderSortCompareDate);
+	break;
     }
-    for (i=j=0; i < infoPtr->number; i++) {
-	if (!infoPtr->hidden[p[i]]) {
-	    p[j++] = p[i];
+
+    if (reverse) {
+	tmpPtr = (int*)malloc(infoPtr->number*sizeof(int));
+	for (i=infoPtr->number-1, j=0; i >= 0; i--) {
+	    if (!infoPtr->hidden[p[i]]) {
+		tmpPtr[j++] = p[i];
+	    }
+	}
+	memcpy(p, tmpPtr, j*sizeof(int));
+	free(tmpPtr);
+    } else {
+	for (i=j=0; i < infoPtr->number; i++) {
+	    if (!infoPtr->hidden[p[i]]) {
+		p[j++] = p[i];
+	    }
 	}
     }
     infoPtr->visible = j;
+
+    /*
+     * Cleanup dataPtr
+     */
+    for (i=0; i<infoPtr->number; i++) {
+	if (!infoPtr->hidden[i]) {
+	    if (needSubject) {
+		free(dataPtr[i].subject);
+	    }
+	    if (needSender) {
+		free(dataPtr[i].sender);
+	    }
+	}
+    }
+    free(dataPtr);
 }
 
 /*
  *----------------------------------------------------------------------
  *
- * RatFolderSortCompare --
+ * RatFolderSortCompar* --
  *
- * 	This is a helper function to RatFolderSort, it is used as
- *	comparision function for qsort. The behaviour is controlled
- *	via two global variables:
- *	rat_SortList -	Points to the list which contains the values to
- *			be sorted.
- *	rat_SortOrder -	Indicates the order: 0 = increasing, 1 = decreasing
+ *	This is the comparison functions used by RatFolderSort. They
+ *	expect to get pointers to integers as argumens. The integers
+ *	pointed at are indexes into a list of SortData structs which can be
+ *	found at the address in baseSortDataPtr.
  *
  * Results:
- *	None.
+ *	An integers describing the order of the compared objects.
  *
  * Side effects:
- *	The presentation order member of the RatFolderInfo structure
- *	is initialized.
+ *	None.
  *
  *
  *----------------------------------------------------------------------
  */
 
 static int
-RatFolderSortCompare(const void *arg1, const void *arg2)
+RatFolderSortCompareDate(const void *arg1, const void *arg2)
 {
-    long val1, val2;
-    int result;
+    return baseSortDataPtr[*((int*)arg1)].date
+	    - baseSortDataPtr[*((int*)arg2)].date;
+}
 
-    val1 = atoi(rat_SortList[*((int*)arg1)]);
-    val2 = atoi(rat_SortList[*((int*)arg2)]);
-    if (val1 > val2) {
-	result = 1;
-    } else if (val2 > val1) {
-	result = -1;
+static int
+RatFolderSortCompareSubject(const void *arg1, const void *arg2)
+{
+    return strcmp(baseSortDataPtr[*((int*)arg1)].subject,
+		  baseSortDataPtr[*((int*)arg2)].subject);
+}
+
+static int
+RatFolderSortCompareSender(const void *arg1, const void *arg2)
+{
+    return strcmp(baseSortDataPtr[*((int*)arg1)].sender,
+		  baseSortDataPtr[*((int*)arg2)].sender);
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * RatFolderSubjCopy --
+ *
+ * Copy a subject line and remove certain constructs (the re:).
+ *
+ * Results:
+ *	Pointer to a new copy of the subject. It is the callers
+ *	responsibility to free this string later.
+ *
+ * Side effects:
+ *	None.
+ *
+ *
+ *----------------------------------------------------------------------
+ */
+
+static char*
+RatFolderSubjCopy(char *subject)
+{
+    char *s = subject, *ret;
+    int len;
+
+    if (s) {
+	/*
+	 * We first try to find the start of the actual text (i.e. without any
+	 * leading Re:'s and whitespaces. Then we find how long the text is
+	 * (ignore trailing whitespaces).
+	 */
+	while (*s) {
+	    while (*s && isspace(*s)) s++;
+	    if (!strncasecmp(s, "re", 2) && (':' == s[2] || isspace(s[2]))) {
+		s += 2;
+		if (*s == ':') {
+		    s++;
+		}
+	    } else {
+		break;
+	    }
+	}
+	len = strlen(s);
+	while (len && isspace(s[len-1])) len--;
+
+	ret = (char*)malloc(len+1);
+	strncpy(ret, s, len);
+	ret[len] = '\0';
     } else {
-	result = 0;
+	ret = (char*)malloc(1);
+	ret[0] = '\0';
     }
-    if (rat_SortOrder) {
-	result *= -1;
-    }
-    return result;
-}
-
-/*
- *----------------------------------------------------------------------
- *
- * RatFolderSubjSame --
- *
- * Checks if the two subjects belongs to the same thread.
- *
- * Results:
- *	True if they do.
- *
- * Side effects:
- *	None.
- *
- *
- *----------------------------------------------------------------------
- */
-
-static int
-RatFolderSubjSame(char *subj1, char *subj2)
-{
-    char *s1 = subj1, *s2 = subj2;
-    int l1, l2;
-
-    /*
-     * For each subject we first try to find the start of the actual text
-     * (i.e. without any leading Re:'s and whitespaces. Then we find how
-     * long the text is (ignore trailing whitespaces).
-     */
-    while (*s1) {
-	while (*s1 && isspace(*s1)) s1++;
-	if (!strncasecmp(s1, "re", 2) && (':' == s1[2] || isspace(s1[2]))) {
-	    s1 += 2;
-	    if (*s1 == ':') {
-		s1++;
-	    }
-	} else {
-	    break;
-	}
-    }
-    l1 = strlen(s1);
-    while (l1 && isspace(s1[l1-1])) l1--;
-    while (*s2) {
-	while (*s2 && isspace(*s2)) s2++;
-	if (!strncasecmp(s2, "re", 2) && (':' == s2[2] || isspace(s2[2]))) {
-	    s2 += 2;
-	    if (*s2 == ':') {
-		s2++;
-	    }
-	} else {
-	    break;
-	}
-    }
-    l2 = strlen(s2);
-    while (l2 && isspace(s2[l2-1])) l2--;
-
-    /*
-     * Now we are ready to compare them. If any of the subject are less
-     * than 10 characters we say that they must match exactly otherwise it
-     * is enough that they match as long as they are.
-     */
-    if (l1 < 10 || l2 < 10) {
-	return (l1 == l2 && !strcmp(s1, s2));
-    }
-    return !strncmp(s1, s2, (l1>l2 ? l2 : l1));
+    return ret;
 }
 
 /*
@@ -831,19 +894,21 @@ RatFolderSubjSame(char *subj1, char *subj2)
  * informations items needs both the messagePtr and the eltPtr. The following
  * table describes which needs which:
  *
- *  Info			envPtr	bodyPtr	eltPtr	size
- *  RAT_FOLDER_SUBJECT		needed	-	-	-
- *  RAT_FOLDER_NAME		needed	-	-	-
- *  RAT_FOLDER_MAIL		needed	-	-	-
- *  RAT_FOLDER_NAME_RECIPIENT	needed	-	-	-
- *  RAT_FOLDER_MAIL_RECIPIENT	needed	-	-	-
- *  RAT_FOLDER_SIZE		-	-	-	needed
- *  RAT_FOLDER_SIZE_F		-	-	-	needed
- *  RAT_FOLDER_DATE_F		needed	-	needed	-
- *  RAT_FOLDER_DATE_N		needed	-	needed	-
+ *  Info			msgPtr	envPtr	bodyPtr	eltPtr	size
+ *  RAT_FOLDER_SUBJECT		-	needed	-	-	-
+ *  RAT_FOLDER_NAME		needed	needed	-	-	-
+ *  RAT_FOLDER_MAIL_REAL	-	needed	-	-	-
+ *  RAT_FOLDER_MAIL		needed	needed	-	-	-
+ *  RAT_FOLDER_NAME_RECIPIENT	needed	needed	-	-	-
+ *  RAT_FOLDER_MAIL_RECIPIENT	needed	needed	-	-	-
+ *  RAT_FOLDER_SIZE		-	-	-	-	needed
+ *  RAT_FOLDER_SIZE_F		-	-	-	-	needed
+ *  RAT_FOLDER_DATE_F		-	needed	-	needed	-
+ *  RAT_FOLDER_DATE_N		-	needed	-	needed	-
  *  RAT_FOLDER_STATUS		    [not supported ]
- *  RAT_FOLDER_TYPE		-	needed	-	-
- *  RAT_FOLDER_PARAMETERS	-	needed	-	-
+ *  RAT_FOLDER_TYPE		-	-	needed	-	-
+ *  RAT_FOLDER_PARAMETERS	-	-	needed	-	-
+ *  RAT_FOLDER_FLAGS		-	-	-	needed	-
  *  RAT_FOLDER_INDEX		    [not supported ]
  *
  * Results:
@@ -858,8 +923,8 @@ RatFolderSubjSame(char *subj1, char *subj2)
  */
 
 char *
-RatGetMsgInfo(Tcl_Interp *interp, RatFolderInfoType type, ENVELOPE *envPtr,
-	BODY *bodyPtr, MESSAGECACHE *eltPtr, int size)
+RatGetMsgInfo(Tcl_Interp *interp, RatFolderInfoType type, MessageInfo *msgPtr,
+	ENVELOPE *envPtr, BODY *bodyPtr, MESSAGECACHE *eltPtr, int size)
 {
     static Tcl_DString ds;
     static int initialized = 0;
@@ -872,18 +937,31 @@ RatGetMsgInfo(Tcl_Interp *interp, RatFolderInfoType type, ENVELOPE *envPtr,
     if (!initialized) {
 	Tcl_DStringInit(&ds);
 	initialized = 1;
-    } else {
-	Tcl_DStringSetLength(&ds, 0);
     }
 
     switch (type) {
 	case RAT_FOLDER_SUBJECT:
 	    return envPtr->subject;
+	case RAT_FOLDER_MAIL_REAL:
+	    Tcl_DStringSetLength(&ds, 0);
+	    for (adrPtr = envPtr->from; adrPtr; adrPtr = adrPtr->next) {
+		if (adrPtr->mailbox && adrPtr->host) {
+		    break;
+		}
+	    }
+	    if (!adrPtr) {
+		return "";
+	    }
+	    Tcl_DStringAppend(&ds, RatAddressMail(adrPtr), -1);
+	    return Tcl_DStringValue(&ds);
 	case RAT_FOLDER_NAME:
 	    if (!envPtr->from) {
 		return "";
 	    }
-	    if (RatAddressIsMe(interp, envPtr->from, 1)) {
+	    if (RAT_ISME_YES == msgPtr->fromMe
+		    || (RAT_ISME_UNKOWN == msgPtr->fromMe
+			&& RatAddressIsMe(interp, envPtr->from, 1))) {
+		msgPtr->fromMe = RAT_ISME_YES;
 		if (envPtr->to && envPtr->to->personal) {
 		    Tcl_DStringSetLength(&ds, 0);
 		    Tcl_DStringAppend(&ds,
@@ -892,18 +970,25 @@ RatGetMsgInfo(Tcl_Interp *interp, RatFolderInfoType type, ENVELOPE *envPtr,
 		    Tcl_DStringAppend(&ds, envPtr->to->personal, -1);
 		    return Tcl_DStringValue(&ds);
 		}
-	    } else if (envPtr->from->personal) {
-		return envPtr->from->personal;
+	    } else {
+		msgPtr->fromMe = RAT_ISME_NO;
+		if (envPtr->from->personal) {
+		    return envPtr->from->personal;
+		}
 	    }
 	    /* fallthrough */
 	case RAT_FOLDER_MAIL:
 	    Tcl_DStringSetLength(&ds, 0);
-	    if (RatAddressIsMe(interp, envPtr->from, 1)) {
+	    if (RAT_ISME_YES == msgPtr->fromMe
+		    || (RAT_ISME_UNKOWN == msgPtr->fromMe
+			&& RatAddressIsMe(interp, envPtr->from, 1))) {
+		msgPtr->fromMe = RAT_ISME_YES;
 		adrPtr = envPtr->to;
 		Tcl_DStringAppend(&ds,
 			Tcl_GetVar2(interp, "t", "to", TCL_GLOBAL_ONLY),-1);
 		Tcl_DStringAppend(&ds, ": ", 2);
 	    } else {
+		msgPtr->fromMe = RAT_ISME_NO;
 		adrPtr = envPtr->from;
 	    }
 	    for (; adrPtr; adrPtr = adrPtr->next) {
@@ -920,7 +1005,10 @@ RatGetMsgInfo(Tcl_Interp *interp, RatFolderInfoType type, ENVELOPE *envPtr,
 	    if (!envPtr->to) {
 		return "";
 	    }
-	    if (RatAddressIsMe(interp, envPtr->to, 1)) {
+	    if (RAT_ISME_YES == msgPtr->toMe
+		    || (RAT_ISME_UNKOWN == msgPtr->toMe
+			&& RatAddressIsMe(interp, envPtr->to, 1))) {
+		msgPtr->toMe = RAT_ISME_YES;
 		if (envPtr->from->personal) {
 		    Tcl_DStringSetLength(&ds, 0);
 		    Tcl_DStringAppend(&ds,
@@ -929,18 +1017,25 @@ RatGetMsgInfo(Tcl_Interp *interp, RatFolderInfoType type, ENVELOPE *envPtr,
 		    Tcl_DStringAppend(&ds, envPtr->from->personal, -1);
 		    return Tcl_DStringValue(&ds);
 		}
-	    } else if (envPtr->to->personal) {
-		return envPtr->to->personal;
+	    } else {
+		msgPtr->toMe = RAT_ISME_NO;
+		if (envPtr->to->personal) {
+		    return envPtr->to->personal;
+		}
 	    }
 	    /* fallthrough */
 	case RAT_FOLDER_MAIL_RECIPIENT:
 	    Tcl_DStringSetLength(&ds, 0);
-	    if (RatAddressIsMe(interp, envPtr->to, 1)) {
+	    if (RAT_ISME_YES == msgPtr->toMe
+		    || (RAT_ISME_UNKOWN == msgPtr->toMe
+			&& RatAddressIsMe(interp, envPtr->to, 1))) {
+		msgPtr->toMe = RAT_ISME_YES;
 		adrPtr = envPtr->from;
 		Tcl_DStringAppend(&ds,
 			Tcl_GetVar2(interp, "t", "from", TCL_GLOBAL_ONLY),-1);
 		Tcl_DStringAppend(&ds, ": ", 2);
 	    } else {
+		msgPtr->toMe = RAT_ISME_NO;
 		adrPtr = envPtr->to;
 	    }
 	    for (; adrPtr; adrPtr = adrPtr->next) {
@@ -958,28 +1053,16 @@ RatGetMsgInfo(Tcl_Interp *interp, RatFolderInfoType type, ENVELOPE *envPtr,
 	    sprintf(Tcl_DStringValue(&ds), "%d", size);
 	    return Tcl_DStringValue(&ds);
 	case RAT_FOLDER_SIZE_F:
-	    Tcl_DStringSetLength(&ds, 256);
-	    sprintf(Tcl_DStringValue(&ds), "RatMangleNumber %d", size);
-	    Tcl_GlobalEval(interp, Tcl_DStringValue(&ds));
-	    Tcl_DStringGetResult(interp, &ds);
-	    return Tcl_DStringValue(&ds);
+	    return RatMangleNumber(size);
 	case RAT_FOLDER_DATE_F:
-	    if (envPtr->date) {
-		mail_parse_date(&dateElt, envPtr->date);
+	    if (envPtr->date && T == mail_parse_date(&dateElt, envPtr->date)) {
 		dateEltPtr = &dateElt;
 	    } else {
 		dateEltPtr = eltPtr;
 	    }
-	    Tcl_DStringSetLength(&ds, 256);
-	    sprintf(Tcl_DStringValue(&ds), "RatFormatDate %d %d %d %d %d %d",
-		    dateEltPtr->year, dateEltPtr->month, dateEltPtr->day,
-		    dateEltPtr->hours, dateEltPtr->minutes,dateEltPtr->seconds);
-	    Tcl_GlobalEval(interp, Tcl_DStringValue(&ds));
-	    Tcl_DStringGetResult(interp, &ds);
-	    return Tcl_DStringValue(&ds);
+	    return RatFormatDate(interp, dateEltPtr->month-1, dateEltPtr->day);
 	case RAT_FOLDER_DATE_N:
-	    if (envPtr->date) {
-		mail_parse_date(&dateElt, envPtr->date);
+	    if (envPtr->date && T == mail_parse_date(&dateElt, envPtr->date)) {
 		dateEltPtr = &dateElt;
 	    } else {
 		dateEltPtr = eltPtr;
@@ -1002,6 +1085,14 @@ RatGetMsgInfo(Tcl_Interp *interp, RatFolderInfoType type, ENVELOPE *envPtr,
 	    time += zonediff;
 	    sprintf(Tcl_DStringValue(&ds), "%ld", time);
 	    return Tcl_DStringValue(&ds);
+	case RAT_FOLDER_DATE_IMAP4:
+	    if (envPtr->date && T == mail_parse_date(&dateElt, envPtr->date)) {
+		dateEltPtr = &dateElt;
+	    } else {
+		dateEltPtr = eltPtr;
+	    }
+	    Tcl_DStringSetLength(&ds, 256);
+	    return mail_date(Tcl_DStringValue(&ds), dateEltPtr);
 	case RAT_FOLDER_TYPE:
 	    Tcl_DStringSetLength(&ds, strlen(bodyPtr->subtype)+2+
 		    strlen(body_types[bodyPtr->type]));
@@ -1010,6 +1101,7 @@ RatGetMsgInfo(Tcl_Interp *interp, RatFolderInfoType type, ENVELOPE *envPtr,
 		    bodyPtr->subtype);
 	    return Tcl_DStringValue(&ds);
 	case RAT_FOLDER_PARAMETERS:
+	    Tcl_DStringSetLength(&ds, 0);
 	    for (parmPtr = bodyPtr->parameter; parmPtr;
 		    parmPtr = parmPtr->next) {
 		Tcl_DStringStartSublist(&ds);
@@ -1043,6 +1135,25 @@ RatGetMsgInfo(Tcl_Interp *interp, RatFolderInfoType type, ENVELOPE *envPtr,
 	    *(Tcl_DStringValue(&ds)) = '\0';
 	    rfc822_write_address(Tcl_DStringValue(&ds), envPtr->reply_to);
 	    return Tcl_DStringValue(&ds);
+	case RAT_FOLDER_FLAGS:
+	    Tcl_DStringSetLength(&ds, 0);
+	    if (eltPtr->seen) {
+		Tcl_DStringAppend(&ds, " " RAT_SEEN_STR, -1);
+	    }
+	    if (eltPtr->deleted) {
+		Tcl_DStringAppend(&ds, " " RAT_DELETED_STR, -1);
+	    }
+	    if (eltPtr->flagged) {
+		Tcl_DStringAppend(&ds, " " RAT_FLAGGED_STR, -1);
+	    }
+	    if (eltPtr->answered) {
+		Tcl_DStringAppend(&ds, " " RAT_ANSWERED_STR, -1);
+	    }
+	    if (Tcl_DStringLength(&ds)) {
+		return Tcl_DStringValue(&ds)+1;
+	    } else {
+		return "";
+	    }
 	case RAT_FOLDER_STATUS:	/*fallthrough */
 	case RAT_FOLDER_INDEX:	/*fallthrough */
 	case RAT_FOLDER_END:
