@@ -3,18 +3,20 @@
  *
  *	This file contains code which implements the message entities.
  *
- * TkRat software and its included text is Copyright 1996 by Martin Forssen.
+ * TkRat software and its included text is Copyright 1996,1997,1998
+ * by Martin Forssn
  *
  * The full text of the legal notice is contained in the file called
  * COPYRIGHT, included with this distribution.
  */
 
 #include "ratFolder.h"
+#include "ratPGP.h"
 #include <signal.h>
 
 /*
- * An array of the above defined structure. It contains one entry for each
- * internal message type (as defined by RatMessageType).
+ * An array of commands. It contains one entry for each internal message
+ * type (as defined by RatMessageType).
  */
 static MessageProcInfo *messageProcInfo;
 
@@ -24,7 +26,14 @@ static MessageProcInfo *messageProcInfo;
  */
 static int numReplies = 0;
 
+/*
+ * The number of message entities created. This is used to create new
+ * unique command names.
+ */
+static int numBodies = 0;
+
 static void RatBodyDelete(Tcl_Interp *interp, BodyInfo *bodyInfoPtr);
+static BodyInfo *RatFindFirstText(BodyInfo *bodyInfoPtr);
 
 
 /*
@@ -103,6 +112,9 @@ RatMessageCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 	    msgPtr->bodyInfoPtr =
 		    (*messageProcInfo[msgPtr->type].createBodyProc)(interp,
 		    msgPtr);
+	    RatPGPBodyCheck(interp, messageProcInfo, &msgPtr->bodyInfoPtr);
+	    Tcl_CreateCommand(interp, msgPtr->bodyInfoPtr->cmdName,
+		    RatBodyCmd, (ClientData) msgPtr->bodyInfoPtr, NULL);
 	}
 	Tcl_SetResult(interp, msgPtr->bodyInfoPtr->cmdName, TCL_STATIC);
 	return TCL_OK;
@@ -157,11 +169,9 @@ RatMessageCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 	 */
 	ENVELOPE *env = (*messageProcInfo[msgPtr->type].envelopeProc)(msgPtr);
 	char handler[32], buf[1024], *subject;
+	BodyInfo *bodyInfoPtr;
 	ADDRESS *adrPtr;
 	Tcl_DString ds;
-	BODY *bodyPtr;
-	char *dataPtr;
-	int len;
 	Tcl_DStringInit(&ds);
 
 	if (argc != 3) {
@@ -188,10 +198,12 @@ RatMessageCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 	    }
 	    for (;adrPtr; adrPtr = adrPtr->next) {
 		RatAddressTranslate(interp, adrPtr);
-		if (Tcl_DStringLength(&ds)) {
-		    Tcl_DStringAppend(&ds, ", ", -1);
+		if (adrPtr->mailbox) {
+		    if (Tcl_DStringLength(&ds)) {
+			Tcl_DStringAppend(&ds, ", ", -1);
+		    }
+		    Tcl_DStringAppend(&ds, RatAddressMail(adrPtr), -1);
 		}
-		Tcl_DStringAppend(&ds, RatAddressMail(adrPtr), -1);
 	    }
 	    Tcl_SetVar2(interp, handler, "to", Tcl_DStringValue(&ds),
 		    TCL_GLOBAL_ONLY);
@@ -226,21 +238,6 @@ RatMessageCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 	    SCANLIST(env->from);
 	    SCANLIST(env->to);
 	    SCANLIST(env->cc);
-	    for (adrPtr = env->sender; adrPtr; adrPtr = adrPtr->next) {
-		if (!strncasecmp("owner-", adrPtr->mailbox, 6) ||
-			(strlen(adrPtr->mailbox) > 6
-			 && !strncasecmp("-owner",
-			     adrPtr->mailbox+strlen(adrPtr->mailbox)-6, 6))) {
-		    continue;
-		}
-		if (numRecipients == numAllocated) {
-		    numAllocated += 16;
-		    recipientPtrPtr = (ADDRESS**)REALLOC(
-			    recipientPtrPtr,
-			    numAllocated*sizeof(ADDRESS*));
-		}
-		recipientPtrPtr[numRecipients++] = adrPtr;
-	    }
 
 	    for (i=0; i<numRecipients; i++) {
 		adrPtr = recipientPtrPtr[i];
@@ -277,9 +274,6 @@ RatMessageCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 		to = recipientPtrPtr[0];
 	    }
 	    if (to) {
-		if (!currentHost) {
-		    RatInitCurrent(interp);
-		}
 		if (strcmp(to->host, currentHost)) {
 		    sprintf(buf, "%s@%s", to->mailbox, to->host);
 		} else {
@@ -308,14 +302,20 @@ RatMessageCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 	    Tcl_SetVar2(interp, handler, "in_reply_to", env->message_id,
 		    TCL_GLOBAL_ONLY);
 	}
-	bodyPtr = (*messageProcInfo[msgPtr->type].fetchFirstTextProc)(msgPtr,
-		&dataPtr, &len);
-	if (bodyPtr) {
+	bodyInfoPtr = RatFindFirstText(msgPtr->bodyInfoPtr);
+	if (bodyInfoPtr) {
 	    char *leader, *newPtr, *srcPtr, *dstPtr, *attrFormat, *attr;
 	    unsigned char *decodedBody;
 	    ListExpression *exprPtr;
 	    RatEncoding encoding;
-	    int outLength, lines;
+	    int outLength, lines, skipSig;
+	    BODY *bodyPtr;
+	    char *dataPtr;
+	    unsigned long len;
+
+	    bodyPtr = bodyInfoPtr->bodyPtr;
+	    dataPtr =(*messageProcInfo[bodyInfoPtr->msgPtr->type].fetchBodyProc)
+		    (bodyInfoPtr, &len);
 
 	    switch(bodyPtr->encoding) {
 	    case ENC7BIT:		encoding = RAT_7BIT; break;
@@ -344,6 +344,8 @@ RatMessageCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 		attr = NULL;
 	    }
 	    leader = Tcl_GetVar2(interp, "option","reply_lead",TCL_GLOBAL_ONLY);
+	    Tcl_GetBoolean(interp, Tcl_GetVar2(interp, "option", "skip_sig",
+		    TCL_GLOBAL_ONLY), &skipSig);
 	    newPtr = (char*)ckalloc(strlen((char*)decodedBody) +
 		    (strlen(leader)+1)*lines+1+(attr?strlen(attr)+2:0));
 	    if (attr) {
@@ -354,6 +356,10 @@ RatMessageCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 		dstPtr = newPtr;
 	    }
 	    for (srcPtr = (char*)decodedBody; *srcPtr;) {
+		if (skipSig && !strncmp("-- ", srcPtr, 3) &&
+			(srcPtr[3] == '\r' || srcPtr[3] == '\n')) {
+		    break;
+		}
 		strcpy(dstPtr, leader);
 		dstPtr += strlen(leader);
 		if (strncmp(leader, srcPtr, strlen(leader))) {
@@ -385,12 +391,12 @@ RatMessageCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 
     } else if ((c == 'c') && (strncmp(argv[1], "copy", length) == 0)
 	    && (length > 1)) {
+	char **listArgv, *folder, *cPtr, flags[128], date[128];
 	MAILSTREAM *stream;
 	struct stat sbuf;
-	char **listArgv, *folder;
 	Tcl_DString ds;
 	STRING string;
-	int listArgc;
+	int listArgc, result, i;
 
 	if (argc != 3) {
 	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
@@ -403,8 +409,10 @@ RatMessageCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 	 * If the destination is dbase then we call RatInsert
 	 */
 	if (!strcmp("dbase", listArgv[1])) {
-	    return RatInsertMsg(interp, msgPtr, listArgv[2], listArgv[4],
-		    listArgv[3]);
+	    result = RatInsertMsg(interp, msgPtr, listArgv[3], listArgv[5],
+		    listArgv[4]);
+	    free(listArgv);
+	    return result;
 	}
 
 	/*
@@ -412,21 +420,22 @@ RatMessageCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 	 * magic.
 	 */
 	if (!strcmp("dynamic", listArgv[1])) {
-	    char *name, **newArgv, *cPtr, i;
+	    char *name, **newArgv;
 
 	    name = (*messageProcInfo[msgPtr->type].getInfoProc)(interp,
-		    (ClientData)msgPtr, RAT_FOLDER_MAIL, 0);
+		    (ClientData)msgPtr, RAT_FOLDER_MAIL_REAL, 0);
 	    if (!name || !strlen(name)) {
 		name = currentMailboxName;
 	    }
 	    newArgv = (char**)malloc(3*sizeof(char*)+strlen(listArgv[0])+
-		    strlen(listArgv[2])+strlen(name)+3);
+		    strlen(listArgv[3])+strlen(name)+3);
 	    newArgv[0] = (char*)newArgv+3*sizeof(char*);
 	    strcpy(newArgv[0], listArgv[0]);
 	    newArgv[1] = "file";
-	    newArgv[2] = newArgv[0]+strlen(newArgv[0])+1;
-	    strcpy(newArgv[2], listArgv[2]);
-	    cPtr = newArgv[2]+strlen(newArgv[2]);
+	    newArgv[2] = "";
+	    newArgv[3] = newArgv[0]+strlen(newArgv[0])+1;
+	    strcpy(newArgv[3], listArgv[3]);
+	    cPtr = newArgv[3]+strlen(newArgv[3]);
 	    *cPtr++ = '/';
 	    for (i=0; name[i] && name[i] != '@'; i++) {
 		*cPtr++ = name[i];
@@ -439,12 +448,12 @@ RatMessageCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 	/*
 	 * Try to create nonexisting files
 	 */
-	if (!strcmp("file", listArgv[1]) && stat(listArgv[2], &sbuf)) {
+	if (!strcmp("file", listArgv[1]) && stat(listArgv[3], &sbuf)) {
 	    int mode, fd;
 
 	    Tcl_GetInt(interp, Tcl_GetVar2(interp, "option", "permissions",
 		    TCL_GLOBAL_ONLY), &mode);
-	    fd = open(listArgv[2], O_CREAT, mode);
+	    fd = open(listArgv[3], O_CREAT, mode);
 	    close(fd);
 	}
 
@@ -452,10 +461,11 @@ RatMessageCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 	 * Try the simple (but common) case where both source
 	 * and destination are c-client messages. 
 	 */
-	folder = RatLindex(interp, listArgv[2], 0);
+	folder = RatLindex(interp, listArgv[3], 0);
 	if (RAT_CCLIENT_MESSAGE == msgPtr->type
 		&& RatStdEasyCopyingOK(msgPtr, listArgv[1], folder,
-			(listArgc > 3 ? listArgv[3] : NULL))) {
+			(listArgc > 4 ? listArgv[4] : NULL))) {
+	    free(listArgv);
 	    return RatStdMessageCopy(interp, msgPtr, folder);
 	}
 
@@ -463,16 +473,29 @@ RatMessageCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 	 * We will need the message text so please retrieve it
 	 */
 	Tcl_DStringInit(&ds);
-	RatMessageGet(interp, msgPtr, &ds);
+	RatMessageGet(interp, msgPtr, &ds, flags, date);
 	INIT(&string,mail_string,Tcl_DStringValue(&ds),Tcl_DStringLength(&ds));
-	stream = OpenStdFolder(interp, folder, listArgv[1], listArgv[3], NULL);
+	stream = OpenStdFolder(interp, folder, listArgv[1], listArgv[4], NULL);
+	free(listArgv);
 	if (stream) {
-	    if (!mail_append(stream, folder, &string)){
-		mail_close(stream);
+	    if (NULL != (cPtr = strstr(flags, RAT_FLAGGED_STR))) {
+		i = strlen(RAT_FLAGGED_STR);
+		if (flags == cPtr) {
+		    if (' ' == cPtr[i]) {
+			i++;
+		    }
+		} else {
+		    cPtr--;
+		    i++;
+		}
+		strcpy(cPtr, cPtr+i);
+	    }
+	    if (!mail_append_full(stream, folder, flags, date, &string)){
+		CloseStdFolder(interp, stream);
 		Tcl_SetResult(interp, "mail_append failed", TCL_STATIC);
 		return TCL_ERROR;
 	    }
-	    mail_close(stream);
+	    CloseStdFolder(interp, stream);
 	}
 	return TCL_OK;
 
@@ -626,23 +649,87 @@ RatMessageSetIndex(Tcl_Interp *interp, char *cmdName, int index)
  *----------------------------------------------------------------------
  */
 
-void
-RatMessageDelete(MessageInfo *msgPtr)
+int
+RatMessageDelete(Tcl_Interp *interp, char *msgName)
 {
+    Tcl_CmdInfo cmdInfo;
+    MessageInfo *msgPtr;
     char buf[256];
 
+    if (0 == Tcl_GetCommandInfo(interp, msgName, &cmdInfo)) {
+	Tcl_AppendResult(interp, "No such message: ", msgName, NULL);
+	return TCL_ERROR;
+    }
+    msgPtr = (MessageInfo*)cmdInfo.clientData;
+
     (*messageProcInfo[msgPtr->type].msgDeleteProc)(msgPtr);
     if (msgPtr->bodyInfoPtr) {
-	RatBodyDelete(msgPtr->interp, msgPtr->bodyInfoPtr);
+	if (msgPtr->bodyInfoPtr->altPtr) {
+	    RatBodyDelete(interp, msgPtr->bodyInfoPtr->altPtr);
+	}
+	if (msgPtr->bodyInfoPtr->decodedTextPtr) {
+	    Tcl_DStringFree(msgPtr->bodyInfoPtr->decodedTextPtr);
+	    free(msgPtr->bodyInfoPtr->decodedTextPtr);
+	}
+	if (msgPtr->bodyInfoPtr->secPtr) {
+	    RatBodyDelete(interp, msgPtr->bodyInfoPtr->secPtr);
+	} else {
+	    RatBodyDelete(interp, msgPtr->bodyInfoPtr);
+	}
     }
     sprintf(buf, "msgInfo_%s", msgPtr->name);
-    Tcl_UnsetVar(msgPtr->interp, buf, TCL_GLOBAL_ONLY);
-    /*
-     * Here we should free msgPtr->name, but Tcl_DeleteCommand is
-     * reading the value stored there when we have returned.
-     * For now on I choose to let this leak.	/MaF
-     */
+    Tcl_UnsetVar(interp, buf, TCL_GLOBAL_ONLY);
+    Tcl_DeleteCommand(interp, msgName);
+    ckfree(msgPtr->name);
     ckfree(msgPtr);
+
+    return TCL_OK;
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * CreateBodyInfo --
+ *
+ *      Create and somewhat initialize a BodyInfo structure.
+ *
+ * Results:
+ *	A pointer to a BodyInfo structure.
+ *
+ * Side effects:
+ *	None.
+ *
+ *
+ *----------------------------------------------------------------------
+ */
+BodyInfo*
+CreateBodyInfo(MessageInfo *msgPtr)
+{
+    BodyInfo *bodyInfoPtr;
+    int pad = sizeof(char*) - sizeof(BodyInfo)%sizeof(char*);
+
+    if (sizeof(char*) == pad) {
+	pad = 0;
+    }
+
+    bodyInfoPtr = (BodyInfo*)ckalloc(sizeof(BodyInfo)+pad+16);
+
+    bodyInfoPtr->cmdName = (char*)bodyInfoPtr + pad + sizeof(BodyInfo);
+
+    sprintf(bodyInfoPtr->cmdName, "RatBody%d", numBodies++);
+    bodyInfoPtr->firstbornPtr = NULL;
+    bodyInfoPtr->nextPtr = NULL;
+    bodyInfoPtr->containedEntity = NULL;
+    bodyInfoPtr->type = msgPtr->type;
+    bodyInfoPtr->msgPtr = msgPtr;
+    bodyInfoPtr->secPtr = NULL;
+    bodyInfoPtr->altPtr = NULL;
+    bodyInfoPtr->decodedTextPtr = NULL;
+    bodyInfoPtr->encoded = 0;
+    bodyInfoPtr->sigStatus = RAT_UNSIGNED;
+    bodyInfoPtr->pgpOutput = NULL;
+    return bodyInfoPtr;
 }
 
 
@@ -669,7 +756,7 @@ RatBodyCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 {
     BodyInfo *bodyInfoPtr = (BodyInfo*) clientData;
     BODY *bodyPtr = bodyInfoPtr->bodyPtr;
-    int length;
+    unsigned long length;
     char c;
 
     if (argc < 2) {
@@ -680,18 +767,24 @@ RatBodyCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
     c = argv[1][0];
     length = strlen(argv[1]);
     if ((c == 'c') && (strncmp(argv[1], "children", length) == 0)
-	    && (length > 1)) {
+	    && (length > 2)) {
 	BodyInfo *partInfoPtr;
+	Tcl_DString resultDS;
 
 	if (TYPEMULTIPART != bodyPtr->type) {
 	    return TCL_OK;
 	}
-	(*messageProcInfo[bodyInfoPtr->msgPtr->type].makeChildrenProc)
+	Tcl_DStringInit(&resultDS);
+	(*messageProcInfo[bodyInfoPtr->type].makeChildrenProc)
 		(interp, bodyInfoPtr);
 	for (partInfoPtr = bodyInfoPtr->firstbornPtr; partInfoPtr;
 		partInfoPtr = partInfoPtr->nextPtr) {
-	    Tcl_AppendElement(interp, partInfoPtr->cmdName);
+	    RatPGPBodyCheck(interp, messageProcInfo, &partInfoPtr);
+	    Tcl_CreateCommand(interp, partInfoPtr->cmdName, RatBodyCmd,
+		    (ClientData) partInfoPtr, NULL);
+	    Tcl_DStringAppendElement(&resultDS, partInfoPtr->cmdName);
 	}
+	Tcl_DStringResult(interp, &resultDS);
 	return TCL_OK;
 
     } else if ((c == 'm') && (strncmp(argv[1], "message", length) == 0)
@@ -705,11 +798,11 @@ RatBodyCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 			TCL_STATIC);
 		return TCL_ERROR;
 	    }
-	    body = (*messageProcInfo[bodyInfoPtr->msgPtr->type].fetchBodyProc)
+	    body = (*messageProcInfo[bodyInfoPtr->type].fetchBodyProc)
 		    (bodyInfoPtr, &length);
 	    if (body && *body) {
 		bodyInfoPtr->containedEntity =
-			RatFrMessageCreate(interp, body, length);
+			RatFrMessageCreate(interp, body, length, NULL);
 		Tcl_SetResult(interp, bodyInfoPtr->containedEntity, TCL_STATIC);
 	    }
 	} else {
@@ -802,7 +895,7 @@ RatBodyCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 	    && (length > 1)) {
 	unsigned char *body, *selection;
 	char *isCharset = "us-ascii", *toCharset = "us-ascii", *alias;
-	int encoded, length, outLength;
+	int encoded, outLength;
 	PARAMETER *parameter;
 	RatEncoding encoding;
 
@@ -841,11 +934,11 @@ RatBodyCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 	default:		encoding = RAT_UNKOWN; break;
 	}
 	body = (unsigned char *)
-		(*messageProcInfo[bodyInfoPtr->msgPtr->type].fetchBodyProc)
+		(*messageProcInfo[bodyInfoPtr->type].fetchBodyProc)
 		(bodyInfoPtr, &length);
 	if (body) {
 	    if (encoded) {
-		selection = (char*)malloc(length+1);
+		selection = (unsigned char*)malloc(length+1);
 		memcpy(selection, body, length);
 		selection[length] = '\0';
 		Tcl_SetResult(interp, (char*)selection, TCL_DYNAMIC);
@@ -891,7 +984,7 @@ RatBodyCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 		    TCL_STATIC);
 	    return TCL_ERROR;
 	}
-	body = (*messageProcInfo[bodyInfoPtr->msgPtr->type].fetchBodyProc)
+	body = (*messageProcInfo[bodyInfoPtr->type].fetchBodyProc)
 		(bodyInfoPtr, &length);
 	if (*body) {
 	    return RatDSNExtract(interp, body, length);
@@ -906,8 +999,9 @@ RatBodyCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 	PARAMETER *parmPtr;
 
 	if (TYPETEXT != bodyPtr->type) {
-	    Tcl_SetResult(interp, "Not a text bodypart", TCL_STATIC);
-	    return TCL_ERROR;
+	    Tcl_AppendElement(interp, "good");
+	    Tcl_AppendElement(interp, "us-ascii");
+	    return TCL_OK;
 	}
 	for (parmPtr = bodyPtr->parameter; parmPtr; parmPtr = parmPtr->next) {
 	    if (!strcasecmp(parmPtr->attribute, "charset")) {
@@ -956,7 +1050,8 @@ RatBodyCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 	PARAMETER *parmPtr;
 
 	for (parmPtr = bodyPtr->parameter; parmPtr; parmPtr = parmPtr->next) {
-	    if (!strcasecmp(parmPtr->attribute, "filename")) {
+	    if (!strcasecmp(parmPtr->attribute, "filename")
+		    || !strcasecmp(parmPtr->attribute, "name")) {
 		Tcl_SetResult(interp, parmPtr->value, TCL_VOLATILE);
 		return TCL_OK;
 	    }
@@ -967,11 +1062,45 @@ RatBodyCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
 	}
 	return TCL_OK;
 
+    } else if ((c == 'e') && (strncmp(argv[1], "encoded", length) == 0)
+	    && (length > 1)) {
+	sprintf(interp->result, "%d", bodyInfoPtr->encoded);
+	return TCL_OK;
+
+    } else if ((c == 's') && (strncmp(argv[1], "sigstatus", length) == 0)
+	    && (length > 1)) {
+	char *status = NULL;
+
+	switch (bodyInfoPtr->sigStatus) {
+	case RAT_UNSIGNED:	status = "pgp_none"; break;
+	case RAT_UNCHECKED:	status = "pgp_unchecked"; break;
+	case RAT_SIG_GOOD:	status = "pgp_good"; break;
+	case RAT_SIG_BAD:	status = "pgp_bad"; break;
+	}
+	Tcl_SetResult(interp, status, TCL_STATIC);
+	return TCL_OK;
+
+    } else if ((c == 'c') && (strncmp(argv[1], "checksig", length) == 0)
+	    && (length > 2)) {
+	RatPGPChecksig(interp, messageProcInfo, bodyInfoPtr);
+	return TCL_OK;
+
+    } else if ((c == 'g') && (strncmp(argv[1], "getPGPOutput", length) == 0)
+	    && (length > 3)) {
+	if (bodyInfoPtr->pgpOutput) {
+	    Tcl_SetResult(interp, Tcl_DStringValue(bodyInfoPtr->pgpOutput),
+		    TCL_VOLATILE);
+	} else {
+	    Tcl_ResetResult(interp);
+	}
+	return TCL_OK;
+
     } else {
 	Tcl_AppendResult(interp, "bad option \"", argv[1],
 		"\": must be one of children, message, type, parameter, id",
 		", description, data, saveData, dsn, getShowCharset",
-		", getShowCommand or filename", (char *) NULL);
+		", getShowCommand, filename, encoded, sigstatus, checksig",
+		" or getPGPOutput", NULL);
 	return TCL_ERROR;
     }
 }
@@ -1000,12 +1129,18 @@ RatBodyDelete(Tcl_Interp *interp, BodyInfo *bodyInfoPtr)
     BodyInfo *siblingInfoPtr, *nextSiblingInfoPtr;
     Tcl_DeleteCommand(interp, bodyInfoPtr->cmdName);
     siblingInfoPtr = bodyInfoPtr->firstbornPtr;
-    (*messageProcInfo[bodyInfoPtr->msgPtr->type].bodyDeleteProc)(bodyInfoPtr);
+    (*messageProcInfo[bodyInfoPtr->type].bodyDeleteProc)(bodyInfoPtr);
     while (siblingInfoPtr) {
 	nextSiblingInfoPtr = siblingInfoPtr->nextPtr;
 	RatBodyDelete(interp, siblingInfoPtr);
 	siblingInfoPtr = nextSiblingInfoPtr;
     }
+    if (bodyInfoPtr->containedEntity) {
+	RatMessageDelete(interp, bodyInfoPtr->containedEntity);
+    }
+    if (bodyInfoPtr->pgpOutput) {
+	free(bodyInfoPtr->pgpOutput);
+    }
     ckfree(bodyInfoPtr);
 }
 
@@ -1068,16 +1203,30 @@ RatMessageSave(Tcl_Interp *interp, Tcl_Channel channel, MessageInfo *msgPtr)
  */
 
 void
-RatMessageGet(Tcl_Interp *interp, MessageInfo *msgPtr, Tcl_DString *ds)
+RatMessageGet(Tcl_Interp *interp, MessageInfo *msgPtr, Tcl_DString *ds,
+	      char *flags, char *date)
 {
     char *data;
+    int seen;
 
     data = (*messageProcInfo[msgPtr->type].getHeadersProc)(interp, msgPtr);
     Tcl_DStringAppend(ds, data, strlen(data));
-    data = "\n";
-    Tcl_DStringAppend(ds, data, strlen(data));
+    seen  = (*msgPtr->folderInfoPtr->getFlagProc)(msgPtr->folderInfoPtr,
+	    interp, msgPtr->msgNo, RAT_SEEN);
     data = (*messageProcInfo[msgPtr->type].fetchTextProc)(interp, msgPtr);
     Tcl_DStringAppend(ds, data, strlen(data));
+    if (!seen) {
+        (*msgPtr->folderInfoPtr->setFlagProc)(msgPtr->folderInfoPtr,
+	    interp, msgPtr->msgNo, RAT_SEEN, 0);
+    }
+    if (flags) {
+	data = (*messageProcInfo[msgPtr->type].getInfoProc)(interp,
+		(ClientData)msgPtr, RAT_FOLDER_FLAGS, 0);
+	strcpy(flags, data);
+	data = (*messageProcInfo[msgPtr->type].getInfoProc)(interp,
+		(ClientData)msgPtr, RAT_FOLDER_DATE_IMAP4, 0);
+	strcpy(date, data);
+    }
 }
 
 /*
@@ -1172,17 +1321,20 @@ RatInsertMsg (Tcl_Interp *interp, MessageInfo *msgPtr, char *keywords,
 		flags = cpystr(elemArgv[1]);
 	    }
 	} else if (!strcasecmp(elemArgv[0], "date")) {
-	    mail_parse_date(&elt, elemArgv[1]);
-	    tm.tm_sec = elt.seconds;
-	    tm.tm_min = elt.minutes;
-	    tm.tm_hour = elt.hours;
-	    tm.tm_mday = elt.day;
-	    tm.tm_mon = elt.month - 1;
-	    tm.tm_year = elt.year+70;
-	    tm.tm_wday = 0;
-	    tm.tm_yday = 0;
-	    tm.tm_isdst = -1;
-	    date = (int)mktime(&tm);
+	    if (T == mail_parse_date(&elt, elemArgv[1])) {
+		tm.tm_sec = elt.seconds;
+		tm.tm_min = elt.minutes;
+		tm.tm_hour = elt.hours;
+		tm.tm_mday = elt.day;
+		tm.tm_mon = elt.month - 1;
+		tm.tm_year = elt.year+70;
+		tm.tm_wday = 0;
+		tm.tm_yday = 0;
+		tm.tm_isdst = -1;
+		date = (int)mktime(&tm);
+	    } else {
+		date = 0;
+	    }
 	}
 	ckfree(elemArgv);
     }
@@ -1193,10 +1345,8 @@ RatInsertMsg (Tcl_Interp *interp, MessageInfo *msgPtr, char *keywords,
     }
     Tcl_DStringInit(&dString);
     eFrom = (*messageProcInfo[msgPtr->type].getEnvelopeProc)(interp, msgPtr);
-    Tcl_DStringAppend(&dString, eFrom, strlen(eFrom));
     header = (*messageProcInfo[msgPtr->type].getHeadersProc)(interp, msgPtr);
     Tcl_DStringAppend(&dString, header, strlen(header));
-    Tcl_DStringAppend(&dString, "\n", -1);
     body = (*messageProcInfo[msgPtr->type].fetchTextProc)(interp, msgPtr);
     Tcl_DStringAppend(&dString, body, strlen(body));
     Tcl_ResetResult(interp);
@@ -1205,8 +1355,9 @@ RatInsertMsg (Tcl_Interp *interp, MessageInfo *msgPtr, char *keywords,
 	exTime = 0;
     }
     result = RatDbInsert(interp, to, from, cc, subject, date, flags, keywords,
-	    exTime, exType, Tcl_DStringValue(&dString),
+	    exTime, exType, eFrom, Tcl_DStringValue(&dString),
 	    Tcl_DStringLength(&dString));
+    Tcl_DStringFree(&dString);
     ckfree(to);
     ckfree(from);
     ckfree(cc);
@@ -1262,7 +1413,7 @@ RatParseList(char *format)
 		continue;
 	    }
 	    buf[bufLen] = '\0';
-	    expPtr->preString[expIndex] = strdup(buf);
+	    expPtr->preString[expIndex] = cpystr(buf);
 	    if ('-' == format[i]) {
 		expPtr->leftJust[expIndex] = 1;
 		i++;
@@ -1280,7 +1431,7 @@ RatParseList(char *format)
 	    case 'm': expPtr->typeList[expIndex++] = RAT_FOLDER_MAIL; break;
 	    case 'r': expPtr->typeList[expIndex++] = RAT_FOLDER_NAME_RECIPIENT;
 		    break;
-	    case 'R': expPtr->typeList[expIndex++] =RAT_FOLDER_MAIL_RECIPIENT;
+	    case 'R': expPtr->typeList[expIndex++] = RAT_FOLDER_MAIL_RECIPIENT;
 		    break;
 	    case 'b': expPtr->typeList[expIndex++] = RAT_FOLDER_SIZE; break;
 	    case 'B': expPtr->typeList[expIndex++] = RAT_FOLDER_SIZE_F; break;
@@ -1296,7 +1447,7 @@ RatParseList(char *format)
     }
     if (bufLen) {
 	buf[bufLen] = '\0';
-	expPtr->postString = strdup(buf);
+	expPtr->postString = cpystr(buf);
     } else {
 	expPtr->postString = NULL;
     }
@@ -1435,10 +1586,11 @@ RatBodySave(Tcl_Interp *interp,Tcl_Channel channel, BodyInfo *bodyInfoPtr,
     BODY *bodyPtr = bodyInfoPtr->bodyPtr;
     RatEncoding encoding;
     unsigned char *body;
-    int length, result = 0, i;
+    int result = 0, i;
+    unsigned long length;
 
     if (NULL == (body = (unsigned char *)
-	    (*messageProcInfo[bodyInfoPtr->msgPtr->type].fetchBodyProc)
+	    (*messageProcInfo[bodyInfoPtr->type].fetchBodyProc)
 	    (bodyInfoPtr, &length))) {
 	Tcl_SetResult(interp, "[Body not available]\n", TCL_STATIC);
 	return TCL_OK;
@@ -1482,3 +1634,40 @@ RatBodySave(Tcl_Interp *interp,Tcl_Channel channel, BodyInfo *bodyInfoPtr,
     }
     return TCL_OK;
 }
+
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * RatFindFirstText --
+ *
+ *      Finds the first text part of a message
+ *
+ * Results:
+ *	A pointer to the BodyInfo of the first text part, or NULL if none
+ *	are found.
+ *
+ * Side effects:
+ *	None.
+ *
+ *
+ *----------------------------------------------------------------------
+ */
+
+static BodyInfo*
+RatFindFirstText(BodyInfo *bodyInfoPtr)
+{
+    BodyInfo *b2Ptr;
+
+    for (; NULL != bodyInfoPtr; bodyInfoPtr = bodyInfoPtr->nextPtr) {
+	if (TYPETEXT == bodyInfoPtr->bodyPtr->type) {
+	    return bodyInfoPtr;
+	}
+	if (bodyInfoPtr->firstbornPtr && (NULL !=
+		(b2Ptr = RatFindFirstText(bodyInfoPtr->firstbornPtr)))) {
+	    return b2Ptr;
+	}
+    }
+    return NULL;
+}
