@@ -4,22 +4,36 @@
  *	This is the subprocess which handles the actual sending of messages.
  *
  *
- * TkRat software and its included text is Copyright 1996 by Martin Forssen.
+ * TkRat software and its included text is Copyright 1996,1997,1998
+ * by Martin Forssn
  *
  * The full text of the legal notice is contained in the file called
  * COPYRIGHT, included with this distribution.
  */
 
 #include <tk.h>
-#include "rat.h"
+#include "ratFolder.h"
+#include "ratPGP.h"
+
+/*
+ * This is a list of outstanding commands
+ */
+typedef struct CmdList {
+    char *cmd;
+    struct CmdList *next;
+} CmdList;
+static CmdList *cmdList = NULL;
+
+static SMTPChannel smtpChannel = NULL;
 
 /*
  * Create bodypart s procedure
  */
-static BODY *RatCreateBody(Tcl_Interp *interp, char *handler, int *errorFlag,
-	char *charset, Tcl_DString *files);
+static BODY *RatCreateBody(Tcl_Interp *interp, char *handler, ENVELOPE *env,
+	int *errorFlag, char *charset, Tcl_DString *files);
 static int RatSenderSend(Tcl_Interp *interp, char *prefix,
 	Tcl_DString *usedVariables, Tcl_DString *files);
+static void RatSenderStanddown(Tcl_Interp *interp);
 
 
 /*
@@ -43,46 +57,85 @@ void
 RatSender(Tcl_Interp *interp)
 {
     Tcl_DString usedVariables, files, result;
-    char buf[1024], **argv;
-    int argc, i;
+    char *buf, **argv, **sendlistArgv;
+    CmdList *cmdPtr;
+    int sendlistArgc, argc, i, s, buflen;
+
+    /*
+     * Clear cached passwords
+     */
+    ClearPGPPass(NULL);
 
     Tcl_DStringInit(&usedVariables);
     Tcl_DStringInit(&files);
     Tcl_DStringInit(&result);
+    buflen = 1024;
+    buf = (char*)malloc(buflen);
 
-    while ((fgets(buf, sizeof(buf), stdin), !feof(stdin))
-	    && !strncmp(buf, "SEND", 4)) {
-	(void)Tcl_SplitList(interp, buf, &argc, &argv);
-	if (TCL_OK == RatSenderSend(interp, argv[2], &usedVariables, &files)) {
-	    Tcl_DStringAppendElement(&result, "SENT");
-	    Tcl_DStringAppendElement(&result, argv[1]);
-	    Tcl_DStringAppendElement(&result, argv[3]);
-	    free(argv);
-	    (void)Tcl_SplitList(interp, Tcl_DStringValue(&files), &argc, &argv);
-	    for (i=0; i<argc; i++) {
-		(void)unlink(argv[i]);
-	    }
+    while (1) {
+	if (cmdList) {
+	    cmdPtr = cmdList;
+	    strcpy(buf, cmdList->cmd);
+	    cmdList = cmdList->next;
+	    free(cmdPtr->cmd);
+	    free(cmdPtr);
 	} else {
-	    Tcl_DStringAppendElement(&result, "FAILED");
-	    Tcl_DStringAppendElement(&result, argv[1]);
-	    Tcl_DStringAppendElement(&result, argv[2]);
-	    Tcl_DStringAppendElement(&result, argv[3]);
-	    Tcl_DStringAppendElement(&result, interp->result);
+	    i = 0;
+	    while (buf[buflen-2] = '\0',
+		    fgets(buf+i, buflen-i, stdin)
+		    && buflen-i-1 == strlen(buf+i)
+		    && buf[buflen-2] != '\n') {
+		i = buflen-1;
+		buflen += 1024;
+		buf = REALLOC(buf, buflen);
+	    }
+	    if (feof(stdin)) {
+		exit(0);
+	    }
 	}
-	free(argv);
-	(void)Tcl_SplitList(interp, Tcl_DStringValue(&usedVariables), &argc,
-		&argv);
-	for (i=0; i<argc; i++) {
-	    (void)Tcl_UnsetVar(interp, argv[i], TCL_GLOBAL_ONLY);
+	
+	if (!strncmp(buf, "SEND", 4)) {
+	    (void)Tcl_SplitList(interp, buf, &sendlistArgc, &sendlistArgv);
+	    for (s=1; s<sendlistArgc; s++) {
+		(void)Tcl_SplitList(interp, sendlistArgv[s], &argc, &argv);
+		if (TCL_OK == RatSenderSend(interp, argv[1], &usedVariables,
+			&files)){
+		    Tcl_DStringAppendElement(&result, "SENT");
+		    Tcl_DStringAppendElement(&result, argv[0]);
+		    Tcl_DStringAppendElement(&result, argv[2]);
+		    free(argv);
+		    Tcl_SplitList(interp, Tcl_DStringValue(&files),&argc,&argv);
+		    for (i=0; i<argc; i++) {
+			(void)unlink(argv[i]);
+		    }
+		} else {
+		    Tcl_DStringAppendElement(&result, "FAILED");
+		    Tcl_DStringAppendElement(&result, argv[0]);
+		    Tcl_DStringAppendElement(&result, argv[1]);
+		    Tcl_DStringAppendElement(&result, argv[2]);
+		    Tcl_DStringAppendElement(&result, interp->result);
+		}
+		free(argv);
+		(void)Tcl_SplitList(interp, Tcl_DStringValue(&usedVariables),
+				    &argc, &argv);
+		for (i=0; i<argc; i++) {
+		    (void)Tcl_UnsetVar(interp, argv[i], TCL_GLOBAL_ONLY);
+		}
+		free(argv);
+		fwrite(Tcl_DStringValue(&result), Tcl_DStringLength(&result)+1,
+		       1, stdout);
+		fflush(stdout);
+		Tcl_DStringSetLength(&result, 0);
+		Tcl_DStringSetLength(&usedVariables, 0);
+		Tcl_DStringSetLength(&files, 0);
+	    }
+	    free(sendlistArgv);
+	    RatSenderStanddown(interp);
+	} else {
+	    exit(0);
 	}
-	free(argv);
-	fprintf(stdout, "%s", Tcl_DStringValue(&result));
-	fputc('\0', stdout);
-	fflush(stdout);
-	Tcl_DStringSetLength(&result, 0);
-	Tcl_DStringSetLength(&usedVariables, 0);
-	Tcl_DStringSetLength(&files, 0);
     }
+    /* Notreached */
     exit(0);
 }
 
@@ -111,8 +164,7 @@ RatSenderSend(Tcl_Interp *interp, char *prefix, Tcl_DString *usedVariablesPtr,
     char *charset = Tcl_GetVar2(interp, "option", "charset", TCL_GLOBAL_ONLY);
     int listArgc, i, errorFlag = 0, requestDSN, useFrom, verbose;
     char *tmp, buf[10000], **listArgv, *handler, *saveTo;
-    SMTPChannel smtpChannel;
-    ENVELOPE *msg;
+    ENVELOPE *env;
     BODY *body;
 
     /*
@@ -121,7 +173,7 @@ RatSenderSend(Tcl_Interp *interp, char *prefix, Tcl_DString *usedVariablesPtr,
     if (TCL_OK != RatHoldExtract(interp, prefix, usedVariablesPtr, filesPtr)) {
 	return TCL_ERROR;
     }
-    handler = strdup(interp->result);
+    handler = cpystr(interp->result);
 
     /*
      * Construct the headers
@@ -131,82 +183,82 @@ RatSenderSend(Tcl_Interp *interp, char *prefix, Tcl_DString *usedVariablesPtr,
 	requestDSN = 0;
     }
     tmp = Tcl_GetVar2(interp, handler, "to", TCL_GLOBAL_ONLY);
-    msg = mail_newenvelope ();
-    rfc822_parse_adrlist(&msg->to, tmp, currentHost);
-    msg->remail = cpystr(Tcl_GetVar2(interp,handler,"remail",TCL_GLOBAL_ONLY));
+    env = mail_newenvelope ();
+    rfc822_parse_adrlist(&env->to, tmp, currentHost);
+    env->remail = cpystr(Tcl_GetVar2(interp,handler,"remail",TCL_GLOBAL_ONLY));
     if ((tmp = Tcl_GetVar2(interp, handler, "return_path", TCL_GLOBAL_ONLY)) 
 	    && !RatIsEmpty(tmp)) {
-	rfc822_parse_adrlist(&msg->return_path, tmp, currentHost);
+	rfc822_parse_adrlist(&env->return_path, tmp, currentHost);
     } else {
-	msg->return_path = mail_newaddr();
-	msg->return_path->personal =
+	env->return_path = mail_newaddr();
+	env->return_path->personal =
 		currentPersonalName ? cpystr(currentPersonalName) : NULL;
-	msg->return_path->mailbox = cpystr(currentMailboxName);
-	msg->return_path->host = cpystr(currentHost);
+	env->return_path->mailbox = cpystr(currentMailboxName);
+	env->return_path->host = cpystr(currentHost);
     }
-    if (NULL==(msg->date=Tcl_GetVar2(interp,handler,"date",TCL_GLOBAL_ONLY))) {
+    if (NULL==(env->date=Tcl_GetVar2(interp,handler,"date",TCL_GLOBAL_ONLY))) {
 	rfc822_date(buf);
-	msg->date = buf;
+	env->date = buf;
     }
-    msg->date = cpystr(msg->date);
+    env->date = cpystr(env->date);
     if (TCL_OK != Tcl_GetBoolean(interp, Tcl_GetVar2(interp, "option",
 	    "use_from", TCL_GLOBAL_ONLY), &useFrom)) {
 	useFrom = 0;
     }
     if (useFrom && (tmp = Tcl_GetVar2(interp, handler, "from", TCL_GLOBAL_ONLY))
 	    && !RatIsEmpty(tmp)) {
-	rfc822_parse_adrlist(&msg->from, tmp, currentHost);
+	rfc822_parse_adrlist(&env->from, tmp, currentHost);
     }
-    if (msg->from) {
+    if (env->from) {
 	ADDRESS *adrPtr;
 
-	for (adrPtr = msg->from; adrPtr; adrPtr = adrPtr->next) {
+	for (adrPtr = env->from; adrPtr; adrPtr = adrPtr->next) {
 	    if (RatAddressIsMe(interp, adrPtr, 0)) {
 		break;
 	    }
 	}
 	if (!adrPtr) {
-	    msg->sender = mail_newaddr();
-	    msg->sender->personal =
+	    env->sender = mail_newaddr();
+	    env->sender->personal =
 		    currentPersonalName ? cpystr(currentPersonalName) : NULL;
-	    msg->sender->mailbox = cpystr(currentMailboxName);
-	    msg->sender->host = cpystr(currentHost);
+	    env->sender->mailbox = cpystr(currentMailboxName);
+	    env->sender->host = cpystr(currentHost);
 	} else {
-	    msg->sender = NULL;
+	    env->sender = NULL;
 	}
     } else {
-	msg->from = mail_newaddr();
-	msg->from->personal =
+	env->from = mail_newaddr();
+	env->from->personal =
 		currentPersonalName ? cpystr(currentPersonalName) : NULL;
-	msg->from->mailbox = cpystr(currentMailboxName);
-	msg->from->host = cpystr(currentHost);
+	env->from->mailbox = cpystr(currentMailboxName);
+	env->from->host = cpystr(currentHost);
     }
     if ((tmp = Tcl_GetVar2(interp, handler, "reply_to", TCL_GLOBAL_ONLY))
 	    && !RatIsEmpty(tmp)) {
-	rfc822_parse_adrlist(&msg->reply_to, tmp, currentHost);
+	rfc822_parse_adrlist(&env->reply_to, tmp, currentHost);
     }
-    msg->subject = (char*)RatEncodeHeaderLine((unsigned char*)
+    env->subject = (char*)RatEncodeHeaderLine((unsigned char*)
 	    Tcl_GetVar2(interp, handler, "subject", TCL_GLOBAL_ONLY),
 	    charset, 9, 1);
     if ((tmp = Tcl_GetVar2(interp, handler, "cc", TCL_GLOBAL_ONLY))
 	    && !RatIsEmpty(tmp)) {
-	rfc822_parse_adrlist(&msg->cc, tmp, currentHost);
+	rfc822_parse_adrlist(&env->cc, tmp, currentHost);
     }
     if ((tmp = Tcl_GetVar2(interp, handler, "bcc", TCL_GLOBAL_ONLY))
 	    && !RatIsEmpty(tmp)) {
-	rfc822_parse_adrlist(&msg->bcc, tmp, currentHost);
+	rfc822_parse_adrlist(&env->bcc, tmp, currentHost);
     }
-    msg->in_reply_to = cpystr(Tcl_GetVar2(interp, handler, "in_reply_to",
+    env->in_reply_to = cpystr(Tcl_GetVar2(interp, handler, "in_reply_to",
 	    TCL_GLOBAL_ONLY));
-    msg->message_id = cpystr(Tcl_GetVar2(interp, handler, "message_id",
+    env->message_id = cpystr(Tcl_GetVar2(interp, handler, "message_id",
 	    TCL_GLOBAL_ONLY));
-    msg->newsgroups = NULL;
+    env->newsgroups = NULL;
 
     /*
      * Construct the body
      */
     if ((tmp = Tcl_GetVar2(interp, handler, "body", TCL_GLOBAL_ONLY))) {
-	body = RatCreateBody(interp, tmp, &errorFlag, charset, filesPtr);
+	body = RatCreateBody(interp, tmp, env, &errorFlag, charset, filesPtr);
     } else {
 	body = mail_newbody ();
     }
@@ -225,6 +277,10 @@ RatSenderSend(Tcl_Interp *interp, char *prefix, Tcl_DString *usedVariablesPtr,
     }
     tmp = Tcl_GetVar2(interp, "option", "sendprot", TCL_GLOBAL_ONLY);
     if (tmp && !strcmp(tmp, "smtp")) {
+	int reuseChannel;
+
+	Tcl_GetBoolean(interp, Tcl_GetVar2(interp, "option", "smtp_reuse",
+		TCL_GLOBAL_ONLY), &reuseChannel);
 	listArgc = 0;
 	if ((tmp = Tcl_GetVar2(interp,"option","smtp_hosts",TCL_GLOBAL_ONLY))) {
 	    Tcl_SplitList(interp, tmp, &listArgc, &listArgv);
@@ -235,7 +291,7 @@ RatSenderSend(Tcl_Interp *interp, char *prefix, Tcl_DString *usedVariablesPtr,
 		          TCL_STATIC);
 	    goto error;
 	}
-	for (i=0, smtpChannel=NULL; !smtpChannel && i<listArgc; i++) {
+	for (i=0; !smtpChannel && i<listArgc; i++) {
 	    smtpChannel = RatSMTPOpen(interp, listArgv[i], verbose);
 	}
 	ckfree(listArgv);
@@ -243,16 +299,20 @@ RatSenderSend(Tcl_Interp *interp, char *prefix, Tcl_DString *usedVariablesPtr,
 	    Tcl_SetResult(interp, "No valid SMTP hosts", TCL_STATIC);
 	    goto error;
 	}
-	if (TCL_OK != RatSMTPSend(interp, smtpChannel, msg, body, requestDSN,
+	if (TCL_OK != RatSMTPSend(interp, smtpChannel, env, body, requestDSN,
 		verbose)) {
 	    Tcl_DString ds;
 	    Tcl_DStringInit(&ds);
 	    Tcl_DStringGetResult(interp, &ds);
 	    RatSMTPClose(interp, smtpChannel, verbose);
+	    smtpChannel = NULL;
 	    Tcl_DStringResult(interp, &ds);
 	    goto error;
 	}
-	RatSMTPClose(interp, smtpChannel, verbose);
+	if (!reuseChannel) {
+	    RatSMTPClose(interp, smtpChannel, verbose);
+	    smtpChannel = NULL;
+	}
 
     } else if (tmp && !strcmp(tmp, "prog")) {
 	Tcl_Channel channel;
@@ -268,33 +328,36 @@ RatSenderSend(Tcl_Interp *interp, char *prefix, Tcl_DString *usedVariablesPtr,
 	}
 	Tcl_DStringInit(&ds);
 	Tcl_DStringAppend(&ds, tmp, -1);
-	for (adrPtr = msg->to; adrPtr; adrPtr = adrPtr->next) {
+	for (adrPtr = env->to; adrPtr; adrPtr = adrPtr->next) {
 	    buf[0] = '\0';
 	    rfc822_address(buf, adrPtr);
 	    Tcl_DStringAppendElement(&ds, buf);
 	}
-	for (adrPtr = msg->cc; adrPtr; adrPtr = adrPtr->next) {
+	for (adrPtr = env->cc; adrPtr; adrPtr = adrPtr->next) {
 	    buf[0] = '\0';
 	    rfc822_address(buf, adrPtr);
 	    Tcl_DStringAppendElement(&ds, buf);
 	}
-	for (adrPtr = msg->bcc; adrPtr; adrPtr = adrPtr->next) {
+	for (adrPtr = env->bcc; adrPtr; adrPtr = adrPtr->next) {
 	    buf[0] = '\0';
 	    rfc822_address(buf, adrPtr);
 	    Tcl_DStringAppendElement(&ds, buf);
 	}
 
+	Tcl_ResetResult(interp);
 	if (TCL_OK != Tcl_SplitList(interp, Tcl_DStringValue(&ds),
 		    &listArgc, &listArgv)
 		|| (NULL == (channel = Tcl_OpenCommandChannel(interp, listArgc,
 		    listArgv, TCL_STDIN|TCL_STDOUT|TCL_STDERR)))) {
-	    Tcl_DStringFree(&ds);
-	    Tcl_SetResult(interp, "Invalid send program", TCL_STATIC);
+	    Tcl_DStringSetLength(&ds, 0);
+	    Tcl_DStringAppend(&ds, "Failed to run send program: ", -1);
+	    Tcl_DStringAppend(&ds, interp->result, -1);
+	    Tcl_DStringResult(interp, &ds);
 	    goto error;
 	} else {
 	    Tcl_DStringFree(&ds);
 	    ckfree(listArgv);
-	    rfc822_output(buf, msg, body, RatTclPuts, channel, i);
+	    rfc822_output(buf, env, body, RatTclPutsSMTP, channel, i);
 	    Tcl_Close(interp, channel);
 	}
     } else {
@@ -308,27 +371,19 @@ RatSenderSend(Tcl_Interp *interp, char *prefix, Tcl_DString *usedVariablesPtr,
 
     saveTo = Tcl_GetVar2(interp, handler, "save_to", TCL_GLOBAL_ONLY);
     if (saveTo && *saveTo) {
+	MESSAGECACHE elt;
 	Tcl_DString saveArg;
 	Tcl_Channel channel;
 	char saveFile[1024];
 	struct tm *tmPtr;
-	int perm;
 	time_t now;
+	int perm;
 
 	Tcl_DStringInit(&saveArg);
 	tmp = Tcl_GetVar2(interp, "option", "tmp", TCL_GLOBAL_ONLY);
 	RatGenId(NULL, interp, 0, NULL);
 	sprintf(saveFile, "%s/rat.%s", tmp, interp->result);
 
-	now = time(NULL);
-	tmPtr = gmtime(&now);
-	sprintf(buf, "From %s@%s %s %s %2d %02d:%02d GMT 19%02d\n",
-		currentMailboxName, currentHost,
-		dayName[tmPtr->tm_wday], monthName[tmPtr->tm_mon],
-		tmPtr->tm_mday, tmPtr->tm_hour, tmPtr->tm_min,
-		tmPtr->tm_year);
-	strcat(buf, "Status: RO\n");
-
 	Tcl_GetInt(interp, Tcl_GetVar2(interp, "option", "permissions",
 		TCL_GLOBAL_ONLY), &perm);
 	if (NULL == (channel=Tcl_OpenFileChannel(interp,saveFile,"a",perm))){
@@ -336,33 +391,45 @@ RatSenderSend(Tcl_Interp *interp, char *prefix, Tcl_DString *usedVariablesPtr,
 		    Tcl_PosixError(interp), (char*) NULL);
 	    goto error;
 	}
-	Tcl_Write(channel, buf, strlen(buf));
-	rfc822_output(buf, msg, body, RatTclPuts, channel, 1);
+	rfc822_output(buf, env, body, RatTclPuts, channel, 1);
 	Tcl_Close(interp, channel);
 
 	Tcl_DStringAppendElement(&saveArg, saveFile);
 	Tcl_DStringAppendElement(&saveArg, saveTo);
 	buf[0] = '\0';
-	rfc822_write_address(buf, msg->to);
+	rfc822_write_address(buf, env->to);
 	Tcl_DStringAppendElement(&saveArg, buf);
-	if (msg->from) {
+	if (env->from) {
 	    buf[0] = '\0';
-	    rfc822_write_address(buf, msg->from);
+	    rfc822_write_address(buf, env->from);
 	} else {
 	    sprintf(buf, "%s@%s", currentMailboxName, currentHost);
 	}
 	Tcl_DStringAppendElement(&saveArg, buf);
 	buf[0] = '\0';
-	rfc822_write_address(buf, msg->cc);
+	rfc822_write_address(buf, env->cc);
 	Tcl_DStringAppendElement(&saveArg, buf);
-	Tcl_DStringAppendElement(&saveArg, msg->subject);
+	Tcl_DStringAppendElement(&saveArg, env->subject);
+	Tcl_DStringAppendElement(&saveArg, RAT_SEEN_STR);
+	now = time(NULL);
+	tmPtr = gmtime(&now);
+	elt.day = tmPtr->tm_mday;
+	elt.month = tmPtr->tm_mon+1;
+	elt.year = tmPtr->tm_year+1900-BASEYEAR;
+	elt.hours = tmPtr->tm_hour;
+	elt.minutes = tmPtr->tm_min;
+	elt.seconds = tmPtr->tm_sec;
+	elt.zoccident = 0;
+	elt.zhours = 0;
+	elt.zminutes = 0;
+	Tcl_DStringAppendElement(&saveArg, mail_date(buf, &elt));
 	fprintf(stdout, "SAVE %s", Tcl_DStringValue(&saveArg));
 	fputc('\0', stdout);
 	fflush(stdout);
 	Tcl_DStringFree(&saveArg);
     }
 
-    mail_free_envelope(&msg);
+    mail_free_envelope(&env);
     mail_free_body(&body);
     return TCL_OK;
 
@@ -370,10 +437,42 @@ error:
     if (verbose) {
 	RatLog(interp, RAT_PARSE, "", 1);
     }
-    mail_free_envelope(&msg);
+    mail_free_envelope(&env);
     mail_free_body(&body);
     return TCL_ERROR;
 }
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * RatSenderStanddown --
+ *
+ *	Closes the open SMTP channel (if open)
+ *
+ * Results:
+ *	None
+ *
+ * Side effects:
+ *      smtpChannel is closed.
+ *
+ *
+ *----------------------------------------------------------------------
+ */
+
+static void
+RatSenderStanddown(Tcl_Interp *interp)
+{
+    int verbose;
+
+    if (smtpChannel) {
+	Tcl_GetInt(interp, Tcl_GetVar2(interp, "option", "smtp_verbose",
+		TCL_GLOBAL_ONLY), &verbose);
+	RatSMTPClose(interp, smtpChannel, verbose);
+	smtpChannel = NULL;
+    }
+}
+
 
 /*
  *----------------------------------------------------------------------
@@ -394,8 +493,8 @@ error:
  *----------------------------------------------------------------------
  */
 static BODY*
-RatCreateBody(Tcl_Interp *interp, char *handler, int *errorFlag, char *charset,
-	Tcl_DString *filesPtr)
+RatCreateBody(Tcl_Interp *interp, char *handler, ENVELOPE *env,
+	int *errorFlag, char *charset, Tcl_DString *filesPtr)
 {
     BODY *body = mail_newbody();
     char *type;
@@ -403,10 +502,12 @@ RatCreateBody(Tcl_Interp *interp, char *handler, int *errorFlag, char *charset,
     char *parameter;
     char buf[1024];
     char *filename = NULL;
+    char *value;
+    int pgp_sign, pgp_encrypt;
 
     if (NULL == (type = Tcl_GetVar2(interp, handler, "type",
 	    TCL_GLOBAL_ONLY))) {
-	RatLog(interp, RAT_ERROR, "Internal error, no bhandler(type)", 0);
+	Tcl_SetResult(interp, "Internal error, no bhandler(type)", TCL_STATIC);
 	(*errorFlag)++;
 	return body;
     }
@@ -430,13 +531,13 @@ RatCreateBody(Tcl_Interp *interp, char *handler, int *errorFlag, char *charset,
 					body->encoding = ENCQUOTEDPRINTABLE;
 	else {
 	    sprintf(buf, "Unkown encoding %s\n", encoding);
+	    Tcl_SetResult(interp, buf, TCL_VOLATILE);
 	    (*errorFlag)++;
-	    RatLog(interp, RAT_ERROR, buf, 0);
 	}
     }
     if (NULL == (body->subtype = Tcl_GetVar2(interp, handler, "subtype",
 	    TCL_GLOBAL_ONLY))) {
-	RatLog(interp, RAT_ERROR, "Internal error, no bhandler(subtype)", 0);
+	Tcl_SetResult(interp,"Internal error, no bhandler(subtype)",TCL_STATIC);
 	(*errorFlag)++;
 	return body;
     }
@@ -450,7 +551,7 @@ RatCreateBody(Tcl_Interp *interp, char *handler, int *errorFlag, char *charset,
 	    if (TCL_ERROR == Tcl_SplitList(interp, parmListArgv[i], &parmArgc,
 		    &parmArgv) || 2 != parmArgc) {
 		sprintf(buf, "Illegal parameter: %s", parmListArgv[i]);
-		RatLog(interp, RAT_ERROR, buf, 0);
+		Tcl_SetResult(interp, buf, TCL_VOLATILE);
 		(*errorFlag)++;
 	    } else {
 		*parmPtrPtr = mail_newbody_parameter();
@@ -483,7 +584,7 @@ RatCreateBody(Tcl_Interp *interp, char *handler, int *errorFlag, char *charset,
 	for (i=0; i<childrenArgc; i++) {
 	    *partPtrPtr = mail_newbody_part();
 	    (*partPtrPtr)->body = *RatCreateBody(interp, childrenArgv[i],
-		    errorFlag, charset, filesPtr);
+		    env, errorFlag, charset, filesPtr);
 	    partPtrPtr = &(*partPtrPtr)->next;
 	}
     } else if (TYPEMESSAGE == body->type) {
@@ -493,14 +594,15 @@ RatCreateBody(Tcl_Interp *interp, char *handler, int *errorFlag, char *charset,
 
 	if (NULL == (filename = Tcl_GetVar2(interp, handler,"filename",
 		TCL_GLOBAL_ONLY))) {
-	    RatLog(interp, RAT_ERROR,"Internal error, no bhandler(filename)",0);
+	    Tcl_SetResult(interp, "Internal error, no bhandler(filename)",
+			  TCL_STATIC);
 	    (*errorFlag)++;
 	    return body;
 	}
 	if (-1 == (fd = open(filename, O_RDONLY))) {
 	    sprintf(buf, "Failed to open file \"%s\": %s", filename,
 		    Tcl_PosixError(interp));
-	    RatLog(interp, RAT_ERROR, buf, 0);
+	    Tcl_SetResult(interp, buf, TCL_VOLATILE);
 	    (*errorFlag)++;
 	    return body;
 	}
@@ -522,23 +624,24 @@ RatCreateBody(Tcl_Interp *interp, char *handler, int *errorFlag, char *charset,
 
 	if (NULL == (filename = Tcl_GetVar2(interp, handler,"filename",
 		TCL_GLOBAL_ONLY))) {
-	    RatLog(interp, RAT_ERROR,"Internal error, no bhandler(filename)",0);
+	    Tcl_SetResult(interp, "Internal error, no bhandler(filename)",
+			  TCL_STATIC);
 	    (*errorFlag)++;
 	    return body;
 	}
 	if (NULL == (fp = fopen(filename, "r"))) {
 	    sprintf(buf, "Failed to open file \"%s\": %s", filename,
 		    Tcl_PosixError(interp));
-	    RatLog(interp, RAT_ERROR, buf, 0);
+	    Tcl_SetResult(interp, buf, TCL_VOLATILE);
 	    (*errorFlag)++;
 	    return body;
 	}
 	fstat(fileno(fp), &statbuf);
 	allocated = (statbuf.st_size*105)/100;
 	data = (char*)ckalloc(allocated);
-	if (TYPETEXT == body->type) {
+	if (ENCBINARY != body->encoding) {
 	    while (EOF != (ci = getc(fp))) {
-		if (len >= allocated-1) {
+		if (len >= allocated-2) {
 		    allocated += 1024;
 		    data = (char*)REALLOC(data, allocated);
 		}
@@ -553,20 +656,86 @@ RatCreateBody(Tcl_Interp *interp, char *handler, int *errorFlag, char *charset,
 	}
 	data[len] = '\0';
 	fclose(fp);
-	body->contents.text.data = (unsigned char*)data;
+	body->contents.text.data = data;
 	body->size.bytes = body->contents.text.size = len;
     }
     if (filename) {
 	int removeFile = 0;
-	char *tmp;
 
-	if ((tmp = Tcl_GetVar2(interp, handler,"removeFile", TCL_GLOBAL_ONLY))){
-	    Tcl_GetBoolean(interp, tmp, &removeFile);
+	if ((value = Tcl_GetVar2(interp,handler,"removeFile",TCL_GLOBAL_ONLY))){
+	    Tcl_GetBoolean(interp, value, &removeFile);
 	}
 	if (removeFile && filesPtr) {
 	    Tcl_DStringAppendElement(filesPtr, filename);
 	}
     }
 
+    /*
+     * Do PGP signing and/or encryption
+     */
+    if (!(value = Tcl_GetVar2(interp, handler, "pgp_sign", TCL_GLOBAL_ONLY)) ||
+         TCL_OK != Tcl_GetBoolean(interp, value, &pgp_sign)) {
+	pgp_sign = 0;
+    }
+    if (!(value = Tcl_GetVar2(interp, handler,"pgp_encrypt",TCL_GLOBAL_ONLY)) ||
+         TCL_OK != Tcl_GetBoolean(interp, value, &pgp_encrypt)) {
+	pgp_encrypt = 0;
+    }
+    if (pgp_sign || pgp_encrypt) {
+	if (pgp_encrypt) {
+	    body = RatPGPEncrypt(interp, env, body, pgp_sign);
+	} else {
+	    body = RatPGPSign(interp, body);
+	}
+	if (!body) {
+	    Tcl_SetResult(interp, "Failed to do PGP operation", TCL_STATIC);
+	    (*errorFlag)++;
+	    return body;
+	}
+    }
+
     return body;
 }
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * RatSendPGPCommand --
+ *
+ *	Sends a PGP command to the master and then waits for a reply.
+ *
+ * Results:
+ *      A pointer to a static buffer containing the command.
+ *
+ * Side effects:
+ *      Other commands that arrive while we are waiting are placed in cmdList
+ *
+ *
+ *----------------------------------------------------------------------
+ */
+
+char*
+RatSendPGPCommand(char *cmd)
+{
+    static char buf[1024];
+    CmdList **cmdPtrPtr;
+
+    fwrite(cmd, strlen(cmd)+1, 1, stdout);
+    fflush(stdout);
+    for (cmdPtrPtr = &cmdList; *cmdPtrPtr; cmdPtrPtr = &(*cmdPtrPtr)->next);
+    while(1) {
+	fgets(buf, sizeof(buf), stdin);
+	if (feof(stdin)) {
+	    exit(0);
+	}
+	buf[strlen(buf)-1] = '\0';
+	if (strncmp("PGP ", buf, 4)) {
+	    *cmdPtrPtr = (CmdList*)malloc(sizeof(CmdList));
+	    (*cmdPtrPtr)->cmd = cpystr(buf);
+	    (*cmdPtrPtr)->next = NULL;
+	    cmdPtrPtr = &(*cmdPtrPtr)->next;
+	} else {
+	    return buf+4;
+	}
+    }
+}
