@@ -4,7 +4,8 @@
  *      This file contains code which implements standard c-client folders.
  *	This means ONLY filefolders at this moment.
  *
- * TkRat software and its included text is Copyright 1996 by Martin Forssen.
+ * TkRat software and its included text is Copyright 1996,1997,1998
+ * by Martin Forssn
  *
  * The full text of the legal notice is contained in the file called
  * COPYRIGHT, included with this distribution.
@@ -53,6 +54,7 @@ typedef struct MMPasswd {
     char *password;
     RatStdFolderType type;
     struct MMPasswd *next;
+    Tcl_TimerToken token;
 } MMPasswd;
 
 /*
@@ -96,7 +98,12 @@ static char *loginUser = NULL;
 /*
  * List of flag names
  */
-static char *stdFlagNames[] = {"\\SEEN","\\DELETED","\\FLAGGED","\\ANSWERED"};
+static char *stdFlagNames[] = {
+    RAT_SEEN_STR,
+    RAT_DELETED_STR,
+    RAT_FLAGGED_STR,
+    RAT_ANSWERED_STR
+};
 
 /*
  * The types of folders. This list is closely related to RatStdFolderType
@@ -125,13 +132,13 @@ static RatUpdateProc Std_UpdateProc;
 static RatInsertProc Std_InsertProc;
 static RatSetFlagProc Std_SetFlagProc;
 static RatGetFlagProc Std_GetFlagProc;
-static RatExpungeProc Std_ExpungeProc;
 static RatInfoProc Std_InfoProc;
 static RatCreateProc Std_CreateProc;
 static MAILSTREAM *OpenAndConvert(Tcl_Interp *interp, RatStdFolderType type,
 	char *host, char *name, char *user);
 static Tcl_CmdProc StdListIMAPFolders;
 static Tcl_TimerProc CloseConnection;
+static Tcl_TimerProc ErasePasswd;
 static Tcl_CmdProc StdJudgeFolder;
 
 
@@ -160,12 +167,13 @@ RatStdFolderInit(Tcl_Interp *interp)
 {
     stdInterp = interp;
     Tcl_InitHashTable(&folderTable, TCL_ONE_WORD_KEYS);
-    mail_link(&bezerkdriver);
+    mail_link(&unixdriver);
     mail_link(&mmdfdriver);
     mail_link(&dummydriver);
     mail_link(&imapdriver);
     mail_link(&pop3driver);
     mail_link(&mhdriver);
+    auth_link(&auth_log);
     Tcl_CreateCommand(interp, "RatListIMAP", StdListIMAPFolders, NULL, NULL);
     Tcl_CreateCommand(interp, "RatImapCreateFolder", StdJudgeFolder,
 	    (ClientData)1, NULL);
@@ -200,7 +208,7 @@ OpenAndConvert(Tcl_Interp *interp, RatStdFolderType type, char *host,
 {
     MAILSTREAM *stream = NULL;
     Connection *connPtr;
-    int addToList = 0, doCache;
+    int doCache = 0;
 
     if ('{' == name[0]) {
 	for (connPtr = connListPtr; connPtr; connPtr = connPtr->next) {
@@ -211,12 +219,11 @@ OpenAndConvert(Tcl_Interp *interp, RatStdFolderType type, char *host,
 	}
 	if (connPtr) {
 	    stream = connPtr->stream;
-	    Tcl_DeleteTimerHandler(connPtr->token);
+	    if (connPtr->token) {
+		Tcl_DeleteTimerHandler(connPtr->token);
+	    }
 	    connPtr->closing = 0;
 	} else {
-	    if (RAT_POP != type) {
-		addToList = 1;
-	    }
 	    Tcl_GetBoolean(stdInterp,
 		    Tcl_GetVar2(stdInterp, "option", "cache_passwd",
 		    TCL_GLOBAL_ONLY), &doCache);
@@ -230,7 +237,7 @@ OpenAndConvert(Tcl_Interp *interp, RatStdFolderType type, char *host,
 	    }
 	}
     }
-    stream = mail_open(stream, name, 0);
+    stream = mail_open(stream, name, NIL);
     if (NIL == stream && name[0] == '/') {
 	FILE *fpIn = NULL, *fpOut;
 	struct stat statbuf;
@@ -297,20 +304,38 @@ OpenAndConvert(Tcl_Interp *interp, RatStdFolderType type, char *host,
 	    }
 	}
     }
-    if (addToList) {
-	connPtr = (Connection*)malloc(sizeof(Connection));
-	connPtr->stream = stream;
-	connPtr->type = type;
-	connPtr->host = strdup(host);
-	connPtr->user = strdup(user);
-	connPtr->closing = 0;
-	connPtr->next = connListPtr;
-	connListPtr = connPtr;
+    if (RAT_IMAP == type && NIL != stream) {
+	Tcl_GetBoolean(stdInterp,
+		Tcl_GetVar2(stdInterp, "option", "cache_conn",
+		TCL_GLOBAL_ONLY), &doCache);
+	if (doCache) {
+	    connPtr = (Connection*)malloc(sizeof(Connection));
+	    connPtr->stream = stream;
+	    connPtr->type = type;
+	    connPtr->host = cpystr(host);
+	    connPtr->user = cpystr(user);
+	    connPtr->closing = 0;
+	    connPtr->next = connListPtr;
+	    connListPtr = connPtr;
+	}
     }
     if (passwdCandidate) {
-	if (passwdCandidate->password) {
+	if (passwdCandidate->password && NIL != stream) {
+	    int timeout;
+
 	    passwdCandidate->next = passwdListPtr;
 	    passwdListPtr = passwdCandidate;
+
+	    Tcl_GetInt(interp,
+		       Tcl_GetVar2(interp, "option", "cache_passwd_timeout",
+		       TCL_GLOBAL_ONLY), &timeout);
+
+	    if (timeout) {
+		passwdCandidate->token = Tcl_CreateTimerHandler(timeout*1000,
+			ErasePasswd, (ClientData)passwdCandidate);
+	    } else {
+		passwdCandidate->token = NULL;
+	    }
 	} else {
 	    free(passwdCandidate);
 	}
@@ -412,7 +437,7 @@ OpenStdFolder(Tcl_Interp *interp, char *name, char *prot, char *user,
 	stdPtr->origName = cpystr(name);
 	stdPtr->type = type;
 	stdPtr->host = host;
-	stdPtr->user = loginUser ? strdup(loginUser) : NULL;
+	stdPtr->user = loginUser ? cpystr(loginUser) : NULL;
     } else if (host && *host) {
 	free(host);
     }
@@ -452,9 +477,6 @@ RatStdFolderCreate(Tcl_Interp *interp, int argc, char *argv[])
      * Now it is time to initialize things
      */
     if (initialize) {
-	if (!currentHost) {
-	    RatInitCurrent(interp);
-	}
 	env_parameters(SET_LOCALHOST, currentHost);
 	initialize = 0;
     }
@@ -493,7 +515,6 @@ RatStdFolderCreate(Tcl_Interp *interp, int argc, char *argv[])
     infoPtr->insertProc = Std_InsertProc;
     infoPtr->setFlagProc = Std_SetFlagProc;
     infoPtr->getFlagProc = Std_GetFlagProc;
-    infoPtr->expungeProc = Std_ExpungeProc;
     infoPtr->infoProc = Std_InfoProc;
     infoPtr->createProc = Std_CreateProc;
     infoPtr->private = (ClientData) stdPtr;
@@ -502,6 +523,49 @@ RatStdFolderCreate(Tcl_Interp *interp, int argc, char *argv[])
     return infoPtr;
 }
 
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * Std_CloseProc --
+ *
+ *      See the documentation for closeProc in ratFolder.h
+ *
+ * Results:
+ *	A standard Tcl result.
+ *
+ * Side effects:
+ *	See the documentation for closeProc in ratFolder.h
+ *
+ *
+ *----------------------------------------------------------------------
+ */
+void
+CloseStdFolder(Tcl_Interp *interp, MAILSTREAM *stream)
+{
+    Connection **connPtrPtr;
+
+    for (connPtrPtr = &connListPtr;
+	    *connPtrPtr && stream != (*connPtrPtr)->stream;
+	    connPtrPtr = &(*connPtrPtr)->next);
+    if (*connPtrPtr) {
+	int timeout;
+
+	Tcl_GetInt(interp, Tcl_GetVar2(interp, "option", "cache_conn_timeout",
+		TCL_GLOBAL_ONLY), &timeout);
+
+	(*connPtrPtr)->closing = 1;
+	if (timeout) {
+	    (*connPtrPtr)->token = Tcl_CreateTimerHandler(timeout*1000,
+		    CloseConnection, (ClientData)*connPtrPtr);
+	} else {
+	    (*connPtrPtr)->token = NULL;
+	}
+    } else {
+	mail_close_full(stream, NIL);
+    }
+}
+
 
 /*
  *----------------------------------------------------------------------
@@ -523,26 +587,19 @@ static int
 Std_CloseProc(RatFolderInfoPtr infoPtr, Tcl_Interp *interp)
 {
     StdFolderInfo *stdPtr = (StdFolderInfo *) infoPtr->private;
-    Connection **connPtrPtr;
+    int expunge;
 
     if (stdPtr->isNotYet) {
 	ckfree(stdPtr);
 	return TCL_OK;
     }
-    mail_expunge(stdPtr->stream);
-    for (connPtrPtr = &connListPtr;
-	    *connPtrPtr && stdPtr->stream != (*connPtrPtr)->stream;
-	    connPtrPtr = &(*connPtrPtr)->next);
-    if (*connPtrPtr) {
-	int keep;
-	Tcl_GetInt(interp, Tcl_GetVar2(interp, "option", "keep_conn",
-		TCL_GLOBAL_ONLY), &keep);
-	(*connPtrPtr)->closing = 1;
-	(*connPtrPtr)->token = Tcl_CreateTimerHandler(keep*1000,
-		CloseConnection, (ClientData)*connPtrPtr);
-    } else {
-	mail_close(stdPtr->stream);
+    Tcl_GetBoolean(interp,
+	   Tcl_GetVar2(interp, "option", "expunge_on_close", TCL_GLOBAL_ONLY),
+	   &expunge);
+    if (expunge) {
+	mail_expunge(stdPtr->stream);
     }
+    CloseStdFolder(interp, stdPtr->stream);
     if (0 == --stdPtr->referenceCount) {
 	Tcl_HashEntry *entryPtr;
 	if ((entryPtr=Tcl_FindHashEntry(&folderTable, (char*)stdPtr->stream))) {
@@ -576,7 +633,7 @@ Std_CloseProc(RatFolderInfoPtr infoPtr, Tcl_Interp *interp)
  *----------------------------------------------------------------------
  */
 static int
-Std_UpdateProc(RatFolderInfoPtr infoPtr, Tcl_Interp *interp)
+Std_UpdateProc(RatFolderInfoPtr infoPtr, Tcl_Interp *interp, int expunge)
 {
     StdFolderInfo *stdPtr = (StdFolderInfo *) infoPtr->private;
 
@@ -605,12 +662,48 @@ Std_UpdateProc(RatFolderInfoPtr infoPtr, Tcl_Interp *interp)
 	Tcl_SetHashValue(entryPtr, (ClientData) stdPtr);
 	stdPtr->isNotYet = 0;
     } else {
-	if (T != mail_ping(stdPtr->stream)) {
-	    RatLog(interp, RAT_FATAL, "Mailbox stream died", 0);
-	    exit(1);
+	if (expunge) {
+	    MESSAGECACHE *cachePtr;
+	    char sequence[16];
+	    int i, offset = 0;
+
+	    if (infoPtr->number) {
+		sprintf(sequence, "1:%d", infoPtr->number);
+		mail_fetchfast_full(stdPtr->stream, sequence, NIL);
+		for (i=0; i<infoPtr->number; i++) {
+		    cachePtr = mail_elt(stdPtr->stream, i+1);
+		    if (cachePtr->deleted) {
+			if (-1 != infoPtr->size) {
+			    infoPtr->size -= cachePtr->rfc822_size;
+			}
+			if (infoPtr->msgCmdPtr[i]) {
+			    RatMessageDelete(interp, infoPtr->msgCmdPtr[i]);
+			}
+			offset++;
+		    } else if (offset) {
+			infoPtr->msgCmdPtr[i-offset] = infoPtr->msgCmdPtr[i];
+			infoPtr->privatePtr[i-offset] = infoPtr->privatePtr[i];
+			if (infoPtr->msgCmdPtr[i]) {
+			    RatMessageSetIndex(interp, infoPtr->msgCmdPtr[i],
+					       i - offset);
+			}
+		    }
+		}
+		for (i=infoPtr->number-offset; i<infoPtr->number; i++) {
+		    infoPtr->msgCmdPtr[i] = NULL;
+		    infoPtr->privatePtr[i] = NULL;
+		}
+	    }
+	    mail_expunge(stdPtr->stream);
+	    infoPtr->number = stdPtr->exists;
+	} else {
+	    if (T != mail_ping(stdPtr->stream)) {
+		RatLog(interp, RAT_FATAL, "Mailbox stream died", 0);
+		exit(1);
+	    }
 	}
     }
-    return stdPtr->exists - infoPtr->number;
+    return stdPtr->exists;
 }
 
 
@@ -635,6 +728,7 @@ Std_InsertProc(RatFolderInfoPtr infoPtr, Tcl_Interp *interp, int argc,
 	char *argv[])
 {
     StdFolderInfo *stdPtr = (StdFolderInfo *) infoPtr->private;
+    char flags[128], date[128];
     Tcl_CmdInfo cmdInfo;
     Tcl_DString ds;
     STRING string;
@@ -667,9 +761,10 @@ Std_InsertProc(RatFolderInfoPtr infoPtr, Tcl_Interp *interp, int argc,
     Tcl_DStringInit(&ds);
     for (i=0; i<argc; i++) {
 	Tcl_GetCommandInfo(interp, argv[i], &cmdInfo);
-	RatMessageGet(interp, (MessageInfo*)cmdInfo.clientData, &ds);
+	RatMessageGet(interp, (MessageInfo*)cmdInfo.clientData, &ds,flags,date);
 	INIT(&string,mail_string,Tcl_DStringValue(&ds),Tcl_DStringLength(&ds));
-	if (!mail_append(stdPtr->stream, stdPtr->origName, &string)){
+	if (!mail_append_full(stdPtr->stream, stdPtr->origName, flags, date,
+			      &string)){
 	    fprintf(stderr, "mail_append failed\n");
 	}
 	Tcl_DStringSetLength(&ds, 0);
@@ -705,11 +800,15 @@ Std_SetFlagProc(RatFolderInfoPtr infoPtr, Tcl_Interp *interp, int index,
     StdFolderInfo *stdPtr = (StdFolderInfo *) infoPtr->private;
     char sequence[8];
 
+    if (stdPtr->stream->rdonly) {
+	return TCL_OK;
+    }
+
     sprintf(sequence, "%d", index+1);
     if (value) {
-	mail_setflag(stdPtr->stream, sequence, stdFlagNames[flag]);
+	mail_setflag_full(stdPtr->stream, sequence, stdFlagNames[flag], NIL);
     } else {
-	mail_clearflag(stdPtr->stream, sequence, stdFlagNames[flag]);
+	mail_clearflag_full(stdPtr->stream, sequence, stdFlagNames[flag], NIL);
     }
     if (logLevel > RAT_WARN) {
 	Tcl_SetResult(interp, logMessage, TCL_VOLATILE);
@@ -747,7 +846,7 @@ Std_GetFlagProc(RatFolderInfoPtr infoPtr, Tcl_Interp *interp, int index,
 
     sprintf(sequence, "%d", index+1);
     logLevel = RAT_BABBLE;
-    (void)mail_fetchstructure(stdPtr->stream, index+1, NIL);
+    (void)mail_fetchstructure_full(stdPtr->stream, index+1, NIL, NIL);
     cachePtr = mail_elt(stdPtr->stream, index+1);
     switch (flag) {
     case RAT_SEEN:	value = cachePtr->seen; break;
@@ -758,63 +857,6 @@ Std_GetFlagProc(RatFolderInfoPtr infoPtr, Tcl_Interp *interp, int index,
     return value;
 }
 
-
-/*
- *----------------------------------------------------------------------
- *
- * Std_ExpungeProc --
- *
- *      See the documentation for expungeProc in ratFolder.h
- *
- * Results:
- *	A standard Tcl result.
- *
- * Side effects:
- *	See the documentation for expungeProc in ratFolder.h
- *
- *
- *----------------------------------------------------------------------
- */
-static int
-Std_ExpungeProc(RatFolderInfoPtr infoPtr, Tcl_Interp *interp)
-{
-    StdFolderInfo *stdPtr = (StdFolderInfo *) infoPtr->private;
-    MESSAGECACHE *cachePtr;
-    char sequence[16];
-    int i, offset = 0;
-
-    if (stdPtr->isNotYet) {
-	return TCL_OK;
-    }
-
-    /* We have to notify all created messages that their index has changed */
-    if (infoPtr->number) {
-	sprintf(sequence, "1:%d", infoPtr->number);
-	mail_fetchfast(stdPtr->stream, sequence);
-	for (i=0; i<infoPtr->number; i++) {
-	    cachePtr = mail_elt(stdPtr->stream, i+1);
-	    if (cachePtr->deleted) {
-		if (-1 != infoPtr->size) {
-		    infoPtr->size -= cachePtr->rfc822_size;
-		}
-		if (infoPtr->msgCmdPtr[i]) {
-		    Tcl_DeleteCommand(interp, infoPtr->msgCmdPtr[i]);
-		}
-		offset++;
-	    } else if (offset) {
-		infoPtr->msgCmdPtr[i-offset] = infoPtr->msgCmdPtr[i];
-		if (infoPtr->msgCmdPtr[i]) {
-		    RatMessageSetIndex(interp,infoPtr->msgCmdPtr[i],i-offset+1);
-		}
-	    }
-	}
-    }
-    infoPtr->number -= offset;
-    stdPtr->exists -= offset;
-    mail_expunge(stdPtr->stream);
-    return TCL_OK;
-}
-
 
 /*
  *----------------------------------------------------------------------
@@ -839,13 +881,11 @@ Std_InfoProc(Tcl_Interp *interp, ClientData clientData, RatFolderInfoType type,
 {
     RatFolderInfo *infoPtr = (RatFolderInfo*)clientData;
     int i = infoPtr->presentationOrder[index];
-    Tcl_CmdInfo info;
 
-    if (NULL == infoPtr->msgCmdPtr[i]) {
+    if (NULL == infoPtr->privatePtr[i]) {
 	infoPtr->msgCmdPtr[i] = Std_CreateProc(infoPtr, interp, i);
     }
-    Tcl_GetCommandInfo(interp, infoPtr->msgCmdPtr[i], &info);
-    return Std_GetInfoProc(interp, (MessageInfo*)info.clientData, type, 0);
+    return Std_GetInfoProc(interp, (ClientData)infoPtr->privatePtr[i], type, 0);
 }
 
 /*
@@ -869,7 +909,7 @@ Std_CreateProc(RatFolderInfoPtr infoPtr, Tcl_Interp *interp, int index)
 {
     StdFolderInfo *stdPtr = (StdFolderInfo *) infoPtr->private;
 
-    return RatStdMessageCreate(interp, infoPtr, stdPtr->stream, index+1,
+    return RatStdMessageCreate(interp, infoPtr, stdPtr->stream, index,
 	    stdPtr->type, stdPtr->host, stdPtr->user);
 }
 
@@ -933,7 +973,7 @@ StdListIMAPFolders(ClientData dummy, Tcl_Interp *interp, int argc, char *argv[])
     mail_list(stream, NIL, spec);
     free(spec);
     if (doClose) {
-	mail_close(stream);
+	mail_close_full(stream, NIL);
     }
 
     /*
@@ -973,9 +1013,18 @@ BuildIMAPList(Mailbox *mPtr, Tcl_DString *dsPtr)
     for (; mPtr; mPtr = nPtr) {
 	Tcl_DStringStartSublist(dsPtr);
 	Tcl_DStringAppendElement(dsPtr, mPtr->name);
-	if (LATT_NOSELECT & mPtr->attributes) {
+	if (mPtr->children) {
 	    Tcl_DStringAppendElement(dsPtr, "0");
 	    Tcl_DStringStartSublist(dsPtr);
+	    if (!(LATT_NOSELECT & mPtr->attributes)) {
+		Tcl_DStringStartSublist(dsPtr);
+		Tcl_DStringAppendElement(dsPtr, mPtr->name);
+		Tcl_DStringAppendElement(dsPtr, "1");
+		Tcl_DStringStartSublist(dsPtr);
+		Tcl_DStringAppendElement(dsPtr, mPtr->spec);
+		Tcl_DStringEndSublist(dsPtr);
+		Tcl_DStringEndSublist(dsPtr);
+	    }
 	    BuildIMAPList(mPtr->children, dsPtr);
 	    Tcl_DStringEndSublist(dsPtr);
 	} else {
@@ -994,6 +1043,37 @@ BuildIMAPList(Mailbox *mPtr, Tcl_DString *dsPtr)
     }
 }
 
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * ErasePasswd --
+ *
+ *      Removes a passwd from the cache
+ *
+ * Results:
+ *	None.
+ *
+ * Side effects:
+ *	The connection list is modified.
+ *
+ *
+ *----------------------------------------------------------------------
+ */
+static void
+ErasePasswd(ClientData clientData)
+{
+    MMPasswd **pwdPtrPtr, *pwdPtr = (MMPasswd*)clientData;
+
+    for (pwdPtrPtr = &passwdListPtr; *pwdPtrPtr != pwdPtr;
+	    pwdPtrPtr = &(*pwdPtrPtr)->next);
+    *pwdPtrPtr = pwdPtr->next;
+    free(pwdPtr->user);
+    memset(pwdPtr->password, '\0', strlen(pwdPtr->password));
+    free(pwdPtr->password);
+    free(pwdPtr);
+}
+
 
 /*
  *----------------------------------------------------------------------
@@ -1016,7 +1096,7 @@ CloseConnection(ClientData clientData)
 {
     Connection **connPtrPtr, *connPtr = (Connection*)clientData;
 
-    mail_close(connPtr->stream);
+    mail_close_full(connPtr->stream, NIL);
     for (connPtrPtr = &connListPtr; *connPtrPtr != connPtr;
 	    connPtrPtr = &(*connPtrPtr)->next);
     *connPtrPtr = connPtr->next;
@@ -1047,7 +1127,7 @@ StdJudgeFolder(ClientData op, Tcl_Interp *interp, int argc, char *argv[])
 {
     MAILSTREAM *stream = NULL;
     Connection *connPtr;
-    int doCache, addToList = 0;
+    int doCachePw, doCacheConn;
     char *host, *cPtr;
 
     if (3 != argc) {
@@ -1064,7 +1144,7 @@ StdJudgeFolder(ClientData op, Tcl_Interp *interp, int argc, char *argv[])
     /*
      * Look for reusable stream
      */
-    for (cPtr = host = strdup(argv[1]);
+    for (cPtr = host = cpystr(argv[1]);
 	    *cPtr && '/' != *cPtr && ':' != *cPtr && '}' != *cPtr;
 	    cPtr++);
     *cPtr++ = '\0';
@@ -1076,14 +1156,15 @@ StdJudgeFolder(ClientData op, Tcl_Interp *interp, int argc, char *argv[])
     }
     if (connPtr) {
 	stream = connPtr->stream;
-	Tcl_DeleteTimerHandler(connPtr->token);
+	if (connPtr->token) {
+	    Tcl_DeleteTimerHandler(connPtr->token);
+	}
 	connPtr->closing = 0;
     } else {
-	addToList = 1;
 	Tcl_GetBoolean(stdInterp,
 		Tcl_GetVar2(stdInterp, "option", "cache_passwd",
-		TCL_GLOBAL_ONLY), &doCache);
-	if (doCache) {
+		TCL_GLOBAL_ONLY), &doCachePw);
+	if (doCachePw) {
 	    passwdCandidate = (MMPasswd*)malloc(sizeof(MMPasswd)+
 		    strlen(host)+1);
 	    passwdCandidate->host = (char*)passwdCandidate+sizeof(MMPasswd);
@@ -1118,36 +1199,55 @@ StdJudgeFolder(ClientData op, Tcl_Interp *interp, int argc, char *argv[])
 	(void)mail_delete(stream, argv[1]);
     }
 
-    /*
-     * Cache this stream if needed, otherwise we close it
-     */
-    if (addToList) {
-	int keep;
+    Tcl_GetBoolean(stdInterp,
+	    Tcl_GetVar2(stdInterp, "option", "cache_conn",
+	    TCL_GLOBAL_ONLY), &doCacheConn);
+    if (doCacheConn) {
+	int timeout;
+
+	Tcl_GetInt(interp, Tcl_GetVar2(interp, "option", "cache_conn_timeout",
+		TCL_GLOBAL_ONLY), &timeout);
 
 	connPtr = (Connection*)malloc(sizeof(Connection));
 	connPtr->stream = stream;
 	connPtr->type = RAT_IMAP;
-	connPtr->host = strdup(host);
-	connPtr->user = strdup(argv[2]);
+	connPtr->host = cpystr(host);
+	connPtr->user = cpystr(loginUser);
 	connPtr->closing = 1;
 	connPtr->next = connListPtr;
-	connListPtr = connPtr;
-	if (passwdCandidate) {
-	    if (passwdCandidate->password) {
-		passwdCandidate->next = passwdListPtr;
-		passwdListPtr = passwdCandidate;
-	    } else {
-		free(passwdCandidate);
-	    }
-	    passwdCandidate = NULL;
+	if (timeout) {
+	    connPtr->token = Tcl_CreateTimerHandler(timeout*1000,
+		    CloseConnection, (ClientData)connPtr);
+	} else {
+	    connPtr->token = NULL;
 	}
-	Tcl_GetInt(interp, Tcl_GetVar2(interp, "option", "keep_conn",
-		TCL_GLOBAL_ONLY), &keep);
-	connPtr->token = Tcl_CreateTimerHandler(keep*1000,
-		CloseConnection, (ClientData)connPtr);
+	connListPtr = connPtr;
     } else {
-	mail_close(stream);
+	mail_close_full(stream, NIL);
     }
+    if (passwdCandidate) {
+	if (passwdCandidate->password) {
+	    int timeout;
+
+	    passwdCandidate->next = passwdListPtr;
+	    passwdListPtr = passwdCandidate;
+
+	    Tcl_GetInt(interp,
+		       Tcl_GetVar2(interp, "option", "cache_passwd_timeout",
+		       TCL_GLOBAL_ONLY), &timeout);
+
+	    if (timeout) {
+		passwdCandidate->token = Tcl_CreateTimerHandler(timeout*1000,
+			ErasePasswd, (ClientData)passwdCandidate);
+	    } else {
+		passwdCandidate->token = NULL;
+	    }
+	} else {
+	    free(passwdCandidate);
+	}
+	passwdCandidate = NULL;
+    }
+
     free(host);
     return TCL_OK;
 }
@@ -1169,15 +1269,19 @@ StdJudgeFolder(ClientData op, Tcl_Interp *interp, int argc, char *argv[])
  *
  *----------------------------------------------------------------------
  */
+
 void
-AppendToIMAP(Tcl_Interp *interp, char *mailboxSpec, char *user, char *msg,
-	     int length)
+AppendToIMAP(Tcl_Interp *interp, char *mailboxSpec, char *user, char *flags,
+	     char *date, char *msg, int length)
 {
-    char buf[1024], *mailbox;
+    char buf[1024], *mailbox, *crlfMsg;
     Connection *connPtr;
     MAILSTREAM *stream;
     int i, doClose;
     STRING msgString;
+    unsigned long crlfLength, buflen = length*1.05;
+
+    crlfMsg = (char*)malloc(buflen);
 
     mailbox = RatLindex(interp, mailboxSpec, 0);
     for (i=0; mailbox[i+1] != '}'; i++) {
@@ -1201,8 +1305,10 @@ AppendToIMAP(Tcl_Interp *interp, char *mailboxSpec, char *user, char *msg,
 	doClose = 1;
     }
 
-    INIT(&msgString, mail_string, msg, length);
-    mail_append(stream, mailbox, &msgString);
+    crlfLength = strcrlfcpy(&crlfMsg, &buflen, msg, length);
+    INIT(&msgString, mail_string, crlfMsg, crlfLength);
+    mail_append_full(stream, mailbox, flags, date, &msgString);
+    free(crlfMsg);
 
     if (doClose) {
 	Connection **connPtrPtr;
@@ -1211,14 +1317,49 @@ AppendToIMAP(Tcl_Interp *interp, char *mailboxSpec, char *user, char *msg,
 		*connPtrPtr && stream != (*connPtrPtr)->stream;
 		connPtrPtr = &(*connPtrPtr)->next);
 	if (*connPtrPtr) {
-	    int keep;
-	    Tcl_GetInt(interp, Tcl_GetVar2(interp, "option", "keep_conn",
-		    TCL_GLOBAL_ONLY), &keep);
+	    int timeout;
+	    Tcl_GetInt(interp, Tcl_GetVar2(interp, "option",
+		    "cache_conn_timeout", TCL_GLOBAL_ONLY), &timeout);
 	    (*connPtrPtr)->closing = 1;
-	    (*connPtrPtr)->token = Tcl_CreateTimerHandler(keep*1000,
+	    (*connPtrPtr)->token = Tcl_CreateTimerHandler(timeout*1000,
 		    CloseConnection, (ClientData)*connPtrPtr);
 	} else {
-	    mail_close(stream);
+	    mail_close_full(stream, NIL);
+	}
+    }
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * ClearStdPasswds --
+ *
+ *      Clear the standard passwords from memory. If freethem is non
+ *	null we also try to deallocate the memory.
+ *
+ * Results:
+ *	None.
+ *
+ * Side effects:
+ *	The specified folder will be modified
+ *
+ *
+ *----------------------------------------------------------------------
+ */
+
+void
+ClearStdPasswds(int freethem)
+{
+    MMPasswd *pwPtr, *nextPtr;
+
+    for (pwPtr = passwdListPtr; pwPtr; pwPtr = nextPtr) {
+	memset(pwPtr->password, '\0', strlen(pwPtr->password));
+	nextPtr = pwPtr->next;
+	if (freethem) {
+	    free(pwPtr->user);
+	    free(pwPtr->password);
+	    free(pwPtr);
 	}
     }
 }
@@ -1245,7 +1386,7 @@ void mm_exists (MAILSTREAM *stream,unsigned long number)
     StdFolderInfo *stdPtr;
 
     if ((entryPtr = Tcl_FindHashEntry(&folderTable, (char*)stream))) {
-	stdPtr = (StdFolderInfo *) Tcl_GetHashValue(entryPtr);
+	stdPtr = (StdFolderInfo *)Tcl_GetHashValue(entryPtr);
 	stdPtr->exists = (int)number;
     }
 }
@@ -1253,6 +1394,13 @@ void mm_exists (MAILSTREAM *stream,unsigned long number)
 
 void mm_expunged (MAILSTREAM *stream,unsigned long number)
 {
+    Tcl_HashEntry *entryPtr;
+    StdFolderInfo *stdPtr;
+
+    if ((entryPtr = Tcl_FindHashEntry(&folderTable, (char*)stream))) {
+	stdPtr = (StdFolderInfo *)Tcl_GetHashValue(entryPtr);
+	stdPtr->exists -= 1;
+    }
 }
 
 
@@ -1291,7 +1439,7 @@ void mm_log (char *string,long errflg)
     if (logMessage) {
 	free(logMessage);
     }
-    logMessage = strdup(string);
+    logMessage = cpystr(string);
     RatLog(stdInterp, logLevel, string, 0);
     if (logLevel > RAT_ERROR) {
 	fprintf(stderr, "c-client error. TkRat is aborting...\n");
@@ -1309,18 +1457,33 @@ void mm_dlog (char *string)
 
 void mm_login (NETMBX *mbPtr, char *user, char *pwd, long trial)
 {
-    char **largv, buf[16];
+    char **largv, buf[1024];
     MMPasswd *pwPtr;
-    int largc;
+    int largc, timeout;
 
     /*
      * Check for cached entry
      */
+    sprintf(buf, "%s:%ld", mbPtr->host, mbPtr->port);
     for (pwPtr = passwdListPtr; pwPtr; pwPtr = pwPtr->next) {
-	if (!strcmp(pwPtr->host, mbPtr->host) && !strcmp(pwPtr->user, loginUser)
+	if ((!strcmp(pwPtr->host, mbPtr->host) || !strcmp(pwPtr->host, buf))
+		&& !strcmp(pwPtr->user, loginUser)
 		&& !strcasecmp(ratStdTypeNames[pwPtr->type], mbPtr->service)) {
 	    strcpy(user, pwPtr->user);
 	    strcpy(pwd, pwPtr->password);
+	    if (pwPtr->token) {
+		Tcl_DeleteTimerHandler(pwPtr->token);
+	    }
+	    Tcl_GetInt(stdInterp,
+		       Tcl_GetVar2(stdInterp, "option", "cache_passwd_timeout",
+		       TCL_GLOBAL_ONLY), &timeout);
+
+	    if (timeout) {
+		pwPtr->token = Tcl_CreateTimerHandler(timeout*1000,
+			ErasePasswd, (ClientData)pwPtr);
+	    } else {
+		pwPtr->token = NULL;
+	    }
 	    return;
 	}
     }
@@ -1334,11 +1497,11 @@ void mm_login (NETMBX *mbPtr, char *user, char *pwd, long trial)
 	return;
     }
     strcpy(user, largv[0]);
-    loginUser = strdup(largv[0]);
+    loginUser = cpystr(largv[0]);
     strcpy(pwd, largv[1]);
 
     if (passwdCandidate) {
-	passwdCandidate->password = strdup(largv[1]);
+	passwdCandidate->password = cpystr(largv[1]);
 	passwdCandidate->user = loginUser;
     }
     ckfree(largv);
@@ -1408,12 +1571,17 @@ mm_list(MAILSTREAM *stream, int delimiter, char *name,long attributes)
 	}
 	if (!found) {
 	    *mPtrPtr = (Mailbox*)malloc(sizeof(Mailbox));
-	    (*mPtrPtr)->name = strdup(elem);
-	    (*mPtrPtr)->attributes = attributes;
-	    if (attributes & LATT_NOSELECT) {
+	    (*mPtrPtr)->name = cpystr(elem);
+	    if (cPtr) {
+		(*mPtrPtr)->attributes = LATT_NOSELECT;
 		(*mPtrPtr)->spec = NULL;
 	    } else {
-		(*mPtrPtr)->spec = strdup(name);
+		(*mPtrPtr)->attributes = attributes;
+		if (attributes & LATT_NOSELECT) {
+		    (*mPtrPtr)->spec = NULL;
+		} else {
+		    (*mPtrPtr)->spec = cpystr(name);
+		}
 	    }
 	    (*mPtrPtr)->children = NULL;
 	    (*mPtrPtr)->next = NULL;
@@ -1432,6 +1600,7 @@ mm_list(MAILSTREAM *stream, int delimiter, char *name,long attributes)
 void
 mm_lsub (MAILSTREAM *stream, int delimiter, char *name, long attributes)
 {
+    mm_list(stream, delimiter, name, attributes);
 }
 
 
