@@ -3,7 +3,8 @@
  *
  *      Declarations of functions used in the Std folder and messages
  *
- * TkRat software and its included text is Copyright 1996 by Martin Forssen.
+ * TkRat software and its included text is Copyright 1996,1997,1998
+ * by Martin Forssn
  *
  * The full text of the legal notice is contained in the file called
  * COPYRIGHT, included with this distribution.
@@ -24,7 +25,6 @@ RatGetEnvelopeProc Std_GetEnvelopeProc;
 RatCreateBodyProc Std_CreateBodyProc;
 RatFetchTextProc Std_FetchTextProc;
 RatEnvelopeProc Std_EnvelopeProc;
-RatFetchFirstTextProc Std_FetchFirstTextProc;
 RatSetIndexProc Std_SetIndexProc;
 RatMsgDeleteProc Std_MsgDeleteProc;
 RatMakeChildrenProc Std_MakeChildrenProc;
