@@ -3,7 +3,8 @@
 # This file contains the code which handles the composing of messages
 #
 #
-#  TkRat software and its included text is Copyright 1996 by Martin Forssen.
+#  TkRat software and its included text is Copyright 1996,1997,1998
+#  by Martin Forssn
 #
 #  The full text of the legal notice is contained in the file called
 #  COPYRIGHT, included with this distribution.
@@ -13,8 +14,8 @@
 set composeWindowList {}
 
 # List of header names we know of
-set composeHeaderList {To Subject Cc Bcc Keywords Content_Description
-		       Reply_To Comments From}
+set composeHeaderList {to subject cc bcc keywords content_description
+		       reply_to comments from}
 
 # List of headers which contains addresses (OBS! must be lower case OBS!)
 set composeAdrHdrList {to cc bcc reply_to}
@@ -64,7 +65,7 @@ proc ComposeForwardInline {msg} {
     set handler composeM[incr idCnt]
     upvar #0 $handler mh
 
-    set msg [ComposeChoose $msg $t(reply_to_which)]
+    set msg [ComposeChoose $msg $t(forward_which)]
     if ![string length $msg] {
 	return 0
     }
@@ -124,10 +125,15 @@ proc ComposeForwardInline {msg} {
 		set length [string length $f]
 	    }
 	}
-	foreach f [string tolower $option(show_header_selection)] {
-	    regsub -all -- - $f _ f
+	foreach field $option(show_header_selection) {
+	    regsub -all -- - [string tolower $field] _ f
 	    if [info exists inline_header($f)] {
-		set preface [format "%s%${length}s: %s\n" $preface $t($f) \
+		if [info exists t($f)] {
+		    set name $t($f)
+		} else {
+		    set name $field
+		}
+		set preface [format "%s%${length}s: %s\n" $preface $name \
 			$inline_header($f)]
 	    }
 	}
@@ -190,13 +196,13 @@ proc ComposeForwardAttachment {msg} {
     set fh [open $hd(filename) w 0600]
     set hd(description) $t(forwarded_message)
     foreach header [$msg headers full] {
-	puts $fh "[lindex $header 0]: [lindex $header 1]"
+	puts $fh "[lindex $header 0]: [lindex $header 1]\r"
 	switch [string tolower [lindex $header 0]] {
 	subject		{ set mh(subject) "Fwd: [lindex $header 1]" }
 	description	{ set hd(description) [lindex $header 1] }
 	}
     }
-    puts $fh ""
+    puts $fh "\r"
     puts -nonewline $fh [$msg rawBody]
     close $fh
     set hd(type) message
@@ -215,7 +221,7 @@ proc ComposeForwardAttachment {msg} {
 # Arguments:
 
 proc ComposeHeld {} {
-    global idCnt t defaultFont
+    global idCnt t b defaultFont
 
     # First see if there are any held messages
     if ![llength [set content [RatHold list]]] {
@@ -240,6 +246,7 @@ proc ComposeHeld {} {
     listbox $w.list.list -yscroll "$w.list.scroll set" -height 10 \
 	    -width 80 -relief sunken -font $defaultFont -selectmode browse \
 	    -exportselection false -highlightthickness 0
+    set b($w.list.list) compose_held_pick
     bind $w.list.list <Double-1> "set ${id}(done) 1"
     pack $w.list.scroll -side right -fill y
     pack $w.list.list -side left -expand 1 -fill both
@@ -266,6 +273,7 @@ proc ComposeHeld {} {
 	    Popup $t(must_select)
 	}
     }
+    unset b($w.list.list)
     unset hd
 }
 
@@ -296,12 +304,16 @@ proc ComposeExtracted {mgh} {
 	    set edit {}
 	    set children $mh(body)
 	}
+	if [info exists bh(pgp_sign)] {
+	    set mh(pgp_sign) $bh(pgp_sign)
+	    set mh(pgp_encrypt) $bh(pgp_encrypt)
+	}
 	if [string length $edit] {
 	    set fh [open [set ${edit}(filename)] r]
 	    set mh(data) [read $fh]
 	    close $fh
 	    if [set ${edit}(removeFile)] {
-		catch "exec rm -f [set ${edit}(filename)]"
+		catch "file delete -force -- [set ${edit}(filename)]"
 	    }
 	}
 	set mh(attachmentList) $children
@@ -320,16 +332,17 @@ proc ComposeExtracted {mgh} {
 # add_sig   -   '1' if we should add the signature
 
 proc DoCompose {handler edit_text add_sig} {
-    global option t composeHeaderList langCharset composeWindowList \
-	   tk_strictMotif
+    global option t b composeHeaderList langCharset composeWindowList \
+	   tk_strictMotif ratCurrentHost env
     upvar #0 $handler mh
 
     # Initialize variables
     foreach i $composeHeaderList {
-	set mh($i) 0
+	set mh(O_$i) 0
     }
-    foreach i $option(compose_headers) {
-	set mh($i) 1
+    regsub -all -- - [string tolower $option(compose_headers)] _ vars
+    foreach i $vars {
+	set mh(O_$i) 1
     }
     set mh(request_dsn) $option(dsn_request)
     if {![info exists mh(from)] || ![string length $mh(from)]} {
@@ -342,8 +355,14 @@ proc DoCompose {handler edit_text add_sig} {
     set mh(save_to) ""
     set mh(doWrap) 1
     set mh(doIndent) 1
-    set mh(copy_attached) 1
     set mh(font) [GetFont $option(charset) 0 {}]
+    set mh(closing) 0
+
+    set mh(copy_attached) $option(copy_attached)
+    if ![info exists mh(pgp_sign)] {
+	set mh(pgp_sign) $option(pgp_sign)
+	set mh(pgp_encrypt) $option(pgp_encrypt)
+    }
 
     # Create window
     set w .$handler
@@ -353,72 +372,103 @@ proc DoCompose {handler edit_text add_sig} {
     wm iconname $w $t(compose_name)
 
     # Menus
+    FindAccelerators a {file edit headers extra admin}
+
     frame $w.menu -relief raised -bd 1
-    menubutton $w.menu.file -menu $w.menu.file.m -text $t(file)
-    menu $w.menu.file.m
-    $w.menu.file.m add command -label $t(insert_file)... \
+    set m $w.menu.file.m
+    menubutton $w.menu.file -menu $m -text $t(file) -underline $a(file)
+    menu $m
+    $m add command -label $t(insert_file)... \
 	    -command "ComposeInsertFile $handler"
-    $w.menu.file.m add separator
-    $w.menu.file.m add checkbutton -label $t(automatic_wrap) \
-	    -variable ${handler}(doWrap)
-    $w.menu.file.m add checkbutton -label $t(automatic_indent) \
+    set b($m,[$m index end]) compose_insert_file
+    $m add separator
+    $m add checkbutton -label $t(automatic_wrap) -variable ${handler}(doWrap)
+    set b($m,[$m index end]) automatic_wrap
+    $m add checkbutton -label $t(automatic_indent) \
 	    -variable ${handler}(doIndent)
-    $w.menu.file.m add separator
-    $w.menu.file.m add command -label $t(abort) \
+    set b($m,[$m index end]) automatic_indent
+    $m add separator
+    $m add command -label $t(abort) \
 	    -command "ComposeBuildStruct $handler; DoCompose2 $w $handler abort"
-    set mh(abort_menu) $w.menu.file.m
-    set mh(abort_index) [$w.menu.file.m index end]
+    set b($m,[$m index end]) abort_compose
+    set mh(abort_menu) [list $m [$m index end]]
     lappend mh(eEditBlock) $w.menu.file
-    menubutton $w.menu.edit -menu $w.menu.edit.m -text $t(edit)
-    menu $w.menu.edit.m -postcommand "ComposePostEdit $handler $w.menu.edit.m"
-    $w.menu.edit.m add command -label $t(undo) -command "ComposeUndo $handler"
-    set mh(undo_menu) $w.menu.edit.m
-    set mh(undo_index) [$w.menu.edit.m index end]
-    $w.menu.edit.m add separator
-    $w.menu.edit.m add command -label $t(cut) \
-	    -command "ComposeEditOP $handler cut"
-    set mh(cut_menu) $w.menu.edit.m
-    set mh(cut_index) [$w.menu.edit.m index end]
-    $w.menu.edit.m add command -label $t(copy) \
-	    -command "ComposeEditOP $handler copy"
-    $w.menu.edit.m add command -label $t(paste) -command "ComposePaste $handler"
-    set mh(paste_menu) $w.menu.edit.m
-    set mh(paste_index) [$w.menu.edit.m index end]
-    $w.menu.edit.m add command -label $t(cut_all) \
-	    -command "ComposeEditOP $handler cut_all"
-    set mh(cut_all_menu) $w.menu.edit.m
-    set mh(cut_all_index) [$w.menu.edit.m index end]
-    $w.menu.edit.m add command -label $t(copy_all) \
+
+    set m $w.menu.edit.m
+    menubutton $w.menu.edit -menu $m -text $t(edit) -underline $a(edit)
+    menu $m -postcommand "ComposePostEdit $handler $m"
+    $m add command -label $t(undo) -command "ComposeUndo $handler"
+    set b($m,[$m index end]) undo
+    set mh(undo_menu) [list $m [$m index end]]
+    $m add separator
+    $m add command -label $t(cut) -command "ComposeEditOP $handler cut"
+    set b($m,[$m index end]) cut
+    set mh(cut_menu) [list $m [$m index end]]
+    $m add command -label $t(copy) -command "ComposeEditOP $handler copy"
+    set b($m,[$m index end]) copy
+    $m add command -label $t(paste) -command "ComposePaste $handler"
+    set b($m,[$m index end]) paste
+    set mh(paste_menu) [list $m [$m index end]]
+    $m add command -label $t(cut_all) -command "ComposeEditOP $handler cut_all"
+    set b($m,[$m index end]) cut_all
+    set mh(cut_all_menu) [list $m [$m index end]]
+    $m add command -label $t(copy_all) \
 	    -command "ComposeEditOP $handler copy_all"
-    $w.menu.edit.m add separator
-    $w.menu.edit.m add command -label $t(wrap_lines) \
+    set b($m,[$m index end]) copy_all
+    $m add separator
+    $m add command -label $t(wrap_lines) \
 	    -command "ComposeWrapSelection $handler"
-    $w.menu.edit.m add separator
-    $w.menu.edit.m add command -label $t(run_through_command)... \
+    set b($m,[$m index end]) wrap_lines
+    $m add separator
+    $m add command -label $t(run_through_command)... \
 	    -command "ComposeSpecifyCmd $handler"
-    menubutton $w.menu.headers -menu $w.menu.headers.m -text $t(headers)
+    set b($m,[$m index end]) run_through_command
+
+    menubutton $w.menu.headers -menu $w.menu.headers.m -text $t(headers) \
+			       -underline $a(headers)
+    set b($w.menu.headers) headers_menu
     menu $w.menu.headers.m
     foreach header $composeHeaderList {
-	$w.menu.headers.m add checkbutton -label $t([string tolower $header]) \
-		-variable ${handler}($header) \
+	$w.menu.headers.m add checkbutton -label $t($header) \
+		-variable ${handler}(O_$header) \
 		-onvalue 1 -offvalue 0 \
 		-command "ComposeBuildHeaderEntries $handler"
     }
-    menubutton $w.menu.extra -menu $w.menu.extra.m -text $t(extra)
-    menu $w.menu.extra.m
-    $w.menu.extra.m add checkbutton \
+
+    set m $w.menu.extra.m
+    menubutton $w.menu.extra -menu $m -text $t(extra) -underline $a(extra)
+    menu $m
+    $m add checkbutton \
 	    -label $t(copy_attached_files) \
 	    -variable ${handler}(copy_attached) \
 	    -onvalue 1 -offvalue 0
-    $w.menu.extra.m add checkbutton \
+    set b($m,[$m index end]) copy_attached_files
+    $m add checkbutton \
 	    -label $t(request_notification) \
 	    -variable ${handler}(request_dsn) \
 	    -onvalue 1 -offvalue 0
-    menubutton $w.menu.admin -menu $w.menu.admin.m -text $t(admin)
-    menu $w.menu.admin.m
-    $w.menu.admin.m add command -label $t(define_keys)... \
-	    -command {KeyDef compose}
-    $w.menu.admin.m add command -label $t(command_list)... -command CmdList
+    set b($m,[$m index end]) request_notification
+    if $option(pgp_enable) {
+	$m add checkbutton \
+		-label $t(sign) \
+		-variable ${handler}(pgp_sign) \
+		-onvalue 1 -offvalue 0
+	set b($m,[$m index end]) pgp_sign
+	$m add checkbutton \
+		-label $t(encrypt) \
+		-variable ${handler}(pgp_encrypt) \
+		-onvalue 1 -offvalue 0
+	set b($m,[$m index end]) pgp_encrypt
+    }
+    
+    set m $w.menu.admin.m
+    menubutton $w.menu.admin -menu $m -text $t(admin) \
+			     -underline $a(admin)
+    menu $m
+    $m add command -label $t(define_keys)... -command {KeyDef compose}
+    set b($m,[$m index end]) define_keys
+    $m add command -label $t(command_list)... -command CmdList
+    set b($m,[$m index end]) command_list
     pack $w.menu.file \
 	 $w.menu.edit \
 	 $w.menu.headers \
@@ -436,6 +486,7 @@ proc DoCompose {handler edit_text add_sig} {
     text $w.body.text -relief sunken -bd 1 -setgrid true \
 	    -yscrollcommand "$w.body.scroll set" -wrap none \
 	    -font $mh(font)
+    set b($w.body.text) compose_body
     Size $w.body.text compose
     pack $w.body.scroll -side right -fill y
     pack $w.body.text -side left -expand yes -fill both
@@ -458,7 +509,8 @@ proc DoCompose {handler edit_text add_sig} {
 	    } else {
 		$w.body.text insert end "\n" {} "-- \n$text" noWrap
 	    }
-	} elseif [file readable $option(signature)] {
+	} elseif {![file isdirectory $option(signature)]
+		&& [file readable $option(signature)]} {
 	    set fh [open $option(signature) r]
 	    $w.body.text insert end "\n" {} "-- \n[read -nonewline $fh]" noWrap
 	    close $fh
@@ -510,11 +562,25 @@ proc DoCompose {handler edit_text add_sig} {
     frame $w.attach
     frame $w.attach.b
     label $w.attach.b.label -text $t(attachments)
-    button $w.attach.b.attach -text $t(attach) -command "Attach $handler"
+    button $w.attach.b.attachf -text $t(attach_file) -command "Attach $handler"
+    set b($w.attach.b.attachf) attach_file
+    menubutton $w.attach.b.attachs -text $t(attach_special) -indicatoron 1 \
+	    -menu $w.attach.b.attachs.m -relief raised
+    set b($w.attach.b.attachf) attach_special
+    if $option(pgp_enable) {
+	menu $w.attach.b.attachs.m
+	$w.attach.b.attachs.m add command -label $t(attach_pgp_keys)... \
+		-command "AttachKeys $handler"
+	set b($w.attach.b.attachf) attach_keys
+    } else {
+	$w.attach.b.attachs configure -state disabled
+    }
     button $w.attach.b.detach -text $t(detach) -state disabled \
 	    -command "Detach $handler $w.attach.b.detach"
+    set b($w.attach.b.detach) detach
     pack $w.attach.b.label -side left -padx 10
-    pack $w.attach.b.attach \
+    pack $w.attach.b.attachf \
+	 $w.attach.b.attachs \
 	 $w.attach.b.detach -side left -padx 5
     frame $w.attach.list
     scrollbar $w.attach.list.scroll -relief sunken -takefocus 0 \
@@ -522,6 +588,7 @@ proc DoCompose {handler edit_text add_sig} {
     listbox $w.attach.list.list -yscroll "$w.attach.list.scroll set" \
 	    -height 3 -relief sunken -bd 1 -font $mh(font) \
 	    -exportselection false -highlightthickness 0 -selectmode extended
+    set b($w.attach.list.list) attachments
     bind $w.attach.list.list <ButtonRelease-1> \
 	    "if { 1 == \[llength \[%W curselection\]\]} {\
 	         $w.attach.b.detach configure -state normal}"
@@ -533,9 +600,9 @@ proc DoCompose {handler edit_text add_sig} {
     if ![info exists mh(attachmentList)] {
 	set mh(attachmentList) {}
     } else {
-	foreach a $mh(attachmentList) {
-	    upvar #0 $a bp
-	    if { "" != $bp(description) } {
+	foreach attachment $mh(attachmentList) {
+	    upvar #0 $attachment bp
+	    if { [info exists bp(description)] && "" != $bp(description) } {
 		set desc $bp(description)
 	    } else {
 		set desc "$bp(type)/$bp(subtype) : $bp(filename)"
@@ -549,19 +616,24 @@ proc DoCompose {handler edit_text add_sig} {
     frame $w.buttons
     button $w.buttons.send -text $t(send) \
 	    -command "DoCompose2 $w $handler send"
+    set b($w.buttons.send) send
     lappend mh(eEditBlock) $w.buttons.send
     menubutton $w.buttons.sendsave -text $t(send_save) -indicatoron 1 \
 	    -menu $w.buttons.sendsave.m -relief raised -underline 0
+    set b($w.buttons.sendsave) sendsave
     menu $w.buttons.sendsave.m -tearoff 0 \
 	    -postcommand "RatSendSavePostMenu $w $w.buttons.sendsave.m $handler"
     lappend mh(eEditBlock) $w.buttons.sendsave
     button $w.buttons.hold -text $t(hold)... -command "ComposeHold $w $handler"
+    set b($w.buttons.hold) hold
     lappend mh(eEditBlock) $w.buttons.hold
     button $w.buttons.edit -text $t(external_editor) \
 	    -command "ComposeEEdit $handler"
+    set b($w.buttons.edit) eedit
     lappend mh(eEditBlock) $w.buttons.abort
     button $w.buttons.abort -text $t(abort) \
 	    -command "ComposeBuildStruct $handler; DoCompose2 $w $handler abort"
+    set b($w.buttons.abort) abort_compose
     lappend mh(eEditBlock) $w.buttons.edit
     pack $w.buttons.send \
 	 $w.buttons.sendsave \
@@ -579,22 +651,29 @@ proc DoCompose {handler edit_text add_sig} {
 
     if { 1 == $edit_text } {
 	focus $w.body.text
-    } else {
+    } elseif [string length $first] {
 	focus $first
     }
     Place $w compose
-    ComposeBind $handler
     lappend composeWindowList $handler
+    ComposeBind $handler
+    bind $mh(composeBody) <Destroy> "DoCompose2 $w $handler abort"
+
+    if {1 >= [llength [split $ratCurrentHost .]] && 0 == $option(force_send)} {
+	$w.buttons.send configure -state disabled
+	$w.buttons.sendsave configure -state disabled
+	Popup $t(no_send_bad_host)
+    }
     return $handler
 }
 proc DoCompose2 {w handler do} {
-    global composeWindowList
+    global composeWindowList option ratCurrentHost b
     upvar #0 $handler mh
 
-    # By moving the focus to the text windge we force all HeaderEntries
-    # to update their variables
-    focus $mh(composeBody)
-    update
+    # Are we already doing this?
+    if {0 == [info exists mh(closing)] || 1 == $mh(closing)} {
+	return
+    }
 
     # Check if A save folder is defined
     if ![string length $mh(save_to)] {
@@ -606,6 +685,18 @@ proc DoCompose2 {w handler do} {
 
     set mh(do) $do
     if { "send" == $do } {
+	set mh(closing) 1
+	# By moving the focus to the text windge we force all HeaderEntries
+	# to update their variables
+	focus $mh(composeBody)
+	update
+
+	# Check for good hostname
+	if {1 >= [llength [split $ratCurrentHost .]]
+		&& 0 == $option(force_send)} {
+	    bell $mh(toplevel)
+	    return
+	}
 	wm withdraw $w
 	if ![ComposeSend $handler] {
 	    set sent 1
@@ -622,15 +713,19 @@ proc DoCompose2 {w handler do} {
 	set doRemove 0
     }
 
-    RecordPos $w compose
-    RecordSize $w.body.text compose
     if [info exists mh(body)] {
 	ComposeFreeBody $mh(body) $doRemove
     }
     set index [lsearch $composeWindowList $handler]
     set composeWindowList [lreplace $composeWindowList $index $index]
+    if [winfo exists $w] {
+	RecordPos $w compose
+	RecordSize $mh(composeBody) compose
+	bind $mh(composeBody) <Destroy> { }
+	foreach bn [array names b $w.*] {unset b($bn)}
+	destroy $w
+    }
     unset mh
-    destroy $w
 }
 
 # RatTextInsert --
@@ -665,10 +760,12 @@ proc RatTextInsert {w h s t} {
     }
     $w insert insert $s $t
     if {$doIndent && ![string compare "\n" $s]} {
-	set lp [expr [lindex [split [$w index insert] .] 0] -1]
+	set lp [expr int([$w index insert])-1]
 	set wl [$w search -regexp {[^ 	]} $lp.0 $lp.end]
-	if {$doIndent && [string length $wl] && [string compare $lp.0 $wl]} {
+	if [string length $wl] {
 	    $w insert insert [$w get $lp.0 $wl]
+	} else {
+	    $w delete $lp.0 $lp.end
 	}
     }
     ComposeWrap $h [expr int([$w index insert])]
@@ -685,7 +782,8 @@ proc RatTextInsert {w h s t} {
 # h -	The handler which describes this text widget
 
 proc RatTextDelete {w h} {
-    if {[$w tag nextrange sel 1.0 end] != ""} {
+    if {[$w tag nextrange sel 1.0 end] != ""
+	    && -1 != [lsearch -exact [$w tag names insert-1c] sel]} {
 	upvar #0 $h hd
 	$w mark set undoStart sel.first
 	$w mark set undoEnd sel.last
@@ -693,13 +791,16 @@ proc RatTextDelete {w h} {
 	set hd(undoTags) [$w tag ranges noWrap]
 	set hd(undoInsert) [$w index insert]
 	$w delete sel.first sel.last
+	ComposeWrap $h [expr int([$w index undoStart])]
     } else {
 	$w delete insert
 	$w see insert
+	ComposeWrap $h [expr int([$w index insert])]
     }
 }
 proc RatTextBackSpace {w h} {
-    if {[$w tag nextrange sel 1.0 end] != ""} {
+    if {[$w tag nextrange sel 1.0 end] != ""
+	    && -1 != [lsearch -exact [$w tag names insert-1c] sel]} {
 	upvar #0 $h hd
 	$w mark set undoStart sel.first
 	$w mark set undoEnd sel.last
@@ -707,9 +808,11 @@ proc RatTextBackSpace {w h} {
 	set hd(undoTags) [$w tag ranges noWrap]
 	set hd(undoInsert) [$w index insert]
 	$w delete sel.first sel.last
+	ComposeWrap $h [expr int([$w index undoStart])]
     } else {
 	$w delete insert-1c
 	$w see insert
+	ComposeWrap $h [expr int([$w index insert])]
     }
 }
 
@@ -734,6 +837,7 @@ proc RatTextKillLine {w h} {
     set hd(undoTags) [$w tag ranges noWrap]
     set hd(undoInsert) [$w index insert]
     $w delete insert $end
+    ComposeWrap $h [expr int([$w index undoStart])]
 }
 
 # ComposeWrap --
@@ -810,7 +914,8 @@ proc ComposeWrap {handler line} {
 	# While the current line and the next line joined does give a line
 	# shorter than 72 characters we joint the next line.
 	set nc [$w search -regexp {[^  ]} $ln.0 $ln.end]
-	while {[string length $nc] && 72 > $len+[RatLL [$w get $nc $ln.end]]} {
+	while {[string length $nc] && 72 > $len+[RatLL [$w get $nc $ln.end]]
+		&& -1 == [lsearch [$w tag names $ln.0] noWrap]} {
 	    $w delete $ln.0 $nc
 	    $w insert $ln.0 " "
 	    $w delete $ln.0-1c
@@ -850,17 +955,13 @@ proc ComposeBind {handler} {
     foreach w "$mh(toplevel) $mh(composeBody)" {
 	RatBind $w compose_key_send "DoCompose2 $mh(toplevel) $handler send"
 	RatBind $w compose_key_abort "ComposeBuildStruct $handler; \
-		DoCompose2 $mh(toplevel) $handler abort" \
-		$mh(abort_menu) $mh(abort_index)
+		DoCompose2 $mh(toplevel) $handler abort" $mh(abort_menu)
 	RatBind $w compose_key_editor "ComposeEEdit $handler"
-	RatBind $w compose_key_undo "ComposeUndo $handler" \
-		$mh(undo_menu) $mh(undo_index)
-	RatBind $w compose_key_cut "ComposeEditOP $handler cut" \
-		$mh(cut_menu) $mh(cut_index)
+	RatBind $w compose_key_undo "ComposeUndo $handler" $mh(undo_menu)
+	RatBind $w compose_key_cut "ComposeEditOP $handler cut" $mh(cut_menu)
 	RatBind $w compose_key_cut_all "ComposeEditOP $handler cut_all" \
-		$mh(cut_all_menu) $mh(cut_all_index)
-	RatBind $w compose_key_paste "ComposePaste $handler" \
-		$mh(paste_menu) $mh(paste_index)
+		$mh(cut_all_menu)
+	RatBind $w compose_key_paste "ComposePaste $handler" $mh(paste_menu)
     }
 }
 
@@ -902,32 +1003,34 @@ proc RatSendSaveDo {w handler save_to} {
 # handler -	The handler for the active compose session
 
 proc ComposeBuildHeaderEntries {handler} {
-    global composeHeaderList composeAdrHdrList t
+    global composeHeaderList composeAdrHdrList t b
     upvar #0 $handler mh
 
     foreach slave [grid slaves $mh(headerFrame)] {
 	destroy $slave
     }
+    set first {}
     set row 0
     grid columnconfigure $mh(headerFrame) 1 -weight 1
 
     foreach header $composeHeaderList {
-	if {0 == $mh($header)} {
+	if {0 == $mh(O_$header)} {
 	    continue
 	}
-	set l [string tolower $header]
-	label $mh(headerFrame).${l}_label -text $t($l):
-	grid $mh(headerFrame).${l}_label -row $row -column 0 -sticky en
-	set w $mh(headerFrame).${l}_entry
-	if {-1 != [lsearch $composeAdrHdrList $l]} {
-	    ComposeBuildHE $w $handler ${handler}($l)
+	label $mh(headerFrame).${header}_label -text $t($header):
+	grid $mh(headerFrame).${header}_label -row $row -column 0 -sticky en
+	set w $mh(headerFrame).${header}_entry
+	if {-1 != [lsearch $composeAdrHdrList $header]} {
+	    ComposeBuildHE $w $handler ${handler}($header)
 	} else {
-	    entry $w -textvariable ${handler}($l) -font $mh(font)
+	    entry $w -textvariable ${handler}($header) -font $mh(font)
 	}
+	set b($w) compose_$header
+	set b($w.t) compose_$header
 	grid $w -row $row -column 1 -sticky we
 	incr row
-	if ![info exist first] {
-	    set first $mh(headerFrame).${l}_entry
+	if {0 == [string length $first]} {
+	    set first $mh(headerFrame).${header}_entry
 	}
     }
 
@@ -966,6 +1069,8 @@ proc ComposeBuildStruct {handler} {
     }
     set bh(removeFile) 1
     if {0 == [llength $mh(attachmentList)]} {
+	set bh(pgp_sign) $mh(pgp_sign)
+	set bh(pgp_encrypt) $mh(pgp_encrypt)
 	set mh(body) $id
     } else {
 	# Create identifier
@@ -973,6 +1078,8 @@ proc ComposeBuildStruct {handler} {
 	upvar #0 $mid ph
 	set mh(body) $mid
 
+	set ph(pgp_sign) $mh(pgp_sign)
+	set ph(pgp_encrypt) $mh(pgp_encrypt)
 	set ph(type) multipart
 	set ph(subtype) mixed
 	set ph(encoding) 7bit
@@ -1029,8 +1136,8 @@ proc ComposeEEdit {handler} {
     set ehandler compose_E[incr idCnt]
     upvar #0 $ehandler eh
 
-    foreach b $mh(eEditBlock) {
-	$b configure -state disabled
+    foreach block $mh(eEditBlock) {
+	$block configure -state disabled
     }
 
     # Write data, change text visible and edit
@@ -1057,14 +1164,14 @@ proc ComposeEEdit2 {handler fname name1 name2 op} {
 	    $mh(composeBody) insert end "$line\n"
 	}
 	close $fh
-	catch "exec rm -f $fname"
-	foreach b $mh(eEditBlock) {
-	    $b configure -state normal
+	catch "file delete -force -- $fname"
+	foreach block $mh(eEditBlock) {
+	    $block configure -state normal
 	}
     }
 
     # Remove the trace
-    trace vdelete $name1($name2) w "ComposeEEdit2 $handler $fname"
+    trace vdelete ${name1}($name2) w "ComposeEEdit2 $handler $fname"
 }
 
 # Attach --
@@ -1075,7 +1182,7 @@ proc ComposeEEdit2 {handler fname name1 name2 op} {
 # handler -	The handler for the active compose session
 
 proc Attach {handler} {
-    global idCnt t valueFont cpProg option
+    global idCnt t b valueFont option
     upvar #0 $handler mh
 
     # Create identifier
@@ -1088,7 +1195,7 @@ proc Attach {handler} {
     # Create toplevel
     toplevel $w
     wm transient $w $mh(toplevel)
-    wm title $w $t(attach)
+    wm title $w $t(attach_file)
 
     frame $w.chooser
     set fh [FileSelectorCreate $w.chooser ${id}(filename) \
@@ -1119,7 +1226,7 @@ proc Attach {handler} {
 
     toplevel $w
     wm transient $w $mh(toplevel)
-    wm title $w $t(attach)
+    wm title $w $t(attach_file)
 
     # Get default type $hd(filename)
     set type [RatType $hd(filename)]
@@ -1140,6 +1247,7 @@ proc Attach {handler} {
     label $w.type.typelabel -width 14 -anchor e -text $t(type):
     menubutton $w.type.type -menu $w.type.type.m -anchor w -width 23 \
 	    -textvariable ${id}(typestring) -relief raised -indicatoron 1
+    set b($w.type.type) type_menu
     menu $w.type.type.m -tearoff 0
     foreach type { {text {plain enriched}}
 		   {image {jpeg gif}}
@@ -1162,6 +1270,7 @@ proc Attach {handler} {
     label $w.type.r.enclabel -anchor e -text $t(current_encoding):
     menubutton $w.type.r.enc -menu $w.type.r.enc.m -anchor w -width 14 \
 	    -textvariable ${id}(encoding) -relief raised -indicatoron 1
+    set b($w.type.r.enc) encoding_menu
     menu $w.type.r.enc.m -tearoff 0
     # Make sure the user can only select the apropriate entries
     switch $hd(encoding) {
@@ -1202,10 +1311,12 @@ proc Attach {handler} {
     entry $w.desc.entry -width 65 -textvariable ${id}(description)
     pack $w.desc.label \
 	 $w.desc.entry -side left
+    set b($w.desc.entry) attach_description
 
     frame $w.id
     label $w.id.label -width 14 -anchor e -text $t(id):
     entry $w.id.entry -width 65 -textvariable ${id}(id)
+    set b($w.id.entry) attach_id
     pack $w.id.label \
 	 $w.id.entry -side left
     
@@ -1230,7 +1341,7 @@ proc Attach {handler} {
 	set hd(removeFile) 0
 	if $mh(copy_attached) {
 	    set fname [RatTildeSubst $option(send_cache)/[RatGenId]]
-	    if [catch {exec $cpProg $hd(filename) $fname} result] {
+	    if [catch {file copy -- $hd(filename) $fname} result] {
 		Popup "$t(failed_to_make_copy): $result"
 	    } else {
 		set hd(filename) $fname
@@ -1249,6 +1360,7 @@ proc Attach {handler} {
     }
 
     RecordPos $w attach2
+    foreach bn [array names b $w.*] {unset b($bn)}
     destroy $w
     focus $oldFocus
 }
@@ -1315,6 +1427,40 @@ proc SubtypeSpec {variable type topwin} {
     focus $oldFocus
 }
 
+# AttachKeys --
+#
+# Attach keys to the message currently being composed
+#
+# Arguments:
+# handler -	The handler for the active compose session
+
+proc AttachKeys {handler} {
+    RatPGPGetIds AttachKeysDo $handler
+}
+proc AttachKeysDo {handler ids} {
+    global idCnt t option
+    upvar #0 $handler mh
+
+    foreach keyid $ids {
+	# Create identifier
+	set id attach[incr idCnt]
+	upvar #0 $id hd
+
+	set hd(type) application
+	set hd(subtype) pgp-keys
+	set hd(encoding) 7bit
+	set hd(description) "$t(pgp_key_for) $keyid"
+	set hd(filename) [RatTildeSubst $option(send_cache)/[RatGenId]]
+	set hd(removeFile) 1
+
+	set f [open $hd(filename) w]
+	puts $f [RatPGP extract $keyid]
+	close $f
+	lappend mh(attachmentList) $id
+	$mh(attachmentListWindow) insert end $hd(description)
+    }
+}
+
 # Detach --
 #
 # Detach a previously attached attachment to a message
@@ -1354,7 +1500,7 @@ proc ComposeFreeBody {handler doRemove} {
     } 
     if [info exists bh(removeFile)] {
 	if {$doRemove && $bh(removeFile)} {
-	    catch {exec rm $bh(filename)}
+	    catch {file delete -- $bh(filename)}
 	}
     }
     unset bh
@@ -1370,7 +1516,7 @@ proc ComposeFreeBody {handler doRemove} {
 
 proc ComposeHold {mainW handler} {
     upvar #0 $handler mh
-    global idCnt t
+    global idCnt t b
 
     # Create identifier
     set id hold[incr idCnt]
@@ -1402,6 +1548,7 @@ proc ComposeHold {mainW handler} {
     OkButtons $w $t(ok) $t(cancel) "set ${id}(done)"
     pack $w.desc -side top -anchor w -padx 5 -pady 5
     pack $w.buttons -side top -pady 5 -fill x
+    set b($w.desc.entry) hold_description
     
     Place $w composeHold
     grab $w
@@ -1420,6 +1567,7 @@ proc ComposeHold {mainW handler} {
 
     RecordPos $w composeHold
     catch {focus $oldFocus}
+    unset b($w.desc.entry)
     destroy $w
     unset hd
 }
@@ -1437,7 +1585,7 @@ proc ComposeHold {mainW handler} {
 #		of the window
 
 proc ComposeChoose {msg info} {
-    global idCnt t option defaultFont langCharset
+    global idCnt t b option defaultFont langCharset
 
     set msgs [ComposeChooseDig [$msg body] $msg]
     while { -1 != [set i [lsearch -exact $msgs {}]]} {
@@ -1506,6 +1654,8 @@ proc ComposeChoose {msg info} {
 	pack $f.r -side left -anchor n
 	pack $f.t
 	pack $f -side top -anchor w
+	set b($f.r) choose_msg
+	set b($f.t) choose_msg
     }
     pack $w.l -side top -expand 1 -fill both
     OkButtons $w $t(ok) $t(cancel) "set ${id}(done)"
@@ -1589,6 +1739,7 @@ proc ComposeBuildHE {w handler textvariable} {
     set hd(lines) 1
     set hd(varname) $textvariable
     set hd(scroll) 0
+    set hd(width) 0
 
     # Do bindings
     bind $w <FocusIn> "focus $w.t"
@@ -1596,12 +1747,12 @@ proc ComposeBuildHE {w handler textvariable} {
     bind $w.t <Tab> {focus [tk_focusNext %W]; break}
     bind $w.t <Shift-Tab> {focus [tk_focusPrev %W]; break}
     bind $w.t <Shift-space> { }
-    bind $w.t <space> {bell; break}
     bind $w.t <KeyRelease-comma> "ComposeHandleHE %W $handler"
     bind $w.t <FocusOut> "ComposeHandleHE %W $handler"
     bind $w.t <Destroy> "unset $handler"
     bind $w.t <ButtonRelease-2> "ComposeHandleHE %W $handler"
     bind $w.t <Control-l> "ComposeHandleHEAlias %W $handler"
+    bind $w.t <Configure> "ComposeHandleHEConfigure %W $handler %w"
 
     # Create error tag
     if {[winfo cells $w.t] > 2} {
@@ -1615,10 +1766,53 @@ proc ComposeBuildHE {w handler textvariable} {
 	set textvar {}
     } else {
 	$w.t insert end $textvar
-	ComposeHandleHE $w.t $handler
     }
 }
 
+# ComposeHandleHEConfigure --
+#
+# Handle configure events in an address entry
+#
+# Arguments:
+# w	  - The text widget
+# handler - The handler which identifies this address widget
+# pixwidth- The width of the text widget (in pixels)
+
+proc ComposeHandleHEConfigure {w handler pixwidth} {
+    global defaultFontWidth
+    upvar #0 $handler hd
+
+    if ![info exists hd(borders)] {
+	set hd(borders) [expr 2*([$w cget -borderwidth] \
+			 +[$w cget -highlightthickness])]
+    }
+    set width [expr ($pixwidth-$hd(borders))/$defaultFontWidth]
+    if {$width == $hd(width)} {
+	return
+    }
+
+    set hd(width) $width
+
+    if {$width > 60} {
+	set cw [expr $width/3]
+	set hd(col1) $cw
+	set hd(col2) [expr $cw*2]
+	set hd(cols) 3
+    } elseif {$width > 40} {
+	set cw [expr $width/2]
+	set hd(col1) $cw
+	set hd(col2) $width
+	set hd(cols) 2
+    } else {
+	set hd(col1) $width
+	set hd(col2) $width
+	set hd(cols) 1
+    }
+    $w configure -tabs [list [expr $defaultFontWidth*$hd(col1)] \
+			     [expr $defaultFontWidth*$hd(col2)]]
+    ComposeHandleHE $w $handler
+}
+
 # ComposeHandleHE --
 #
 # Handle events in an address entry
@@ -1633,6 +1827,10 @@ proc ComposeHandleHE {w handler} {
 
     set var [string trim [$w get 1.0 end]]
 
+    if {$hd(col1) < 1} {
+	return
+    }
+
     $w delete 1.0 end
     set hd(column) 1
     set hd(lines) 1
@@ -1646,16 +1844,18 @@ proc ComposeHandleHE {w handler} {
 	    set tag {}
 	    set adr $adr2
 	}
-	if {(2==$hd(column) && 48 < [string length $adr])
-		|| (3==$hd(column) && 24 < [string length $adr])} {
+	if {(2==$hd(column) &&
+			[expr $hd(width)-$hd(col1)] <= [string length $adr])
+		|| (3==$hd(column) &&
+			[expr $hd(width)-$hd(col2)] <= [string length $adr])} {
 	    $w insert end \n
 	    incr hd(lines)
 	    set hd(column) 1
 	}
 	$w insert end $adr $tag , {}
-	set i [expr [string length $adr]/24+1]
+	set i [expr [string length $adr]/$hd(col1)+1]
 	incr hd(column) $i
-	if {3 < $hd(column)} {
+	if {$hd(cols) < $hd(column)} {
 	    $w insert end \n
 	    incr hd(lines)
 	    set hd(column) 1
@@ -1699,7 +1899,7 @@ proc ComposeHandleHEAlias {w handler} {
 	    $w insert end ,
 	}
 	$w insert end $alias
-	ComposeHandleHE $w $handler 1
+	ComposeHandleHE $w $handler
     }
 }
 
@@ -1713,15 +1913,17 @@ proc SendDeferred {} {
     global t ratDeferred deferred langCharset valueFont ratSenderSending
 
     set w .deferred
-    set deferred(w) $w
-    set deferred(to_send) $ratDeferred
-    set deferred(sent) 0
 
     # Allow only one window at a time.
     if [winfo exists $w] {
+	incr deferred(to_send) [RatSend sendDeferred]
 	return
     }
 
+    set deferred(w) $w
+    set deferred(sent) 0
+    set deferred(oldDeferred) $ratDeferred
+
     toplevel $w
     wm title $w $t(send_deferred)
 
@@ -1740,7 +1942,7 @@ proc SendDeferred {} {
     trace variable ratDeferred w SendDeferredUpdate
     trace variable ratSenderSending w SendDeferredUpdate
 
-    RatSend sendDeferred
+    set deferred(to_send) [RatSend sendDeferred]
     return $w
 }
 
@@ -1761,11 +1963,12 @@ proc SendDeferredUpdate {name1 name2 op} {
 	RatSend init
 	return
     }
-    if {$deferred(to_send) > $ratDeferred} {
+    if {$ratDeferred < $deferred(oldDeferred)} {
 	incr deferred(sent) 1
+	incr deferred(to_send) -1
     }
-    set deferred(to_send) $ratDeferred
-    if {0 == $ratDeferred && 0 == $ratSenderSending} {
+    set deferred(oldDeferred) $ratDeferred
+    if {0 == $deferred(to_send) || 0 == $ratSenderSending} {
 	RecordPos $deferred(w) sendDeferred
 	destroy $deferred(w)
 	trace vdelete ratDeferred w SendDeferredUpdate
@@ -2012,7 +2215,7 @@ proc ComposeWrapSelection {handler} {
 # handler -	The handler for the active compose session
 
 proc ComposeSpecifyCmd {handler} {
-    global idCnt t defaultFont cmdArrayId cmdList cmdName cmdCmd
+    global idCnt t b defaultFont cmdArrayId cmdList cmdName cmdCmd
     upvar #0 $handler mh
 
     # Create identifier
@@ -2033,7 +2236,7 @@ proc ComposeSpecifyCmd {handler} {
     entry $w.s.entry -width 20 -textvariable ${id}(saveAs) \
 	    -font $defaultFont
     bind $w.s.entry <KeyRelease> \
-	    "if \[string length ${id}(saveAs)\] { \
+	    "if {0 < \[string length ${id}(saveAs)\]} { \
 		 set ${id}(doSave) 1 \
 	     } else { \
 		 set ${id}(doSave) 0 \
@@ -2042,10 +2245,13 @@ proc ComposeSpecifyCmd {handler} {
     pack $w.s.label -side left -padx 5 -anchor s
     pack $w.s.entry \
 	 $w.s.but -side right -pady 5
+    set b($w.s.but) save_cmd_as
+    set b($w.s.entry) save_cmd_as
 
     # The text widget and the buttons
     text $w.t -relief sunken -bd 1 -wrap word -setgrid 1 -font $defaultFont
     Size $w.t giveCmd
+    set b($w.t) command_to_run_through
     OkButtons $w $t(ok) $t(cancel) "set ${id}(done)"
 
     pack $w.s -side top -anchor w -fill x
@@ -2121,8 +2327,8 @@ proc ComposeRunCmd {handler cmd} {
     $hd(composeBody) insert end "\n\n\n\t$t(command_is_running)..."
 
     # Disable compose window
-    foreach b $hd(eEditBlock) {
-	$b configure -state disabled
+    foreach block $hd(eEditBlock) {
+	$block configure -state disabled
     }
     $hd(composeBody) configure -state disabled
 
@@ -2135,8 +2341,8 @@ proc ComposeRunCmdDone {handler name name1 name2 op} {
     upvar #0 $handler hd
 
     if [info exists hd] {
-	foreach b $hd(eEditBlock) {
-	    $b configure -state normal
+	foreach block $hd(eEditBlock) {
+	    $block configure -state normal
 	}
 	$hd(composeBody) configure -state normal
 	$hd(composeBody) delete 1.0 end
@@ -2155,7 +2361,7 @@ proc ComposeRunCmdDone {handler name name1 name2 op} {
 	    close $fh
 	}
     }
-    catch {exec rm -f $name.in $name.out}
+    catch {file delete -force -- $name.in $name.out}
     trace vdelete $name1($name2) w "ComposeRunCmdDone $handler $name"
 }
 
@@ -2202,7 +2408,7 @@ proc CmdRead {} {
 # Arguments:
 
 proc CmdList {} {
-    global cmdList cmdName idCnt t defaultFont
+    global cmdList cmdName idCnt t b defaultFont
 
     # Create identifier
     set id ccmd[incr idCnt]
@@ -2230,6 +2436,7 @@ proc CmdList {} {
     pack $w.l.scroll -side right -fill y
     pack $w.l.list -expand 1 -fill both
     Size $w.l.list cmdList
+    set b($w.l.list) saved_commands
 
     # The buttons
     frame $w.b
@@ -2240,6 +2447,9 @@ proc CmdList {} {
     pack $w.b.apply \
 	 $w.b.delete \
 	 $w.b.close -side top -pady 5 -padx 5
+    set b($w.b.apply) apply_changes_to_cmd
+    set b($w.b.delete) delete_command
+    set b($w.b.close) dismiss
 
     # The command content
     text $w.t \
@@ -2248,7 +2458,9 @@ proc CmdList {} {
 	    -wrap word \
 	    -font $defaultFont \
 	    -width 40 \
-	    -height 4
+	    -height 4 \
+	    -state disabled
+    set b($w.t) command_content
 
     # Pack them all
     pack $w.t -side bottom
@@ -2269,6 +2481,7 @@ proc CmdList {} {
     bind $w.l.list <ButtonRelease-1> "\
 	    $w.b.apply configure -state disabled; \
 	    $w.b.delete configure -state normal; \
+	    $w.t configure -state normal; \
 	    $w.t delete 1.0 end; \
 	    $w.t insert 1.0 \$cmdCmd(\[lindex \$cmdList \[%W index @%x,%y\]\])"
     bind $w.t <KeyRelease> "CmdTextCheck $w"
@@ -2301,6 +2514,7 @@ proc CmdDelete {w handler} {
 
     # Clear the text
     $w.t delete 1.0 end
+    $w.t configure -state disabled
 
     # Disable the buttons
     $w.b.apply configure -state disabled
@@ -2360,3 +2574,25 @@ proc CmdClose {w handler} {
     unset hd
     destroy $w
 }
+
+# CheckHostName --
+#
+# The user has just changed our local host name. Check if it is okay
+#
+# Arguments:
+
+proc CheckHostName {} {
+    global ratCurrentHost composeWindowList option
+
+    if {1 >= [llength [split $ratCurrentHost .]] && 0 == $option(force_send)} {
+	set state disabled
+    } else {
+	set state normal
+    }
+    foreach hd $composeWindowList {
+	upvar #0 ${hd}(toplevel) w
+
+	$w.buttons.send configure -state $state
+	$w.buttons.sendsave configure -state $state
+    }
+}
