@@ -3,7 +3,8 @@
 # This file contains code which handles the actual displaying of a message
 #
 #
-#  TkRat software and its included text is Copyright 1996 by Martin Forssen.
+#  TkRat software and its included text is Copyright 1996,1997,1998
+#  by Martin Forssn
 #
 #  The full text of the legal notice is contained in the file called
 #  COPYRIGHT, included with this distribution.
@@ -36,21 +37,36 @@ proc ShowNothing handler {
 # Arguments:
 # handler -	The handler which identifies the show text widget
 # msg     -	Message to show
+# browse  -	True if we should use browse mode showing
 
-proc Show {handler msg} {
-    upvar #0 hd$handler fh
-    global option langCharset
+proc Show {handler msg browse} {
+    global option b langCharset
+    upvar #0 $handler fh \
+    	     msgInfo_$msg msgInfo
 
     set fh(id) 0
     set fh(current) $msg
+    set fh(sigstatus) pgp_none
+    set fh(pgpOutput) ""
+    set fh(browse) $browse
 
-    # Enable updates
+    # Enable updates & clear data
     $handler configure -state normal
+    $handler delete 0.0 end
+    eval "$handler tag delete [$handler tag names]"
 
     # Reset the tags to a known status
-    $handler tag delete [$handler tag names]
     $handler tag configure HeaderName -font [GetFont $langCharset 0 bold]
     $handler tag configure Center -justify center -spacing1 5m -spacing3 5m
+    if { 4 < [winfo cells $handler]} {
+	$handler tag configure URL -foreground $option(url_color) -underline 1
+	$handler tag configure Found -background #ffff80
+    } else {
+	$handler tag configure URL -underline 1
+	$handler tag configure Found -borderwidth 2 -relief raised
+    }
+    $handler tag bind URL <Enter> "$handler config -cursor hand2"
+    $handler tag bind URL <Leave> "$handler config -cursor xterm"
 
     # Delete old subwindows
     foreach slave [winfo children $handler] {
@@ -67,16 +83,21 @@ proc Show {handler msg} {
 	unset fh($n)
     }
     set fh(lastType) ""
-
-    # Clear the display
-    $handler delete 0.0 end
+    foreach bn [array names b $handler.*] {unset b($bn)}
 
     # Show the message
     ShowMessage 1 $handler $msg
     BuildStructMenu $handler $msg
 
+    # Prepare for URL searching
+    $handler mark set searched 1.0
+
     # Don't allow the user to change this
     $handler configure -state disabled
+
+    # Return the signature status
+    set msgInfo(pgp,signed_parts) $fh(signed_parts)
+    return [list $fh(sigstatus) $fh(pgpOutput)]
 }
 
 # InsertHeader --
@@ -90,7 +111,7 @@ proc Show {handler msg} {
 
 proc InsertHeader {gtag w header {width 1}} {
     global t
-    upvar #0 hd$w fh
+    upvar #0 $w fh
 
     set hn [lindex $header 0]
     regsub -all -- - [string tolower $hn] _ hni
@@ -133,7 +154,7 @@ proc InsertHeader {gtag w header {width 1}} {
 
 proc ShowMessage {first handler msg} {
     global option t
-    upvar #0 hd$handler fh \
+    upvar #0 $handler fh \
     	     msgInfo_$msg msgInfo
 
     set tag t[incr fh(id)]
@@ -142,8 +163,10 @@ proc ShowMessage {first handler msg} {
     }
     if {0 == $first} {
 	$handler tag bind $tag <3> "tk_popup $fh(struct_menu) %X %Y \
-			        \[lsearch \[set hd${handler}(struct_list)\] \
+			        \[lsearch \[set ${handler}(struct_list)\] \
 			        $msg\]"
+    } else {
+	set fh(signed_parts) {}
     }
 
     if ![string length $msg] {
@@ -182,6 +205,26 @@ proc ShowMessage {first handler msg} {
     default	{ }
     }
 
+    if $fh(browse) {
+	if ![info exists fh(browse,$msg)] {
+	    if {1 == $first} {
+		set fh(browse,$msg) $fh(browse)
+	    } else {
+		set fh(browse,$msg) 0
+	    }
+	}
+    } else {
+	set fh(browse,$msg) 0
+    }
+
+    if $fh(browse,$msg) {
+	button $handler.download -text $t(show_body) \
+		-command "Show $handler $msg 0" -cursor top_left_arrow
+	$handler window create end -window $handler.download -padx 20 -pady 20
+	set fh(sigstatus) pgp_none
+	return
+    }
+
     # Insert the body
     set body [$msg body]
     if ![info exists msgInfo(show,$body)] {
@@ -196,6 +239,11 @@ proc ShowMessage {first handler msg} {
         set msgInfo(show,$body) 0
     }
     ShowBody $handler $body $msg
+
+    if {1 == $first && ![string compare pgp_good $fh(sigstatus)] &&
+	    [string compare pgp_good [$body sigstatus]]} {
+	set fh(sigstatus) pgp_part
+    }
 }
 
 # ShowBody --
@@ -210,13 +258,24 @@ proc ShowMessage {first handler msg} {
 
 proc ShowBody {handler body msg} {
     global option
-    upvar #0 hd$handler fh \
+    upvar #0 $handler fh \
     	     msgInfo_$msg msgInfo
 
     if !$msgInfo(show,$body) {
         return
     }
 
+    # Update signature status
+    if [string compare pgp_none [set sigstatus [$body sigstatus]]] {
+	set fh(sigstatus) $sigstatus
+	lappend fh(signed_parts) $body
+	if [string length $fh(pgpOutput)] {
+	    set fh(pgpOutput) "$fh(pgpOutput)\n\n[$body getPGPOutput]"
+	} else {
+	    set fh(pgpOutput) [$body getPGPOutput]
+	}
+    }
+
     # We will need this later on.
     set type [string tolower [$body type]]
 
@@ -242,6 +301,8 @@ proc ShowBody {handler body msg} {
     message/rfc822		{ ShowMessage 0 $handler [$body message] }
     message/delivery-status	{ ShowDSN $handler $body }
     image/gif           	{ ShowImage $handler $body gif }
+    application/pgp-keys	{ ShowPGPKeys $handler $body }
+    application/pgp		{ ShowTextPlain $handler $body $msg}
     default			{ ShowDefault $handler $body }
     }
 }
@@ -256,13 +317,13 @@ proc ShowBody {handler body msg} {
 # msg     -	The message name
 
 proc ShowTextPlain {handler body msg} {
-    global option defaultFont
-    upvar #0 hd$handler fh \
+    global option b defaultFont
+    upvar #0 $handler fh \
     	     msgInfo_$msg msgInfo
 
     set tag t[incr fh(id)]
     $handler tag bind $tag <3> "tk_popup $fh(struct_menu) %X %Y \
-			        \[lsearch \[set hd${handler}(struct_list)\] \
+			        \[lsearch \[set ${handler}(struct_list)\] \
 			        $body\]"
     set action [$body getShowCharset]
     if ![string compare good [lindex $action 0]] {
@@ -282,8 +343,10 @@ proc ShowTextPlain {handler body msg} {
 	    	-text "$t(here_is_text) '$charset' $t(which_cant_be_shown)"
 	label $w.modlab -text $t(mode):
 	menubutton $w.mode -menu $w.mode.m -indicatoron 1 -relief raised \
-	    	-textvariable hd${handler}(mode,$body)
+	    	-textvariable ${handler}(mode,$body)
+	set b($w.mode) howto_display
 	button $w.save -text $t(save_to_file) -command "SaveBody $body"
+	set b($w.save) save_bodypart
 	pack $w.type -side top -anchor w
 	pack $w.modlab $w.mode -side left
 	pack $w.save -side left -padx 20
@@ -291,7 +354,7 @@ proc ShowTextPlain {handler body msg} {
 	menu $w.mode.m -tearoff 0
 	foreach s {dont_show show_sevenbit show_all} {
 	    $w.mode.m add command -label $t($s) \
-		    -command "ShowTextCharset $body $msg hd$handler $handler \
+		    -command "ShowTextCharset $body $msg $handler $handler \
 					      $tag $s"
 	    if {[string length $t($s)] > $l} {
 		set l [string length $t($s)]
@@ -304,7 +367,7 @@ proc ShowTextPlain {handler body msg} {
 	$handler window create insert -window $w -pady 5
 	foreach win [concat $w [pack slaves $w]] {
 	    bind $win <3> "tk_popup $fh(struct_menu) %X %Y \
-			   \[lsearch \[set hd${handler}(struct_list)\] \
+			   \[lsearch \[set ${handler}(struct_list)\] \
 			   $body\]"
 	}
 	$handler insert insert "\n" $tag
@@ -312,7 +375,7 @@ proc ShowTextPlain {handler body msg} {
 	$handler mark set ${body}_e insert
 	$handler mark gravity ${body}_s left
 	$handler mark gravity ${body}_e right
-	ShowTextCharset $body $msg hd$handler $handler $tag \
+	ShowTextCharset $body $msg $handler $handler $tag \
 		$msgInfo(show,$body,how)
     }
 }
@@ -362,8 +425,8 @@ proc ShowTextCharset {body msg hd w tag mode} {
 # msg     -	The message name
 
 proc ShowTextOther {handler body msg} {
-    global t defaultFont langCharset
-    upvar #0 hd$handler fh \
+    global t b defaultFont langCharset
+    upvar #0 $handler fh \
     	     msgInfo_$msg msgInfo
 
     set type [$body type]
@@ -407,7 +470,8 @@ proc ShowTextOther {handler body msg} {
     pack $w.text -side top -anchor w
 
     if [string length [lindex $mailcap 0]] {
-	button $w.view_ext -text $t(view) -command "RunMailcap $body {$mailcap}"
+	button $w.view_ext -text $t(view) \
+		-command "RunMailcap $body [list $mailcap]"
 	pack $w.view_ext -side left -padx 10
     }
     if ![info exists msgInfo(show,$body,tp)] {
@@ -425,11 +489,14 @@ proc ShowTextOther {handler body msg} {
     button $w.save -text $t(save_to_file) -command "SaveBody $body"
     button $w.view_int -text $t(view_source) -command "ShowSource $body"
     pack $w.tp $w.save $w.view_int -side left -padx 5
+    set b($w.tp) view_as_text
+    set b($w.save) save_bodypart
+    set b($w.view_int) view_source
 
     $handler window create insert -window $w -padx 5 -pady 5
     foreach win [concat $w [pack slaves $w]] {
 	bind $win <3> "tk_popup $fh(struct_menu) %X %Y \
-		       \[lsearch \[set hd${handler}(struct_list)\] \
+		       \[lsearch \[set ${handler}(struct_list)\] \
 		       $body\]"
     }
     $handler insert insert "\n"
@@ -451,7 +518,7 @@ proc ShowTextOther {handler body msg} {
 # msg     -	The message name
 
 proc ShowTextOtherDo {handler body msg} {
-    upvar #0 hd$handler fh \
+    upvar #0 $handler fh \
     	     msgInfo_$msg msgInfo
     
     $handler configure -state normal
@@ -474,7 +541,7 @@ proc ShowTextOtherDo {handler body msg} {
 
 proc ShowMultiAlt {handler body msg} {
     global GoodTypes
-    upvar #0 hd$handler fh \
+    upvar #0 $handler fh \
     	     msgInfo_$msg msgInfo
     set found 0
 
@@ -520,7 +587,7 @@ proc ShowMultiAlt {handler body msg} {
 
 proc ShowMultiMixed {handler body msg} {
     global option t
-    upvar #0 hd$handler fh \
+    upvar #0 $handler fh \
     	     msgInfo_$msg msgInfo
 
     foreach child [$body children] {
@@ -542,7 +609,7 @@ proc ShowMultiMixed {handler body msg} {
 
 proc ShowImage {handler body subtype} {
     global option
-    upvar #0 hd$handler fh
+    upvar #0 $handler fh
 
     set tag t[incr fh(id)]
     set filename $option(tmp)/rat.[RatGenId]
@@ -550,9 +617,15 @@ proc ShowImage {handler body subtype} {
     $body saveData $fid 0 0
     close $fid
     switch $subtype {
-    gif    { set img [image create photo -format gif -file $filename] }
+    gif    {
+	    if [catch {image create photo -format gif -file $filename} img] {
+		Popup $img
+		ShowDefault $handler $body
+		return
+	    }
+	}
     }
-    exec rm -f $filename
+    file delete -force -- $filename
     set c [canvas $handler.c[incr fh(id)] -width [image width $img] \
 	    -height [image height $img]]
     $c create image 0 0 -image $img -anchor nw
@@ -561,10 +634,10 @@ proc ShowImage {handler body subtype} {
     $handler insert insert "\n" $tag
     lappend fh(show_cleanup_cmds) "image delete $img"
     $handler tag bind $tag <3> "tk_popup $fh(struct_menu) %X %Y \
-				 \[lsearch \[set hd${handler}(struct_list)\] \
+				 \[lsearch \[set ${handler}(struct_list)\] \
 				 $body\]"
     bind $c <3> "tk_popup $fh(struct_menu) %X %Y \
-			  \[lsearch \[set hd${handler}(struct_list)\] \
+			  \[lsearch \[set ${handler}(struct_list)\] \
 			  $body\]"
 }
 
@@ -579,11 +652,11 @@ proc ShowImage {handler body subtype} {
 
 proc ShowDSN {handler body} {
     global t
-    upvar #0 hd$handler fh
+    upvar #0 $handler fh
 
     set tag t[incr fh(id)]
     $handler tag bind $tag <3> "tk_popup $fh(struct_menu) %X %Y \
-				 \[lsearch \[set hd${handler}(struct_list)\] \
+				 \[lsearch \[set ${handler}(struct_list)\] \
 				 $body\]"
     $handler insert insert \n $tag
     set dsn [$body dsn]
@@ -609,6 +682,48 @@ proc ShowDSN {handler body} {
     $handler insert insert \n $tag
 }
 
+# ShowPGPKeys --
+#
+# Handles embedded pgp keys
+#
+# Arguments:
+# handler -	The handler which identifies the show text widget
+# body    -	The bodypart to show
+
+proc ShowPGPKeys {handler body} {
+    global t b
+    upvar #0 $handler fh
+
+    set mailcap [$body findShowCommand]
+
+    set w $handler.w[incr fh(id)]
+    frame $w -relief raised -bd 2 -cursor top_left_arrow
+    if [string length [lindex $mailcap 4]] {
+	if ![catch {image create bitmap $body.icon -file [lindex $mailcap 4]}] {
+	    label $w.icon -image $body.icon
+	    pack $w.icon -side left
+	}
+    }
+    label $w.label -relief flat -text $t(embedded_pgp_keys)
+    pack $w.label -side top -anchor w
+
+    button $w.add -text $t(add_to_keyring) \
+	    -command "RatPGP add \[$body data 0\]"
+    button $w.save -text $t(save_to_file) -command "SaveBody $body"
+    button $w.view_int -text $t(view_source) -command "ShowSource $body"
+    set b($w.add) add_to_keyring
+    set b($w.save) save_bodypart
+    set b($w.view_int) view_source
+    pack $w.add $w.save $w.view_int -side left -padx 5
+    $handler window create insert -window $w -padx 5 -pady 5
+    foreach win [concat $w [pack slaves $w]] {
+	bind $win <3> "tk_popup $fh(struct_menu) %X %Y \
+		       \[lsearch \[set ${handler}(struct_list)\] \
+		       $body\]"
+    }
+    $handler insert insert "\n"
+}
+
 # ShowDefault --
 #
 # The default type shower, just iserts a marker that here is an object
@@ -620,7 +735,7 @@ proc ShowDSN {handler body} {
 
 proc ShowDefault {handler body} {
     global t defaultFont langCharset
-    upvar #0 hd$handler fh
+    upvar #0 $handler fh
 
     set type [$body type]
     set typename [string tolower [lindex $type 0]/[lindex $type 1]]
@@ -663,15 +778,18 @@ proc ShowDefault {handler body} {
 
     if [string length [lindex $mailcap 0]] {
 	button $w.view_ext -text $t(view) -command "RunMailcap $body {$mailcap}"
+	set b($w.view_ext) run_mailcap
 	pack $w.view_ext -side left -padx 10
     }
     button $w.save -text $t(save_to_file) -command "SaveBody $body"
     button $w.view_int -text $t(view_source) -command "ShowSource $body"
+    set b($w.save) save_bodypart
+    set b($w.view_int) view_source
     pack $w.save $w.view_int -side left -padx 5
     $handler window create insert -window $w -padx 5 -pady 5
     foreach win [concat $w [pack slaves $w]] {
 	bind $win <3> "tk_popup $fh(struct_menu) %X %Y \
-		       \[lsearch \[set hd${handler}(struct_list)\] \
+		       \[lsearch \[set ${handler}(struct_list)\] \
 		       $body\]"
     }
     $handler insert insert "\n"
@@ -688,6 +806,17 @@ proc ShowHome {handler} {
     $handler yview moveto 0
 }
 
+# ShowBottom --
+#
+# Scroll to the bottom of the document
+#
+# Arguments:
+# handler -	The handler which identifies the show text widget
+
+proc ShowBottom {handler} {
+    $handler yview moveto 1
+}
+
 # ShowPageDown --
 #
 # Move the show one page down
@@ -710,6 +839,28 @@ proc ShowPageUp {handler} {
     $handler yview scroll -1 pages
 }
 
+# ShowLineDown --
+#
+# Move the show one line down
+#
+# Arguments:
+# handler -	The handler which identifies the show text widget
+
+proc ShowLineDown {handler} {
+    $handler yview scroll 1 units
+}
+
+# ShowLineUp --
+#
+# Move the show one line up
+#
+# Arguments:
+# handler -	The handler which identifies the show text widget
+
+proc ShowLineUp {handler} {
+    $handler yview scroll -1 units
+}
+
 # SaveBody --
 #
 # Save a bodypart to file.
@@ -776,7 +927,7 @@ proc SaveBody {body} {
 	}
 	if {0 == $action} {
 	    set mode [GetMode $id]
-	    if { 0 == [catch "open $hd(name) w $mode" fh]} {
+	    if { 0 == [catch [list open $hd(name) w $mode] fh]} {
 		$body saveData $fh false $hd(convertNL)
 		close $fh
 		# Defeat the umask
@@ -803,7 +954,7 @@ proc SaveBody {body} {
 # msg     -	The message we should display the structure of
 
 proc BuildStructMenu {handler msg} {
-    upvar #0 hd$handler fh
+    upvar #0 $handler fh
 
     # Clear the old menu
     $fh(struct_menu) delete 0 end
@@ -815,6 +966,8 @@ proc BuildStructMenu {handler msg} {
     # Check if we got anything
     if ![string length $msg] { return }
 
+    if $fh(browse,$msg) { return }
+
     BuildStructEntry $handler $fh(struct_menu) $msg [$msg body] ""
 }
 
@@ -830,8 +983,8 @@ proc BuildStructMenu {handler msg} {
 # preamble-	The preamble to add before this entry
 
 proc BuildStructEntry {handler m msg body preamble} {
-    global t showParams
-    upvar #0 hd$handler fh \
+    global b
+    upvar #0 $handler fh \
     	     msgInfo_$msg msgInfo
 
     if ![string length $body] { return }
@@ -841,36 +994,11 @@ proc BuildStructEntry {handler m msg body preamble} {
     set typepair [$body type]
     set type [string tolower [lindex $typepair 0]/[lindex $typepair 1]]
     $m add cascade -label $preamble$type -menu $sm
+    set b($m,[$m index end]) bodypart_entry
     menu $sm -tearoff 0 \
-    	     -disabledforeground [lindex [$m configure -activeforeground] 3]
+    	     -disabledforeground [$m cget -activeforeground] \
+	     -postcommand "PopulateStructEntry $handler $sm $msg $body $type"
 
-    if [string length [$body description]] {
-        $sm add command -label [$body description] -state disabled
-    }
-    # Do parameters
-    if [info exists showParams($type)] {
-        set sp $showParams($type)
-        foreach param $sp {
-            set value [$body parameter $param]
-            if [string length $param] {
-		$sm add command -label "$t($param): $value" \
-				-state disabled
-            }
-        }
-    } else {
-        set sp {{}}
-    }
-    $sm add command -label "$t(size): [RatMangleNumber [$body size]]" \
-    		    -state disabled
-    $sm add separator
-    if ![info exists msgInfo(show,$body)] {
-	set msgInfo(show,$body) 1
-    }
-    $sm add checkbutton -label $t(show) -onvalue 1 -offvalue 0 \
-    	    -variable msgInfo_${msg}(show,$body) \
-    	    -command "Show $handler $fh(current)"
-    $sm add command -label $t(save_to_file)... -command "SaveBody $body"
-    $sm add command -label $t(view_source)... -command "ShowSource $body"
 
     # See if we have children to show
     switch -glob $type {
@@ -892,6 +1020,63 @@ proc BuildStructEntry {handler m msg body preamble} {
     }
 }
 
+# PopulateStructEntry --
+#
+# Populates a struct menu entry
+#
+# Arguments:
+# handler - The handler which identifies the show text widget
+# m       - Menu to populate
+# msg  	  - Message the body belongs to
+# body 	  - Body to use
+# type 	  - Type of body
+
+proc PopulateStructEntry {handler m msg body type} {
+    global t b showParams
+    upvar #0 $handler fh \
+    	     msgInfo_$msg msgInfo
+
+    $m delete 0 end
+
+    if [string length [$body description]] {
+        $m add command -label [$body description] -state disabled
+    }
+    set sigstatus [$body sigstatus]
+    if [string compare pgp_none $sigstatus] {
+	$m add command -label "$t(signature): $t($sigstatus)" -state disabled
+    }
+    if [$body encoded] {
+	$m add command -label "$t(decoded)" -state disabled
+    }
+
+    # Do parameters
+    if [info exists showParams($type)] {
+        set sp $showParams($type)
+        foreach param $sp {
+            set value [$body parameter $param]
+            if [string length $param] {
+		$m add command -label "$t($param): $value" -state disabled
+            }
+        }
+    } else {
+        set sp {{}}
+    }
+    $m add command -label "$t(size): [RatMangleNumber [$body size]]" \
+    		    -state disabled
+    $m add separator
+    if ![info exists msgInfo(show,$body)] {
+	set msgInfo(show,$body) 1
+    }
+    $m add checkbutton -label $t(show) -onvalue 1 -offvalue 0 \
+    	    -variable msgInfo_${msg}(show,$body) \
+    	    -command "Show $handler $fh(current) 0"
+    set b($m,[$m index end]) bodypart_show
+    $m add command -label $t(save_to_file)... -command "SaveBody $body"
+    set b($m,[$m index end]) bodypart_save
+    $m add command -label $t(view_source)... -command "ShowSource $body"
+    set b($m,[$m index end]) view_source
+}
+
 # RunMailcap --
 #
 # Runs the mailcap entry for a body
@@ -950,7 +1135,7 @@ proc RunMailcap {body mailcap} {
 	Place $w extView
 	$w.text insert insert [eval "exec $cmd"]
 	$w.text configure -state disabled
-	exec rm -f $rm(fileName) &
+	file delete -force -- $rm(fileName) &
 	unset rm
 	return
     } else {
@@ -972,7 +1157,106 @@ proc RunMailcap {body mailcap} {
 proc RunMailcapDone {name1 name2 op} {
     upvar #0 $name1 rm
 
-    exec rm -f $rm(fileName) &
+    file delete -force -- $rm(fileName) &
     unset rm
 }
 
+# RatShowURL --
+#
+# Invoke an URL browser.
+#
+# Arguments:
+# url     -	URL to browse
+# win     -	Text widget containing URL
+# tag     -	Tag of url in text widget
+# x, y	  -	Mouse coordinates
+
+proc RatShowURL {url win tag x y} {
+    global option
+
+    # Flash URL for feedback
+    $win tag configure $tag -underline 1
+    update idletasks
+
+    # Check if we should abort
+    if {-1 == [lsearch -exact [$win tag names @$x,$y] $tag]} {
+	return
+    }
+
+    # Start viewer
+    switch $option(url_viewer) {
+	RatUP {
+		if [catch {RatUP_ShowURL $url} error] {
+		    Popup error
+		}
+	    }
+	netscape {
+		if [catch {eval {exec $option(netscape)} $option(netscape_args)\
+			{-remote "openURL($url)"} >&/dev/null}] {
+		    if [catch {eval {exec $option(netscape)} \
+			    $option(netscape_args) {$url} &} error] {
+			Popup $error
+		    }
+		}
+	    }
+    }
+}
+
+# RatFindURL --
+#
+# Search for an URL in a part of a text widget
+#
+# Arguments:
+# t	- Text widget
+# start - Index to start searching at
+# end   - Index to end searching at
+
+proc RatFindURL {t start end} {
+    global option
+
+    set exp "([join $option(urlprot) |]):/+\[^ \n\t\"\>]+\[^ \n\t,\.\)>\'\"\]"
+    set found [$t search -nocase -regexp -count len $exp $start $end]
+    if [string length $found] {
+	return [list $found $found+${len}c]
+    } else {
+	return {}
+    }
+}
+
+# RatScrollShow --
+#
+# Scroll the show text widget and search for URL's if needed
+#
+# Arguments:
+# t       - The text widget
+# scrollw - The scrollbar widget
+# sfrac   - The new start of shown section
+# efrac   - The new end of shown section
+
+proc RatScrollShow {t scrollw sfrac efrac} {
+    $scrollw set $sfrac $efrac
+
+    if [$t compare searched >= @0,20000] {
+	return
+    }
+
+    upvar #0 $t fh
+
+    set start searched
+    set end [$t index @0,20000+1c+20l]
+
+    set result [RatFindURL $t $start $end]
+    while {[llength $result] == 2} {
+	set tag t[incr fh(id)]
+	set s [lindex $result 0]
+	set e [lindex $result 1]
+	$t tag add URL $s $e
+	$t tag add $tag $s $e
+	regsub % [$t get $s $e] %% url
+	$t tag bind $tag <ButtonRelease-1> [list RatShowURL $url %W $tag %x %y]
+	$t tag bind $tag <1> \
+		"$t tag configure $tag -underline 0; update idletasks"
+	set result [RatFindURL $t $e $end]
+    }
+    $t mark set searched $end
+}
