@@ -5,7 +5,8 @@
 # Each virtual folder is described by one entry in the vFolderDef array.
 #
 #
-#  TkRat software and its included text is Copyright 1996 by Martin Forssen.
+#  TkRat software and its included text is Copyright 1996,1997,1998
+#  by Martin Forssn
 #
 #  The full text of the legal notice is contained in the file called
 #  COPYRIGHT, included with this distribution.
@@ -18,7 +19,7 @@
 # Arguments:
 
 proc VFolderDef {} {
-    global t vFolderRowHeight vFolderFrameHeight vFolderChanged
+    global t b vFolderRowHeight vFolderFrameHeight vFolderChanged
 
     set vFolderChanged 0
 
@@ -31,43 +32,61 @@ proc VFolderDef {} {
 	return
     }
     upvar #0 $id hd
+    set hd(w) $w
     set hd(done) 0
+    set hd(dragging) 0
     set oldFocus [focus]
 
-
     # Create toplevel
     toplevel $w
     wm title $w $t(vfolderdef)
 
     # Populate window
+    FindAccelerators a {window folder import}
     frame $w.mbar -relief raised -bd 1
-    menubutton $w.mbar.w -menu $w.mbar.w.m -text $t(window)
-    menu $w.mbar.w.m
-    $w.mbar.w.m add command -label $t(close) \
-	    -command "RecordPos $w vFolderDef; \
-	    RecordSize $w.canvas vFolderDef; destroy $w; unset $id"
-    menubutton $w.mbar.folder -menu $w.mbar.folder.m -text $t(folder)
-    menu $w.mbar.folder.m
-    $w.mbar.folder.m add command -label $t(create_submenu)... \
+    set m $w.mbar.w.m
+    menubutton $w.mbar.w -menu $m -text $t(window) -underline $a(window)
+    menu $m
+    $m add command -label $t(close) -command "VFolderWinClose $id"
+    set b($m,[$m index end]) dismiss
+
+    set m $w.mbar.folder.m
+    menubutton $w.mbar.folder -menu $m -text $t(folder) -underline $a(folder)
+    set b($w.mbar.folder) vd_foldermenu
+    menu $m
+    $m add command -label $t(create_submenu)... \
 	    -command {set vFolderCurrent {0 end}; VFolderCreateMenu}
-    $w.mbar.folder.m add command -label $t(new_file)... \
+    set b($m,[$m index end]) vd_createsub
+    $m add command -label $t(new_file)... \
 	    -command {set vFolderCurrent {0 end}; VFolderNew file}
-    $w.mbar.folder.m add command -label $t(new_mh)... \
+    set b($m,[$m index end]) vd_newfile
+    $m add command -label $t(new_mh)... \
 	    -command {set vFolderCurrent {0 end}; VFolderNew mh}
-    $w.mbar.folder.m add command -label $t(new_dbase)... \
+    set b($m,[$m index end]) vd_newmh
+    $m add command -label $t(new_dbase)... \
 	    -command {set vFolderCurrent {0 end}; VFolderNew dbase}
-    $w.mbar.folder.m add command -label $t(new_imap)... \
+    set b($m,[$m index end]) vd_newdbase
+    $m add command -label $t(new_imap)... \
 	    -command {set vFolderCurrent {0 end}; VFolderNew imap}
-    $w.mbar.folder.m add command -label $t(new_pop3)... \
+    set b($m,[$m index end]) vd_newimap
+    $m add command -label $t(new_pop3)... \
 	    -command {set vFolderCurrent {0 end}; VFolderNew pop3}
-    $w.mbar.folder.m add command -label $t(new_dynamic)... \
+    set b($m,[$m index end]) vd_newpop
+    $m add command -label $t(new_dynamic)... \
 	    -command {set vFolderCurrent {0 end}; VFolderNew dynamic}
-    menubutton $w.mbar.import -menu $w.mbar.import.m -text $t(import)
-    menu $w.mbar.import.m
-    $w.mbar.import.m add command -label $t(directory)... \
-	    -command VFolderImportDirectory
-    $w.mbar.import.m add command -label $t(import_IMAP)... \
-	    -command VFolderImportIMAP
+    set b($m,[$m index end]) vd_newdynamic
+
+    set m $w.mbar.import.m
+    menubutton $w.mbar.import -menu $m -text $t(import) -underline $a(import)
+    set b($w.mbar.import) vd_importmenu
+    menu $m
+    $m add command -label $t(directory)... -command VFolderImportDirectory
+    set b($m,[$m index end]) vd_importdir
+    $m add command -label "MH $t(directory)..." \
+	    -command VFolderImportMHDirectory
+    set b($m,[$m index end]) vd_importmh
+    $m add command -label $t(import_IMAP)... -command VFolderImportIMAP
+    set b($m,[$m index end]) vd_importimap
 
     pack $w.mbar.w \
 	 $w.mbar.folder \
@@ -90,19 +109,30 @@ proc VFolderDef {} {
 
     menu $w.m -tearoff 0
     $w.m add command -label $t(edit)... -command EditVFolderEntry
+    set b($w.m,[$w.m index end]) vd_edit
     $w.m add command -label $t(delete)... -command VFolderDelete
+    set b($w.m,[$w.m index end]) vd_delete
     $w.m add command -label $t(set_to_inbox) \
 	    -command {VFolderSetToSpecial Inbox}
+    set b($w.m,[$w.m index end]) vd_setinbox
     $w.m add command -label $t(set_to_save_outgoing) \
 	    -command {VFolderSetToSpecial Save}
+    set b($w.m,[$w.m index end]) vd_setsaveout
     $w.m add separator
     $w.m add command -label $t(create_submenu)... -command VFolderCreateMenu
+    set b($w.m,[$w.m index end]) vd_createsub
     $w.m add command -label $t(new_file)... -command {VFolderNew file}
+    set b($w.m,[$w.m index end]) vd_newfile
     $w.m add command -label $t(new_mh)... -command {VFolderNew mh}
+    set b($w.m,[$w.m index end]) vd_newmh
     $w.m add command -label $t(new_dbase)... -command {VFolderNew dbase}
+    set b($w.m,[$w.m index end]) vd_newdbase
     $w.m add command -label $t(new_imap)... -command {VFolderNew imap}
+    set b($w.m,[$w.m index end]) vd_newimap
     $w.m add command -label $t(new_pop3)... -command {VFolderNew pop3}
+    set b($w.m,[$w.m index end]) vd_newpop
     $w.m add command -label $t(new_dynamic)... -command {VFolderNew dynamic}
+    set b($w.m,[$w.m index end]) vd_newdynamic
 
     # Find out how big the listboxes are
     listbox $w.canvas.l -height 1 -width 1 -highlightthickness 0
@@ -120,6 +150,24 @@ proc VFolderDef {} {
     Place $w vFolderDef
 }
 
+# VFolderWinClose --
+#
+# Close a vfolderdef window
+#
+# Arguments:
+# handler - The handler identifying the window
+
+proc VFolderWinClose {handler} {
+    upvar #0 $handler hd
+    global b
+
+    RecordPos $hd(w) vFolderDef
+    RecordSize $hd(w).canvas vFolderDef
+    destroy $hd(w)
+    foreach bn [array names b $hd(w).*] {unset b($bn)}
+    unset hd
+}
+
 # RedrawVFolderList --
 #
 # Redraws the list of vfolders. This is done by first clearing the canvas
@@ -157,7 +205,7 @@ proc RedrawVFolderList {} {
 
 proc DrawVFolderStruct {sIdent column} {
     global vFolderX vFolderY option vFolderCI vFolderStruct vFolderSave \
-	   vFolderRowHeight vFolderFrameHeight maxX maxY vFolderInbox 
+	   vFolderRowHeight vFolderFrameHeight maxX maxY vFolderInbox b
 
     set c .vfolderdef.canvas
     set l $c.l$vFolderCI
@@ -165,7 +213,8 @@ proc DrawVFolderStruct {sIdent column} {
     listbox $l -width $option(vfoldername_width) -relief groove \
 	    -height [llength $vFolderStruct($sIdent)] -highlightthickness 0 \
 	    -selectmode single
-    bind $l <1> "VFolderPostMenu %X %Y %W $sIdent \[%W index @%x,%y\]"
+    bind $l <1> "VFolderPostMenu $c %X %Y %W $sIdent \[%W index @%x,%y\]; break"
+    set b($l) vd_list
     set x [lindex $vFolderX $column]
     set y [lindex $vFolderY $column]
     set id [$c create window $x $y -anchor nw -window $l -tags $sIdent]
@@ -214,9 +263,11 @@ proc DrawVFolderStruct {sIdent column} {
     if {[lindex $bbox 3] > $maxY} {
 	set maxY [lindex $bbox 3]
     }
-    bind $l <3> "VFolderStartDrag $c %x %y $id $sIdent \[%W index @%x,%y\]"
+    if {0 < [$l size]} {
+	bind $l <3> "VFolderStartDrag $c %x %y $id $sIdent \[%W index @%x,%y\]"
+    }
     bind $l <B3-Motion> "VFolderDrag $c %x %y"
-    bind $l <B3-ButtonRelease> "VFolderStopDrag $c"
+    bind $l <ButtonRelease-3> "VFolderStopDrag $c"
     set vFolderY [lreplace $vFolderY $column $column [expr [lindex $bbox 3]+10]]
     return [lindex $bbox 1]
 }
@@ -225,19 +276,23 @@ proc DrawVFolderStruct {sIdent column} {
 # Post a the menu for a given entry. The arguments are: the root window
 # coordinates, the listbox, the ident of the vFolderStruct and the index
 # of the active member.
-# VFolderPostMenu --
 #
 # Post the menu for the given entry.
 #
 # Arguments:
+# c	 -	The canvas widget command name
 # x, y   -	Position of pointer, this is where we will popup the menu
 # listW  -	Name of the list widget the entry belongs to
 # sIdent -	Identifier of the vFolderStruct the entry belongs to
 # index  -	Index in the above mentioned vFolderStruct
 
-proc VFolderPostMenu {x y listW sIdent index} {
+proc VFolderPostMenu {c x y listW sIdent index} {
     global vFolderCurrent vFolderStruct vFolderInbox vFolderSave t
 
+    if [winfo exists $c.l] {
+	return
+    }
+
     set vFolderCurrent [list $sIdent $index]
     set elem [lindex $vFolderStruct($sIdent) $index]
     if ![llength $elem] {
@@ -306,6 +361,7 @@ proc VFolderStartDrag {c x y id sIdent index} {
 
 proc VFolderDrag {c x y} {
     global vFLastX vFLastY
+    if ![winfo exists $c.l] return
     $c move drag [expr $x-$vFLastX] [expr $y-$vFLastY]
     set vFLastX $x
     set vFLastY $y
@@ -323,6 +379,7 @@ proc VFolderStopDrag {c} {
     global vFLastX vFLastY vFolderRowHeight vFolderFrameHeight \
 	   vFOffsetX vFOffsetY vFolderStruct vFolderCurrent vFolderChanged
 
+    if ![winfo exists $c.l] return
     set x [expr $vFLastX+$vFOffsetX]
     set y [expr $vFLastY+$vFOffsetY]
     destroy $c.l
@@ -333,12 +390,13 @@ proc VFolderStopDrag {c} {
     if {1 != [llength $id]} {
 	return
     }
-    if {"window" != [$c type $id]} {
+    if [string compare window [$c type $id]] {
 	return
     }
     set w [$c itemcget $id -window]
     set iy [expr $y-[lindex [$c bbox $id] 1]]
-    set indexN [expr int(($iy-$vFolderFrameHeight/2) /$vFolderRowHeight)]
+    set indexN [expr int(($iy-$vFolderFrameHeight/2+$vFolderRowHeight/2)/ \
+	    $vFolderRowHeight)]
     set sIdentN [$c itemcget $id -tags]
     if { 1 < [llength $sIdentN] } {
 	set sIdentN [lreplace $sIdentN \
@@ -347,12 +405,40 @@ proc VFolderStopDrag {c} {
     set sIdent [lindex $vFolderCurrent 0]
     set index [lindex $vFolderCurrent 1]
     set item [lindex $vFolderStruct($sIdent) $index]
+    if ![string compare struct [lindex $item 0]] {
+	if {[lindex $item 1] == $sIdentN
+		|| [VFolderIsDescendant $sIdentN [lindex $item 1]]} {
+	    return
+	}
+    }
     set vFolderStruct($sIdent) [lreplace $vFolderStruct($sIdent) $index $index]
     set vFolderStruct($sIdentN) [linsert $vFolderStruct($sIdentN) $indexN $item]
     incr vFolderChanged
     RedrawVFolderList
 }
 
+# VFolderIsDescendant --
+#
+# Checks if one vfolder structure is descendant of another.
+#
+# Arguments:
+# brat -	Structure to check for
+# candidate -	Possible parent
+
+proc VFolderIsDescendant {brat candidate} {
+    global vFolderStruct
+
+    foreach c $vFolderStruct($candidate) {
+	if ![string compare struct [lindex $c 0]] {
+	    if {$brat == [lindex $c 1]
+		    || 1 == [VFolderIsDescendant $brat [lindex $c 1]]} {
+		return 1
+	    }
+	}
+    }
+    return 0
+}
+
 # VFolderNew --
 # This procedure is invoked when the user wants to create a new vfolder.
 # The struct it should be inserted into is specified via the vFolderCurrent
@@ -366,14 +452,14 @@ proc VFolderNew {type} {
 	   vFolderDef option env
 
     switch $type {
-    file      {set vFolderDef($vFolderDefIdent) {{} file {}}}
-    mh        {set vFolderDef($vFolderDefIdent) {{} mh {}}}
+    file      {set vFolderDef($vFolderDefIdent) {{} file {} {}}}
+    mh        {set vFolderDef($vFolderDefIdent) {{} mh {} {}}}
     dbase     {set vFolderDef($vFolderDefIdent) \
-	        [list {} dbase {} $option(def_extype) $option(def_exdate)]}
+	        [list {} dbase {} {} $option(def_extype) $option(def_exdate)]}
     imap      {set vFolderDef($vFolderDefIdent) \
-	        "{} imap \\\\{:$option(imap_port)/imap\\\\} $env(USER)"}
-    pop3      {set vFolderDef($vFolderDefIdent) {{} pop3 \\{/pop3\\} {}}}
-    dynamic   {set vFolderDef($vFolderDefIdent) {{} dynamic {} sender}}
+	        "{} imap {} \\\\{:$option(imap_port)/imap\\\\} $env(USER)"}
+    pop3      {set vFolderDef($vFolderDefIdent) {{} pop3 {} \\{/pop3\\} {}}}
+    dynamic   {set vFolderDef($vFolderDefIdent) {{} dynamic {} {} sender}}
     }
     set vFolderName [EditVFolder $vFolderDefIdent]
     if ![string length $vFolderName] {
@@ -396,7 +482,7 @@ proc VFolderNew {type} {
 # Arguments:
 
 proc VFolderCreateMenu {} {
-    global vFolderCurrent t idCnt
+    global vFolderCurrent t b idCnt
 
     # Create identifier
     set id cf[incr idCnt]
@@ -412,6 +498,7 @@ proc VFolderCreateMenu {} {
     # Folder name
     frame $w.name
     label $w.name.label -text $t(name):
+    set b($w.name) vd_menuname
     entry $w.name.entry -textvariable ${id}(name) -width 20
     pack $w.name.entry \
 	 $w.name.label -side right
@@ -425,6 +512,7 @@ proc VFolderCreateMenu {} {
     grab $w
     tkwait variable ${id}(done)
     RecordPos $w vFolderCreateMenu
+    unset b($w.name)
     destroy $w
     focus $oldFocus
 
@@ -449,7 +537,7 @@ proc VFolderCreateMenu {} {
 # Arguments:
 
 proc EditVFolderEntry {} {
-    global t idCnt vFolderCurrent vFolderStruct vFolderChanged vFolderDef
+    global t b idCnt vFolderCurrent vFolderStruct vFolderChanged vFolderDef
 
     set sIdent [lindex $vFolderCurrent 0]
     set index [lindex $vFolderCurrent 1]
@@ -474,6 +562,7 @@ proc EditVFolderEntry {} {
 	    entry $w.name.entry -textvariable ${id}(name) -width 20
 	    pack $w.name.label $w.name.entry -side left
 	    pack $w.name -side top -expand 1 -padx 5 -pady 5
+	    set b($w.name.entry) vd_menuname
 	    OkButtons $w $t(ok) $t(cancel) "set ${id}(done)"
 	    pack $w.buttons -fill both -padx 5 -pady 5
 
@@ -497,6 +586,7 @@ proc EditVFolderEntry {} {
 	    RecordPos $w editVfolderEntry
 	    destroy $w
 	    focus $oldFocus
+	    unset b($w.name.entry)
 	    unset hd
 	}
     vfolder {
@@ -545,8 +635,11 @@ proc VFolderDelete {} {
 		return
 	    } 
 	    if {0 == $ret} {
-		catch {RatImapDeleteFolder [lindex [lindex $vd 2] 0] \
-			[lindex $vd 3]}
+		if [catch {RatImapDeleteFolder [lindex [lindex $vd 3] 0] \
+			[lindex $vd 4]}] {
+		    Popup $t(delete_failed)
+		    return
+		}
 	    }
 	}
 	unset vFolderDef([lindex $elem 1])
@@ -610,7 +703,7 @@ proc VFolderDeleteStruct {op sIdent} {
 		    	    $t(dont_delete)]
 		}
 		if {0 == $op} {
-		    catch {RatImapDeleteFolder [lindex [lindex $vd 2] 0] \
+		    catch {RatImapDeleteFolder [lindex [lindex $vd 3] 0] \
 			    [lindex $vd 3]}
 		}
 	    }
@@ -649,18 +742,28 @@ proc EditVFolder {ident} {
     pack $w.name.label $w.name.entry -side left
     pack $w.name -side top -expand 1 -padx 5 -pady 5
     bind $w <Return> {}
+    set b($w.name.entry) vd_name
+
+    # Features
+    array set hd [lindex $def 2]
+    if ![info exists hd(browse)] {
+	set hd(browse) 0
+    }
+    checkbutton $w.browse -text $t(default_to_browse) -variable ${id}(browse) \
+	    -anchor w
+    pack $w.browse -side top -fill x -padx 10
 
     # Insert type specific widgets
     switch [lindex $def 1] {
     file {
-	    set hd(fname) [lindex $def 2]
+	    set hd(fname) [lindex $def 3]
 	    frame $w.chooser -relief sunken -bd 1
 	    set fh [FileSelectorCreate $w.chooser ${id}(fname) \
 		    "set ${id}(done) 1" 0 0]
 	    pack $w.chooser -side top -fill both -expand 1 -padx 5 -pady 5
 	}
     mh {
-	    set hd(mh_name) [string range [lindex $def 2] 4 end]
+	    set hd(mh_name) [string range [lindex $def 3] 4 end]
 	    frame $w.mhn
 	    label $w.mhn.label -text $t(mh_name):
 	    entry $w.mhn.entry -textvariable ${id}(mh_name) -width 20
@@ -668,39 +771,43 @@ proc EditVFolder {ident} {
 	    pack $w.mhn.label \
 		 $w.mhn.entry -side left -pady 5
 	    pack $w.mhn -side top
+	    set b($w.mhn.entry) vd_mhname
 	}
     dbase {
-	    set hd(keywords) [lindex $def 2]
-	    set hd(extype) [lindex $def 3]
-	    set hd(exdate) [lindex $def 4]
+	    set hd(keywords) [lindex $def 3]
+	    set hd(extype) [lindex $def 4]
+	    set hd(exdate) [lindex $def 5]
 	    frame $w.keywords
 	    label $w.keywords.label -text $t(keywords):
 	    entry $w.keywords.entry -textvariable ${id}(keywords) -width 20
 	    pack $w.keywords.entry \
 		 $w.keywords.label -side right
+	    set b($w.keywords.entry) keywords
 	    frame $w.extype
 	    label $w.extype.label -text $t(extype):
 	    frame $w.extype.b
 	    radiobutton $w.extype.b.none -text $t(none) \
 		    -variable ${id}(extype) -value none
+	    set b($w.extype.b.none) exp_none
 	    radiobutton $w.extype.b.remove -text $t(remove) \
 		    -variable ${id}(extype) -value remove
+	    set b($w.extype.b.none) exp_remove
 	    radiobutton $w.extype.b.incoming -text $t(add_incoming) \
 		    -variable ${id}(extype) -value incoming
+	    set b($w.extype.b.none) exp_incoming
 	    radiobutton $w.extype.b.backup -text $t(backup) \
 		    -variable ${id}(extype) -value backup
-	    radiobutton $w.extype.b.custom -variable ${id}(extype) \
-		    -value custom -state disabled
+	    set b($w.extype.b.none) exp_backup
 	    pack $w.extype.b.none \
 		 $w.extype.b.remove \
 		 $w.extype.b.incoming \
-		 $w.extype.b.backup \
-		 $w.extype.b.custom -side top -anchor w
+		 $w.extype.b.backup -side top -anchor w
 	    pack $w.extype.b \
 		 $w.extype.label -side right -anchor nw
 	    frame $w.exdate
 	    label $w.exdate.label -text $t(exdate):
 	    entry $w.exdate.entry -textvariable ${id}(exdate) -width 20
+	    set b($w.exdate.entry) exp_date
 	    pack $w.exdate.entry \
 		 $w.exdate.label -side right
 	    pack $w.keywords \
@@ -708,8 +815,8 @@ proc EditVFolder {ident} {
 		 $w.exdate -side top -fill both
 	}
     imap {
-	    set hd(user) [lindex $def 3]
-	    set h [lindex [lindex $def 2] 0]
+	    set hd(user) [lindex $def 4]
+	    set h [lindex [lindex $def 3] 0]
 	    regsub {(/|\}).+} [string trim $h \{] {} host
 	    set hlist [split $host :]
 	    set hd(host) [lindex $hlist 0]
@@ -719,54 +826,35 @@ proc EditVFolder {ident} {
 	    regsub {.+\}} $h {} hd(mbox)
 	    set oldHost $hd(host)
 	    set oldMbox $hd(mbox)
-	    frame $w.host
-	    label $w.host.label -text $t(host): -width 17 -anchor e
-	    entry $w.host.entry -textvariable ${id}(host) -width 30
-	    pack $w.host.label $w.host.entry -side left
-	    frame $w.user
-	    label $w.user.label -text $t(user): -width 17 -anchor e
-	    entry $w.user.entry -textvariable ${id}(user) -width 30
-	    pack $w.user.label $w.user.entry -side left
-	    frame $w.mbox
-	    label $w.mbox.label -text $t(mbox): -width 17 -anchor e
-	    entry $w.mbox.entry -textvariable ${id}(mbox) -width 30
-	    pack $w.mbox.label $w.mbox.entry -side left
-	    frame $w.port
-	    label $w.port.label -text $t(port): -width 17 -anchor e
-	    entry $w.port.entry -textvariable ${id}(port) -width 30
-	    pack $w.port.label $w.port.entry -side left
-	    pack $w.host \
-		 $w.user \
-		 $w.mbox \
-		 $w.port -side top -fill both
+	    foreach e {host user mbox port} {
+		frame $w.$e
+		label $w.$e.label -text $t($e): -width 17 -anchor e
+		entry $w.$e.entry -textvariable ${id}($e) -width 30
+		pack $w.$e.label $w.$e.entry -side left
+		pack $w.$e -side top -fill both
+		set b($w.$e.entry) vd_$e
+	    }
 	}
     pop3 {
-	    set hd(user) [lindex $def 3]
-	    set h [lindex [lindex $def 2] 0]
+	    set hd(user) [lindex $def 4]
+	    set h [lindex [lindex $def 3] 0]
 	    regsub {(/|\}).+} [string trim $h \{] {} host
 	    set hlist [split $host :]
 	    set hd(host) [lindex $hlist 0]
 	    if {2 == [llength $hlist]} {
 		set hd(port) [lindex $hlist 1]
 	    }
-	    frame $w.host
-	    label $w.host.label -text $t(host): -width 17 -anchor e
-	    entry $w.host.entry -textvariable ${id}(host) -width 30
-	    pack $w.host.label $w.host.entry -side left
-	    frame $w.user
-	    label $w.user.label -text $t(user): -width 17 -anchor e
-	    entry $w.user.entry -textvariable ${id}(user) -width 30
-	    pack $w.user.label $w.user.entry -side left
-	    frame $w.port
-	    label $w.port.label -text $t(port): -width 17 -anchor e
-	    entry $w.port.entry -textvariable ${id}(port) -width 30
-	    pack $w.port.label $w.port.entry -side left
-	    pack $w.host \
-		 $w.user \
-		 $w.port -side top -fill both
+	    foreach e {host user port} {
+		frame $w.$e
+		label $w.$e.label -text $t($e): -width 17 -anchor e
+		entry $w.$e.entry -textvariable ${id}($e) -width 30
+		pack $w.$e.label $w.$e.entry -side left
+		pack $w.$e -side top -fill both
+		set b($w.$e.entry) vd_$e
+	    }
 	}
     dynamic {
-	    set hd(fname) [lindex $def 2]
+	    set hd(fname) [lindex $def 3]
 	    frame $w.chooser -relief sunken -bd 1
 	    set fh [FileSelectorCreate $w.chooser ${id}(fname) \
 		    "set ${id}(done) 1" 1 1]
@@ -785,10 +873,12 @@ proc EditVFolder {ident} {
 	tkwait variable ${id}(done)
 
 	if { 1 == $hd(done) } {
+	    set hd(name) [string trim $hd(name)]
 	    if ![string length $hd(name)] {
 		Popup $t(need_name)
 		continue
 	    }
+	    set features [list browse $hd(browse)]
 	    switch [lindex $def 1] {
 	    file {
 		    FileSelectorCleanup $fh
@@ -798,14 +888,15 @@ proc EditVFolder {ident} {
 			continue
 		    }
 		    FileSelectorDone $fh
-		    set vFolderDef($ident) [list $hd(name) file $hd(fname)]
+		    set vFolderDef($ident) [list $hd(name) file $features \
+			    $hd(fname)]
 		}
 	    mh {
 		    if {![string length $hd(mh_name)]} {
 		        Popup $t(need_mh_name)
 			continue
 		    }
-		    set vFolderDef($ident) [list $hd(name) mh \
+		    set vFolderDef($ident) [list $hd(name) mh $features \
 			    "#MH/$hd(mh_name)"]
 		}
 	    dbase {
@@ -813,7 +904,7 @@ proc EditVFolder {ident} {
 		        Popup $t(need_keyword)
 			continue
 		    }
-		    set vFolderDef($ident) [list $hd(name) dbase \
+		    set vFolderDef($ident) [list $hd(name) dbase $features \
 			    $hd(keywords) $hd(extype) $hd(exdate)]
 		}
 	    imap {
@@ -827,7 +918,7 @@ proc EditVFolder {ident} {
 		    } else {
 			set host $hd(host)
 		    }
-		    set vFolderDef($ident) [list $hd(name) imap \
+		    set vFolderDef($ident) [list $hd(name) imap $features \
 			    \{{$host}$hd(mbox)\} $hd(user)]
 		    if {[string compare $oldHost $hd(host)]
 			    || [string compare $oldMbox $hd(mbox)]} {
@@ -845,7 +936,7 @@ proc EditVFolder {ident} {
 		    } else {
 			set host $hd(host)
 		    }
-		    set vFolderDef($ident) [list $hd(name) pop3 \
+		    set vFolderDef($ident) [list $hd(name) pop3 $features \
 			    \{{$host/pop3}\} $hd(user)]
 		}
 	    dynamic {
@@ -856,7 +947,7 @@ proc EditVFolder {ident} {
 		    }
 		    FileSelectorDone $fh
 		    set vFolderDef($ident) \
-			    [list $hd(name) dynamic $hd(fname) sender]
+			    [list $hd(name) dynamic $features $hd(fname) sender]
 		}
 	    }
 	    incr vFolderChanged
@@ -865,12 +956,14 @@ proc EditVFolder {ident} {
 	    focus $oldFocus
 	    set name $hd(name)
 	    unset hd
+	    foreach bn [array names b $w.*] {unset b($bn)}
 	    return $name
 	} else {
 	    RecordPos $w editVfolder
 	    destroy $w
 	    focus $oldFocus
 	    unset hd
+	    foreach bn [array names b $w.*] {unset b($bn)}
 	    return ""
 	}
     }
@@ -940,7 +1033,7 @@ proc VFolderWindowDestroy {} {
 
 proc VFolderImportDirectory {} {
     global vFolderDefIdent vFolderStruct vFolderChanged vFolderStructIdent\
-	   vFolderDef t idCnt
+	   vFolderDef t b idCnt
 
     # Create identifier
     set id cf[incr idCnt]
@@ -958,15 +1051,16 @@ proc VFolderImportDirectory {} {
     set hd(fname) {}
     frame $w.chooser -relief sunken -bd 1
     set fh [FileSelectorCreate $w.chooser ${id}(fname) "set ${id}(done) 1" 1 1]
-    pack $w.chooser -side top -fill both -expand 1 -padx 5 -pady 5
 
     # The patterns
     frame $w.pat
     label $w.pat.lab -text $t(patterns_to_use)
     label $w.pat.labf -text $t(files):
     entry $w.pat.entf -textvariable ${id}(file_pattern)
+    set b($w.pat.entf) vd_filepat
     label $w.pat.labd -text $t(directories):
     entry $w.pat.entd -textvariable ${id}(dir_pattern)
+    set b($w.pat.entd) vd_dirpat
     grid $w.pat.lab -columnspan 2
     grid $w.pat.labf -row 1 -column 0 -sticky e
     grid $w.pat.entf -row 1 -column 1 -sticky ew
@@ -982,6 +1076,8 @@ proc VFolderImportDirectory {} {
     grab $w
     tkwait variable ${id}(done)
     RecordPos $w vFolderImportDirectory
+    unset b($w.pat.entf)
+    unset b($w.pat.entd)
     destroy $w
 
     if {1 == $hd(done)} {
@@ -995,7 +1091,7 @@ proc VFolderImportDirectory {} {
 	    VFolderScanDir $sid $hd(fname) $hd(file_pattern) $hd(dir_pattern)
 	} else {
 	    lappend vFolderStruct(0) [list vfolder $vFolderDefIdent $name]
-	    set vFolderDef($vFolderDefIdent) [list $name file $hd(fname)]
+	    set vFolderDef($vFolderDefIdent) [list $name file {} $hd(fname)]
 	    incr vFolderDefIdent
 	}
 	RedrawVFolderList
@@ -1003,6 +1099,7 @@ proc VFolderImportDirectory {} {
     unset hd
 }
 
+
 # VFolderScanDir --
 #
 # Scan a directory and build a structure from it.
@@ -1035,7 +1132,7 @@ proc VFolderScanDir {sid dirName file_pattern dir_pattern} {
 		continue
 	    }
 	    lappend vFolderStruct($sid) [list vfolder $vFolderDefIdent $name]
-	    set vFolderDef($vFolderDefIdent) [list $name file $f]
+	    set vFolderDef($vFolderDefIdent) [list $name file {} $f]
 	    incr vFolderDefIdent
 	} else {
 	    Popup "$f $t(ignored_not_file_not_dir)"
@@ -1043,6 +1140,125 @@ proc VFolderScanDir {sid dirName file_pattern dir_pattern} {
     }
 }
 
+# VFolderImportMHDirectory --
+#
+# Import a folder structure from disk (i.e. the folders are stored in
+# directories on the disk.
+#
+# Arguments:
+
+proc VFolderImportMHDirectory {} {
+    global vFolderDefIdent vFolderStruct vFolderChanged vFolderStructIdent\
+	   vFolderDef t idCnt
+
+    # Create identifier
+    set id cf[incr idCnt]
+    set w .$id
+    upvar #0 $id hd
+    set hd(done) 0
+    set hd(file_pattern) {*}
+    set hd(dir_pattern) {*}
+
+    # Create toplevel
+    toplevel $w
+    wm title $w $t(import_directory)
+
+    # Create fileselector
+    set hd(fname) {}
+    frame $w.chooser -relief sunken -bd 1
+    set fh [FileSelectorCreate $w.chooser ${id}(fname) "set ${id}(done) 1" 1 1]
+
+    # The patterns
+    frame $w.pat
+    label $w.pat.lab -text $t(patterns_to_use)
+    label $w.pat.labf -text $t(files): -anchor e
+    entry $w.pat.entf -textvariable ${id}(file_pattern)
+    label $w.pat.labd -text $t(directories): -anchor e
+    entry $w.pat.entd -textvariable ${id}(dir_pattern)
+    grid $w.pat.lab -columnspan 2
+    grid $w.pat.labf $w.pat.entf -sticky ew
+    grid $w.pat.labd $w.pat.entd -sticky ew
+    pack $w.pat
+
+    OkButtons $w $t(ok) $t(cancel) "set ${id}(done)"
+    pack $w.buttons \
+    	 $w.chooser -side bottom -fill both -expand 1 -padx 5 -pady 5
+
+    Place $w vFolderImportDirectory
+    grab $w
+    tkwait variable ${id}(done)
+    RecordPos $w vFolderImportDirectory
+    destroy $w
+
+    if {1 == $hd(done)} {
+	FileSelectorDone $fh
+	incr vFolderChanged
+	set sid [incr vFolderStructIdent]
+
+	set name [file tail [string trimright $hd(fname) /]]
+	if [VFolderMHHasSubfolders $hd(fname)] {
+	    lappend vFolderStruct(0) [list struct $sid "$name/"]
+	    VFolderMHScanDir $sid $hd(fname) $hd(file_pattern) $hd(dir_pattern)
+	}
+	RedrawVFolderList
+    }
+    unset hd
+}
+
+proc VFolderMHHasSubfolders { dirName } {
+    if { [file tail $dirName] == "{}" } {
+       return 0
+    }
+    if [catch "glob $dirName/*/."] {
+        return 0
+    } else {
+        return 1
+    }
+}
+
+# VFolderMHScanDir --
+#
+# Scan a directory and build a structure from it.
+#
+# Arguments:
+# sid	       - Structure id for our struct entry
+# dirName      - Name of the directory
+# file_pattern - pattern to match files against
+# dir_pattern  - pattern to match directories against
+
+proc VFolderMHScanDir {sid dirName file_pattern dir_pattern {folder_prefix {}}} {
+    global vFolderDefIdent vFolderStruct vFolderDef vFolderStructIdent t 
+
+    set vFolderStruct($sid) {}
+    foreach f [lsort [glob -nocomplain $dirName/*]] {
+	set name [file tail $f]
+	if ![file readable $f] {
+	    Popup "$t(cant_read): $f"
+	    continue
+	}
+	if [VFolderMHHasSubfolders $f] {
+	    if ![string match $dir_pattern $name] {
+		continue
+	    }
+	    lappend vFolderStruct($sid) \
+		    [list struct [incr vFolderStructIdent] "$name/"]
+	    VFolderMHScanDir $vFolderStructIdent $f $file_pattern $dir_pattern $name/
+	    lappend vFolderStruct($sid) [list vfolder $vFolderDefIdent $name]
+	    set vFolderDef($vFolderDefIdent) [list $name mh {} "#MH$folder_prefix$name"]
+	    incr vFolderDefIdent
+	} elseif [file isfile $f] {
+	    # nothing to do...
+	} else {
+	    if ![string match $file_pattern $name] {
+		continue
+	    }
+	    lappend vFolderStruct($sid) [list vfolder $vFolderDefIdent $name]
+	    set vFolderDef($vFolderDefIdent) [list $name mh {} "#MH/$folder_prefix/$name"]
+	    incr vFolderDefIdent
+	}
+    }
+}
+
 # VFolderImportIMAP --
 #
 # Import a folder structure from an IMAP server.
@@ -1050,7 +1266,7 @@ proc VFolderScanDir {sid dirName file_pattern dir_pattern} {
 # Arguments:
 
 proc VFolderImportIMAP {} {
-    global t env idCnt vFolderChanged option \
+    global t b env idCnt vFolderChanged option \
 	   vFolderStruct vFolderStructIdent
 
     # Create identifier
@@ -1068,30 +1284,15 @@ proc VFolderImportIMAP {} {
     set hd(port) $option(imap_port)
     set hd(user) $env(USER)
     set hd(pattern) {*}
-    set row 0
-    label $w.host_label -text $t(host):
-    entry $w.host_entry -textvariable ${id}(host) -width 32
-    grid $w.host_label -row $row -sticky e -padx 5 -pady 5
-    grid $w.host_entry -row $row -column 1 -sticky ew -padx 5
-    incr row
-    label $w.port_label -text $t(port):
-    entry $w.port_entry -textvariable ${id}(port) -width 32
-    grid $w.port_label -row $row -sticky e -padx 5 -pady 5
-    grid $w.port_entry -row $row -column 1 -sticky ew -padx 5
-    incr row
-    label $w.user_label -text $t(user):
-    entry $w.user_entry -textvariable ${id}(user) -width 32
-    grid $w.user_label -row $row -sticky e -padx 5 -pady 5
-    grid $w.user_entry -row $row -column 1 -sticky ew -padx 5
-    incr row
-    label $w.pattern_label -text $t(pattern):
-    entry $w.pattern_entry -textvariable ${id}(pattern) -width 32
-    grid $w.pattern_label -row $row -sticky e -padx 5 -pady 5
-    grid $w.pattern_entry -row $row -column 1 -sticky ew -padx 5
-    incr row
+    foreach e {host port user pattern} {
+	label $w.${e}_label -text $t($e): -anchor e
+	entry $w.${e}_entry -textvariable ${id}($e) -width 32
+	set b($w.${e}_entry) vd_$e
+	grid $w.${e}_label $w.${e}_entry -sticky ew -padx 5
+    }
 
     OkButtons $w $t(ok) $t(cancel) "set ${id}(done)"
-    grid $w.buttons -row $row -columnspan 2 -pady 10 -sticky ew
+    grid $w.buttons -columnspan 2 -pady 10 -sticky ew
 
     Place $w vFolderImportIMAP
     focus $w.host_entry
@@ -1144,7 +1345,7 @@ proc VFolderBuildIMAPStruct {flist sIdent user} {
 	if [lindex $f 1] {
 	    lappend vFolderStruct($sIdent) [list vfolder $vFolderDefIdent \
 		    [lindex $f 0]]
-	    set vFolderDef($vFolderDefIdent) [list [lindex $f 0] imap \
+	    set vFolderDef($vFolderDefIdent) [list [lindex $f 0] imap {} \
 		    [lindex $f 2] $user]
 	    incr vFolderDefIdent
 	} else {
